"""
Definition of Classes implementing functionalities for cuts in the online and offline selection process.

Felix Metzner 2019
"""

import itertools

from enum import Enum, unique
from typing import Union, Optional, List, Tuple, Dict

from templatefitter.plotter.histogram_variable import HistVariable

from analysistools.processing.observables_utils import makeROOTCompatible, invert_make_ROOT_compatible


__all__ = [
    "CutType",
    "CutInfo",
]


class CutType:
    smaller_than_cut_direction = "SMALLER"  # type: str
    greater_than_cut_direction = "GREATER"  # type: str

    __cut_type_string_mapping = {
        "NONE": "",
        "EQUAL": "==",
        "SMALLER_CUT": "<",
        "GREATER_CUT": ">",
        "SE_CUT": "<=",
        "GE_CUT": ">=",
        "G_S_CUT": ("<", "<"),
        "G_SE_CUT": ("<", "<="),
        "GE_S_CUT": ("<=", "<"),
        "GE_SE_CUT": ("<=", "<="),
    }  # type: Dict[str, Union[str, Tuple[str, str]]]

    __cut_type_key_mapping = {
        v: k for k, v in __cut_type_string_mapping.items()
    }  # type: Dict[Union[str, Tuple[str, str]], str]

    __direction_mapping = {
        "NONE": None,
        "EQUAL": None,
        "SMALLER_CUT": smaller_than_cut_direction,
        "GREATER_CUT": greater_than_cut_direction,
        "SE_CUT": smaller_than_cut_direction,
        "GE_CUT": greater_than_cut_direction,
        "G_S_CUT": (greater_than_cut_direction, smaller_than_cut_direction),
        "G_SE_CUT": (greater_than_cut_direction, smaller_than_cut_direction),
        "GE_S_CUT": (greater_than_cut_direction, smaller_than_cut_direction),
        "GE_SE_CUT": (greater_than_cut_direction, smaller_than_cut_direction),
    }  # type: Dict[str, Union[None, str, Tuple[str, str]]]

    def __init__(
        self,
        type_str: Union[str, Tuple[str, str]],
    ) -> None:
        assert self.__direction_mapping.keys() == self.__cut_type_string_mapping.keys()
        assert isinstance(type_str, (str, tuple)), type(type_str)
        if isinstance(type_str, tuple):
            assert len(type_str) == 2, type_str
            assert all(isinstance(ts, str) for ts in type_str), type_str

        assert type_str in self.__cut_type_string_mapping.values(), type_str

        self.__type_str = type_str
        self.__type_key = self.__cut_type_key_mapping[type_str]
        self.__cut_direction = self.get_cut_direction(type_str=type_str)

    @staticmethod
    def get_cut_direction(type_str: Union[str, Tuple[str, str]]) -> Union[None, str, Tuple[str, str]]:
        return CutType.__direction_mapping[CutType.__cut_type_key_mapping[type_str]]

    @property
    def type_str(self) -> Union[str, Tuple[str, str]]:
        return self.__type_str

    @property
    def type_key(self) -> str:
        return self.__type_key

    @property
    def cut_direction(self) -> Union[None, str, Tuple[str, str]]:
        return self.__cut_direction

    @property
    def is_smaller_than_cut(self) -> bool:
        return self.__cut_direction is self.smaller_than_cut_direction

    @property
    def is_greater_than_cut(self) -> bool:
        return self.__cut_direction is self.greater_than_cut_direction

    @property
    def is_two_sided_cut(self) -> bool:
        if isinstance(self.type_str, tuple):
            assert isinstance(self.cut_direction, tuple)
            return True
        return False

    def __str__(self) -> str:
        return str(self.type_str)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CutType):
            raise NotImplementedError(f"other must be of type CutType, but is of type {type(other).__name__}")
        if self.type_str != other.type_str:
            return False
        assert self.cut_direction == other.cut_direction
        assert self.type_key == other.type_key

        return True


@unique
class CutTypes(Enum):
    NO_CUT = CutType(type_str="")
    EQUAL_CUT = CutType(type_str="==")
    SMALLER_CUT = CutType(type_str="<")
    GREATER_CUT = CutType(type_str=">")
    SE_CUT = CutType(type_str="<=")
    GE_CUT = CutType(type_str=">=")
    G_S_CUT = CutType(type_str=("<", "<"))
    G_SE_CUT = CutType(type_str=("<", "<="))
    GE_S_CUT = CutType(type_str=("<=", "<"))
    GE_SE_CUT = CutType(type_str=("<=", "<="))

    @classmethod
    def allowed_cut_type_values(cls) -> Tuple[CutType, ...]:
        return tuple([member.value for member in cls.__members__.values()])


class CutInfo:
    """
    Definition of a information set for cuts used for the Selection:
    Contains:
      - cut_name: Name of the cut.
      - cut_string: The corresponding cut string.
      - cut_label: The latex label for the cut.
      - cut_var: The corresponding variable.
      - cut_var_latex_label: The latex label for the corresponding variable as used in cut_label.
      - cut_type: The cut type, e.g. smaller, greater, equals, abs etc.
      - cut_limits: The limits where the cut is done.
      - latex_unit: The SI unit as used in the cut string!.
      - cut_abs: Boolean indicating whether to use the absolute value of the variable for the cut. Default = False.
      - cut_online: Whether cut is performed online or not. Default = False.
    """

    def __init__(
        self,
        cut_str: str,
        cut_label: str,
        cut_name: str,
        cut_var: str,
        cut_var_latex_label: str,
        cut_type: Union[str, Tuple[str, str], CutType],
        cut_limits: Union[float, Tuple[float, float]],
        latex_unit: Optional[str] = None,
        cut_abs: bool = False,
        cut_online: bool = False,
    ) -> None:

        self.__cut_str = cut_str
        self.__cut_label = cut_label
        self.__cut_var_latex_label = cut_var_latex_label
        self.__latex_unit = latex_unit
        self.__cut_name = cut_name
        self.__cut_var = cut_var
        self.__cut_type = cut_type if isinstance(cut_type, CutType) else CutType(type_str=cut_type)
        self.__cut_limits = cut_limits
        self.__cut_abs = cut_abs
        self.__cut_online = cut_online

        self.check_valid()

    @property
    def cut_str(self) -> str:
        return self.__cut_str

    @property
    def cut_str_root(self) -> str:
        return self.generate_cut_str(make_root_compatible=True)

    @property
    def cut_label(self) -> str:
        return self.__cut_label

    @property
    def cut_name(self) -> str:
        return self.__cut_name

    @property
    def cut_var(self) -> str:
        return self.__cut_var

    @property
    def cut_var_root(self) -> str:
        return makeROOTCompatible(variables=[self.__cut_var])[0]

    @property
    def cut_var_latex_label(self) -> str:
        return self.__cut_var_latex_label

    @property
    def latex_unit(self) -> Optional[str]:
        return self.__latex_unit

    @property
    def cut_type(self) -> CutType:
        return self.__cut_type

    @property
    def cut_limits(self) -> Union[float, Tuple[float, float]]:
        return self.__cut_limits

    @property
    def cut_abs(self) -> bool:
        return self.__cut_abs

    @property
    def cut_online(self) -> bool:
        return self.__cut_online

    def check_valid(self) -> None:
        assert self.cut_type in CutTypes.allowed_cut_type_values(), (self.cut_type, CutTypes.allowed_cut_type_values())
        if isinstance(self.cut_limits, tuple):
            assert isinstance(self.cut_type.type_str, type(self.cut_limits)), (self.cut_type.type_str, self.cut_limits)
            assert self.cut_limits[0] < self.cut_limits[1], self.cut_limits
        assert self.cut_str in self.generate_cut_strings(), self.cut_str

    @staticmethod
    def get_abs_value_var_for_var(var: str) -> str:
        return f"abs({var})"

    def generate_cut_strings(self, make_root_compatible: bool = False) -> List[str]:
        var = self.cut_var

        if CutTypes(self.cut_type) == CutTypes.NO_CUT:
            assert isinstance(CutTypes.NO_CUT.value.type_str, str), CutTypes.NO_CUT.value.type_str
            return [CutTypes.NO_CUT.value.type_str]

        is_already_abs = var.startswith("abs(") or var.startswith("abs__bo")
        if self.cut_abs and not is_already_abs:
            var = self.get_abs_value_var_for_var(var=var)

        if make_root_compatible:
            var = makeROOTCompatible(variables=[var])[0]

        if not self.cut_type.is_two_sided_cut:
            assert isinstance(self.cut_type.type_str, str), self.cut_type.type_str
            assert isinstance(self.cut_limits, (float, int)), self.cut_limits
            cut_strings = [
                f"{var} {self.cut_type.type_str} {limit}" for limit in self.float_str_conversions(self.cut_limits)
            ]
        else:
            assert isinstance(self.cut_type.type_str, tuple), self.cut_type.type_str
            assert isinstance(self.cut_limits, tuple), self.cut_limits
            cut_strings = [
                f"{item[0]} {self.cut_type.type_str[0]} {var} {self.cut_type.type_str[1]} {item[1]}"
                for item in itertools.product(
                    self.float_str_conversions(self.cut_limits[0]),
                    self.float_str_conversions(self.cut_limits[1]),
                )
            ]

        return cut_strings

    def generate_cut_str(self, make_root_compatible: bool = False) -> str:
        return self.generate_cut_strings(make_root_compatible=make_root_compatible)[0]

    @staticmethod
    def float_str_conversions(f: float) -> List[str]:
        if f % 1 == 0:
            return [str(float(f)), str(float(f)).strip("0"), str(float(f)).strip(".0")]
        else:
            return [str(float(f))]

    def scale_cut(self, percentage: float = 0.1) -> "CutInfo":
        assert 0.0 <= percentage <= 1.0, percentage
        if CutTypes(self.cut_type) is CutTypes.EQUAL_CUT:
            cut_str = ""
            cut_type = CutTypes.NO_CUT.value
            cut_limits = 0.0  # type: Union[float, Tuple[float, float]]
        elif self.cut_type.cut_direction is CutType.smaller_than_cut_direction:
            assert isinstance(self.cut_limits, float), self.cut_limits
            temp_cut_limits = self.cut_limits  # type: Union[float, Tuple[float, float]]
            scale_factor = (1.0 + percentage) if self.cut_limits > 0.0 else (1.0 - percentage)
            self.__cut_limits = scale_factor * self.cut_limits
            cut_str = self.generate_cut_str()
            cut_type = self.cut_type
            cut_limits = self.cut_limits
            self.__cut_limits = temp_cut_limits
        elif self.cut_type.cut_direction is CutType.greater_than_cut_direction:
            assert isinstance(self.cut_limits, float), self.cut_limits
            temp_cut_limits = self.cut_limits
            scale_factor = (1.0 - percentage) if self.cut_limits > 0.0 else (1.0 + percentage)
            self.__cut_limits = scale_factor * self.cut_limits
            cut_str = self.generate_cut_str()
            cut_type = self.cut_type
            cut_limits = self.cut_limits
            self.__cut_limits = temp_cut_limits
        else:
            assert isinstance(self.cut_limits, tuple), self.cut_limits
            temp_cut_limits = self.cut_limits
            left_scale_factor = (1.0 - percentage) if self.cut_limits[0] > 0.0 else (1.0 + percentage)
            right_scale_factor = (1.0 + percentage) if self.cut_limits[1] > 0.0 else (1.0 - percentage)
            self.__cut_limits = (left_scale_factor * self.cut_limits[0], right_scale_factor * self.cut_limits[1])
            cut_str = self.generate_cut_str()
            cut_type = self.cut_type
            cut_limits = self.cut_limits
            self.__cut_limits = temp_cut_limits

        return CutInfo(
            cut_str=cut_str,
            cut_label=self.cut_label,
            cut_name=self.cut_name,
            cut_var=self.cut_var,
            cut_var_latex_label=self.cut_var_latex_label,
            latex_unit=self.latex_unit,
            cut_type=cut_type.type_str,
            cut_limits=cut_limits,
            cut_abs=self.cut_abs,
            cut_online=self.cut_online,
        )

    @staticmethod
    def create_fake_cut_info_from_hist_variable(
        hist_var: HistVariable,
        invert_root_compatible_df_label: bool = False,  # Either this or the option below can be turned on!
        make_cut_name_root_compatible: bool = False,
    ) -> "CutInfo":
        if make_cut_name_root_compatible is invert_root_compatible_df_label is True:
            raise RuntimeError(
                "The arguments make_cut_name_root_compatible and invert_root_compatible_df_label "
                "should not both be set to True at the same time!"
            )

        cut_name = hist_var.df_label  # type: str
        if make_cut_name_root_compatible:
            [cut_name] = makeROOTCompatible(variables=[hist_var.df_label])

        cut_var = hist_var.df_label  # type: str
        if invert_root_compatible_df_label:
            [cut_var] = invert_make_ROOT_compatible(variables=[hist_var.df_label])

        return CutInfo(
            cut_str="",
            cut_label=hist_var.variable_name,
            cut_name=cut_name,
            cut_var=cut_var,
            cut_var_latex_label=hist_var.variable_name,
            cut_type=CutTypes.NO_CUT.value,
            cut_limits=0,
            latex_unit=hist_var.unit,
            cut_abs=False,
            cut_online=False,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CutInfo):
            raise NotImplementedError()
        if self.cut_str not in other.generate_cut_strings():
            return False
        if self.cut_label != other.cut_label:
            return False
        if self.cut_name != other.cut_name:
            return False
        if self.cut_var != other.cut_var:
            return False
        if self.cut_abs != other.cut_abs:
            return False
        if self.cut_online != other.cut_online:
            return False
        return True

    def __str__(self) -> str:
        return (
            (self.cut_name + ":\n")
            + ("Cut String: " + self.cut_str + "\n")
            + ("Cut Label: " + self.cut_label + "\n")
            + ("Cut Variable: " + self.cut_var + "\n")
            + ("Cut Type: " + str(self.cut_type) + "\n")
            + ("Cut Limits: " + str(self.cut_limits) + "\n")
            + ("Cut Abs: " + str(self.cut_abs) + "\n")
            + ("Cut Online: " + str(self.cut_online) + "\n")
        )
