"""
Providing latex strings for particle decay processes.

Felix Metzner 2021
"""

from typing import List, Dict

from rdstar.latex_labels.latex_particle_names import LatexParticleNames as LPNs
from rdstar.latex_labels.latex_string_methods import LatexObsNameMethods as LoNM


__all__ = [
    "LatexDecayProcesses",
]


class LatexDecayProcesses:
    # fmt: off

    # Some Tags
    _bkg_tag = r"$\mathrm{Bkg}$"  # type: str
    _hadronic_tag = r"$\mathrm{Hadronic}$"  # type: str

    # Decay Strings without Charges
    b_to_d_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d_st_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    b_to_d_1_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_one_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d0st_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_sszero_st_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_dprime1_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_prime_one_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_dst_2_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_two_st_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    b_to_d_tau_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.tau_str, LPNs.nu_tau_str])  # type: str
    b_to_d_st_tau_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.tau_str, LPNs.nu_tau_str])  # type: str

    b_to_d_1_tau_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_one_str, LPNs.tau_str, LPNs.nu_tau_str])  # type: str
    b_to_d0st_tau_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_sszero_st_str, LPNs.tau_str, LPNs.nu_tau_str])  # type: str
    b_to_dprime1_tau_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_prime_one_str, LPNs.tau_str, LPNs.nu_tau_str])  # type: str
    b_to_dst_2_tau_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_two_st_str, LPNs.tau_str, LPNs.nu_tau_str])  # type: str

    d_1_to_d_pi_pi = LoNM.get_decay(from_part=LPNs.d_one_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str])  # type: str
    _sub_decay_d_1_to_d_pi_pi = LoNM.get_as_sub_decay(decay=d_1_to_d_pi_pi, with_mother=True)  # type: str
    b_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_1_to_d_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    b_to_d_eta_l_nu_gap = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.eta_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d_st_eta_l_nu_gap = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.eta_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    b_to_d_pi_pi_l_nu_gap = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d_st_pi_pi_l_nu_gap = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    d_ss0_to_d_eta = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_str, LPNs.eta_str])  # type: str
    _sub_decay_d_ss0_to_d_eta = LoNM.get_as_sub_decay(decay=d_ss0_to_d_eta, with_mother=True)  # type: str
    b_to_d_eta_l_nu_gap_via_d_st_0 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_ss0_to_d_eta, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    d_p1_to_d_st_eta = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_st_str, LPNs.eta_str])  # type: str
    _sub_decay_d_p1_to_d_st_eta = LoNM.get_as_sub_decay(decay=d_p1_to_d_st_eta, with_mother=True)  # type: str
    b_to_d_st_eta_l_nu_gap_via_d_prime_1 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_p1_to_d_st_eta, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    d_p1_to_d_pi_pi = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str])  # type: str
    _sub_decay_d_p1_to_d_pi_pi = LoNM.get_as_sub_decay(decay=d_p1_to_d_pi_pi, with_mother=True)  # type: str
    d_ss0_to_d_pi_pi = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str])  # type: str
    _sub_decay_d_ss0_to_d_pi_pi = LoNM.get_as_sub_decay(decay=d_ss0_to_d_pi_pi, with_mother=True)  # type: str
    b_to_d_pi_pi_l_nu_gap_via_d_prime_1 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_p1_to_d_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d_pi_pi_l_nu_gap_via_d_st_0 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_ss0_to_d_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    d_p1_to_d_st_pi_pi = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str])  # type: str
    _sub_decay_d_p1_to_d_st_pi_pi = LoNM.get_as_sub_decay(decay=d_p1_to_d_st_pi_pi, with_mother=True)  # type: str
    d_ss0_to_d_st_pi_pi = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str])  # type: str
    _sub_decay_d_ss0_to_d_st_pi_pi = LoNM.get_as_sub_decay(decay=d_ss0_to_d_st_pi_pi, with_mother=True)  # type: str
    b_to_d_st_pi_pi_l_nu_gap_via_d_prime_1 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_p1_to_d_st_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d_st_pi_pi_l_nu_gap_via_d_st_0 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_ss0_to_d_st_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    dst_2_to_dst_pi_pi = LoNM.get_decay(from_part=LPNs.d_two_st_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str])  # type: str
    _sub_decay_dst_2_to_dst_pi_pi = LoNM.get_as_sub_decay(decay=dst_2_to_dst_pi_pi, with_mother=True)  # type: str
    b_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_NC = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dst_2_to_dst_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    b_to_d_s_kaon_l_nu_NC = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_strange_str, LPNs.k_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_d_st_s_kaon_l_nu_NC = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_strange_st_str, LPNs.k_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    b_to_dstst_l_nu = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_st_str, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    dstst_to_d_wwost_pi_zero = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])  # type: str
    dstst_to_d_wwost_pi_pm = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str])  # type: str
    dstst_to_d_wwost_pi_plus = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])  # type: str
    dstst_to_d_wwost_pi_minus = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_minus_str])  # type: str
    _sub_decay_dstst_to_d_wwost_pi_zero = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_zero, with_mother=True)  # type: str
    _sub_decay_dstst_to_d_wwost_pi_pm = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_pm, with_mother=True)  # type: str
    _sub_decay_dstst_to_d_wwost_pi_plus = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_plus, with_mother=True)  # type: str
    _sub_decay_dstst_to_d_wwost_pi_minus = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_minus, with_mother=True)  # type: str
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_plus = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_plus, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_minus = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_minus, LPNs.lepton_str, LPNs.nu_lep_str])  # type: str

    #
    # SigID and GenID Matched Labels (Only products, no charge, with Neutrinos)
    to_d_l_nu = LoNM.get_products_from_decay(process_label=b_to_d_l_nu)  # type: str
    to_d_st_l_nu = LoNM.get_products_from_decay(process_label=b_to_d_st_l_nu)  # type: str

    to_d_1_l_nu = LoNM.get_products_from_decay(process_label=b_to_d_1_l_nu)  # type: str
    to_d0st_l_nu = LoNM.get_products_from_decay(process_label=b_to_d0st_l_nu)  # type: str
    to_dprime1_l_nu = LoNM.get_products_from_decay(process_label=b_to_dprime1_l_nu)  # type: str
    to_dst_2_l_nu = LoNM.get_products_from_decay(process_label=b_to_dst_2_l_nu)  # type: str

    to_d_tau_nu = LoNM.get_products_from_decay(process_label=b_to_d_tau_nu)  # type: str
    to_d_st_tau_nu = LoNM.get_products_from_decay(process_label=b_to_d_st_tau_nu)  # type: str

    to_d_1_tau_nu = LoNM.get_products_from_decay(process_label=b_to_d_1_tau_nu)  # type: str
    to_d0st_tau_nu = LoNM.get_products_from_decay(process_label=b_to_d0st_tau_nu)  # type: str
    to_dprime1_tau_nu = LoNM.get_products_from_decay(process_label=b_to_dprime1_tau_nu)  # type: str
    to_dst_2_tau_nu = LoNM.get_products_from_decay(process_label=b_to_dst_2_tau_nu)  # type: str

    to_dstst_l_nu = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu)  # type: str

    to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero)  # type: str
    to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm)  # type: str
    to_dstst_l_nu_with_dstst_to_d_wwost_pi_plus = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_plus)  # type: str
    to_dstst_l_nu_with_dstst_to_d_wwost_pi_minus = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_minus)  # type: str

    to_hadronic_bkg_txt = _hadronic_tag[:-1] + r"\;" + _bkg_tag[1:]  # type: str
    to_b_bbar_bkg = LPNs.b_str[1:] + LoNM.get_bare_var_str(LoNM.get_anti_particle(particle_str=LPNs.b_str)) + r"\;" + _bkg_tag[1:]  # type: str
    to_continuum_txt = r"$\mathrm{Continuum}$"  # type: str

    to_hadronic_tau_bkg_txt = _hadronic_tag[:-1] + r"\;" + LoNM.get_bare_var_str(LoNM.get_anti_particle(particle_str=LPNs.tau_str)) + r"\;" + _bkg_tag[1:]  # type: str
    to_other_bkg_txt = r"$\mathrm{Other}$"  # type: str
    to_inconclusive_txt = r"$\mathrm{Inconclusive}$"  # type: str

    # General Reco Labels
    all_modes_label = r"$\mathrm{All}\,\mathrm{Reconstruction}\,\mathrm{Modes}$"  # type: str
    all_modes_label_short = r"$\mathrm{All}\,\mathrm{Reco.}\,\mathrm{Modes}$"  # type: str

    # Reconstruction Labels (with Charge, no Neutrinos)
    reco_b0_to_dm_lp = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str])  # type: str
    reco_b0_to_d_stm_lp = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str])  # type: str
    reco_b0_to_dm_ep = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.electron_plus_str])  # type: str
    reco_b0_to_d_stm_ep = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.electron_plus_str])  # type: str
    reco_b0_to_dm_mup = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.muon_plus_str])  # type: str
    reco_b0_to_d_stm_mup = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.muon_plus_str])  # type: str
    reco_b0_to_dm_taup = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str])  # type: str
    reco_b0_to_d_stm_taup = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str])  # type: str

    reco_bm_to_d0_lm = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.lepton_minus_str])  # type: str
    reco_bm_to_d_st0_lm = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.lepton_minus_str])  # type: str
    reco_bm_to_d0_em = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.electron_minus_str])  # type: str
    reco_bm_to_d_st0_em = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.electron_minus_str])  # type: str
    reco_bm_to_d0_mum = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.muon_minus_str])  # type: str
    reco_bm_to_d_st0_mum = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.muon_minus_str])  # type: str
    reco_bm_to_d0_taum = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.tau_minus_str])  # type: str
    reco_bm_to_d_st0_taum = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.tau_minus_str])  # type: str

    reco_b0_to_dm_lp_pi0, reco_b0_to_dm_pi0_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_lp_pi0, reco_b0_to_d_stm_pi0_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_dm_ep_pi0, reco_b0_to_dm_pi0_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.electron_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_ep_pi0, reco_b0_to_d_stm_pi0_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.electron_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_dm_mup_pi0, reco_b0_to_dm_pi0_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.muon_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_mup_pi0, reco_b0_to_d_stm_pi0_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.muon_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_dm_taup_pi0, reco_b0_to_dm_pi0_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_taup_pi0, reco_b0_to_d_stm_pi0_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str, LPNs.pi_zero_str])  # type: str, str

    reco_bm_to_d0_lm_pi0, reco_bm_to_d0_pi0_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.lepton_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_lm_pi0, reco_bm_to_d_st0_pi0_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.lepton_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d0_em_pi0, reco_bm_to_d0_pi0_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.electron_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_em_pi0, reco_bm_to_d_st0_pi0_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.electron_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d0_mum_pi0, reco_bm_to_d0_pi0_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.muon_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_mum_pi0, reco_bm_to_d_st0_pi0_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.muon_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d0_taum_pi0, reco_bm_to_d0_pi0_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.tau_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_taum_pi0, reco_bm_to_d_st0_pi0_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.tau_minus_str, LPNs.pi_zero_str])  # type: str, str

    reco_b0_to_anti_d0_lp_pim, reco_b0_to_anti_d0_pim_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.lepton_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_lp_pim, reco_b0_to_anti_d_st0_pim_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.lepton_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d0_ep_pim, reco_b0_to_anti_d0_pim_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.electron_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_ep_pim, reco_b0_to_anti_d_st0_pim_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.electron_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d0_mup_pim, reco_b0_to_anti_d0_pim_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.muon_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_mup_pim, reco_b0_to_anti_d_st0_pim_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.muon_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d0_taup_pim, reco_b0_to_anti_d0_pim_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.tau_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_taup_pim, reco_b0_to_anti_d_st0_pim_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.tau_plus_str, LPNs.pi_minus_str])  # type: str, str

    reco_bm_to_dp_lm_pim, reco_bm_to_dp_pim_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.lepton_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_lm_pim, reco_bm_to_d_stp_pim_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.lepton_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_dp_em_pim, reco_bm_to_dp_pim_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.electron_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_em_pim, reco_bm_to_d_stp_pim_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.electron_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_dp_mum_pim, reco_bm_to_dp_pim_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.muon_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_mum_pim, reco_bm_to_d_stp_pim_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.muon_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_dp_taum_pim, reco_bm_to_dp_pim_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.tau_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_taum_pim, reco_bm_to_d_stp_pim_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.tau_minus_str, LPNs.pi_minus_str])  # type: str, str

    #
    # Products Only Reconstruction Labels (with Charge, no Neutrinos)
    reco_to_dm_lp = LoNM.get_decay_products(to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str])  # type: str
    reco_to_d_stm_lp = LoNM.get_decay_products(to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str])  # type: str

    reco_to_dm_taup = LoNM.get_decay_products(to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str])  # type: str
    reco_to_d_stm_taup = LoNM.get_decay_products(to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str])  # type: str

    #
    # Reconstruction Labels (no Charge, no Neutrinos)
    reco_b_to_d_l = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.lepton_str])  # type: str
    reco_b_to_d_st_l = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.lepton_str])  # type: str

    reco_b_to_d_tau = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.tau_str])  # type: str
    reco_b_to_d_st_tau = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.tau_str])  # type: str

    #
    # Products Only Reconstruction Labels (no Charge, no Neutrinos)
    reco_to_d_l = LoNM.get_decay_products(to_parts=[LPNs.d_str, LPNs.lepton_str])  # type: str
    reco_to_d_st_l = LoNM.get_decay_products(to_parts=[LPNs.d_st_str, LPNs.lepton_str])  # type: str

    reco_to_d_tau = LoNM.get_decay_products(to_parts=[LPNs.d_str, LPNs.tau_str])  # type: str
    reco_to_d_st_tau = LoNM.get_decay_products(to_parts=[LPNs.d_st_str, LPNs.tau_str])  # type: str

    # Upsilon(4s) Reco Modes
    u4s_to_bp_tag_bm_sig = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_plus_str, LPNs.b_sig_minus_str])  # type: str
    u4s_to_b0_tag_anti_b0_sig = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_anti_neutral_str])  # type: str
    u4s_to_b0_tag_bm_sig = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_minus_str])  # type: str
    u4s_to_b0_tag_bp_sig = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_plus_str])  # type: str
    u4s_to_bp_tag_b0_sig = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_plus_str, LPNs.b_sig_neutral_str])  # type: str
    u4s_to_bm_tag_b0_sig = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_minus_str, LPNs.b_sig_neutral_str])  # type: str

    #
    # D(*) Reco Modes
    # D*0 Channels
    dst0_to_d0_pi0 = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.pi_zero_str])  # type: str
    dst0_to_d0_gamma = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.gamma_str])  # type: str
    # D*+ Channels
    dstp_to_d0_pip = LoNM.get_decay(from_part=LPNs.d_st_plus_str, to_parts=[LPNs.d_zero_str, LPNs.pi_plus_str])  # type: str
    dstp_to_dp_pi0 = LoNM.get_decay(from_part=LPNs.d_st_plus_str, to_parts=[LPNs.d_plus_str, LPNs.pi_zero_str])  # type: str
    # D+ Channels
    dp_to_kaon_m_pip_pip = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str])  # type: str
    dp_to_kaon_m_pip_pip_pi0 = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_zero_str])  # type: str
    dp_to_kaon_m_pip_pip_pip_pim = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str])  # type: str
    dp_to_kaon_s0_pip = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str])  # type: str
    dp_to_kaon_s0_pip_pi0 = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_zero_str])  # type: str
    dp_to_kaon_s0_pip_pip_pim = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str])  # type: str
    dp_to_kaon_s0_kaon_p = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.k_plus_str])  # type: str
    dp_to_kaon_p_kaon_m_pip = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_plus_str, LPNs.k_minus_str, LPNs.pi_plus_str])  # type: str
    # D0 Channels
    d0_to_kaon_m_pip = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str])  # type: str
    d0_to_kaon_m_pip_pi0 = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_zero_str])  # type: str
    d0_to_kaon_m_pip_pip_pim = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str])  # type: str
    d0_to_kaon_m_pip_pip_pim_pi0 = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.pi_zero_str])  # type: str
    d0_to_kaon_s0_pi0 = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.pi_zero_str])  # type: str
    d0_to_kaon_s0_pip_pim = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_minus_str])  # type: str
    d0_to_kaon_s0_pip_pim_pi0 = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.pi_zero_str])  # type: str
    d0_to_kaon_m_kaon_p = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.k_plus_str])  # type: str

    #
    # The following strings are just for test purposes:
    __test_str_1 = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.electron_str, LPNs.anti_nu_e_str])  # type: str
    # fmt: on

    @staticmethod
    def get_unfiltered_attributes_mapping() -> Dict[str, str]:
        _method_names = [
            "get_unfiltered_attributes_mapping",
            "get_attributes_mapping",
        ]  # type: List[str]

        return {
            attr_name: val for attr_name, val in LatexDecayProcesses.__dict__.items() if not (attr_name in _method_names)
        }

    @staticmethod
    def get_attributes_mapping() -> Dict[str, str]:
        return {
            attr_name: val
            for attr_name, val in LatexDecayProcesses.get_unfiltered_attributes_mapping().items()
            if not (attr_name.endswith("__") or attr_name.startswith("_"))
        }


if __name__ == "__main__":
    from collections import Counter
    from matplotlib import pyplot as plt

    print("Testing some stuff:")

    all_latex_process_keys = [
        process_name
        for process_name in LatexDecayProcesses.get_unfiltered_attributes_mapping().keys()
        if not (process_name.endswith("__"))
    ]
    all_process_labels = [LatexDecayProcesses.__dict__[process_name] for process_name in all_latex_process_keys]

    assert len(set(all_process_labels)) == len(all_process_labels), (
        len(set(all_process_labels)),
        len(all_process_labels),
        [v for v, c in Counter(all_process_labels).items() if c > 1],
    )

    test_text = ""
    for process_name in all_latex_process_keys:
        process_label = LatexDecayProcesses.__dict__[process_name]
        print("\t- " + process_name + "  ->  " + process_label)
        assert process_label.count("{") == process_label.count("}")
        assert process_label.count("$") % 2 == 0, process_label.count("$")

        test_text += f"{process_label}\n"

        fig, ax = plt.subplots(figsize=(6, 4), tight_layout=True)
        try:
            ax.annotate(text=process_label, xy=(0.5, 0.5), xytext=(0.6, 0.6))
            plt.show()
        except:  # noqa: E722  # Ignore this flake8 error for this test.
            plt.close(fig)
            fig.clf()
            raise ValueError("Error in " + process_name + "  ->  " + process_label)

        plt.close(fig)
        fig.clf()

    fig, ax = plt.subplots(figsize=(6, 4), tight_layout=True)
    ax.annotate(text=test_text, xy=(0.5, 0.5), xytext=(0.6, 0.6))
    plt.show()

    # latex_process_label_attrs = LatexDecayProcesses.get_attributes_mapping()
    # from rdstar.latex_labels.latex_output_utilities import create_latex_command_file, \
    #     get_latex_cmd_to_attributes_mapping
    # create_latex_command_file(
    #     output_path="/home/metzner/Downloads/rdstar_latex_process_labels.tex",
    #     command_mapping_dict=get_latex_cmd_to_attributes_mapping(attributes_mapping=latex_process_label_attrs),
    #     command_prefix=None,
    #     overwrite=True
    # )
