"""
Provides task flow to validate the issues found in the mc production by the mc_production_issue_finder
by running the b2bii conversion on the events and evaluating the number of events.

Felix Metzner 2021
"""

import os
import json
import basf2
import numpy as np
import pandas as pd
import b2luigi as luigi

from textops import cat, grep
from typing import List, Dict, Any, Optional, Generator

from b2luigi.basf2_helper import Basf2PathTask
from b2luigi.core.utils import get_log_file_dir as get_b2luigi_log_file_dir_for_task

from modularAnalysis import fillParticleListsFromMC

from analysistools.processing import path_utilities
from analysistools.dataframe_tools.hdf5_utils import load_root_file

from rdstar.utilities import PathType
from rdstar.online_analysis import b2bii_utils
from rdstar.mc_production.check_mc.mc_production_issue_finder import MCProductionEvaluator


__all__ = [
    "MCIssueValidatorMaster",
]


class MdstEventCounterTask(Basf2PathTask):
    mdst_file_path = luigi.Parameter(positional=False, hashed=True)  # type: PathType

    output_file_name = "mc_eval_check_n_tuple.root"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = "s"

    @staticmethod
    def get_basic_config() -> Dict[str, Any]:
        return {
            "use_b2bii": True,
            "own_local_mc_db_path": "/home/belle2/metzner/databases/RDStar_b2bii/mc_beamparams/localdb/",
            "custom_fei_db_path": None,
        }

    def create_path(self) -> basf2.Path:
        path = basf2.create_path()

        path_utilities.setup_environment_vars()
        path_utilities.setup_databases(self.get_basic_config())

        b2bii_utils.b2bii_read_and_convert(
            path=path,
            input_files=[self.mdst_file_path],
        )

        fillParticleListsFromMC([("Upsilon(4S):mc", "")], addDaughters=True, path=path)

        var_list = [
            "expNum",
            "runNum",
            "evtNum",
        ]  # type: List[str]

        path.add_module(
            "VariablesToNtuple",
            fileName=self.get_output_file_name(self.output_file_name),
            particleList="Upsilon(4S):mc",
            variables=var_list,
        )

        return path

    def output(self) -> Generator[Dict[str, luigi.LocalTarget], None, None]:
        yield self.add_to_output(self.output_file_name)


class MCIssueValidatorMaster(luigi.Task):
    info_dict_path = luigi.Parameter(positional=False, hashed=True)  # type: PathType

    n_tuple_event_count_file_name = "n_tuple_event_count_dict.json"  # type: str
    log_event_count_file_name = "basf2_log_event_count_dict.json"  # type: str
    log_initial_event_count_file_name = "basf2_log_initial_event_count_dict.json"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.log_dict = MCProductionEvaluator.read_info_from_file(input_path=self.info_dict_path)[0]
        log_file_paths = [p for d in self.log_dict.values() for p in d.keys()]  # type: List[PathType]
        self.mdst_file_paths = [self.log_path_to_mdst_path(log_path=lfp) for lfp in log_file_paths]

        self.input_dict = {
            k: list(next(dd).values())[0].path for d in self.input() if isinstance(d, dict) for k, dd in d.items()
        }
        assert set(self.input_dict.keys()) == set(self.mdst_file_paths)

    @staticmethod
    def log_path_to_mdst_path(log_path: PathType) -> PathType:
        assert os.path.exists(log_path), log_path
        path = str(log_path).replace("/gsim/log/", "/gsim/mdst/")  # type: str
        path_wo_ext, _ = os.path.splitext(path)
        mdst_path = path_wo_ext + ".mdst"
        assert os.path.isfile(mdst_path), mdst_path
        return mdst_path

    @staticmethod
    def mdst_path_to_log_path(mdst_path: PathType) -> PathType:
        path = str(mdst_path).replace("/gsim/mdst/", "/gsim/log/")  # type: str
        path_wo_ext, _ = os.path.splitext(path)
        log_path = path_wo_ext + ".log"
        return log_path

    def write_event_count_to_file(
        self,
        output_name: str,
        event_count_dict: Dict[str, int],
    ) -> None:
        with open(self.get_output_file_name(output_name), "w") as out_file:
            json.dump(event_count_dict, out_file, indent=2)

    def run(self) -> None:
        self.get_event_count_from_n_tuples()
        self.get_event_count_from_basf2_log()

    def get_event_count_from_n_tuples(self) -> None:
        events_dict = {}  # type: Dict[str, int]
        for mdst_file_path in self.mdst_file_paths:
            assert str(mdst_file_path) not in events_dict.keys(), mdst_file_path

            n_tuple_path = self.input_dict[mdst_file_path]
            assert os.path.isfile(n_tuple_path), (mdst_file_path, n_tuple_path)
            df = load_root_file(file_path=n_tuple_path, use_uproot=True)  # type: pd.DataFrame

            events_dict[str(mdst_file_path)] = int(len(np.unique(df.index)))

        self.write_event_count_to_file(event_count_dict=events_dict, output_name=self.n_tuple_event_count_file_name)

    def get_event_count_from_basf2_log(self) -> None:
        events_dict = {}  # type: Dict[str, int]
        initial_events_dict = {}  # type: Dict[str, int]

        for mdst_file_path in self.mdst_file_paths:
            assert mdst_file_path not in events_dict.keys(), mdst_file_path

            log_file_dir = get_b2luigi_log_file_dir_for_task(task=MdstEventCounterTask(mdst_file_path=mdst_file_path))
            std_out_log_path = os.path.join(log_file_dir, "stdout")

            search_strings = ["B2BIIConvertMdst", "B2BIIFixMdst"]  # type: List[str]
            search_results = [None, None]  # type: List[Optional[int]]

            for res_index, search_str in enumerate(search_strings):
                in_lines = std_out_log_path | cat() | grep(search_str).tolist()
                in_lines = [line for line in in_lines if line.startswith(search_str)]
                assert len(in_lines) == 1, in_lines
                in_line = in_lines[0]  # type: str
                n_events = in_line.split(sep="|", maxsplit=2)[1].replace(" ", "")
                assert n_events.isdigit(), n_events
                search_results[res_index] = int(n_events)

            assert len(search_results) == 2, search_results
            assert all(isinstance(res, int) for res in search_results), search_results

            event_count = search_results[0]
            initial_event_count = search_results[1]
            assert isinstance(event_count, int), event_count
            assert isinstance(initial_event_count, int), initial_event_count

            events_dict[str(mdst_file_path)] = event_count
            initial_events_dict[str(mdst_file_path)] = initial_event_count

        self.write_event_count_to_file(
            event_count_dict=events_dict,
            output_name=self.log_event_count_file_name,
        )
        self.write_event_count_to_file(
            event_count_dict=initial_events_dict,
            output_name=self.log_initial_event_count_file_name,
        )

    def output(self) -> Generator[Dict[str, luigi.LocalTarget], None, None]:
        yield self.add_to_output(self.n_tuple_event_count_file_name)
        yield self.add_to_output(self.log_event_count_file_name)
        yield self.add_to_output(self.log_initial_event_count_file_name)

    def requires(self) -> Generator[luigi.Task, None, None]:
        for mdst_file_path in self.mdst_file_paths:
            yield {
                mdst_file_path: MdstEventCounterTask(
                    mdst_file_path=mdst_file_path,
                )
            }


if __name__ == "__main__":
    log_info_dict_path = MCProductionEvaluator.get_log_file_path()  # type: PathType

    num_workers = 400
    luigi.process(
        MCIssueValidatorMaster(info_dict_path=log_info_dict_path),
        workers=num_workers,
    )
