"""
Current B meson decay branching fractions to which we reweight in this analysis.

    The Branching Fraction values for B -> D l nu and B -> D* l nu are calculated from the already precise
    measurements for the respective charged and neutral B decay BF average by assuming isospin symmetry.

    For the B -> D** l nu decays and the processes filling the Gap between exclusive and inclusive BF measurements,
    only three measurements and none, respectively, are available. Thus, we rly heavily on the assumption of
    isospin symmetry for the calculation of partial and total branching fractions for these processes.

    The BF values used for the Gap description are dependent on all other BF values considered for the MC cocktail
    and hence must be computed accordingly.

    Some background information:
    - Overview page in Belle II Confluence: https://confluence.desy.de/pages/viewpage.action?pageId=202394372
    - Additional infos can be found in
        * https://arxiv.org/pdf/1909.12524 (BFs by HFLAV)
        * https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/ (PDG 2020 Averages by HFLAV)
        * https://pdglive.lbl.gov/DataBlock.action?node=S042DT1&init=0 (PDG lifetime ratio tau_B+ / tau_B0)
        * https://arxiv.org/pdf/1202.1834
        * https://arxiv.org/abs/1402.2849
        * https://arxiv.org/abs/1611.05624
    - This package provides BF values for
        * B -> D l nu
        * B -> D* l nu
        * B -> D** l nu
        * Gap, which covers the difference between the sum of exclusive modes and the inclusive semi-leptonic BF
    - D** summarizes the heavier 1P states of the D meson, which are the broad D*_0 and D'_1 (also called D*_1)
        and the narrow states D_1 and D*_2
    - For D_1 we use f_D1 = BF(D_1 -> D*- pi+) / BF(D_1 -> D0 pi+ pi-) which can be calculated from the
        ratio values of an LHCb measurement (https://arxiv.org/pdf/1109.6831.pdf, page 23)
    - We use the ratio f_pipi = BF(D** -> D(*)- pi+ pi-) / BF(D** -> D(*) pi pi) = 1/2 +/- 1/6, where the
        uncertainty covers assumptions for pure 3-body decays as well as intermediate decays via the rho and f0 states.

    - Some older info on D** branching fraction estimates can be found in Table V in https://arxiv.org/pdf/1606.09300.pdf
    - The branching fraction (bf) of the modes involving `D_1` are smaller, because the sub modes `D_1 -> D pi pi`
      where the D_1 decays into two pions is excluded, as this decay is listed separately in the gap modes GapD1Mixed
      and GapD1Charged. Because the gap case combines the electron and muon decay, one has to consider the relation
          BF(B0 -> D_1 l nu) == 0.5 * (BF(BZD1E) + BF(BZD1Mu) + BF(GapD1Mixed))

Felix Metzner 2021
"""

import os
import numpy as np
import pandas as pd

from uncertainties import ufloat, UFloat
from typing import Optional, Tuple, List, Dict

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.latex_labels.latex_output_utilities import escape_special_chars
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses as LDPs
from rdstar.mc_reweighting.BFReweighting.bf_utilities import (
    BFMeasurement,
    IsoSpinVarianceWeightedAvgCalculator,
    SimpleIsoSpinVarianceWeightedAvgCalculator,
    assume_100_percent_uncertainty,
    split_up_by_bfs,
    info_str,
    create_latex_value_file,
)


__all__ = [
    # B -> D l nu
    "bf_b0_to_d_l_nu_avg",
    "bf_bp_to_d_l_nu_avg",
    # B -> D tau nu
    "bf_b0_to_d_tau_nu_avg",
    "bf_bp_to_d_tau_nu_avg",
    # B -> D* l nu
    "bf_b0_to_d_st_l_nu_avg",
    "bf_bp_to_d_st_l_nu_avg",
    # B -> D* tau nu
    "bf_b0_to_d_st_tau_nu_avg",
    "bf_bp_to_d_st_tau_nu_avg",
    # B -> D** l nu
    "bf_b0_to_d0st_l_nu_our_cal",
    "bf_bp_to_d0st_l_nu_our_cal",
    "bf_b0_to_dprime1_l_nu_our_cal",
    "bf_bp_to_dprime1_l_nu_our_cal",
    "bf_b0_to_dst_2_l_nu_our_cal",
    "bf_bp_to_dst_2_l_nu_our_cal",
    "bf_b0_to_d_1_l_nu_our_cal",
    "bf_bp_to_d_1_l_nu_our_cal",
    # B -> D** tau nu
    "bf_b0_to_d0st_tau_nu_from_belle2_dec",
    "bf_bp_to_d0st_tau_nu_from_belle2_dec",
    "bf_b0_to_dprime1_tau_nu_from_belle2_dec",
    "bf_bp_to_dprime1_tau_nu_from_belle2_dec",
    "bf_b0_to_dst_2_tau_nu_from_belle2_dec",
    "bf_bp_to_dst_2_tau_nu_from_belle2_dec",
    "bf_b0_to_d_1_tau_nu_from_belle2_dec",
    "bf_bp_to_d_1_tau_nu_from_belle2_dec",
    # Gap
    "bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap",
    "bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap",
    "bf_b0_to_d_eta_l_nu_gap",
    "bf_bp_to_d_eta_l_nu_gap",
    "bf_b0_to_d_st_eta_l_nu_gap",
    "bf_bp_to_d_st_eta_l_nu_gap",
    "bf_b0_to_d_pi_pi_l_nu_our_gap",
    "bf_bp_to_d_pi_pi_l_nu_our_gap",
    "bf_b0_to_d_st_pi_pi_l_nu_our_gap",
    "bf_bp_to_d_st_pi_pi_l_nu_our_gap",
    # Alternative Gap
    "bf_b0_to_d_eta_l_nu_gap_via_d_st_0",
    "bf_bp_to_d_eta_l_nu_gap_via_d_st_0",
    "bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1",
    "bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1",
    "bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0",
    "bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1",
    "bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0",
    "bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1",
    "bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0",
    "bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1",
    "bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0",
    "bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1",
    #
    # Utilities / Other stuff
    "overview_table",
    "latex_overview_table",
    "write_bf_values_to_latex",
    "creat_latex_variable_table",
]

# fmt: off
# We need to assign values to what we call:
#  B0:
#   - DSME = bf_b0_to_d_st_l_nu_avg
#   - DME = bf_b0_to_d_l_nu_avg
#   - BZD1E = bf_b0_to_d_1_l_nu_our_cal
#   - BZD0sE = bf_b0_to_d0st_l_nu_our_cal
#   - BZDp1E = bf_b0_to_dprime1_l_nu_our_cal
#   - BZD2sE = bf_b0_to_dst_2_l_nu_our_cal
#   - DSMMu = bf_b0_to_d_st_l_nu_avg
#   - DMMu = bf_b0_to_d_l_nu_avg
#   - BZD1Mu = bf_b0_to_d_1_l_nu_our_cal
#   - BZD0sMu = bf_b0_to_d0st_l_nu_our_cal
#   - BZDp1Mu = bf_b0_to_dprime1_l_nu_our_cal
#   - BZD2sMu = bf_b0_to_dst_2_l_nu_our_cal
#   - DSMTau = bf_b0_to_d_st_tau_nu_avg
#   - DMTau = bf_b0_to_d_tau_nu_avg
#   - TBZD1Tau = bf_b0_to_d_1_tau_nu_from_belle2_dec
#   - TBZD0sTau = bf_b0_to_d0st_tau_nu_from_belle2_dec
#   - TBZDp1Tau = bf_b0_to_dprime1_tau_nu_from_belle2_dec
#   - TBZD2sTau = bf_b0_to_dst_2_tau_nu_from_belle2_dec
#   - GapD1Mixed = 2 * bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDEtaMixed = 2 * bf_b0_to_d_eta_l_nu_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstEtaMixed = 2 * bf_b0_to_d_st_eta_l_nu_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDPiPiMixed = 2 * bf_b0_to_d_pi_pi_l_nu_our_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstPiPiMixed = 2 * bf_b0_to_d_st_pi_pi_l_nu_our_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDEtaViaD0sMixed = 2 * bf_b0_to_d_eta_l_nu_gap_via_d_st_0 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstEtaViaDp1Mixed = 2 * bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDPiPiViaDp1Mixed = 2 * bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDPiPiViaD0sMixed = 2 * bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstPiPiViaDp1Mixed = 2 * bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstPiPiViaD0sMixed = 2 * bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0 (Factor 2 because of l = e, mu in Gap MC)
#  B+:
#   - DSZE = bf_bp_to_d_st_l_nu_avg
#   - DZE = bf_bp_to_d_l_nu_avg
#   - BMD1E = bf_bp_to_d_1_l_nu_our_cal
#   - BMD0sE = bf_bp_to_d0st_l_nu_our_cal
#   - BMDp1E = bf_bp_to_dprime1_l_nu_our_cal
#   - BMD2sE = bf_bp_to_dst_2_l_nu_our_cal
#   - DSZMu = bf_bp_to_d_st_l_nu_avg
#   - DZMu = bf_bp_to_d_l_nu_avg
#   - BMD1Mu = bf_bp_to_d_1_l_nu_our_cal
#   - BMD0sMu = bf_bp_to_d0st_l_nu_our_cal
#   - BMDp1Mu = bf_bp_to_dprime1_l_nu_our_cal
#   - BMD2sMu = bf_bp_to_dst_2_l_nu_our_cal
#   - DSZTau = bf_bp_to_d_st_tau_nu_avg
#   - DZTau = bf_bp_to_d_tau_nu_avg
#   - TBMD1Tau = bf_bp_to_d_1_tau_nu_from_belle2_dec
#   - TBMD0sTau = bf_bp_to_d0st_tau_nu_from_belle2_dec
#   - TBMDp1Tau = bf_bp_to_dprime1_tau_nu_from_belle2_dec
#   - TBMD2sTau = bf_bp_to_dst_2_tau_nu_from_belle2_dec
#   - GapD1Charged = 2 * bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDEtaCharged = 2 * bf_bp_to_d_eta_l_nu_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstEtaCharged = 2 * bf_bp_to_d_st_eta_l_nu_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDPiPiCharged = 2 * bf_bp_to_d_pi_pi_l_nu_our_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstPiPiCharged = 2 * bf_bp_to_d_st_pi_pi_l_nu_our_gap (Factor 2 because of l = e, mu in Gap MC)
#   - GapDEtaViaD0sCharged = 2 * bf_bp_to_d_eta_l_nu_gap_via_d_st_0 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstEtaViaDp1Charged = 2 * bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDPiPiViaDp1Charged = 2 * bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDPiPiViaD0sCharged = 2 * bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstPiPiViaDp1Charged = 2 * bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1 (Factor 2 because of l = e, mu in Gap MC)
#   - GapDstPiPiViaD0sCharged = 2 * bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0 (Factor 2 because of l = e, mu in Gap MC)
# fmt: on

# Constants
# fmt: off
tau_p0 = ufloat(1.076, 0.004)  # type: UFloat  # lifetime ratio tau_B+ / tau_B0 (https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/)
f_pi = 2.0 / 3.0  # type: float  # f_pi = BF(D** -> D(*)- pi+) / BF(D** -> D(*) pi) = 2/3
two_body_iso_spin_ratio = 1.0 / f_pi  # type: float  # = 1 / f_pi
f_d1 = ufloat(2.32, 0.54)  # type: UFloat  # f_D1 = BF(D_1 -> D*- pi+) / BF(D_1 -> D0 pi+ pi-)
f_pipi = ufloat(1.0 / 2.0, 1.0 / 6.0)  # type: UFloat  # BF(D** -> D(*)- pi+ pi-) / BF(D** -> D(*) pi pi)
three_body_iso_spin_ratio = 1.0 / f_pipi  # type: float  # = 1 / f_pipi

# Theory R(D) and R(D*)
sm_r_d = ufloat(nominal_value=0.299, std_dev=0.003)  # type: UFloat  # R(D) from HFLAV Eq. 224: https://arxiv.org/pdf/1909.12524
sm_r_d_st = ufloat(nominal_value=0.258, std_dev=0.005)  # type: UFloat  # R(D*) from HFLAV Eq. 225: https://arxiv.org/pdf/1909.12524

# Additional required measurement for D*_2 meson case (see https://confluence.desy.de/pages/viewpage.action?pageId=202394372 for details):
ratio_f_dst_2 = ufloat(1.54, 0.15)  # type: UFloat  # ratio f_D*_2 = BF(D*_2 - > D- pi+) / BF(D*_2 - > D*- pi+)
ratio_f_dst_2_to_D_pi_pi = ufloat(2.79, 1.47)  # type: UFloat  # ratio f_D*_2_to_Dpipi = BF(D*_2 - > D*- pi+) / BF(D*_2 - > D0 pi+ pi-)

# Unused constants
ratio_fpm_over_f00 = ufloat(nominal_value=1.058, std_dev=0.024)  # type: UFloat  # f+- / f00 from https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/
ratio_fpm = ufloat(nominal_value=0.514, std_dev=0.006)  # type: UFloat  # f+- from https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/
ratio_f00 = ufloat(nominal_value=0.486, std_dev=0.006)  # type: UFloat  # f00 from https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/
# fmt: on

# fmt: off
# Measured BF values (taken from the PDG Live Database on 4th May 2021)
bf_b0_to_d_l_nu_pdg_eval = BFMeasurement(central_value=2.31e-2, stat_uncert=0.04e-2, sys_uncert=0.09e-2)  # B0 -> D+ l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=37&parCode=S042&home=MXXX045
bf_bp_to_d_l_nu_pdg_eval = BFMeasurement(central_value=2.35e-2, stat_uncert=0.03e-2, sys_uncert=0.09e-2)  # B- -> D0 l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=145&parCode=S041&home=MXXX045
# bf_b0_to_d_st_l_nu_pdg_eval = BFMeasurement(central_value=5.06e-2, stat_uncert=0.02e-2, sys_uncert=0.12e-2)  # B0 -> D*+ l- nu as used by Florian and Stephan
bf_b0_to_d_st_l_nu_pdg_eval = BFMeasurement(central_value=5.05e-2, stat_uncert=0.02e-2, sys_uncert=0.14e-2)  # B0 -> D*+ l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=181&parCode=S042&home=MXXX045
bf_bp_to_d_st_l_nu_pdg_eval = BFMeasurement(central_value=5.66e-2, stat_uncert=0.07e-2, sys_uncert=0.21e-2)  # B- -> D*0 l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=146&parCode=S041&home=MXXX045

bf_bp_to_d0_0_st_l_nu_with_d0_0_st_to_dp_pim = BFMeasurement(central_value=0.28e-2, stat_uncert=0.03e-2, sys_uncert=0.04e-2)  # B- -> D*_0^0 (-> D+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 77
bf_bp_to_dprime10_l_nu_with_dprime10_to_dstp_pim = BFMeasurement(central_value=0.28e-2, stat_uncert=0.06e-2, sys_uncert=0.0)  # B- -> D'1^0 (-> D*+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 76 w/o Belle measurement!
bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip = BFMeasurement(central_value=0.077e-2, stat_uncert=0.006e-2, sys_uncert=0.004e-2)  # B- -> D*_2^0 (-> D*+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 75
bf_bp_to_d0_1_l_nu_with_d0_1_to_d_stp_pim = BFMeasurement(central_value=0.281e-2, stat_uncert=0.01e-2, sys_uncert=0.015e-2)  # B- -> D0_1 (-> D*+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 74
# fmt: on

# Isospin variance weighted average calculation assuming 0 correlation between measurements
# B -> D l nu
b_to_d_l_nu_calculator = IsoSpinVarianceWeightedAvgCalculator(
    neutral_b_bf=bf_b0_to_d_l_nu_pdg_eval,
    charged_b_bf=bf_bp_to_d_l_nu_pdg_eval,
    lifetime_ratio_tau_p0=tau_p0,
    correlation_rho_sys=0.0,
)
bf_b0_to_d_l_nu_avg = b_to_d_l_nu_calculator.average_neutral_b_bf  # type: UFloat
bf_bp_to_d_l_nu_avg = b_to_d_l_nu_calculator.average_charged_b_bf  # type: UFloat

bf_b0_to_d_tau_nu_avg = bf_b0_to_d_l_nu_avg * sm_r_d  # type: UFloat
bf_bp_to_d_tau_nu_avg = bf_bp_to_d_l_nu_avg * sm_r_d  # type: UFloat

# B -> D* l nu
b_to_d_st_l_nu_calculator = IsoSpinVarianceWeightedAvgCalculator(
    neutral_b_bf=bf_b0_to_d_st_l_nu_pdg_eval,
    charged_b_bf=bf_bp_to_d_st_l_nu_pdg_eval,
    lifetime_ratio_tau_p0=tau_p0,
    correlation_rho_sys=0.0,
)
bf_b0_to_d_st_l_nu_avg = b_to_d_st_l_nu_calculator.average_neutral_b_bf  # type: UFloat
bf_bp_to_d_st_l_nu_avg = b_to_d_st_l_nu_calculator.average_charged_b_bf  # type: UFloat

bf_b0_to_d_st_tau_nu_avg = bf_b0_to_d_st_l_nu_avg * sm_r_d_st  # type: UFloat
bf_bp_to_d_st_tau_nu_avg = bf_bp_to_d_st_l_nu_avg * sm_r_d_st  # type: UFloat

# B -> D** l nu
# B -> D*_0 l nu
_bf_bp_to_d0st_l_nu = bf_bp_to_d0_0_st_l_nu_with_d0_0_st_to_dp_pim.ufloat * two_body_iso_spin_ratio  # type: UFloat
_bf_b0_to_d0st_l_nu = _bf_bp_to_d0st_l_nu / tau_p0  # type: UFloat

bf_b0_to_d0st_l_nu_our_cal = _bf_b0_to_d0st_l_nu  # type: UFloat
bf_bp_to_d0st_l_nu_our_cal = _bf_bp_to_d0st_l_nu  # type: UFloat

# B -> D'_1 l nu   (D'_1 == D*_1)
_bf_bp_to_dprime1_l_nu = bf_bp_to_dprime10_l_nu_with_dprime10_to_dstp_pim.ufloat * two_body_iso_spin_ratio  # type: UFloat
_bf_b0_to_dprime1_l_nu = _bf_bp_to_dprime1_l_nu / tau_p0  # type: UFloat

bf_b0_to_dprime1_l_nu_our_cal = _bf_b0_to_dprime1_l_nu  # type: UFloat
bf_bp_to_dprime1_l_nu_our_cal = _bf_bp_to_dprime1_l_nu  # type: UFloat

# B -> D*_2 l nu
# - B -> D*_2 (-> Dpipi) l nu   =>  This is component is not considered (simulated) in the Belle I or Belle II MC!
_bf_bp_dst_2_l_nu_with_dst_2_to_dst_pi_pi = (
    bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip.ufloat * three_body_iso_spin_ratio / ratio_f_dst_2_to_D_pi_pi
)  # type: UFloat
bf_b0_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered = _bf_bp_dst_2_l_nu_with_dst_2_to_dst_pi_pi / tau_p0
bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered = _bf_bp_dst_2_l_nu_with_dst_2_to_dst_pi_pi

# - B -> D*_2 (-> D(*)pi) l nu
_bf_bp_to_dst_2_l_nu_with_dst_2_to_d_pi = (
    bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip.ufloat * ratio_f_dst_2 * two_body_iso_spin_ratio
)  # type: UFloat
_bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi = (
    bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip.ufloat * two_body_iso_spin_ratio
)  # type: UFloat

_bf_bp_to_dst_2_l_nu_with_dst_2_to_Dpi_or_Dstpi = (
    bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip.ufloat * (1.0 + ratio_f_dst_2) * two_body_iso_spin_ratio
)  # type: UFloat
_bf_b0_to_dst_2_l_nu_with_dst_2_to_Dpi_or_Dstpi = _bf_bp_to_dst_2_l_nu_with_dst_2_to_Dpi_or_Dstpi / tau_p0  # type: UFloat

bf_b0_to_dst_2_l_nu_our_cal = (
    _bf_b0_to_dst_2_l_nu_with_dst_2_to_Dpi_or_Dstpi
)  # type: UFloat  # Dst_2 -> D pi pi not considered!
bf_bp_to_dst_2_l_nu_our_cal = (
    _bf_bp_to_dst_2_l_nu_with_dst_2_to_Dpi_or_Dstpi
)  # type: UFloat  # Dst_2 -> D pi pi not considered!

# B -> D_1 l nu
# D1 -> D*pi (Part for our D** Analysis)
_bf_bp_to_d0_1_l_nu_with_d0_1_to_dstp_pim = bf_bp_to_d0_1_l_nu_with_d0_1_to_d_stp_pim.ufloat  # type: UFloat
_bf_bp_to_d_1_l_nu_with_d_1_to_dst_pi = (
    _bf_bp_to_d0_1_l_nu_with_d0_1_to_dstp_pim * two_body_iso_spin_ratio
)  # type: UFloat

bf_b0_to_d_1_l_nu_our_cal = _bf_bp_to_d_1_l_nu_with_d_1_to_dst_pi / tau_p0  # type: UFloat
bf_bp_to_d_1_l_nu_our_cal = _bf_bp_to_d_1_l_nu_with_d_1_to_dst_pi  # type: UFloat

# D1 -> Dpipi (Gap part)
_bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi = (
    _bf_bp_to_d0_1_l_nu_with_d0_1_to_dstp_pim * three_body_iso_spin_ratio / f_d1
)  # type: UFloat

bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap = _bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi / tau_p0  # type: UFloat
bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap = _bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi  # type: UFloat

# Gap

# Measurements to B -> D(*) pi l nu (resonant and non-resonant)
# fmt: off
bf_b0_to_d_pi_l_nu_hfalv_eval = BFMeasurement(central_value=0.406e-2, stat_uncert=0.047e-2, sys_uncert=0.0)  # B0 -> D0 pi+ l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_bp_to_d_pi_l_nu_hflav_eval = BFMeasurement(central_value=0.443e-2, stat_uncert=0.037e-2, sys_uncert=0.0)  # B- -> D+ pi- l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_b0_to_d_st_pi_l_nu_hflav_eval = BFMeasurement(central_value=0.565e-2, stat_uncert=0.061e-2, sys_uncert=0.0)  # B0 -> D*0 pi+ l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_bp_to_d_st_pi_l_nu_hflav_eval = BFMeasurement(central_value=0.589e-2, stat_uncert=0.044e-2, sys_uncert=0.0)  # B- -> D*+ pi- l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
# fmt: on

# Overall B -> D pi l nu (resonant and non-resonant)
b_to_d_pi_l_nu_calculator = SimpleIsoSpinVarianceWeightedAvgCalculator(
    neutral_b_bf=bf_b0_to_d_pi_l_nu_hfalv_eval,
    charged_b_bf=bf_bp_to_d_pi_l_nu_hflav_eval,
    two_body_iso_spin_ratio=two_body_iso_spin_ratio,
    lifetime_ratio_tau_p0=tau_p0,
)
_bf_b0_to_d_pi_l_nu_avg = b_to_d_pi_l_nu_calculator.average_neutral_b_bf  # type: UFloat
_bf_bp_to_d_pi_l_nu_avg = b_to_d_pi_l_nu_calculator.average_charged_b_bf  # type: UFloat

# Overall B -> D* pi l nu (resonant and non-resonant)
b_to_d_st_pi_l_nu_calculator = SimpleIsoSpinVarianceWeightedAvgCalculator(
    neutral_b_bf=bf_b0_to_d_st_pi_l_nu_hflav_eval,
    charged_b_bf=bf_bp_to_d_st_pi_l_nu_hflav_eval,
    two_body_iso_spin_ratio=two_body_iso_spin_ratio,
    lifetime_ratio_tau_p0=tau_p0,
)
_bf_b0_to_d_st_pi_l_nu_avg = b_to_d_st_pi_l_nu_calculator.average_neutral_b_bf  # type: UFloat
_bf_bp_to_d_st_pi_l_nu_avg = b_to_d_st_pi_l_nu_calculator.average_charged_b_bf  # type: UFloat

# Non-resonant B -> D pi l nu -> Can be ignored:
_bf_bp_to_d_pi_l_nu_non_res = (
    _bf_bp_to_d_pi_l_nu_avg - bf_bp_to_d0st_l_nu_our_cal - _bf_bp_to_dst_2_l_nu_with_dst_2_to_d_pi
)  # type: UFloat

bf_b0_to_d_pi_l_nu_our_non_res = _bf_bp_to_d_pi_l_nu_non_res / tau_p0  # type: UFloat
bf_bp_to_d_pi_l_nu_our_non_res = _bf_bp_to_d_pi_l_nu_non_res  # type: UFloat
# This space left to contributions of non-resonant processes is so small that we can ignore them!

# Non-resonant B -> D* pi l nu -> Can be ignored:
_bf_bp_to_d_st_pi_l_nu_non_res = (
    _bf_bp_to_d_st_pi_l_nu_avg
    - bf_bp_to_dprime1_l_nu_our_cal
    - _bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi
    - _bf_bp_to_d_1_l_nu_with_d_1_to_dst_pi
)  # type: UFloat

bf_b0_to_d_st_pi_l_nu_our_non_res = _bf_bp_to_d_st_pi_l_nu_non_res / tau_p0  # type: UFloat
bf_bp_to_d_st_pi_l_nu_our_non_res = _bf_bp_to_d_st_pi_l_nu_non_res  # type: UFloat
# This space left to contributions of non-resonant processes is so small that we can ignore them!

# Overall B -> D(*) pi pi l nu (resonant and non-resonant) from non-resonant BaBar measurement
# fmt: off
ratio_Dpipi_to_D_babar = BFMeasurement(central_value=0.067, stat_uncert=0.010, sys_uncert=0.008)  # R_pipi from https://arxiv.org/abs/1507.08303 Table 2
ratio_Dstpipi_to_Dst_babar = BFMeasurement(central_value=0.019, stat_uncert=0.005, sys_uncert=0.004)  # R*_pipi from https://arxiv.org/abs/1507.08303 Table 2

bf_b_to_d_pi_pi_l_nu_babar_non_res = BFMeasurement(central_value=152e-5, stat_uncert=23e-5, sys_uncert=np.sqrt(18e-5**2 + 7e-5**2))  # B -> D pi+ pi- l- nu from https://arxiv.org/abs/1507.08303 Table 2
bf_b_to_d_st_pi_pi_l_nu_babar_non_res = BFMeasurement(central_value=108e-5, stat_uncert=28e-5, sys_uncert=np.sqrt(23e-5**2 + 4e-5**2))  # B -> D* pi+ pi- l- nu from https://arxiv.org/abs/1507.08303 Table 2
# fmt: on

_bf_bp_to_d_pi_pi_l_nu_all = bf_b_to_d_pi_pi_l_nu_babar_non_res.ufloat * three_body_iso_spin_ratio  # type: UFloat
_bf_bp_to_d_st_pi_pi_l_nu_all = bf_b_to_d_st_pi_pi_l_nu_babar_non_res.ufloat * three_body_iso_spin_ratio  # type: UFloat

# Remaining B -> D(*) pi pi l nu (from above)
_bf_bp_to_d_pi_pi_l_nu_remaining = _bf_bp_to_d_pi_pi_l_nu_all - _bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi  # type: UFloat

bf_b0_to_d_pi_pi_l_nu_our_gap = _bf_bp_to_d_pi_pi_l_nu_remaining / tau_p0  # type: UFloat
bf_bp_to_d_pi_pi_l_nu_our_gap = _bf_bp_to_d_pi_pi_l_nu_remaining  # type: UFloat

# No contributions from P1 states to subtract here:
_bf_bp_to_d_st_pi_pi_l_nu_remaining = _bf_bp_to_d_st_pi_pi_l_nu_all  # type: UFloat

bf_b0_to_d_st_pi_pi_l_nu_our_gap = _bf_bp_to_d_st_pi_pi_l_nu_remaining / tau_p0  # type: UFloat
bf_bp_to_d_st_pi_pi_l_nu_our_gap = _bf_bp_to_d_st_pi_pi_l_nu_remaining  # type: UFloat

# Other modes: B+ -> D_s(*) K l+ nu
# These modes are here not available in the Belle I MC and furthermore don't have large contributions (BF ~ 0.03e-2)
# Thus, they are not considered further.
# fmt: off
bf_b_to_d_s_kaon_l_nu_measurement = BFMeasurement(central_value=3.0e-4, stat_uncert=0.9e-4, sys_uncert=0.5e-4 * (1.1 + 0.8))  # B+ -> D_s- K+ l+ nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=600&parCode=S041&home=MXXX045
bf_b_to_d_st_s_kaon_l_nu_measurement = BFMeasurement(central_value=2.9e-4, stat_uncert=1.6e-4, sys_uncert=0.5e-4 * (1.1 + 1.0))  # B+ -> D*_s- K+ l+ nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=571&parCode=S041&home=MXXX045
# fmt: on

bf_b0_to_d_s_kaon_l_nu_pdg = bf_b_to_d_s_kaon_l_nu_measurement.ufloat / tau_p0  # type: UFloat
bf_bp_to_d_s_kaon_l_nu_pdg = bf_b_to_d_s_kaon_l_nu_measurement.ufloat  # type: UFloat
bf_b0_to_d_st_s_kaon_l_nu_pdg = bf_b_to_d_st_s_kaon_l_nu_measurement.ufloat / tau_p0  # type: UFloat
bf_bp_to_d_st_s_kaon_l_nu_pdg = bf_b_to_d_st_s_kaon_l_nu_measurement.ufloat  # type: UFloat

# Sum exclusive:
sum_bp_to_xc_l_nu_exclusive = (
    bf_bp_to_d_l_nu_avg
    + bf_bp_to_d_st_l_nu_avg
    + bf_bp_to_d0st_l_nu_our_cal
    + bf_bp_to_dprime1_l_nu_our_cal
    + bf_bp_to_dst_2_l_nu_our_cal
    + bf_bp_to_d_1_l_nu_our_cal
)  # type: UFloat
extended_sum_bp_to_xc_l_nu_exclusive = (
    sum_bp_to_xc_l_nu_exclusive
    + bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap  # This is part of the B -> D_1 l nu BF, but in the D1 Gap MC in our case
    + bf_bp_to_d_pi_pi_l_nu_our_gap  # We consider this already in our B -> D pi pi l nu Gap Component
    + bf_bp_to_d_st_pi_pi_l_nu_our_gap  # We consider this already in our B -> D* pi pi l nu Gap Component
)  # type: UFloat
# Sum_Bp_Xc_excl = BF_Bp_D1_Dspi_Dpipi + BF_Bp_D2_Dpi_Dspi + BF_Bp_D0 + BF_Bp_D1p

# Inclusive:
# fmt: off
bf_b_to_xc_l_nu_hflav_combined = BFMeasurement(central_value=10.65e-2, stat_uncert=0.16e-2, sys_uncert=0.0)  # Inclusive B -> X_c l nu from Eq. 205 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_b0_to_xc_l_nu_inclusive = BFMeasurement(central_value=10.08e-2, stat_uncert=0.30e-2, sys_uncert=0.22e-2)  # Inclusive B0 -> X_c l nu from PDG https://pdglive.lbl.gov/BranchingRatio.action?desig=375&parCode=S042&home=MXXX045
bf_bp_to_xc_l_nu_inclusive = BFMeasurement(central_value=10.79e-2, stat_uncert=0.25e-2, sys_uncert=0.27e-2)  # Inclusive B+ -> X_c l nu from PDG https://pdglive.lbl.gov/BranchingRatio.action?desig=473&parCode=S041&home=MXXX045
# fmt: on

gap_b0 = bf_b0_to_xc_l_nu_inclusive.ufloat - extended_sum_bp_to_xc_l_nu_exclusive / tau_p0  # type: UFloat
gap_bp = bf_bp_to_xc_l_nu_inclusive.ufloat - extended_sum_bp_to_xc_l_nu_exclusive  # type: UFloat

# Assuming remaining Gap to be B -> D(*) eta l nu (assuming a 50/50 split for D and D* and 100% uncertainty)
# B -> D eta l nu
_bf_b0_to_d_eta_l_nu = 0.5 * gap_b0  # type: UFloat
_bf_bp_to_d_eta_l_nu = 0.5 * gap_bp  # type: UFloat
bf_b0_to_d_eta_l_nu_gap = assume_100_percent_uncertainty(bf_value=_bf_b0_to_d_eta_l_nu)  # type: UFloat
bf_bp_to_d_eta_l_nu_gap = assume_100_percent_uncertainty(bf_value=_bf_bp_to_d_eta_l_nu)  # type: UFloat

# B -> D* eta l nu
_bf_b0_to_d_st_eta_l_nu = 0.5 * gap_b0  # type: UFloat
_bf_bp_to_d_st_eta_l_nu = 0.5 * gap_bp  # type: UFloat
bf_b0_to_d_st_eta_l_nu_gap = assume_100_percent_uncertainty(bf_value=_bf_b0_to_d_st_eta_l_nu)  # type: UFloat
bf_bp_to_d_st_eta_l_nu_gap = assume_100_percent_uncertainty(bf_value=_bf_bp_to_d_st_eta_l_nu)  # type: UFloat


# Alternative Gap Description (via D** resonances)
# B -> D eta l nu via D** variant (D eta via D*_0 and D* eta via D'_1)
_b0_d_eta_via_dss_bfs = split_up_by_bfs(
    bf_value=gap_b0,
    base_bfs=(bf_b0_to_d0st_l_nu_our_cal, bf_b0_to_dprime1_l_nu_our_cal),
)  # type: Tuple[UFloat, ...]
bf_b0_to_d_eta_l_nu_gap_via_d_st_0 = assume_100_percent_uncertainty(bf_value=_b0_d_eta_via_dss_bfs[0])  # type: UFloat
bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1 = assume_100_percent_uncertainty(bf_value=_b0_d_eta_via_dss_bfs[1])

_bp_d_eta_via_dss_bfs = split_up_by_bfs(
    bf_value=gap_bp,
    base_bfs=(bf_bp_to_d0st_l_nu_our_cal, bf_bp_to_dprime1_l_nu_our_cal),
)  # type: Tuple[UFloat, ...]
bf_bp_to_d_eta_l_nu_gap_via_d_st_0 = assume_100_percent_uncertainty(bf_value=_bp_d_eta_via_dss_bfs[0])  # type: UFloat
bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1 = assume_100_percent_uncertainty(bf_value=_bp_d_eta_via_dss_bfs[1])

# B -> D pi pi l nu via D** variant (D eta via D*_0 or D'_1)
_b0_d_pi_pi_via_dss_bfs = split_up_by_bfs(
    bf_value=bf_b0_to_d_pi_pi_l_nu_our_gap,
    base_bfs=(bf_b0_to_d0st_l_nu_our_cal, bf_b0_to_dprime1_l_nu_our_cal),
)  # type: Tuple[UFloat, ...]
bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0 = assume_100_percent_uncertainty(bf_value=_b0_d_pi_pi_via_dss_bfs[0])
bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1 = assume_100_percent_uncertainty(bf_value=_b0_d_pi_pi_via_dss_bfs[1])

_bp_d_pi_pi_via_dss_bfs = split_up_by_bfs(
    bf_value=bf_bp_to_d_pi_pi_l_nu_our_gap,
    base_bfs=(bf_bp_to_d0st_l_nu_our_cal, bf_bp_to_dprime1_l_nu_our_cal),
)  # type: Tuple[UFloat, ...]
bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0 = assume_100_percent_uncertainty(bf_value=_bp_d_pi_pi_via_dss_bfs[0])
bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1 = assume_100_percent_uncertainty(bf_value=_bp_d_pi_pi_via_dss_bfs[1])

# B -> D* pi pi l nu via D** variant (D eta via D*_0 or D'_1)
_b0_d_st_pi_pi_via_dss_bfs = split_up_by_bfs(
    bf_value=bf_b0_to_d_st_pi_pi_l_nu_our_gap,
    base_bfs=(bf_b0_to_d0st_l_nu_our_cal, bf_b0_to_dprime1_l_nu_our_cal),
)  # type: Tuple[UFloat, ...]
bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0 = assume_100_percent_uncertainty(bf_value=_b0_d_st_pi_pi_via_dss_bfs[0])
bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1 = assume_100_percent_uncertainty(bf_value=_b0_d_st_pi_pi_via_dss_bfs[1])

_bp_d_st_pi_pi_via_dss_bfs = split_up_by_bfs(
    bf_value=bf_bp_to_d_st_pi_pi_l_nu_our_gap,
    base_bfs=(bf_bp_to_d0st_l_nu_our_cal, bf_bp_to_dprime1_l_nu_our_cal),
)  # type: Tuple[UFloat, ...]
bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0 = assume_100_percent_uncertainty(bf_value=_bp_d_st_pi_pi_via_dss_bfs[0])
bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1 = assume_100_percent_uncertainty(bf_value=_bp_d_st_pi_pi_via_dss_bfs[1])

# B -> D** tau nu Modes  (Values taken from Belle II MC file)
# Assuming 100% uncertainty
bf_b0_to_d0st_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0013, std_dev=0.0013)  # type: UFloat
bf_bp_to_d0st_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0013, std_dev=0.0013)  # type: UFloat
bf_b0_to_dprime1_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0020, std_dev=0.0020)  # type: UFloat
bf_bp_to_dprime1_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0020, std_dev=0.0020)  # type: UFloat
bf_b0_to_dst_2_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0020, std_dev=0.0020)  # type: UFloat
bf_bp_to_dst_2_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0020, std_dev=0.0020)  # type: UFloat
bf_b0_to_d_1_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0013, std_dev=0.0013)  # type: UFloat
bf_bp_to_d_1_tau_nu_from_belle2_dec = ufloat(nominal_value=0.0013, std_dev=0.0013)  # type: UFloat

# fmt: off
overview_table = pd.DataFrame({
    "process_label": ["d_st_l_nu_avg", "d_l_nu_avg", "d_1_l_nu_our_cal", "d0st_l_nu_our_cal", "dprime1_l_nu_our_cal", "dst_2_l_nu_our_cal", "d_st_tau_nu_avg", "d_tau_nu_avg", "d_1_tau_nu_from_belle2_dec", "d0st_tau_nu_from_belle2_dec", "dprime1_tau_nu_from_belle2_dec", "dst_2_tau_nu_from_belle2_dec", "d_1_l_nu_with_d_1_to_d_pi_pi_gap", "d_eta_l_nu_gap", "d_st_eta_l_nu_gap", "d_pi_pi_l_nu_our_gap", "d_st_pi_pi_l_nu_our_gap", "d_eta_l_nu_gap_via_d_st_0", "d_st_eta_l_nu_gap_via_d_prime_1", "d_pi_pi_l_nu_gap_via_d_prime_1", "d_pi_pi_l_nu_gap_via_d_st_0", "d_st_pi_pi_l_nu_gap_via_d_prime_1", "d_st_pi_pi_l_nu_gap_via_d_st_0", "dst_2_l_nu_with_dst_2_to_dst_pi_pi_NC", "d_s_kaon_l_nu_NC", "d_st_s_kaon_l_nu_NC"],
    "bp": [bf_bp_to_d_st_l_nu_avg, bf_bp_to_d_l_nu_avg, bf_bp_to_d_1_l_nu_our_cal, bf_bp_to_d0st_l_nu_our_cal, bf_bp_to_dprime1_l_nu_our_cal, bf_bp_to_dst_2_l_nu_our_cal, bf_bp_to_d_st_tau_nu_avg, bf_bp_to_d_tau_nu_avg, bf_bp_to_d_1_tau_nu_from_belle2_dec, bf_bp_to_d0st_tau_nu_from_belle2_dec, bf_bp_to_dprime1_tau_nu_from_belle2_dec, bf_bp_to_dst_2_tau_nu_from_belle2_dec, bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap, bf_bp_to_d_eta_l_nu_gap, bf_bp_to_d_st_eta_l_nu_gap, bf_bp_to_d_pi_pi_l_nu_our_gap, bf_bp_to_d_st_pi_pi_l_nu_our_gap, bf_bp_to_d_eta_l_nu_gap_via_d_st_0, bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1, bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1, bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0, bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1, bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0, bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered, bf_bp_to_d_s_kaon_l_nu_pdg, bf_bp_to_d_st_s_kaon_l_nu_pdg],
    "b0": [bf_b0_to_d_st_l_nu_avg, bf_b0_to_d_l_nu_avg, bf_b0_to_d_1_l_nu_our_cal, bf_b0_to_d0st_l_nu_our_cal, bf_b0_to_dprime1_l_nu_our_cal, bf_b0_to_dst_2_l_nu_our_cal, bf_b0_to_d_st_tau_nu_avg, bf_b0_to_d_tau_nu_avg, bf_b0_to_d_1_tau_nu_from_belle2_dec, bf_b0_to_d0st_tau_nu_from_belle2_dec, bf_b0_to_dprime1_tau_nu_from_belle2_dec, bf_b0_to_dst_2_tau_nu_from_belle2_dec, bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap, bf_b0_to_d_eta_l_nu_gap, bf_b0_to_d_st_eta_l_nu_gap, bf_b0_to_d_pi_pi_l_nu_our_gap, bf_b0_to_d_st_pi_pi_l_nu_our_gap, bf_b0_to_d_eta_l_nu_gap_via_d_st_0, bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1, bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1, bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0, bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1, bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0, bf_b0_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered, bf_b0_to_d_s_kaon_l_nu_pdg, bf_b0_to_d_st_s_kaon_l_nu_pdg],
})  # type: pd.DataFrame
# fmt: on
overview_table.set_index(keys="process_label", inplace=True, verify_integrity=True)

cols_latex_mapping = {
    "process_label": r"$\,$",
    "bp": LatexParticleNames.b_plus_str,
    "b0": LatexParticleNames.b_zero_str,
}  # type: Dict[str, str]

latex_overview_table = overview_table.copy(deep=True)  # type: pd.DataFrame
latex_overview_table.index.names = [cols_latex_mapping["process_label"]]
latex_overview_table = latex_overview_table.rename(columns={k: v for k, v in cols_latex_mapping.items()})

processes_latex_mapping = {
    "d_st_l_nu_avg": LDPs.b_to_d_st_l_nu,
    "d_l_nu_avg": LDPs.b_to_d_l_nu,
    "d_1_l_nu_our_cal": LDPs.b_to_d_1_l_nu,
    "d0st_l_nu_our_cal": LDPs.b_to_d0st_l_nu,
    "dprime1_l_nu_our_cal": LDPs.b_to_dprime1_l_nu,
    "dst_2_l_nu_our_cal": LDPs.b_to_dst_2_l_nu,
    "d_st_tau_nu_avg": LDPs.b_to_d_st_tau_nu,
    "d_tau_nu_avg": LDPs.b_to_d_tau_nu,
    "d_1_tau_nu_from_belle2_dec": LDPs.b_to_d_1_tau_nu,
    "d0st_tau_nu_from_belle2_dec": LDPs.b_to_d0st_tau_nu,
    "dprime1_tau_nu_from_belle2_dec": LDPs.b_to_dprime1_tau_nu,
    "dst_2_tau_nu_from_belle2_dec": LDPs.b_to_dst_2_tau_nu,
    "d_1_l_nu_with_d_1_to_d_pi_pi_gap": LDPs.b_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap,
    "d_eta_l_nu_gap": LDPs.b_to_d_eta_l_nu_gap,
    "d_st_eta_l_nu_gap": LDPs.b_to_d_st_eta_l_nu_gap,
    "d_pi_pi_l_nu_our_gap": LDPs.b_to_d_pi_pi_l_nu_gap,
    "d_st_pi_pi_l_nu_our_gap": LDPs.b_to_d_st_pi_pi_l_nu_gap,
    "d_eta_l_nu_gap_via_d_st_0": LDPs.b_to_d_eta_l_nu_gap_via_d_st_0,
    "d_st_eta_l_nu_gap_via_d_prime_1": LDPs.b_to_d_st_eta_l_nu_gap_via_d_prime_1,
    "d_pi_pi_l_nu_gap_via_d_prime_1": LDPs.b_to_d_pi_pi_l_nu_gap_via_d_prime_1,
    "d_pi_pi_l_nu_gap_via_d_st_0": LDPs.b_to_d_pi_pi_l_nu_gap_via_d_st_0,
    "d_st_pi_pi_l_nu_gap_via_d_prime_1": LDPs.b_to_d_st_pi_pi_l_nu_gap_via_d_prime_1,
    "d_st_pi_pi_l_nu_gap_via_d_st_0": LDPs.b_to_d_st_pi_pi_l_nu_gap_via_d_st_0,
    "dst_2_l_nu_with_dst_2_to_dst_pi_pi_NC": LDPs.b_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_NC,
    "d_s_kaon_l_nu_NC": LDPs.b_to_d_s_kaon_l_nu_NC,
    "d_st_s_kaon_l_nu_NC": LDPs.b_to_d_st_s_kaon_l_nu_NC,
}  # type: Dict[str, str]

latex_overview_table = latex_overview_table.T
latex_overview_table = latex_overview_table.rename(columns={k: v for k, v in processes_latex_mapping.items()})
latex_overview_table = latex_overview_table.T

tex_bf_values_dict = {
    "bf_b0_to_d_l_nu_avg": bf_b0_to_d_l_nu_avg,
    "bf_bp_to_d_l_nu_avg": bf_bp_to_d_l_nu_avg,
    "bf_b0_to_d_tau_nu_avg": bf_b0_to_d_tau_nu_avg,
    "bf_bp_to_d_tau_nu_avg": bf_bp_to_d_tau_nu_avg,
    "bf_b0_to_d_st_l_nu_avg": bf_b0_to_d_st_l_nu_avg,
    "bf_bp_to_d_st_l_nu_avg": bf_bp_to_d_st_l_nu_avg,
    "bf_b0_to_d_st_tau_nu_avg": bf_b0_to_d_st_tau_nu_avg,
    "bf_bp_to_d_st_tau_nu_avg": bf_bp_to_d_st_tau_nu_avg,
    "bf_b0_to_d0st_l_nu_our_cal": bf_b0_to_d0st_l_nu_our_cal,
    "bf_bp_to_d0st_l_nu_our_cal": bf_bp_to_d0st_l_nu_our_cal,
    "bf_b0_to_dprime1_l_nu_our_cal": bf_b0_to_dprime1_l_nu_our_cal,
    "bf_bp_to_dprime1_l_nu_our_cal": bf_bp_to_dprime1_l_nu_our_cal,
    "bf_b0_to_dst_2_l_nu_our_cal": bf_b0_to_dst_2_l_nu_our_cal,
    "bf_bp_to_dst_2_l_nu_our_cal": bf_bp_to_dst_2_l_nu_our_cal,
    "bf_b0_to_d_1_l_nu_our_cal": bf_b0_to_d_1_l_nu_our_cal,
    "bf_bp_to_d_1_l_nu_our_cal": bf_bp_to_d_1_l_nu_our_cal,
    "bf_b0_to_d0st_tau_nu_from_belle2_dec": bf_b0_to_d0st_tau_nu_from_belle2_dec,
    "bf_bp_to_d0st_tau_nu_from_belle2_dec": bf_bp_to_d0st_tau_nu_from_belle2_dec,
    "bf_b0_to_dprime1_tau_nu_from_belle2_dec": bf_b0_to_dprime1_tau_nu_from_belle2_dec,
    "bf_bp_to_dprime1_tau_nu_from_belle2_dec": bf_bp_to_dprime1_tau_nu_from_belle2_dec,
    "bf_b0_to_dst_2_tau_nu_from_belle2_dec": bf_b0_to_dst_2_tau_nu_from_belle2_dec,
    "bf_bp_to_dst_2_tau_nu_from_belle2_dec": bf_bp_to_dst_2_tau_nu_from_belle2_dec,
    "bf_b0_to_d_1_tau_nu_from_belle2_dec": bf_b0_to_d_1_tau_nu_from_belle2_dec,
    "bf_bp_to_d_1_tau_nu_from_belle2_dec": bf_bp_to_d_1_tau_nu_from_belle2_dec,
    "bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap": bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap,
    "bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap": bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap,
    "bf_b0_to_d_eta_l_nu_gap": bf_b0_to_d_eta_l_nu_gap,
    "bf_bp_to_d_eta_l_nu_gap": bf_bp_to_d_eta_l_nu_gap,
    "bf_b0_to_d_st_eta_l_nu_gap": bf_b0_to_d_st_eta_l_nu_gap,
    "bf_bp_to_d_st_eta_l_nu_gap": bf_bp_to_d_st_eta_l_nu_gap,
    "bf_b0_to_d_pi_pi_l_nu_our_gap": bf_b0_to_d_pi_pi_l_nu_our_gap,
    "bf_bp_to_d_pi_pi_l_nu_our_gap": bf_bp_to_d_pi_pi_l_nu_our_gap,
    "bf_b0_to_d_st_pi_pi_l_nu_our_gap": bf_b0_to_d_st_pi_pi_l_nu_our_gap,
    "bf_bp_to_d_st_pi_pi_l_nu_our_gap": bf_bp_to_d_st_pi_pi_l_nu_our_gap,
    "bf_b0_to_d_eta_l_nu_gap_via_d_st_0": bf_b0_to_d_eta_l_nu_gap_via_d_st_0,
    "bf_bp_to_d_eta_l_nu_gap_via_d_st_0": bf_bp_to_d_eta_l_nu_gap_via_d_st_0,
    "bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1": bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1,
    "bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1": bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1,
    "bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0": bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0,
    "bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1": bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1,
    "bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0": bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0,
    "bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1": bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1,
    "bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0": bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0,
    "bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1": bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1,
    "bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0": bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0,
    "bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1": bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1,
    "bf_b0_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered": bf_b0_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered,
    "bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered": bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered,
    "bf_b0_to_d_s_kaon_l_nu_pdg": bf_b0_to_d_s_kaon_l_nu_pdg,
    "bf_bp_to_d_s_kaon_l_nu_pdg": bf_bp_to_d_s_kaon_l_nu_pdg,
    "bf_b0_to_d_st_s_kaon_l_nu_pdg": bf_b0_to_d_st_s_kaon_l_nu_pdg,
    "bf_bp_to_d_st_s_kaon_l_nu_pdg": bf_bp_to_d_st_s_kaon_l_nu_pdg,
}  # type: Dict[str, UFloat]


def creat_latex_variable_table(
    output_path: Optional[PathType] = None,
    use_cell_color: bool = True,
    overwrite: bool = False,
) -> str:
    split_after = (2, 6, 8, 12, 17, 23)  # type: Tuple[int, ...]
    latex_colors = (
        "kit-blue",
        "kit-red",
        "kit-green",
        "kit-lightgreen",
        "kit-cyan",
        "kit-purple",
        "kit-gray",
    )  # type: Tuple[str, ...]
    color_alpha = 20  # type: int
    assert len(split_after) + 1 == len(latex_colors), (len(split_after), len(latex_colors))

    labels_to_bp_vals = (
        ("b_to_d_st_l_nu", "bf_bp_to_d_st_l_nu_avg"),
        ("b_to_d_l_nu", "bf_bp_to_d_l_nu_avg"),
        ("b_to_d_1_l_nu", "bf_bp_to_d_1_l_nu_our_cal"),
        ("b_to_d0st_l_nu", "bf_bp_to_d0st_l_nu_our_cal"),
        ("b_to_dprime1_l_nu", "bf_bp_to_dprime1_l_nu_our_cal"),
        ("b_to_dst_2_l_nu", "bf_bp_to_dst_2_l_nu_our_cal"),
        ("b_to_d_st_tau_nu", "bf_bp_to_d_st_tau_nu_avg"),
        ("b_to_d_tau_nu", "bf_bp_to_d_tau_nu_avg"),
        ("b_to_d_1_tau_nu", "bf_bp_to_d_1_tau_nu_from_belle2_dec"),
        ("b_to_d0st_tau_nu", "bf_bp_to_d0st_tau_nu_from_belle2_dec"),
        ("b_to_dprime1_tau_nu", "bf_bp_to_dprime1_tau_nu_from_belle2_dec"),
        ("b_to_dst_2_tau_nu", "bf_bp_to_dst_2_tau_nu_from_belle2_dec"),
        ("b_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap", "bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap"),
        ("b_to_d_eta_l_nu_gap", "bf_bp_to_d_eta_l_nu_gap"),
        ("b_to_d_st_eta_l_nu_gap", "bf_bp_to_d_st_eta_l_nu_gap"),
        ("b_to_d_pi_pi_l_nu_gap", "bf_bp_to_d_pi_pi_l_nu_our_gap"),
        ("b_to_d_st_pi_pi_l_nu_gap", "bf_bp_to_d_st_pi_pi_l_nu_our_gap"),
        ("b_to_d_eta_l_nu_gap_via_d_st_0", "bf_bp_to_d_eta_l_nu_gap_via_d_st_0"),
        ("b_to_d_st_eta_l_nu_gap_via_d_prime_1", "bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1"),
        ("b_to_d_pi_pi_l_nu_gap_via_d_prime_1", "bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1"),
        ("b_to_d_pi_pi_l_nu_gap_via_d_st_0", "bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0"),
        ("b_to_d_st_pi_pi_l_nu_gap_via_d_prime_1", "bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1"),
        ("b_to_d_st_pi_pi_l_nu_gap_via_d_st_0", "bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0"),
        ("b_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_NC", "bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_not_considered"),
        ("b_to_d_s_kaon_l_nu_NC", "bf_bp_to_d_s_kaon_l_nu_pdg"),
        ("b_to_d_st_s_kaon_l_nu_NC", "bf_bp_to_d_st_s_kaon_l_nu_pdg"),
    )  # type: Tuple[Tuple[str, str], ...]

    assert all(v1 in LDPs.get_attributes_mapping().keys() for v1, _ in labels_to_bp_vals), [
        v1 for v1, _ in labels_to_bp_vals if v1 not in LDPs.get_attributes_mapping().keys()
    ]

    _process_latex_mapping = {k: "\\" + escape_special_chars(s=k) + "{}" for k, _ in labels_to_bp_vals}
    assert len(_process_latex_mapping.keys()) == len(set(_process_latex_mapping.keys())), _process_latex_mapping

    var_table = pd.DataFrame(
        {
            "process_label": [_process_latex_mapping[k] for k, _ in labels_to_bp_vals],
            "bp": ["$\\" + escape_special_chars(s=bp_v) + "{}$" for _, bp_v in labels_to_bp_vals],
            "b0": [
                "$\\" + escape_special_chars(bp_v.replace("bf_bp_", "bf_b0_")) + "{}$" for _, bp_v in labels_to_bp_vals
            ],
        }
    )  # type: pd.DataFrame

    all_cell_values = [uv for c in var_table.columns for uv in var_table[c].unique()]  # type: List[str]

    _cols_latex_mapping = {
        "process_label": r"Process",
        "bp": LatexParticleNames.b_plus_str,
        "b0": LatexParticleNames.b_zero_str,
    }  # type: Dict[str, str]

    latex_var_table = var_table.copy(deep=True)  # type: pd.DataFrame
    latex_var_table = latex_var_table.rename(columns={k: v for k, v in _cols_latex_mapping.items()})

    with pd.option_context("max_colwidth", 1000):
        converted_latex_table = latex_var_table.to_latex(
            escape=False,
            column_format="l|ll",
            index=False,
        )

    new_converted_latex_table = []  # type: List[str]
    row_split = converted_latex_table.split("\\\\")
    for row_num, row in enumerate(row_split):
        this_row = row  # type: str
        if r"\midrule" in row:
            before_mid_rule, after_mid_rule = row.split(r"\midrule")  # type: str, str
            new_converted_latex_table.append(before_mid_rule + r"\midrule")
            this_row = after_mid_rule

        sec_count = "".join(new_converted_latex_table).count(r"\midrule")  # type: int

        if use_cell_color:
            line_start = ""  # type: str
            cell_color_str = "\\cellcolor{" + latex_colors[sec_count - 1] + f"!{color_alpha}" + "}"  # type: str
            for cell_value in all_cell_values:
                this_row = this_row.replace(cell_value, cell_color_str + " " + cell_value)
        else:
            row_color_str = "\\rowcolor{" + latex_colors[sec_count - 1] + f"!{color_alpha}" + "}"  # type: str
            line_start = "" if sec_count == 0 or row_num == len(row_split) - 1 else "\n" + row_color_str

        line_end = "\\\\" if row_num < len(row_split) - 1 else ""  # type: str
        new_converted_latex_table.append(line_start + this_row + line_end)
        if row_num in split_after:
            new_converted_latex_table.append("\n" + r"\midrule")

    converted_latex_table = "".join(new_converted_latex_table)
    converted_latex_table = converted_latex_table.replace(r"\midrule", r"\Xhline{0.5pt}")
    converted_latex_table = converted_latex_table.replace(r"\toprule", r"\Xhline{0.8pt}")
    converted_latex_table = converted_latex_table.replace(r"\bottomrule", r"\Xhline{0.8pt}")

    if output_path is not None:
        if not overwrite and os.path.exists(output_path):
            print(f"File {output_path} already exists and will not be overwritten!")
        else:
            with open(output_path, "w") as latex_file:
                latex_file.write(converted_latex_table)

    return converted_latex_table


def write_bf_values_to_latex(
    output_path: PathType,
    in_percent: bool = True,
    overwrite: bool = True,
) -> None:
    create_latex_value_file(
        path=output_path,
        bf_dict=tex_bf_values_dict,
        in_percent=in_percent,
        overwrite=overwrite,
    )


if __name__ == "__main__":
    print(b_to_d_l_nu_calculator.info_str(decay_id_str="B -> D l nu"))
    print(b_to_d_st_l_nu_calculator.info_str(decay_id_str="B -> D* l nu"))
    print(info_str(decay_id_str="B0 -> D*_0 l nu", bf_value=bf_b0_to_d0st_l_nu_our_cal))
    print(info_str(decay_id_str="B+ -> D*_0 l nu", bf_value=bf_bp_to_d0st_l_nu_our_cal))
    print(info_str(decay_id_str="B0 -> D'_1 l nu", bf_value=bf_b0_to_dprime1_l_nu_our_cal))
    print(info_str(decay_id_str="B+ -> D'_1 l nu", bf_value=bf_bp_to_dprime1_l_nu_our_cal))
    print(info_str(decay_id_str="B0 -> D*_2 l nu", bf_value=bf_b0_to_dst_2_l_nu_our_cal))
    print(info_str(decay_id_str="B+ -> D*_2 l nu", bf_value=bf_bp_to_dst_2_l_nu_our_cal))
    print(info_str(decay_id_str="B0 -> D_1 l nu", bf_value=bf_b0_to_d_1_l_nu_our_cal))
    print(info_str(decay_id_str="B+ -> D_1 l nu", bf_value=bf_bp_to_d_1_l_nu_our_cal))

    # print(b_to_d_pi_l_nu_calculator.info_str(decay_id_str="B -> D pi l nu"))
    # print(b_to_d_st_pi_l_nu_calculator.info_str(decay_id_str="B -> D* pi l nu"))

    print("Gap Composition:")

    print("Gap Contribution of D1 decays with two pions (this is in our Gap samples, which is why it's listed here):")
    print(info_str(decay_id_str="B0 -> D1 (-> D pi pi) l nu", bf_value=bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap))
    print(info_str(decay_id_str="B+ -> D1 (-> D pi pi) l nu", bf_value=bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap))

    print("Non-resonant Decays covered in Gap samples:")
    print(
        "The following values for the remaining B -> D(*) pi l nu BF indicate "
        "that the space left to contributions of non-resonant processes is so small that we can ignore them:"
    )
    print(info_str(decay_id_str="B0 -> D pi l nu (non-resonant)", bf_value=bf_b0_to_d_pi_l_nu_our_non_res))
    print(info_str(decay_id_str="B+ -> D pi l nu (non-resonant)", bf_value=bf_bp_to_d_pi_l_nu_our_non_res))
    print(info_str(decay_id_str="B0 -> D* pi l nu (non-resonant)", bf_value=bf_b0_to_d_st_pi_l_nu_our_non_res))
    print(info_str(decay_id_str="B+ -> D* pi l nu (non-resonant)", bf_value=bf_bp_to_d_st_pi_l_nu_our_non_res))

    print(
        "The following values for the remaining B -> D pi pi l nu BF indicate "
        "that the space left to contributions from other processes is so small, but we consider them as a Gap Component:"
    )
    print(info_str(decay_id_str="B0 -> D pi pi l nu (Gap Comp.)", bf_value=bf_b0_to_d_pi_pi_l_nu_our_gap))
    print(info_str(decay_id_str="B+ -> D pi pi l nu (Gap Comp.)", bf_value=bf_bp_to_d_pi_pi_l_nu_our_gap))
    print(
        "The following for the remaining values of B -> D* pi pi l nu BF indicate "
        "that there IS some space left for other contributions because there are no contributions from P1 states. "
        "Hance, we consider them as Gap Component:"
    )
    print(info_str(decay_id_str="B0 -> D* pi pi l nu (Gap Comp.)", bf_value=bf_b0_to_d_st_pi_pi_l_nu_our_gap))
    print(info_str(decay_id_str="B+ -> D* pi pi l nu (Gap Comp.)", bf_value=bf_bp_to_d_st_pi_pi_l_nu_our_gap))

    print(
        info_str(
            decay_id_str="the sum of exclusive B0 -> Xc l nu",
            bf_value=extended_sum_bp_to_xc_l_nu_exclusive / tau_p0,
        )
    )
    print(info_str(decay_id_str="the sum of exclusive B+ -> Xc l nu", bf_value=extended_sum_bp_to_xc_l_nu_exclusive))

    print("Remaining Gap to be filled:")
    print(info_str(decay_id_str="the remaining Gap in B0 -> Xc l nu", bf_value=gap_b0))
    print(info_str(decay_id_str="the remaining Gap in B+ -> Xc l nu", bf_value=gap_bp))

    print(info_str(decay_id_str="B0 -> D eta l nu to fill 50% of the gap", bf_value=bf_b0_to_d_eta_l_nu_gap))
    print(info_str(decay_id_str="B+ -> D eta l nu to fill 50% of the gap", bf_value=bf_bp_to_d_eta_l_nu_gap))
    print(info_str(decay_id_str="B0 -> D* eta l nu to fill 50% of the gap", bf_value=bf_b0_to_d_st_eta_l_nu_gap))
    print(info_str(decay_id_str="B+ -> D* eta l nu to fill 50% of the gap", bf_value=bf_bp_to_d_st_eta_l_nu_gap))

    print("Alternative Gap Definitions:")
    print("Gap D(*) eta via D** (instead of non-resonant D(*) eta):")
    print(info_str(decay_id_str="B0 -> D*_0 (-> D eta) l nu", bf_value=bf_b0_to_d_eta_l_nu_gap_via_d_st_0))
    print(info_str(decay_id_str="B+ -> D*_0 (-> D eta) l nu", bf_value=bf_bp_to_d_eta_l_nu_gap_via_d_st_0))
    print(info_str(decay_id_str="B0 -> D'_1 (-> D* eta) l nu", bf_value=bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1))
    print(info_str(decay_id_str="B+ -> D'_1 (-> D* eta) l nu", bf_value=bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1))

    print(info_str(decay_id_str="B0 -> D*_0 (-> D pi pi) l nu", bf_value=bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0))
    print(info_str(decay_id_str="B+ -> D*_0 (-> D pi pi) l nu", bf_value=bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0))
    print(info_str(decay_id_str="B0 -> D'_1 (-> D pi pi) l nu", bf_value=bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1))
    print(info_str(decay_id_str="B+ -> D'_1 (-> D pi pi) l nu", bf_value=bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1))
    print(info_str(decay_id_str="B0 -> D*_0 (-> D* pi pi) l nu", bf_value=bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0))
    print(info_str(decay_id_str="B+ -> D*_0 (-> D* pi pi) l nu", bf_value=bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0))
    print(info_str(decay_id_str="B0 -> D'_1 (-> D* pi pi) l nu", bf_value=bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1))
    print(info_str(decay_id_str="B+ -> D'_1 (-> D* pi pi) l nu", bf_value=bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1))

    print("Tau Modes:")
    print(info_str(decay_id_str="B0 -> D tau nu", bf_value=bf_b0_to_d_tau_nu_avg))
    print(info_str(decay_id_str="B+ -> D tau nu", bf_value=bf_bp_to_d_tau_nu_avg))
    print(info_str(decay_id_str="B0 -> D* tau nu", bf_value=bf_b0_to_d_st_tau_nu_avg))
    print(info_str(decay_id_str="B+ -> D* tau nu", bf_value=bf_bp_to_d_st_tau_nu_avg))
    print("Tau Modes of D** Decays:")

    print(info_str(decay_id_str="B0 -> D*_0 tau nu", bf_value=bf_b0_to_d0st_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B+ -> D*_0 tau nu", bf_value=bf_bp_to_d0st_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B0 -> D'_1 tau nu", bf_value=bf_b0_to_dprime1_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B+ -> D'_1 tau nu", bf_value=bf_bp_to_dprime1_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B0 -> D*_2 tau nu", bf_value=bf_b0_to_dst_2_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B+ -> D*_2 tau nu", bf_value=bf_bp_to_dst_2_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B0 -> D_1 tau nu", bf_value=bf_b0_to_d_1_tau_nu_from_belle2_dec))
    print(info_str(decay_id_str="B+ -> D_1 tau nu", bf_value=bf_bp_to_d_1_tau_nu_from_belle2_dec))

    print(overview_table)
    print(latex_overview_table)

    print(creat_latex_variable_table())

    # _ = creat_latex_variable_table(
    #     output_path="/home/metzner/Downloads/rdstar_latex_bf_values_table.tex",
    #     overwrite=True,
    # )

    # write_bf_values_to_latex(
    #     output_path="/home/metzner/Downloads/rdstar_latex_bf_values.tex",
    #     in_percent=True,
    #     overwrite=True,
    # )
