"""
Tool for the Reweighting of Branching Fractions of D -> X

Patrick Ecker 2020
"""

import copy
import logging
import warnings
import itertools
import numpy as np
import pandas as pd

from typing import Dict, List, Tuple
from pandas.core.common import SettingWithCopyWarning

from analysistools.processing.observables_utils import makeROOTCompatible

import rdstar.mc_reweighting.BFReweighting.d_bfs as d_meson_bf_info_module
from rdstar.mc_reweighting.BFReweighting.d_reweighting_utils import convert_weight_dict_for_comparison, PDGBFDictType

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "DBFReweighter",
]


class DBFReweighter:
    _required_cols = [
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc",
    ]  # type: List[str]

    _d_depth_conditions = {
        1: (
            "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc",
            "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc",
        ),
        2: (
            "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc",
            "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc",
        ),
        3: (
            "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc__bc",
            "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc__bc",
        ),
    }  # type: Dict[int, Tuple[str, str]]

    _d_pdgs = [411, 421, -411, -421]  # type: List[int]

    _d_pdgs_weight_dict_mapping = {
        411: d_meson_bf_info_module.d_charged_bf_weight_dict,
        421: d_meson_bf_info_module.d_zero_bf_weight_dict,
        -411: d_meson_bf_info_module.anti_d_charged_bf_weight_dict,
        -421: d_meson_bf_info_module.anti_d_zero_bf_weight_dict,
        -999: {},
    }  # type: Dict[int, PDGBFDictType]

    _temp_sig_b_mdst_index_col = "temp_sig_b_mdst_index"

    _sig_d_depth = "sig_d_depth"
    _tag_d_depth = "tag_d_depth"

    _sig_d_pdg = "sig_d_pdg"
    _tag_d_pdg = "tag_d_pdg"

    _sig_decay_col = "sig_d_decay"
    _tag_decay_col = "tag_d_decay"

    d_bf_weight_col = "d_bf_weight"

    def __init__(self, df: pd.DataFrame, signal_side_only: bool = False) -> None:
        self.df = df
        self.signal_side_only = signal_side_only
        assert all(col in self.df.columns for col in DBFReweighter.get_required_columns())

    @staticmethod
    def get_required_columns() -> List[str]:
        req_cols = copy.copy(DBFReweighter._required_cols)
        req_cols.extend(
            [
                c
                for a, b in itertools.product([1, 2, 3], [1, 2])
                for c in DBFReweighter._generate_d_daughter_vars(depth=a, b_mdst=b)
            ]
        )
        req_cols.extend([c for v in DBFReweighter._d_depth_conditions.values() for c in v])
        return list(set(req_cols))

    @staticmethod
    def _generate_d_daughter_vars(depth: int, b_mdst: int) -> List[str]:
        all_str = []
        mc_daughter_str = "mcDaughter(0," * depth
        brackets = ")" * depth
        for i in range(0, 7):
            var_str = f"genUpsilon4S(mcDaughter({b_mdst - 1}, {mc_daughter_str}mcDaughter({i},PDG))){brackets}"
            all_str.append(var_str)
        return makeROOTCompatible(all_str)

    def _generate_depth_condition(self, b_mdst: int) -> Tuple[pd.Series, ...]:
        all_conds = []
        for depth in range(1, 4):
            assert depth in self._d_depth_conditions.keys(), depth
            cond = self.df[self._d_depth_conditions[depth][b_mdst - 1]].isin(self._d_pdgs)
            all_conds.append(cond)
        return tuple(all_conds)

    def _get_d_pdg_var(self, b_mdst: int, depth: int):
        return self._d_depth_conditions[depth][b_mdst - 1]

    def _set_d_depth(self) -> None:
        b_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        b_mdst_tau_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
        required_cols = [b_mdst_col, b_mdst_tau_col]
        assert all(col in self._required_cols for col in required_cols), required_cols

        # define temporary columns
        temp_b1_d_depth_col = "b1_d_depth"
        temp_b2_d_depth_col = "b2_d_depth"
        temp_b1_d_pdg_col = "b1_d_pdg"
        temp_b2_d_pdg_col = "b2_d_pdg"
        temp_columns = [temp_b1_d_depth_col, temp_b2_d_depth_col, temp_b1_d_pdg_col, temp_b2_d_pdg_col]
        assert all([col not in self.df.columns for col in temp_columns]), [c for c in temp_columns]

        # Daughter D Depth
        combos = {
            1: (temp_b1_d_depth_col, temp_b1_d_pdg_col),
            2: (temp_b2_d_depth_col, temp_b2_d_pdg_col),
        }
        for b_mdst, column in combos.items():
            cond1, cond2, cond3 = self._generate_depth_condition(b_mdst=b_mdst)
            self.df.loc[:, column[0]] = np.nan
            self.df.loc[cond1 & (~cond2) & (~cond3), column[0]] = 1
            self.df.loc[(~cond1) & cond2 & (~cond3), column[0]] = 2
            self.df.loc[(~cond1) & (~cond2) & cond3, column[0]] = 3
            self.df.loc[(~cond1) & (~cond2) & (~cond3), column[0]] = -999
            assert not pd.isnull(self.df[column[0]]).any(), column[0]

            self.df.loc[:, column[1]] = np.nan
            self.df.loc[cond1 & (~cond2) & (~cond3), column[1]] = self.df[self._get_d_pdg_var(b_mdst=b_mdst, depth=1)]
            self.df.loc[(~cond1) & cond2 & (~cond3), column[1]] = self.df[self._get_d_pdg_var(b_mdst=b_mdst, depth=2)]
            self.df.loc[(~cond1) & (~cond2) & cond3, column[1]] = self.df[self._get_d_pdg_var(b_mdst=b_mdst, depth=3)]
            self.df.loc[(~cond1) & (~cond2) & (~cond3), column[1]] = -999
            assert not pd.isnull(self.df[column[1]]).any(), column[1]

        sig_d_depth_col = self._sig_d_depth
        assert sig_d_depth_col not in self.df.columns, sig_d_depth_col
        sig_d_pdg_col = self._sig_d_pdg
        assert sig_d_pdg_col not in self.df.columns, sig_d_pdg_col
        if not self.signal_side_only:
            tag_d_depth_col = self._tag_d_depth
            assert tag_d_depth_col not in self.df.columns, tag_d_depth_col
            tag_d_pdg_col = self._tag_d_pdg
            assert tag_d_pdg_col not in self.df.columns, tag_d_pdg_col

            self.df.loc[:, sig_d_depth_col] = self.df[temp_b2_d_depth_col]
            self.df.loc[:, sig_d_pdg_col] = self.df[temp_b2_d_pdg_col]
            self.df.loc[:, tag_d_depth_col] = self.df[temp_b1_d_depth_col]
            self.df.loc[:, tag_d_pdg_col] = self.df[temp_b1_d_pdg_col]
            for col in [sig_d_depth_col, tag_d_depth_col, sig_d_pdg_col, tag_d_pdg_col]:
                assert not pd.isnull(self.df[col]).any(), col
        else:
            # Now check for the Signal B Meson
            cond = self.df[b_mdst_col].isin([1, 2])
            cond_tau = self.df[b_mdst_tau_col].isin([1, 2])

            self.df.loc[:, self._temp_sig_b_mdst_index_col] = np.nan
            self.df.loc[cond, self._temp_sig_b_mdst_index_col] = self.df[b_mdst_col]
            self.df.loc[cond_tau, self._temp_sig_b_mdst_index_col] = self.df[b_mdst_tau_col]
            self.df.loc[(~cond) & (~cond_tau), self._temp_sig_b_mdst_index_col] = 0
            assert not pd.isnull(self.df[self._temp_sig_b_mdst_index_col]).any(), self._temp_sig_b_mdst_index_col

            cond_b1 = self.df[self._temp_sig_b_mdst_index_col] == 1
            cond_b2 = self.df[self._temp_sig_b_mdst_index_col] == 2

            self.df.loc[:, sig_d_depth_col] = np.nan
            self.df.loc[cond_b1, sig_d_depth_col] = self.df[temp_b1_d_depth_col]
            self.df.loc[cond_b2, sig_d_depth_col] = self.df[temp_b2_d_depth_col]
            self.df.loc[(~cond_b1) & (~cond_b2), sig_d_depth_col] = -999
            assert not pd.isnull(self.df[sig_d_depth_col]).any(), sig_d_depth_col

            self.df.loc[:, sig_d_pdg_col] = np.nan
            self.df.loc[cond_b1, sig_d_pdg_col] = self.df[temp_b1_d_pdg_col]
            self.df.loc[cond_b2, sig_d_pdg_col] = self.df[temp_b2_d_pdg_col]
            self.df.loc[(~cond_b1) & (~cond_b2), sig_d_pdg_col] = -999
            assert not pd.isnull(self.df[sig_d_pdg_col]).any(), sig_d_pdg_col

        for col in temp_columns:
            del self.df[col]

    def _generate_conditions_decay_vars_mapping(
        self,
        depth_var: str,
        sig: bool = True,
    ) -> Dict[Tuple[str, ...], pd.Series]:
        cond_decay_vars_mapping = {}
        if self.signal_side_only:
            cond_b1 = self.df[self._temp_sig_b_mdst_index_col] == 1
            cond_b2 = self.df[self._temp_sig_b_mdst_index_col] == 2
            for b_mdst, cond in {1: cond_b1, 2: cond_b2}.items():
                for i in range(1, 4):
                    cond_decay_vars_mapping.update(
                        {
                            tuple(self._generate_d_daughter_vars(depth=i, b_mdst=b_mdst)): (
                                cond & (self.df[depth_var] == i)
                            )
                        }
                    )
        else:
            for i in range(1, 4):
                cond_decay_vars_mapping.update(
                    {tuple(self._generate_d_daughter_vars(depth=i, b_mdst=2 if sig else 1)): (self.df[depth_var] == i)}
                )
        return cond_decay_vars_mapping

    def _write_decay_to_df(
        self,
        decay_col: str,
        conditions_decay_vars_mapping: Dict[Tuple[str, ...], pd.Series],
    ) -> None:
        self.df.loc[:, decay_col] = np.nan
        for decay_vars, conds in conditions_decay_vars_mapping.items():
            for var in decay_vars:
                self.df.loc[:, "temp" + var] = self.df[var].astype(dtype=int, copy=True)
                # set every photon to -999
                self.df.loc[self.df["temp" + var] == 22, "temp" + var] = -999
                # change pdg codes for f_0 and f'_0
                self.df.loc[self.df["temp" + var] == 10221, "temp" + var] = 9010221
                self.df.loc[self.df["temp" + var] == 10331, "temp" + var] = 30221
            assert (self.df["temp" + decay_vars[-1]] == -999).any()

            # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
            with warnings.catch_warnings(record=True) as warning:
                warnings.simplefilter("always")
                # Actual command:
                self.df["temp"] = list(zip(*[self.df["temp" + var] for var in decay_vars]))
                for w in warning:
                    assert issubclass(w.category, SettingWithCopyWarning)
                    assert "A value is trying to be set on a copy" in str(w.message)

            self.df.loc[conds, decay_col] = self.df["temp"]
            for var in decay_vars:
                del self.df["temp" + var]
            del self.df["temp"]
        self.df.loc[(self.df[decay_col].isnull()), decay_col] = -999
        assert not pd.isnull(self.df[decay_col]).any(), decay_col

    def _set_decay(self) -> None:
        assert self._sig_decay_col not in self.df.columns, self._sig_decay_col

        sig_conditions_decay_vars_mapping = self._generate_conditions_decay_vars_mapping(
            depth_var=self._sig_d_depth,
        )
        self._write_decay_to_df(
            decay_col=self._sig_decay_col,
            conditions_decay_vars_mapping=sig_conditions_decay_vars_mapping,
        )

        if not self.signal_side_only:
            assert self._tag_decay_col not in self.df.columns, self._tag_decay_col

            tag_conditions_decay_vars_mapping = self._generate_conditions_decay_vars_mapping(
                depth_var=self._tag_d_depth,
                sig=False,
            )
            self._write_decay_to_df(
                decay_col=self._tag_decay_col,
                conditions_decay_vars_mapping=tag_conditions_decay_vars_mapping,
            )

    def _set_weights(self) -> None:
        assert self.d_bf_weight_col not in self.df.columns, self.d_bf_weight_col

        self.df.loc[:, self.d_bf_weight_col] = np.nan
        number_of_exceptions = 0
        for d_pdg, d_weight_dict in self._d_pdgs_weight_dict_mapping.items():
            weights = convert_weight_dict_for_comparison(d_weight_dict)
            d_cond = self.df[self._sig_d_pdg] == d_pdg
            self.df.loc[d_cond, self.d_bf_weight_col] = self.df[self._sig_decay_col].map(weights)

            # make this check, cause there are some weird decays in the Belle MC
            cond_not_in_weights = ~self.df[self._sig_decay_col].isin(weights.keys())
            exceptions = self.df.loc[cond_not_in_weights & d_cond, self._sig_decay_col].values
            number_of_exceptions += len(exceptions)
            for decay in exceptions:
                logging.warning(f"Signal Decay {decay} not in weight dict! Setting weight to 0!")
            self.df.loc[cond_not_in_weights & d_cond, self.d_bf_weight_col] = 0

        assert not pd.isnull(self.df[self.d_bf_weight_col]).any(), self.d_bf_weight_col
        assert len(self.df.loc[self.df[self.d_bf_weight_col] == 0]) == number_of_exceptions

        if not self.signal_side_only:
            for d_pdg, d_weight_dict in self._d_pdgs_weight_dict_mapping.items():
                weights = convert_weight_dict_for_comparison(d_weight_dict)
                d_cond = self.df[self._tag_d_pdg] == d_pdg
                cond_weight = self.df[self.d_bf_weight_col] != 0
                self.df.loc[d_cond & cond_weight, self.d_bf_weight_col] *= self.df[self._tag_decay_col].map(weights)

                # make this check, cause there are some weird decays in the Belle MC
                cond_not_in_weights = ~self.df[self._tag_decay_col].isin(weights.keys())
                exceptions = self.df.loc[cond_not_in_weights & d_cond & cond_weight, self._tag_decay_col].values
                number_of_exceptions += len(exceptions)
                for decay in exceptions:
                    logging.warning(f"Tag Decay {decay} not in weight dict! Setting weight to 0!")
                self.df.loc[cond_not_in_weights & d_cond, self.d_bf_weight_col] = 0

            assert not pd.isnull(self.df[self.d_bf_weight_col]).any(), self.d_bf_weight_col
            assert len(self.df.loc[self.df[self.d_bf_weight_col] == 0]) == number_of_exceptions

    def _finalize_weights(self) -> None:
        columns_for_del = [self._sig_d_depth]
        if not self.signal_side_only:
            columns_for_del.append(self._tag_d_depth)
        else:
            columns_for_del.append(self._temp_sig_b_mdst_index_col)
        for column in columns_for_del:
            del self.df[column]

    def do_reweighting(self) -> None:
        self._set_d_depth()
        self._set_decay()
        self._set_weights()
        self._finalize_weights()
