"""
B -> D** l nu FF Reweighting (ISGW2 -> LLSW) with RBF reweighting technique
"""

import os
import copy
import rdstar
import logging
import numpy as np
import pandas as pd

from enum import Enum
from typing import Union, Optional, Tuple, List, Dict

from analysistools.reweighting.rbf_reweighter import RbfReweighter

from rdstar.utilities import PathType
from rdstar.offline_analysis.mc_ids import gen_b_ids
from rdstar.mc_reweighting.utility import FFReweightingBase

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "DststType",
    "BtoDststFFReweighting",
]


class DststType(Enum):
    DZEROSTAR = "Dzerostar"
    DPRIMEONE = "Dprimeone"
    DONE = "Done"
    DTWOSTAR = "Dtwostar"


VariationPathDictType = Dict[Union[str, DststType], List[Tuple[PathType, PathType]]]
NominalPathDictType = Dict[Union[str, DststType], PathType]


default_rbf_nom_paths = {
    DststType.DONE: "rbf_Done_nominal.json",
    DststType.DTWOSTAR: "rbf_Dtwostar_nominal.json",
    DststType.DZEROSTAR: "rbf_Dzerostar_nominal.json",
    DststType.DPRIMEONE: "rbf_Dprimeone_nominal.json",
}  # type: NominalPathDictType

default_rbf_variation_paths = {
    DststType.DONE: [
        ("rbf_Done_up_0.json", "rbf_Done_down_0.json"),
        ("rbf_Done_up_1.json", "rbf_Done_down_1.json"),
        ("rbf_Done_up_2.json", "rbf_Done_down_2.json"),
        ("rbf_Done_up_3.json", "rbf_Done_down_3.json"),
    ],
    DststType.DTWOSTAR: [
        ("rbf_Dtwostar_up_0.json", "rbf_Dtwostar_down_0.json"),
        ("rbf_Dtwostar_up_1.json", "rbf_Dtwostar_down_1.json"),
        ("rbf_Dtwostar_up_2.json", "rbf_Dtwostar_down_2.json"),
        ("rbf_Dtwostar_up_3.json", "rbf_Dtwostar_down_3.json"),
    ],
    DststType.DZEROSTAR: [
        ("rbf_Dzerostar_up_0.json", "rbf_Dzerostar_down_0.json"),
        ("rbf_Dzerostar_up_1.json", "rbf_Dzerostar_down_1.json"),
        ("rbf_Dzerostar_up_2.json", "rbf_Dzerostar_down_2.json"),
    ],
    DststType.DPRIMEONE: [
        ("rbf_Dprimeone_up_0.json", "rbf_Dprimeone_down_0.json"),
        ("rbf_Dprimeone_up_1.json", "rbf_Dprimeone_down_1.json"),
        ("rbf_Dprimeone_up_2.json", "rbf_Dprimeone_down_2.json"),
    ],
}  # type: VariationPathDictType


class BtoDststFFReweighting(FFReweightingBase):
    dstst_gen_ids = {
        DststType.DONE: (gen_b_ids["Donel"].id, gen_b_ids["Donetau"].id),
        DststType.DTWOSTAR: (gen_b_ids["Dtwostarl"].id, gen_b_ids["Dtwostartau"].id),
        DststType.DPRIMEONE: (gen_b_ids["Dprimeonel"].id, gen_b_ids["Dprimeonetau"].id),
        DststType.DZEROSTAR: (gen_b_ids["Dzerostarl"].id, gen_b_ids["Dzerostartau"].id),
    }  # type: Dict[DststType, Tuple[int, int]]

    def __init__(
        self,
        rbf_nominal_paths: Optional[NominalPathDictType] = None,
        rbf_variation_paths: Optional[VariationPathDictType] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        pkg_dir = os.path.dirname(rdstar.__file__)  # type: PathType
        self._rbf_source_dir = os.path.join(pkg_dir, "data", "dstst_reweighting_data")

        self.b1_is_btodststl = None
        self.b2_is_btodststl = None

        if rbf_nominal_paths is None:
            self.rbf_nominal_paths = copy.deepcopy(default_rbf_nom_paths)
        else:
            self.rbf_nominal_paths = rbf_nominal_paths

        if rbf_variation_paths is None:
            self.rbf_variation_paths = copy.deepcopy(default_rbf_variation_paths)
        else:
            self.rbf_variation_paths = rbf_variation_paths

        self.rbf_B1_variables = {
            DststType.DONE: [self.b1_w, self.b1_costhetal, self.b1_costhetav],
            DststType.DTWOSTAR: [self.b1_w, self.b1_costhetal],
            DststType.DZEROSTAR: [self.b1_w],
            DststType.DPRIMEONE: [self.b1_w, self.b1_costhetal],
        }  # type: Dict[DststType, List[str]]

        self.rbf_B2_variables = {
            DststType.DONE: [self.b2_w, self.b2_costhetal, self.b2_costhetav],
            DststType.DTWOSTAR: [self.b2_w, self.b2_costhetal],
            DststType.DZEROSTAR: [self.b2_w],
            DststType.DPRIMEONE: [self.b2_w, self.b2_costhetal],
        }  # type: Dict[DststType, List[str]]

        self.nominal_weight_keys = list()  # type: List[str]
        self.variation_weight_keys = list()  # type: List[str]

    @staticmethod
    def get_b_to_dstst_l_nu_gen_ids() -> Tuple[int, ...]:
        ids = tuple([gen_id for gen_ids in BtoDststFFReweighting.dstst_gen_ids.values() for gen_id in gen_ids])
        gen_b_id_values = [d.id for d in gen_b_ids.values()]  # type: List[int]
        assert all(gen_id in gen_b_id_values for gen_id in ids)
        return ids

    def _add_nominal_weights(self, df: pd.DataFrame, dstst_type: DststType) -> None:
        nominal_weight_key = f"{self.weight_name_str}_nom"
        self.nominal_weight_keys.append(nominal_weight_key)

        df.loc[:, nominal_weight_key] = np.nan

        nom_path = self.rbf_nominal_paths[dstst_type]
        nom_rbf = RbfReweighter.import_from_json(os.path.join(self._rbf_source_dir, nom_path))

        rw_b1_vars, rw_b2_vars = self._get_rbf_vars(df, dstst_type)

        df.loc[self.b1_is_btodststl, nominal_weight_key] = nom_rbf(*rw_b1_vars)
        df.loc[self.b2_is_btodststl, nominal_weight_key] = nom_rbf(*rw_b2_vars)

        df.loc[df[nominal_weight_key].isnull(), nominal_weight_key] = 1.0

    def _get_rbf_vars(self, df: pd.DataFrame, dstst_type: DststType) -> Tuple[List[pd.Series], List[pd.Series]]:
        rw_b1_vars = [df.loc[self.b1_is_btodststl, var].values for var in self.rbf_B1_variables[dstst_type]]
        rw_b2_vars = [df.loc[self.b2_is_btodststl, var].values for var in self.rbf_B2_variables[dstst_type]]

        return rw_b1_vars, rw_b2_vars

    def _add_varied_weights(self, df: pd.DataFrame, dstst_type: DststType) -> None:
        var_paths = self.rbf_variation_paths[dstst_type]

        for i, path in enumerate(var_paths):
            up_rbf = RbfReweighter.import_from_json(os.path.join(self._rbf_source_dir, path[0]))
            down_rbf = RbfReweighter.import_from_json(os.path.join(self._rbf_source_dir, path[1]))

            up_key = f"{self.weight_name_str}_var_{i}_up"  # type: str
            down_key = f"{self.weight_name_str}_var_{i}_down"  # type: str

            rw_b1_vars, rw_b2_vars = self._get_rbf_vars(df, dstst_type)

            self.variation_weight_keys.append(up_key)
            self.variation_weight_keys.append(down_key)

            df.loc[:, up_key] = np.nan
            df.loc[self.b1_is_btodststl, up_key] = up_rbf(*rw_b1_vars)
            df.loc[self.b2_is_btodststl, up_key] = up_rbf(*rw_b2_vars)
            df.loc[df[up_key].isnull(), up_key] = 1.0

            df.loc[:, down_key] = np.nan
            df.loc[self.b1_is_btodststl, down_key] = down_rbf(*rw_b1_vars)
            df.loc[self.b2_is_btodststl, down_key] = down_rbf(*rw_b2_vars)
            df.loc[df[down_key].isnull(), down_key] = 1.0

    def add_weights(self, df: pd.DataFrame, dstst_type: Optional[DststType] = None) -> None:
        assert dstst_type is not None and dstst_type in DststType, dstst_type

        logging.info(f"Adding weights for {dstst_type} to candidates...")
        self._weight_name_str = self.get_dstst_weight_name(dstst_type=dstst_type.value)

        self.b1_is_btodststl = (df.GEN_B1_ID.isin(self.dstst_gen_ids[dstst_type])) & (
            (df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1)
        )
        self.b2_is_btodststl = (df.GEN_B2_ID.isin(self.dstst_gen_ids[dstst_type])) & (
            (df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2)
        )

        self._add_nominal_weights(df, dstst_type)
        self._add_varied_weights(df, dstst_type)
        self._write_weight_info(df, conditions=(self.b1_is_btodststl, self.b2_is_btodststl))

    @staticmethod
    def get_dstst_weight_name(dstst_type: str) -> str:
        return f"weight_ff_{dstst_type}"

    @staticmethod
    def get_nominal_dstst_weight_name(dstst_type: str) -> str:
        return f"{BtoDststFFReweighting.get_dstst_weight_name(dstst_type=dstst_type)}_nom"

    @staticmethod
    def get_var_dstst_weight_name(dstst_type: str) -> str:
        return f"{BtoDststFFReweighting.get_dstst_weight_name(dstst_type=dstst_type)}_var_"

    @staticmethod
    def add_neutral_dstst_weight_columns(
        df: pd.DataFrame,
        mask: Optional[pd.Series] = None,
        rbf_variation_path_dict: Optional[VariationPathDictType] = None,
    ) -> None:
        rbf_var_dict = default_rbf_variation_paths if rbf_variation_path_dict is None else rbf_variation_path_dict
        if mask is None:
            mask = pd.Series(True, index=df.index)
        else:
            assert len(mask.index) == len(df.index), (len(mask.index), len(df.index))

        for dstst_type in DststType:
            nom_column_name = BtoDststFFReweighting.get_nominal_dstst_weight_name(dstst_type.value)
            assert nom_column_name not in df.columns or pd.isnull(df.loc[mask, nom_column_name]).all()
            df.loc[mask, nom_column_name] = 1.0
            for variation in range(len(rbf_var_dict[dstst_type])):
                base_col_name = BtoDststFFReweighting.get_var_dstst_weight_name(dstst_type.value)
                up_var_column_name = f"{base_col_name}{variation}_up"
                down_var_column_name = f"{base_col_name}{variation}_down"
                assert up_var_column_name not in df.columns or pd.isnull(df.loc[mask, up_var_column_name]).all()
                assert down_var_column_name not in df.columns or pd.isnull(df.loc[mask, down_var_column_name]).all()
                df.loc[mask, up_var_column_name] = 1.0
                df.loc[mask, down_var_column_name] = 1.0
