"""
Master Task taking care of the calculation and evaluation of weights for the B -> D** l nu reweighting.

Felix Metzner, Maximilian Welsch (2018 - 2020)
"""

import os
import b2luigi
import logging

from typing import Generator

from rdstar.utilities import PathType
from rdstar.mc_reweighting.BtoDststlnu.b_to_dstst_weights_generation_tasks import DststReweightingMaster
from rdstar.mc_reweighting.BtoDststlnu.b_to_dstst_weights_analysis_tasks import DststWeightEvaluationWrapperTask

__all__ = [
    "DststReweightingAndWeightAnalysisMasterTask",
]

logging.getLogger(__name__).addHandler(logging.NullHandler())


class DststReweightingAndWeightAnalysisMasterTask(b2luigi.WrapperTask):
    n_events = b2luigi.IntParameter(positional=False)  # type: int
    base_dir_path = b2luigi.Parameter(positional=False, hashed=True)  # type: str

    result_dir_name = "results"  # type: str
    log_dir_name = "logs"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.result_dir = os.path.join(self.base_dir_path, self.result_dir_name)  # type: PathType
        self.log_dir = os.path.join(self.base_dir_path, self.log_dir_name)  # type: PathType

        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

    def requires(self) -> Generator[b2luigi.Task, None, None]:
        self.clone(DststReweightingMaster, result_dir=self.result_dir, log_dir=self.log_dir)

        for decay_channel in DststReweightingMaster.decay_channels:
            yield self.clone(
                DststWeightEvaluationWrapperTask,
                decay_channel=decay_channel,
                result_dir=self.result_dir,
                log_dir=self.log_dir,
            )


if __name__ == "__main__":
    n_events = 2000000  # type: int
    reweighting_base_dir_path = "/ceph/fmetzner/dstst_reweighting_for_rdstar_analysis_v2/"  # type: str

    b2luigi.process(
        DststReweightingAndWeightAnalysisMasterTask(n_events=n_events, base_dir_path=reweighting_base_dir_path),
        workers=18,
    )
