"""
Base class for analysis plots.

Felix Metzner 2020
"""

import os
import logging
import pandas as pd

from abc import ABC, abstractmethod
from typing import List, Dict, Any, Callable

from rdstar.utilities import PathType

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "AnalysisPlotBaseClass",
]


class AnalysisPlotBaseClass(ABC):
    _plotter_id_key = ""  # type: str
    base_dir_name = ""  # type: str

    def __init__(
        self,
        input_dictionary: Dict[str, PathType],
        output_dir_path: PathType,
        selection_name: str,
        cs_tag: str,
        **kwargs,
    ) -> None:
        self.input_dictionary = input_dictionary  # type: Dict[str, PathType]
        self.__output_dir_path = output_dir_path  # type: PathType
        self.selection_name = selection_name  # type: str
        self.cs_tag = cs_tag  # type: str

    @abstractmethod
    def create_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_plotter_options(**kwargs) -> Dict[str, List[Any]]:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_required_task_type() -> Callable:
        raise NotImplementedError()

    @abstractmethod
    def get_relevant_input(self) -> Dict[str, PathType]:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def plotter_id_key() -> str:
        raise NotImplementedError()

    @property
    def output_dir_path(self) -> PathType:
        if not self.base_dir_name:
            raise RuntimeError(f"Plotter {self.__class__.__name__} has no base_dir_name {self.base_dir_name}")
        return os.path.join(self.__output_dir_path, self.base_dir_name)

    def load_df(self, dry_run: bool) -> pd.DataFrame:
        if dry_run:
            return pd.DataFrame()
        else:
            input_dict = self.get_relevant_input()  # type: Dict[str, PathType]
            assert all(os.path.isfile(v) for v in input_dict.values()), [
                (k, v) for k, v in input_dict.items() if not os.path.isfile(v)
            ]
            if len(input_dict) > 1:
                df = pd.concat([pd.read_hdf(d) for d in input_dict.values()], ignore_index=True, sort=False)
            elif len(input_dict) == 1:
                path = list(input_dict.values())[0]
                df = pd.read_hdf(path_or_buf=path)
            else:
                raise RuntimeError(f"Got no input information! input_information was: {input_dict}")
            assert isinstance(df, pd.DataFrame)
            return df
