"""
Defines qSquared residual plot.

Felix Metzner 2020
"""

import os
import logging
import itertools
import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from typing import Tuple, List, Dict, NamedTuple, Any, Callable

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors, ylabel_pos

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.rdstar_samples import private_data_types
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable
from rdstar.offline_analysis.histogram_variables import (
    q2_residual_with_q2reco,
    q2_residual_with_q2reco2,
    q2_residual_with_q2reco3,
    q2_residual_with_q2reco4,
    q2_resolution_with_q2reco,
    q2_resolution_with_q2reco2,
    q2_resolution_with_q2reco3,
    q2_resolution_with_q2reco4,
)

from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.analysis_plots.analysis_plot_base import AnalysisPlotBaseClass
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask


logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "QSquaredResidualPlotter",
]


class MaskLabelColorCombination(NamedTuple):
    mask: pd.Series
    label: str
    tag: str
    color: str


class QSquaredResidualPlotter(AnalysisPlotBaseClass):
    _plotter_id_key = "q_Squared_Residuals"  # type: str
    base_dir_name = "q_Squared_Residual_Plots"  # type: str

    # Existing columns
    data_type_col_name = "data_type"  # type: str
    d_st_data_type_keys = [dt for dt in private_data_types if dt.startswith("DS")]  # type: List[str]
    d_st_combination_key = "SignalEff"  # type: str

    input_appendix = f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"  # type: str

    normed_options = (
        True,
        False,
    )  # type: Tuple[bool, bool]

    residual_label_base_string = LatexObsNames.q2_resi  # type: str
    resolution_label_base_string = LatexObsNames.q2_reso  # type: str

    label_base_string_map = {
        False: residual_label_base_string,
        True: resolution_label_base_string,
    }  # type: Dict[bool, str]

    q2_histogram_variables = (
        (q2_residual_with_q2reco, False),
        (q2_residual_with_q2reco2, False),
        (q2_residual_with_q2reco3, False),
        (q2_residual_with_q2reco4, False),
        (q2_resolution_with_q2reco, True),
        (q2_resolution_with_q2reco2, True),
        (q2_resolution_with_q2reco3, True),
        (q2_resolution_with_q2reco4, True),
    )  # type: Tuple[Tuple[HistVariable, bool], ...]

    dst_reco_mode_ids = [15, 16, 17, 18]  # type: List[int]
    tag_sig_prob_cut = 0.1  # type: float

    def __init__(
        self,
        output_dir_path: PathType,
        input_dictionary: Dict[str, PathType],
        selection_name: str,
        cs_tag: str,
        with_taus: bool,
        only_dst_reco_channels: bool,
        arb_units: bool,
    ) -> None:
        super().__init__(
            output_dir_path=output_dir_path,
            input_dictionary=input_dictionary,
            selection_name=selection_name,
            cs_tag=cs_tag,
        )

        self.with_taus = with_taus  # type: bool
        self.only_dst_reco_channels = only_dst_reco_channels  # type: bool
        self.arb_units = arb_units  # type: bool

    @staticmethod
    def plotter_id_key() -> str:
        if not QSquaredResidualPlotter._plotter_id_key:
            raise RuntimeError(
                f"Plotter {QSquaredResidualPlotter.__class__.__name__} has no _plotter_id_key"
                f" {QSquaredResidualPlotter._plotter_id_key}"
            )
        return QSquaredResidualPlotter._plotter_id_key

    @staticmethod
    def get_plotter_options(**kwargs) -> Dict[str, List[Any]]:
        return {
            "with_taus": [True, False],
            "only_dst_reco_channels": [True, False],
            "arb_units": [True, False],
        }

    @staticmethod
    def get_required_task_type() -> Callable:
        return ResolutionCorrectionTask

    def get_relevant_input(self) -> Dict[str, PathType]:
        key_beginning = f"{self.d_st_combination_key}_{self.input_appendix}_"  # type: str
        key_ending = f"_{self.selection_name}_with_CS_{self.cs_tag}"  # type: str

        relevant_input_dict = {
            k: p
            for k, p in self.input_dictionary.items()
            if (k.startswith(key_beginning) and k.endswith(key_ending) and os.path.splitext(p)[1] == ".h5")
        }
        assert len(relevant_input_dict) == 1, (
            relevant_input_dict,
            key_beginning,
            key_ending,
            self.input_dictionary.keys(),
        )
        return relevant_input_dict

    def create_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        df = self.load_df(dry_run=dry_run)  # type: pd.DataFrame

        if not dry_run:
            QSquaredResidualVariable.add_q_squared_residual_definitions(df=df)
            mask_label_color_combinations = self.get_mask_label_color_combinations(df=df)
        else:
            mask_label_color_combinations = []

        set_matplotlibrc_params()
        for q2_var, is_resolution in self.q2_histogram_variables:
            if not dry_run:
                assert q2_var.df_label in df.columns, q2_var.df_label

            for normed in self.normed_options:
                q2_residual_plot_key, q2_residual_comparison_plot_key = self.get_file_name_and_append_to_output_dict(
                    output_dict=output_dict,
                    normed=normed,
                    q_squared_hist_variable=q2_var,
                )

                if dry_run:
                    continue

                mlc1 = mask_label_color_combinations[0]
                assert not df.loc[mlc1.mask, q2_var.df_label].isnull().any(), (q2_var.df_label, mlc1.label)

                shp1 = SimpleHistogramPlot(variable=q2_var)
                self.add_component_to_plot(hp=shp1, df=df, mlc=mlc1, is_resolution=is_resolution)
                self.plot_on_new_figure(
                    shp=shp1,
                    normed=normed,
                    key=q2_residual_plot_key,
                    is_resolution=is_resolution,
                )

                shp2 = SimpleHistogramPlot(variable=q2_var)
                for mlc in mask_label_color_combinations:
                    assert not df.loc[mlc.mask, q2_var.df_label].isnull().any(), (q2_var.df_label, mlc.label)
                    self.add_component_to_plot(hp=shp2, df=df, mlc=mlc, is_resolution=is_resolution)
                self.plot_on_new_figure(
                    shp=shp2,
                    normed=normed,
                    key=q2_residual_comparison_plot_key,
                    is_resolution=is_resolution,
                    add_legend=True,
                )

                for mlc in mask_label_color_combinations[1:]:
                    combination_key = self.get_file_name_for_comparison_plot(
                        normed=normed,
                        q_squared_hist_variable=q2_var,
                        tag_1=mlc1.tag,
                        tag_2=mlc.tag,
                    )  # type: str

                    shp3 = SimpleHistogramPlot(variable=q2_var)
                    self.add_component_to_plot(hp=shp3, df=df, mlc=mlc1, is_resolution=is_resolution)
                    self.add_component_to_plot(hp=shp3, df=df, mlc=mlc, is_resolution=is_resolution)
                    self.plot_on_new_figure(
                        shp=shp3,
                        normed=normed,
                        key=combination_key,
                        is_resolution=is_resolution,
                        add_legend=True,
                    )

        return output_dict

    def add_component_to_plot(
        self,
        hp: SimpleHistogramPlot,
        df: pd.DataFrame,
        mlc: MaskLabelColorCombination,
        is_resolution: bool,
    ) -> None:
        v_min, v_max = hp.variable.scope  # type: float, float
        value_mask = (df[hp.variable.df_label] >= v_min) & (df[hp.variable.df_label] <= v_max)  # type: pd.Series
        rms = np.sqrt(np.mean(np.square(df.loc[mlc.mask & value_mask, hp.variable.df_label].values)))  # type: float
        hp.add_component(
            label=self.label_base_string_map[is_resolution] + mlc.label + f" (RMS={rms:.2f})",
            data=df.loc[mlc.mask],
            color=mlc.color,
            hist_type="step",
        )

    def plot_on_new_figure(
        self,
        shp: SimpleHistogramPlot,
        normed: bool,
        key: str,
        is_resolution: bool,
        add_legend: bool = False,
    ) -> None:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 5), dpi=200)
        shp.plot_on(ax=ax, draw_legend=False, normed=normed)

        if add_legend:
            ax.legend(frameon=False, loc="best")
            y_axis_scale = 1.2 if "_comparison_" in key else 1.5
        else:
            y_axis_scale = 1.1

        y_limits = ax.get_ylim()
        ax.set_ylim(bottom=y_limits[0], top=y_axis_scale * y_limits[1])

        if is_resolution:
            ax.set_title(label=LatexObsNames.fitting_q2 + r" Resolution", loc="left")
        else:
            ax.set_title(label=LatexObsNames.fitting_q2 + r" Residual", loc="left")
        ax.set_title(label=LatexDecayProcesses.b_to_d_st_l_nu, loc="right")

        if self.arb_units:
            normed_tag = " (Normalized)" if normed else ""
            ax.set_ylabel(f"Arb. Units{normed_tag}", ylabel_pos)
        shp.write_hist_data_to_file(file_path=os.path.join(self.output_dir_path, f"{key}_hist_data.h5"))
        export(fig=fig, filename=key, target_dir=self.output_dir_path, close_figure=True)

    def get_file_name_and_append_to_output_dict(
        self,
        output_dict: Dict[str, PathType],
        normed: bool,
        q_squared_hist_variable: HistVariable,
    ) -> Tuple[str, str]:
        base_name = f"qSquared_{q_squared_hist_variable.df_label}"  # type: str
        suffix = self.get_suffix_tag(normed=normed)  # type: str
        q2_residual_plot_name = f"{base_name}_gen_matched{suffix}"  # type: str
        q2_residual_comparison_plot_name = f"{base_name}_various_matchings{suffix}"  # type: str

        keys = [q2_residual_plot_name, q2_residual_comparison_plot_name]
        extensions = ["pdf", "png"]

        for key, extension in itertools.product(keys, extensions):
            plot_path = os.path.join(self.output_dir_path, f"{key}.{extension}")
            output_key = f"{key}_{extension}"
            assert output_key not in output_dict.keys(), (key, extension, output_key)
            assert plot_path not in output_dict.values(), (key, extension, plot_path)
            output_dict.update({output_key: plot_path})

        return q2_residual_plot_name, q2_residual_comparison_plot_name

    def get_file_name_for_comparison_plot(
        self,
        normed: bool,
        q_squared_hist_variable: HistVariable,
        tag_1: str,
        tag_2: str,
    ) -> str:
        suffix = self.get_suffix_tag(normed=normed)  # type: str
        return f"qSquared_{q_squared_hist_variable.df_label}_comparison_{tag_1}_vs_{tag_2}{suffix}"

    def get_mask_label_color_combinations(self, df: pd.DataFrame) -> List[MaskLabelColorCombination]:
        is_sig_mask = df.SIG_ID.isin([2, 4])  # type: pd.Series
        is_dst_reco_mask = df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(self.dst_reco_mode_ids)
        is_dst_mask = df[self.data_type_col_name].isin(self.d_st_data_type_keys)  # type: pd.Series
        is_gen_mask = (df.GEN_B1_ID == 4) | (df.GEN_B2_ID == 4)  # type: pd.Series

        base_mask = is_sig_mask & is_dst_mask & is_gen_mask  # type: pd.Series

        if not self.with_taus:
            is_decay_via_lep_mask = df[QSquaredResidualVariable.b_mdst_col_name].isin(
                QSquaredResidualVariable.valid_b_mdst_indices
            )  # type: pd.Series
            base_mask &= is_decay_via_lep_mask

        if self.only_dst_reco_channels:
            base_mask &= is_dst_reco_mask

        is_sig_b_signal_mask = df.daughter__bo1__cm__spisSignalAcceptMissingNeutrino__bc == 1  # type: pd.Series
        is_tag_b_signal_mask = df.daughter__bo0__cm__spisSignalAcceptMissingNeutrino__bc == 1  # type: pd.Series
        is_y4s_signal_mask = df.isSignalAcceptMissingNeutrino == 1  # type: pd.Series
        tag_sig_prob_mask = (
            df.daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc >= self.tag_sig_prob_cut
        )  # type: pd.Series

        mask_label_color_combinations = [
            MaskLabelColorCombination(
                mask=base_mask,
                label="",
                tag="default",
                color=KITColors.kit_green,
            ),
            MaskLabelColorCombination(
                mask=base_mask & is_sig_b_signal_mask,
                label=r" w/ " + LatexParticleNames.b_sig_str + " isSig",
                tag="sig_is_sig",
                color=KITColors.kit_blue,
            ),
            MaskLabelColorCombination(
                mask=base_mask & is_tag_b_signal_mask,
                label=r" w/ " + LatexParticleNames.b_tag_str + " isSig",
                tag="tag_is_sig",
                color=KITColors.kit_red,
            ),
            MaskLabelColorCombination(
                mask=base_mask & is_y4s_signal_mask,
                label=" w/ " + LatexParticleNames.y4s_str + " isSig",
                tag="y4s_is_sig",
                color=KITColors.kit_orange,
            ),
            MaskLabelColorCombination(
                mask=base_mask & tag_sig_prob_mask,
                label=r" w/ " + LatexObsNames.fei_prob[:-1] + " >\,{self.tag_sig_prob_cut}$",  # noqa: W605
                tag="tag_cut",
                color=KITColors.kit_maygreen,
            ),
            MaskLabelColorCombination(
                mask=base_mask & tag_sig_prob_mask & is_sig_b_signal_mask,
                label=r" w/ "
                + LatexParticleNames.b_sig_str
                + " isSig and "
                + LatexObsNames.fei_prob[:-1]
                + "\,>\,$"  # noqa: W605
                + rf"${self.tag_sig_prob_cut}$",  # noqa: W605
                tag="sig_is_sig_and_tag_cut",
                color=KITColors.kit_purple,
            ),
        ]  # type: List[MaskLabelColorCombination]
        return mask_label_color_combinations

    def get_suffix_tag(self, normed: bool) -> str:
        tau_tag = "_with_taus" if self.with_taus else "_light_leptons"  # type: str
        reco_tag = "_only_dst_modes" if self.only_dst_reco_channels else ""  # type: str
        normed_tag = "_normed" if normed else ""  # type: str
        if self.arb_units:
            normed_tag += "_arb_units"

        return f"{tau_tag}{reco_tag}{normed_tag}"
