"""
Defines plots for several weights applied.

Patrick Ecker 2020
"""

import os
import logging
import itertools
import pandas as pd

from typing import Union, Tuple, List, Dict, Any, Callable, Optional, Iterable, NamedTuple

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.utilities import PathType
from rdstar.offline_analysis.rdstar_samples import main_sample_keys
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.analysis_plots.analysis_plot_base import AnalysisPlotBaseClass
from rdstar.offline_analysis.histogram_variables import missing_mass_res_corr_q2sb, pl_star
from rdstar.offline_analysis.weight_management.weight_information import WeightManager, WeightTypeInfo
from rdstar.offline_analysis.validation.sideband_validation import plot_data_mc_comparison, SidebandInfo
from rdstar.offline_analysis.rdstar_plots import plot_stacked_for_all_modes, DecayModeDictType, DecayTitleDictType

from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "WeightPlotter",
]


class WeightPlotSettings(NamedTuple):
    dec_modes: Optional[DecayModeDictType] = None
    dec_titles: Optional[DecayTitleDictType] = None
    plot_order: Optional[List] = None


class WeightPlotter(AnalysisPlotBaseClass):
    _plotter_id_key = "weight_analysis_plotter"

    _plotting_variables = {
        "missing_mass": missing_mass_res_corr_q2sb,
        "pl_star": pl_star,
    }  # type: Dict[str, HistVariable]

    _sidebands = [
        None,
        "q2_sideband",
    ]  # type: List[Union[str, None]]

    _plot_settings = {
        "default": None,
    }  # type: Dict[str, Union[WeightPlotSettings, None]]

    base_dir_name = "WeightPlots"

    weight_appendix = f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"  # type: str

    data_key = "DataY4S"  # type: str

    combined_weight_column_name = "combined_weight_for_validation"  # type: str

    weights_for_validation = WeightManager.weight_collection.weights_for_validation  # type: Tuple[WeightTypeInfo, ...]

    sideband_info_text = {
        "q2_sideband": {
            "sideband_title": rf"{LatexObsNames.fitting_q2} Sideband",
            "sideband_text": rf"{LatexObsNames.fitting_q2}$ < 4.0$ {LatexUnits.gev_sq}",
        }
    }  # type: Dict[str, Dict[str, str]]

    def __init__(
        self,
        output_dir_path: PathType,
        input_dictionary: Dict[str, PathType],
        selection_name: str,
        cs_tag: str,
        variable: str,
        plot_settings_name: str,
        sideband_name: Optional[str] = None,
    ) -> None:
        super().__init__(
            output_dir_path=output_dir_path,
            input_dictionary=input_dictionary,
            selection_name=selection_name,
            cs_tag=cs_tag,
        )

        self.variable = WeightPlotter._plotting_variables[variable]  # type: HistVariable

        self.sideband_name = sideband_name  # type: Optional[str]
        if self.sideband_name:
            assert self.sideband_name in self.sideband_info_text.keys()

        self.plot_settings_name = plot_settings_name  # type: str

        plot_settings = WeightPlotter._plot_settings[self.plot_settings_name]
        if plot_settings is None:
            plot_settings = WeightPlotSettings()

        self.dec_modes = plot_settings.dec_modes  # type: Optional[DecayModeDictType]
        self.dec_titles = plot_settings.dec_titles  # type: Optional[DecayTitleDictType]
        self.plot_order = plot_settings.plot_order  # type: Optional[List]

    @staticmethod
    def plotter_id_key() -> str:
        if not WeightPlotter._plotter_id_key:
            raise RuntimeError(
                f"Plotter {WeightPlotter.__class__.__name__} has no _plotter_id_key" f" {WeightPlotter._plotter_id_key}"
            )
        return WeightPlotter._plotter_id_key

    @staticmethod
    def get_plotter_options(**kwargs) -> Dict[str, List[Any]]:
        selection_name_argument_name = "selection_name"  # type: str
        cs_tag_argument_name = "cs_tag"  # type: str
        if selection_name_argument_name not in kwargs.keys():
            raise TypeError(f"Missing keyword argument '{selection_name_argument_name}' in call to get_plotter_options.")
        if cs_tag_argument_name not in kwargs.keys():
            raise TypeError(f"Missing keyword argument '{cs_tag_argument_name}' in call to get_plotter_options.")

        selection_name = kwargs[selection_name_argument_name]  # type: str
        cs_tag = kwargs[cs_tag_argument_name]  # type: str
        if not isinstance(selection_name, str):
            raise TypeError(
                f"Expected argument {selection_name_argument_name} to be of type str, "
                f"but got type {type(selection_name).__name__}"
            )
        if not isinstance(cs_tag, str):
            raise TypeError(
                f"Expected argument {cs_tag_argument_name} to be of type str, " f"but got type {type(cs_tag).__name__}"
            )

        selection_tags = SelectionSetups.get_selection_tags(selection_name=selection_name, cs_tag=cs_tag)
        sb_mapping = {sb: "" if sb is None else f"_{sb}" for sb in WeightPlotter._sidebands}
        available_sidebands = [
            sb for sb, sb_tag in sb_mapping.items() if f"{selection_name}_with_CS_{cs_tag}{sb_tag}" in selection_tags
        ]

        return {
            "variable": list(WeightPlotter._plotting_variables.keys()),
            "sideband_name": available_sidebands,
            "plot_settings_name": list(WeightPlotter._plot_settings.keys()),
        }

    @staticmethod
    def get_required_task_type() -> Callable:
        return ResolutionCorrectionTask

    def get_relevant_input(self) -> Dict[str, PathType]:
        sideband_appendix = f"_{self.sideband_name}" if self.sideband_name else ""
        comb_appendix = f"{self.weight_appendix}_{self.selection_name}"
        keys = [f"{dt}_{comb_appendix}_with_CS_{self.cs_tag}{sideband_appendix}" for dt in main_sample_keys]

        relevant_input_dict = {
            k: p for k, p in self.input_dictionary.items() if (k in keys and os.path.splitext(p)[1] == ".h5")
        }
        assert len(relevant_input_dict) == len(main_sample_keys), (
            len(relevant_input_dict),
            len(main_sample_keys),
            relevant_input_dict.keys(),
        )
        return relevant_input_dict

    def load_data(self, dry_run: bool = False) -> pd.DataFrame:
        if dry_run or not self.sideband_name:
            return pd.DataFrame()

        assert self.sideband_name is not None, f"Data should be used on sideband, but was used on {self.sideband_name}!"

        sideband_appendix = f"_{self.sideband_name}"
        comb_appendix = f"{self.weight_appendix}_{self.selection_name}"
        key = f"{self.data_key}_{comb_appendix}_with_CS_{self.cs_tag}{sideband_appendix}"

        df = pd.read_hdf(self.input_dictionary[key])
        assert isinstance(df, pd.DataFrame)

        return df

    def create_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        df_mc = self.load_df(dry_run=dry_run)  # type: pd.DataFrame
        df_data = self.load_data(dry_run=dry_run)  # type: pd.DataFrame

        output_list = self.plot_with_weight_for_all_modes(
            df=df_mc,
            weight_info=None,
            dry_run=dry_run,
        )
        assert all([k not in output_dict.keys() for k, p in output_list]), (output_dict.keys(), output_list)
        output_dict.update({k: p for k, p in output_list})

        all_weight_combinations = self.get_all_weight_info_combinations()
        for weight_infos in all_weight_combinations:
            output_list = self.plot_with_list_of_weights_for_all_modes(
                df=df_mc,
                weight_infos=weight_infos,
                dry_run=dry_run,
            )
            assert all([k not in output_dict.keys() for k, p in output_list]), (output_dict.keys(), output_list)
            output_dict.update({k: p for k, p in output_list})

            if self.sideband_name is not None:
                plot_output_dict = self.plot_data_mc_comparison_for_list_of_weights(
                    df_mc=df_mc,
                    df_data=df_data,
                    weight_infos=weight_infos,
                    dry_run=dry_run,
                )
                assert all([k not in output_dict.keys() for k in plot_output_dict.keys()]), (
                    output_dict.keys(),
                    plot_output_dict.keys,
                )
                output_dict.update(plot_output_dict)

        return output_dict

    def get_all_weight_info_combinations(
        self,
        min_number: int = 1,
        minimal_setup: bool = True,
    ) -> List[Tuple[WeightTypeInfo, ...]]:
        all_weight_combinations = []
        if minimal_setup:
            actual_weights = []
            for weight in self.weights_for_validation:
                actual_weights.append(weight)
                if len(actual_weights) >= min_number:
                    all_weight_combinations.append(tuple(actual_weights))
        else:
            for i in range(min_number, len(self.weights_for_validation) + 1):
                combos = [x for x in itertools.combinations(self.weights_for_validation, i)]
                all_weight_combinations.extend(combos)
        return all_weight_combinations

    def plot_with_weight_for_all_modes(
        self,
        df: pd.DataFrame,
        weight_info: Optional[WeightTypeInfo] = None,
        dry_run: bool = False,
    ) -> List[Tuple[str, PathType]]:
        sideband_appendix = f"_{self.sideband_name}" if self.sideband_name is not None else ""
        weight = f"_{weight_info.name}" if weight_info else ""
        base_output_key = f"weight_analysis_{self.selection_name}{sideband_appendix}_{self.variable.df_label}{weight}_{self.plot_settings_name}"

        output_list = [
            (k, p)
            for _, _, k, p in plot_stacked_for_all_modes(
                df=df,
                variable=self.variable,
                weight_column=weight_info.column_name if weight_info else None,
                additional_text=weight_info.latex_str if weight_info else None,
                dec_modes=self.dec_modes,
                dec_titles=self.dec_titles,
                plot_order=self.plot_order,
                output_name=base_output_key,
                base_output_path=self.output_dir_path,
                close_figure=True,
                dry_run=dry_run,
            )
        ]

        return output_list

    def create_combined_weight_column(
        self,
        df: pd.DataFrame,
        weight_infos: Iterable[WeightTypeInfo],
        dry_run: bool = False,
    ) -> None:
        if not dry_run:
            assert self.combined_weight_column_name not in df.columns, self.combined_weight_column_name
            scale_weight_column = WeightManager.weight_collection.__getitem__("scale_weight").column_name
            df.loc[:, self.combined_weight_column_name] = df[scale_weight_column]
            for weight_info in weight_infos:
                assert weight_info.column_name in df.columns, weight_info.column_name
                df[self.combined_weight_column_name] *= df[weight_info.column_name]

    def reset_combined_weight_column(
        self,
        df: pd.DataFrame,
        dry_run: bool = False,
    ) -> None:
        if not dry_run:
            assert self.combined_weight_column_name in df.columns, self.combined_weight_column_name
            del df[self.combined_weight_column_name]

    def plot_with_list_of_weights_for_all_modes(
        self,
        df: pd.DataFrame,
        weight_infos: Iterable[WeightTypeInfo],
        dry_run: bool = False,
    ) -> List[Tuple[str, PathType]]:
        sideband_appendix = f"_{self.sideband_name}" if self.sideband_name is not None else ""
        all_weights = "_".join([weight_info.name for weight_info in weight_infos])
        weight = f"_{all_weights}"
        base_output_key = f"weight_analysis_{self.selection_name}{sideband_appendix}_{self.variable.df_label}{weight}_{self.plot_settings_name}"

        self.create_combined_weight_column(df=df, weight_infos=weight_infos, dry_run=dry_run)

        weight_str = "\n".join([weight_info.latex_str for weight_info in weight_infos])

        output_list = [
            (k, p)
            for _, _, k, p in plot_stacked_for_all_modes(
                df=df,
                variable=self.variable,
                weight_column=self.combined_weight_column_name,
                additional_text=weight_str,
                dec_modes=self.dec_modes,
                dec_titles=self.dec_titles,
                plot_order=self.plot_order,
                output_name=base_output_key,
                base_output_path=self.output_dir_path,
                close_figure=True,
                dry_run=dry_run,
            )
        ]

        self.reset_combined_weight_column(df=df, dry_run=dry_run)

        return output_list

    def create_sideband_info(
        self,
        used_weights: Iterable[WeightTypeInfo],
    ) -> SidebandInfo:
        assert self.sideband_name is not None
        weight_names_appendix = "_".join([weight_info.name for weight_info in used_weights])  # type: str
        return SidebandInfo(
            key=self.sideband_name,
            hist_var=self.variable,
            title=self.sideband_info_text[self.sideband_name]["sideband_title"],
            sb_text=self.sideband_info_text[self.sideband_name]["sideband_text"],
            file_name_appendix=f"{weight_names_appendix}_{self.plot_settings_name}",
        )

    def plot_data_mc_comparison_for_list_of_weights(
        self,
        df_mc: pd.DataFrame,
        df_data: pd.DataFrame,
        weight_infos: Iterable[WeightTypeInfo],
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        sideband_info = self.create_sideband_info(used_weights=weight_infos)

        self.create_combined_weight_column(
            df=df_mc,
            weight_infos=weight_infos,
            dry_run=dry_run,
        )

        output_dict = plot_data_mc_comparison(
            mc_df=df_mc,
            data_df=df_data,
            weight_col=self.combined_weight_column_name,
            dec_modes=self.dec_modes,
            dec_titles=self.dec_titles,
            plot_order=self.plot_order,
            ratio_type="vs_uncert",
            sideband_info=sideband_info,
            fig_size=(11.0, 10.0),
            export_path=self.output_dir_path,
            dry_run=dry_run,
        )
        assert output_dict is not None

        self.reset_combined_weight_column(df=df_mc, dry_run=dry_run)

        return output_dict
