"""
Tagging Correction for FEI at Belle.

Two versions are provided:
 1) Simple approach described in Judith Schwab's KIT Master Thesis: https://publish.etp.kit.edu/record/21422
 2) Inclusive approach provided by Kilian: https://stash.desy.de/users/klieret/repos/tagcal_applicator/browse

Felix Metzner 2019
"""

import numpy as np
import pandas as pd

from typing import Optional, List, Dict

from tagcal_applicator.tags import get_tags
from tagcal_applicator.convenience import load_cfs_from_tag
from tagcal_applicator.applicator import CfsApplicator, CfsApplicatorResult

__all__ = [
    "TagCorrection",
    "InclusiveTagCorrection",
]


class TagCorrection:
    tagging_correction_name = "__tagging_fei_corr_factor__"  # type: str
    tagging_corr_uncert_name = "__tagging_fei_corr_uncert__"  # type: str

    def __init__(self, correct_only_true_b: bool = False) -> None:
        self.u4s_decay_id_col = "extraInfo__bodecayModeID__bc"  # type: str
        self.tag_b_pdg_col = "daughter__bo0__cm__spPDG__bc"  # type: str
        self.tag_b_mc_pdg_col = "daughter__bo0__cm__spmatchedMC__boPDG__bc__bc"  # type: str

        self.correct_only_true_b = correct_only_true_b

        self.charged_tag_ids = [1001, 1010, 1020]  # type: List[int]
        self.neutral_tag_ids = [1000, 1011, 1021]  # type: List[int]
        self.charged_PDG = 521  # type: int
        self.neutral_PDG = 511  # type: int

        # Correction Factors from Judith Schwab's KIT Master Thesis
        self.charged_factor = 0.810  # type: float
        self.charged_uncert = np.sqrt(0.013 ** 2 + 0.054 ** 2)  # type: float

        self.neutral_factor = 0.853  # type: float
        self.neutral_uncert = np.sqrt(0.064 ** 2 + 0.058 ** 2)  # type: float

    def apply_tagging_correction(self, sample: pd.DataFrame) -> None:
        assert self.u4s_decay_id_col in sample.columns, self.u4s_decay_id_col
        assert self.tag_b_pdg_col in sample.columns, self.tag_b_pdg_col
        assert self.tag_b_mc_pdg_col in sample.columns, self.tag_b_mc_pdg_col

        assert self.tagging_correction_name not in sample.columns, self.tagging_correction_name
        assert self.tagging_corr_uncert_name not in sample.columns, self.tagging_corr_uncert_name

        neutral_mask = sample[self.u4s_decay_id_col].isin(self.neutral_tag_ids)  # type: pd.Series
        charged_mask = sample[self.u4s_decay_id_col].isin(self.charged_tag_ids)  # type: pd.Series
        if self.correct_only_true_b:
            neutral_mask = neutral_mask & (abs(sample[self.tag_b_mc_pdg_col]) == self.neutral_PDG)
            charged_mask = charged_mask & (abs(sample[self.tag_b_mc_pdg_col]) == self.charged_PDG)

        sample.loc[:, self.tagging_correction_name] = 1.0
        sample.loc[:, self.tagging_corr_uncert_name] = 0.0

        sample.loc[neutral_mask, self.tagging_correction_name] = self.neutral_factor
        sample.loc[neutral_mask, self.tagging_corr_uncert_name] = self.neutral_uncert

        sample.loc[charged_mask, self.tagging_correction_name] = self.charged_factor
        sample.loc[charged_mask, self.tagging_corr_uncert_name] = self.charged_uncert

        assert len(sample[self.tagging_correction_name].unique()) == 2
        assert len(sample[self.tagging_corr_uncert_name].unique()) == 2
        assert self.charged_factor in sample[self.tagging_correction_name].unique()
        assert self.neutral_factor in sample[self.tagging_correction_name].unique()
        assert self.charged_uncert in sample[self.tagging_corr_uncert_name].unique()
        assert self.neutral_uncert in sample[self.tagging_corr_uncert_name].unique()


# TODO: Use weights and systematics!
class InclusiveTagCorrection:
    default_tag_calibration_name = "210318_q2_9bins"  # type: str

    tagging_correction_name = "__tagging_inclusive_fei_corr_factor__"  # type: str
    tagging_corr_uncert_prefix = "__tagging_inclusive_fei_corr_uncert_"  # type: str
    tagging_corr_uncert_up_name = "__tagging_inclusive_fei_corr_uncert_up__"  # type: str
    tagging_corr_uncert_down_name = "__tagging_inclusive_fei_corr_uncert_down__"  # type: str

    def __init__(self, tag_calibration_name: Optional[str] = None, correct_only_true_b: bool = False) -> None:
        if tag_calibration_name is None:
            self.tag_cal_name = self.default_tag_calibration_name  # type: str
        else:
            if tag_calibration_name not in get_tags():
                # We don't support calibration factors from files, atm...
                raise KeyError(f"A inclusive tag calibration with the tag name {tag_calibration_name} is not known!")
            self.tag_cal_name = tag_calibration_name

        self.u4s_decay_id_col = "extraInfo__bodecayModeID__bc"  # type: str
        self.tag_b_pdg_col = "daughter__bo0__cm__spPDG__bc"  # type: str
        self.tag_b_mc_pdg_col = "daughter__bo0__cm__spmatchedMC__boPDG__bc__bc"  # type: str
        self.fei_decay_mode_id_col = "daughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str
        self.fei_output_col = "daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc"  # type: str

        self.temp_tag_cal_column_name = "__tmp_inclusig_fei_tag_calib_column__"  # type: str

        self.correct_only_true_b = correct_only_true_b  # type: bool

        self.charged_tag_ids = [1001, 1010, 1020]  # type: List[int]
        self.neutral_tag_ids = [1000, 1011, 1021]  # type: List[int]
        self.charged_PDG = 521  # type: int
        self.neutral_PDG = 511  # type: int

    def apply_tagging_correction(self, sample: pd.DataFrame) -> None:
        assert self.u4s_decay_id_col in sample.columns, self.u4s_decay_id_col
        assert self.tag_b_pdg_col in sample.columns, self.tag_b_pdg_col
        assert self.tag_b_mc_pdg_col in sample.columns, self.tag_b_mc_pdg_col
        assert self.fei_decay_mode_id_col in sample.columns, self.fei_decay_mode_id_col
        assert self.fei_output_col in sample.columns, self.fei_output_col

        assert self.tagging_correction_name not in sample.columns, self.tagging_correction_name
        assert self.tagging_corr_uncert_up_name not in sample.columns, self.tagging_corr_uncert_up_name
        assert self.tagging_corr_uncert_down_name not in sample.columns, self.tagging_corr_uncert_down_name
        assert self.temp_tag_cal_column_name not in sample.columns, self.temp_tag_cal_column_name

        cfs = load_cfs_from_tag(tag=self.tag_cal_name)
        cfs_applicator = CfsApplicator(cfs=cfs)

        fei_tag_cal_column_dict = {
            "placeholder_from_tagcal_applicator": self.temp_tag_cal_column_name,
            "fei_dch_id_column": self.fei_decay_mode_id_col,
            "tag_flavor": self.tag_b_pdg_col,
            "fei_classifier_output_column": self.fei_output_col,
        }  # type: Dict[str, str]

        result = cfs_applicator.get_weights(
            mcdf=sample,
            query_format_dict=fei_tag_cal_column_dict,
            show_progress=False,
        )  # type: CfsApplicatorResult

        sample.loc[:, self.tagging_correction_name] = result.weight_nom
        sample.loc[:, self.tagging_corr_uncert_up_name] = result.weight_err_up
        sample.loc[:, self.tagging_corr_uncert_down_name] = result.weight_err_down

        if self.correct_only_true_b:
            neutral_mask = sample[self.u4s_decay_id_col].isin(self.neutral_tag_ids)  # type: pd.Series
            charged_mask = sample[self.u4s_decay_id_col].isin(self.charged_tag_ids)  # type: pd.Series
            not_mc_neutral_mask = abs(sample[self.tag_b_mc_pdg_col]) != self.neutral_PDG  # type: pd.Series
            not_mc_charged_mask = abs(sample[self.tag_b_mc_pdg_col]) != self.charged_PDG  # type: pd.Series
            not_true_neutral_mask = neutral_mask & not_mc_neutral_mask  # type: pd.Series
            not_true_charged_mask = charged_mask & not_mc_charged_mask  # type: pd.Series

            not_correct_b_tag = not_true_charged_mask | not_true_neutral_mask  # type: pd.Series

            sample.loc[not_correct_b_tag, self.tagging_correction_name] = 1.0
            sample.loc[not_correct_b_tag, self.tagging_corr_uncert_up_name] = 0.0
            sample.loc[not_correct_b_tag, self.tagging_corr_uncert_down_name] = 0.0

        assert not np.any(np.isnan(sample[self.tagging_correction_name].unique()))
        assert not np.any(np.isnan(sample[self.tagging_corr_uncert_up_name].unique()))
        assert not np.any(np.isnan(sample[self.tagging_corr_uncert_down_name].unique()))
