from typing import Dict, List, Optional

from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import (
    deltaM_charged_B_charged_pi_fit_case,
    deltaM_neutral_B_charged_pi_fit_case,
    deltaM_charged_B_neutral_pi_fit_case,
    deltaM_neutral_B_neutral_pi_fit_case,
    twoD_deltaM_charged_B_charged_pi_fit_case,
    twoD_deltaM_charged_B_neutral_pi_fit_case,
    twoD_deltaM_neutral_B_charged_pi_fit_case,
    twoD_deltaM_neutral_B_neutral_pi_fit_case,
)
from rdstar.offline_analysis.dss_fitting.utilities import dss_bf_fitting_utils
from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils
from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import DataTypeFraction

__all__ = [
    "BFFittingSetup",
    "BFFittingSetupCollection",
]


class BFFittingSetup:
    def __init__(
        self,
        name: str,
        fit_case: FitCase,
        decay_titles: Dict[str, str],
        yield_parameters: List[str],
        norm_parameter: str,
        norm_bf: float,
        sig_id_data_type_mapping: Dict[int, DataTypeFraction],
        fix_nui_params: Optional[bool] = True,
        nominal_bfs: Optional[Dict[str, float]] = None,
    ) -> None:

        self._name = name
        self._fit_case = fit_case
        self._decay_titles = decay_titles
        self._yield_parameters = yield_parameters
        self._norm_parameter = norm_parameter
        self._sig_id_data_type_mapping = sig_id_data_type_mapping
        self._norm_bf = norm_bf
        self._fix_nui_params = fix_nui_params
        self._nominal_bfs = nominal_bfs

    @property
    def name(self) -> str:
        return self._name

    @property
    def fit_case(self) -> FitCase:
        return self._fit_case

    @property
    def decay_titles(self) -> Dict[str, str]:
        return self._decay_titles

    @property
    def yield_parameters(self) -> List[str]:
        return self._yield_parameters

    @property
    def norm_parameter(self) -> str:
        return self._norm_parameter

    @property
    def sig_id_data_type_mapping(self) -> Dict[int, DataTypeFraction]:
        return self._sig_id_data_type_mapping

    @property
    def norm_bf(self) -> float:
        return self._norm_bf

    @property
    def fix_nui_params(self) -> bool:
        assert self._fix_nui_params is not None
        return self._fix_nui_params

    @property
    def nominal_bfs(self) -> Dict[str, float]:
        assert self._nominal_bfs is not None
        return self._nominal_bfs


deltaM_charged_B_charged_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="deltaM_charged_B_charged_pi_bf_asimov_fitting_setup",
    fit_case=deltaM_charged_B_charged_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBzeroDs",
    norm_bf=(0.05110 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.charged_B_charged_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

deltaM_neutral_B_charged_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="deltaM_neutral_B_charged_pi_bf_asimov_fitting_setup",
    fit_case=deltaM_neutral_B_charged_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBchargedDs",
    norm_bf=(0.05490 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.neutral_B_charged_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

deltaM_charged_B_neutral_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="deltaM_charged_B_neutral_pi_bf_asimov_fitting_setup",
    fit_case=deltaM_charged_B_neutral_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedDs",
    norm_bf=(0.05490 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.charged_B_neutral_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

deltaM_neutral_B_neutral_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="deltaM_neutral_B_neutral_pi_bf_asimov_fitting_setup",
    fit_case=deltaM_neutral_B_neutral_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroDs",
    norm_bf=(0.05110 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.neutral_B_neutral_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

twoD_charged_B_charged_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="twoD_charged_B_charged_pi_bf_asimov_fitting_setup",
    fit_case=twoD_deltaM_charged_B_charged_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBzeroDs",
    norm_bf=(0.05110 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.charged_B_charged_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

twoD_neutral_B_charged_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="twoD_neutral_B_charged_pi_bf_asimov_fitting_setup",
    fit_case=twoD_deltaM_neutral_B_charged_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBchargedDs",
    norm_bf=(0.05490 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.neutral_B_charged_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

twoD_charged_B_neutral_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="twoD_charged_B_neutral_pi_bf_asimov_fitting_setup",
    fit_case=twoD_deltaM_charged_B_neutral_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedDs",
    norm_bf=(0.05490 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.charged_B_neutral_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

twoD_neutral_B_neutral_pi_bf_asimov_fitting_setup = BFFittingSetup(
    name="twoD_neutral_B_neutral_pi_bf_asimov_fitting_setup",
    fit_case=twoD_deltaM_neutral_B_neutral_pi_fit_case,
    decay_titles=fitting_utils.dss_labels,
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroDs",
    norm_bf=(0.05110 * 2),
    sig_id_data_type_mapping=dss_bf_fitting_utils.neutral_B_neutral_pi_sig_dss_data_type_mapping,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)


class BFFittingSetupCollection:
    available_setups = {
        deltaM_charged_B_charged_pi_bf_asimov_fitting_setup.name: deltaM_charged_B_charged_pi_bf_asimov_fitting_setup,
        deltaM_neutral_B_charged_pi_bf_asimov_fitting_setup.name: deltaM_neutral_B_charged_pi_bf_asimov_fitting_setup,
        deltaM_charged_B_neutral_pi_bf_asimov_fitting_setup.name: deltaM_charged_B_neutral_pi_bf_asimov_fitting_setup,
        deltaM_neutral_B_neutral_pi_bf_asimov_fitting_setup.name: deltaM_neutral_B_neutral_pi_bf_asimov_fitting_setup,
        twoD_charged_B_charged_pi_bf_asimov_fitting_setup.name: twoD_charged_B_charged_pi_bf_asimov_fitting_setup,
        twoD_charged_B_neutral_pi_bf_asimov_fitting_setup.name: twoD_charged_B_neutral_pi_bf_asimov_fitting_setup,
        twoD_neutral_B_charged_pi_bf_asimov_fitting_setup.name: twoD_neutral_B_charged_pi_bf_asimov_fitting_setup,
        twoD_neutral_B_neutral_pi_bf_asimov_fitting_setup.name: twoD_neutral_B_neutral_pi_bf_asimov_fitting_setup,
    }  # type: Dict[str, BFFittingSetup]
