from typing import Dict, Union, Tuple

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.dss_fitting.utilities import dss_bf_fitting_utils
from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetupManager,
)

from templatefitter.plotter.histogram_variable import HistVariable

__all__ = [
    "deltaM_charged_B_charged_pi_fit_case",
    "deltaM_neutral_B_charged_pi_fit_case",
    "deltaM_charged_B_neutral_pi_fit_case",
    "deltaM_neutral_B_neutral_pi_fit_case",
    "twoD_deltaM_charged_B_charged_pi_fit_case",
    "twoD_deltaM_charged_B_neutral_pi_fit_case",
    "twoD_deltaM_neutral_B_charged_pi_fit_case",
    "twoD_deltaM_neutral_B_neutral_pi_fit_case",
]

delta_M_for_bf_fitting_hvar = HistVariable(
    df_label="daughter__bo1__cm__spformula__bodaughterInvM__bo0__cm__sp2__bc__sp__mi__spdaughter__bo0__cm__spInvM__bc__bc__bc",
    n_bins=60,
    scope=(0.2, 0.9),
    var_name=LatexObsNames.d_st_st_delta_m,
    unit=LatexUnits.gev,
)

m2Recoil_for_bf_fitting_var = HistVariable(
    df_label=MissingMassResolutionCorrector.get_corrected_col_name(
        correction_setup_name=ResolutionCorrectionSetupManager.setup_name_to_use
    ),
    n_bins=10,
    scope=(-1.2, 0.4),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

delta_M_for_bf_fitting_binning_dict = {
    "DZ": 20,
    "DM": 20,
    "DSZ": 20,
    "DSM": 20,
    "DZPiZ": 30,
    "DMPiZ": 30,
    "DSZPiZ": 30,
    "DSMPiZ": 30,
    "DZPiM": 40,
    "DPPiM": 40,
    "DSZPiM": 40,
    "DSPPiM": 40,
    "all_modes": 20,
}  # type: Dict[str, Union[int, Tuple[float, ...]]]

delta_M_for_bf_fitting_scope_dict = {
    "DZPiZ": (0.1, 0.9),
    "DMPiZ": (0.1, 0.9),
    "DSZPiZ": (0.2, 0.9),
    "DSMPiZ": (0.2, 0.9),
    "DZPiM": (0.15, 0.9),
    "DPPiM": (0.1, 0.9),
    "DSZPiM": (0.2, 0.9),
    "DSPPiM": (0.2, 0.9),
}  # type: Dict[str, Tuple[float, float]]

m2Recoil_for_bf_fitting_binning_dict = {
    "DZ": 10,
    "DM": 10,
    "DSZ": 10,
    "DSM": 10,
    "DZPiZ": 3,
    "DMPiZ": 3,
    "DSZPiZ": 3,
    "DSMPiZ": 2,
    "DZPiM": 4,
    "DPPiM": 3,
    "DSZPiM": 3,
    "DSPPiM": 3,
    "all_modes": 20,
}  # type: Dict[str, Union[int, Tuple[float, ...]]]

m2Recoil_for_bf_fitting_scope_dict = {
    "DZPiZ": (-0.4, 0.4),
    "DMPiZ": (-0.4, 0.4),
    "DSZPiZ": (-0.4, 0.4),
    "DSMPiZ": (-0.4, 0.4),
    "DZPiM": (-0.67, 0.4),
    "DPPiM": (-0.4, 0.4),
    "DSZPiM": (-0.4, 0.4),
    "DSPPiM": (-0.4, 0.4),
}  # type: Dict[str, Tuple[float, float]]

deltaM_charged_B_charged_pi_fit_case = FitCase(
    name="charged_B_charged_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar,),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.charged_B_charged_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

deltaM_neutral_B_charged_pi_fit_case = FitCase(
    name="neutral_B_charged_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar,),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_charged_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.neutral_B_charged_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

deltaM_charged_B_neutral_pi_fit_case = FitCase(
    name="charged_B_neutral_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar,),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_neutral_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.charged_B_neutral_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

deltaM_neutral_B_neutral_pi_fit_case = FitCase(
    name="neutral_B_neutral_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar,),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_neutral_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.neutral_B_neutral_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

# 2D Fit cases
twoD_deltaM_charged_B_charged_pi_fit_case = FitCase(
    name="twoD_charged_B_charged_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar, m2Recoil_for_bf_fitting_var),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict, m2Recoil_for_bf_fitting_binning_dict),
    scope_by_mode_dicts=(delta_M_for_bf_fitting_scope_dict, m2Recoil_for_bf_fitting_scope_dict),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.charged_B_charged_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

twoD_deltaM_neutral_B_charged_pi_fit_case = FitCase(
    name="twoD_neutral_B_charged_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar, m2Recoil_for_bf_fitting_var),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict, m2Recoil_for_bf_fitting_binning_dict),
    scope_by_mode_dicts=(delta_M_for_bf_fitting_scope_dict, m2Recoil_for_bf_fitting_scope_dict),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_charged_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.neutral_B_charged_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

twoD_deltaM_charged_B_neutral_pi_fit_case = FitCase(
    name="twoD_charged_B_neutral_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar, m2Recoil_for_bf_fitting_var),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict, m2Recoil_for_bf_fitting_binning_dict),
    scope_by_mode_dicts=(delta_M_for_bf_fitting_scope_dict, m2Recoil_for_bf_fitting_scope_dict),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_neutral_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.charged_B_neutral_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)

twoD_deltaM_neutral_B_neutral_pi_fit_case = FitCase(
    name="twoD_neutral_B_neutral_pi_dss_bf_asimov",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(delta_M_for_bf_fitting_hvar, m2Recoil_for_bf_fitting_var),
    binning_by_mode_dicts=(delta_M_for_bf_fitting_binning_dict, m2Recoil_for_bf_fitting_binning_dict),
    scope_by_mode_dicts=(delta_M_for_bf_fitting_scope_dict, m2Recoil_for_bf_fitting_scope_dict),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_neutral_pi_modes.items()],
    template_setup=dss_bf_fitting_utils.neutral_B_neutral_pi_sig_dss_plot_order,
    plot_colors=dss_bf_fitting_utils.sig_dss_colors,
    plot_labels=dss_bf_fitting_utils.sig_dss_labels,
    sig_id_column_name="dss_sig_id",
    sig_queries=dss_bf_fitting_utils.sig_dss_components,
)
