from typing import Optional, Dict

from rdstar.offline_analysis.fitting.fit_case import FitCase, BinningDictType, ScopeDictType
import rdstar.offline_analysis.histogram_variables as rdstar_hist_vars
from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils
from rdstar.offline_analysis.dss_fitting.utilities import dss_ff_study_utils

__all__ = [
    "FitCaseDict",
    # Asimov
    "reduced_dss_asimov_fit_case_dict_for_DPiZ",
    "reduced_dss_asimov_fit_case_dict_for_DSPiZ",
    "reduced_dss_asimov_fit_case_dict_for_DPiP",
    "reduced_dss_asimov_fit_case_dict_for_DSPiP",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DPiP",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiP",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DPiZ",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiZ",
    # Data
]


class FitCaseDict:
    def __init__(
        self,
        pivot_fit_case: FitCase,
        special_scope_per_bin: Optional[Dict[int, ScopeDictType]] = None,
        special_binning_per_bin: Optional[Dict[int, BinningDictType]] = None,
    ) -> None:

        self._pivot_fit_case = pivot_fit_case
        self._special_scope_per_bin = special_scope_per_bin
        self._special_binning_per_bin = special_binning_per_bin

    def get_scope_in_bin(
        self,
        bin_number: int,
    ) -> ScopeDictType:
        if not self._special_scope_per_bin:
            return self._pivot_fit_case.scope_by_mode_dicts
        if bin_number not in self._special_scope_per_bin.keys():
            return self._pivot_fit_case.scope_by_mode_dicts
        else:
            return self._special_scope_per_bin[bin_number]

    def get_binning_in_bin(
        self,
        bin_number: int,
    ) -> BinningDictType:
        if not self._special_binning_per_bin:
            return self._pivot_fit_case.binning_by_mode_dicts
        if bin_number not in self._special_binning_per_bin.keys():
            return self._pivot_fit_case.binning_by_mode_dicts
        else:
            return self._special_binning_per_bin[bin_number]

    def get_fit_case_in_bin(
        self,
        bin_number: int,
    ) -> FitCase:
        scope_by_mode_dicts = self.get_scope_in_bin(bin_number=bin_number)
        binning_by_mode_dicts = self.get_binning_in_bin(bin_number=bin_number)
        return FitCase(
            name=self.pivot_fit_case.name,
            base_dir_name=self.pivot_fit_case.base_dir_name,
            sample_key=self.pivot_fit_case.sample_key,
            output_tag=self.pivot_fit_case.output_tag,
            plot_info_text=self.pivot_fit_case.plot_info_text,
            asimov_data=self.pivot_fit_case.asimov_data,
            round_asimov_data=self.pivot_fit_case.round_asimov_data,
            hist_variables=self.pivot_fit_case.hist_variables,
            binning_by_mode_dicts=binning_by_mode_dicts,
            scope_by_mode_dicts=scope_by_mode_dicts,
            weight_column_name=self.pivot_fit_case.weight_column_name,
            systematics_cols=self.pivot_fit_case.systematics_cols,
            reco_mode_infos=self.pivot_fit_case.reco_mode_infos,
            template_setup=self.pivot_fit_case.template_setup,
            plot_colors=self.pivot_fit_case.plot_colors,
            plot_labels=self.pivot_fit_case.plot_labels,
            sig_id_column_name=self.pivot_fit_case.sig_id_column_name,
            sig_queries=self.pivot_fit_case.sig_queries,
        )

    @property
    def pivot_fit_case(self) -> FitCase:
        return self._pivot_fit_case

    @property
    def name(self) -> str:
        return self.pivot_fit_case.name


# Asimov Fit Cases and Dicts
# q^2
reduced_dss_asimov_fit_case_for_DPiZ = FitCase(
    name="reduced_dss_asimov_for_DPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiZ.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DPiZ,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DPiZ = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DPiZ,
    special_scope_per_bin={i: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,) for i in range(0, 3)},
)

reduced_dss_asimov_fit_case_for_DSPiZ = FitCase(
    name="reduced_dss_asimov_for_DSPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiZ.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DSPiZ,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DSPiZ = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DSPiZ,
    special_scope_per_bin={i: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,) for i in range(0, 3)},
)

reduced_dss_asimov_fit_case_for_DPiP = FitCase(
    name="reduced_dss_asimov_for_DPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiP.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DPiP,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DPiP = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DPiP,
    special_scope_per_bin={i: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,) for i in range(0, 3)},
)

reduced_dss_asimov_fit_case_for_DSPiP = FitCase(
    name="reduced_dss_asimov_for_DSPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiP.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DSPiP,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DSPiP = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DSPiP,
    special_scope_per_bin={i: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,) for i in range(0, 3)},
)

# pStarL
pStarL_reduced_dss_asimov_fit_case_for_DPiZ = FitCase(
    name="pStarL_reduced_dss_asimov_for_DPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiZ.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DPiZ,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DPiZ = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DPiZ,
    special_scope_per_bin={
        5: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,),
        6: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
        7: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
    },
)

pStarL_reduced_dss_asimov_fit_case_for_DSPiZ = FitCase(
    name="pStarL_reduced_dss_asimov_for_DSPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiZ.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DSPiZ,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiZ = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DSPiZ,
    special_scope_per_bin={
        5: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,),
        6: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
        7: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
    },
)

pStarL_reduced_dss_asimov_fit_case_for_DPiP = FitCase(
    name="pStarL_reduced_dss_asimov_for_DPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiP.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DPiP,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DPiP = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DPiP,
    special_scope_per_bin={
        5: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,),
        6: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
        7: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
    },
)

pStarL_reduced_dss_asimov_fit_case_for_DSPiP = FitCase(
    name="pStarL_reduced_dss_asimov_for_DSPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb,),
    binning_by_mode_dicts=(fitting_utils.dss_missing_mass_fine_binning_dict,),
    scope_by_mode_dicts=(None,),
    weight_column_name=fitting_utils.final_weight_column,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiP.items()],
    template_setup=dss_ff_study_utils.reduced_advanced_dss_plot_order_for_DSPiP,
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiP = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DSPiP,
    special_scope_per_bin={
        5: (fitting_utils.dss_missing_mass_scope_dict_for_bin0,),
        6: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
        7: (fitting_utils.dss_missing_mass_scope_dict_for_advanced_bin0,),
    },
)
