import os
import logging
import b2luigi as luigi
import numpy as np
import matplotlib.pyplot as plt
import copy

from luigi.util import inherits
from typing import Dict, Generator, Tuple, List, Union

from rdstar.offline_analysis.tasks import utilities
from rdstar.utilities import PathType

from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetupCollection
from rdstar.offline_analysis.dss_fitting.tasks.bf_fitting_task import BFFittingTask
from rdstar.offline_analysis.dss_fitting.bf_calculator import BFCalculator
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import EfficiencyNumbersDict

from templatefitter.fitter import MinimizeResult
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter import plot_style

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "BFCalculationTask",
]


@inherits(BFFittingTask)
class BFCalculationTask(utilities.BaseOfflineTask):
    fitting_setup_name = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "bf_results", self.fitting_setup_name)

        self.fitting_setup = BFFittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = self.fitting_setup.fit_case
        self.fit_output_tag = self.fit_case.output_tag

        self.yield_parameters = self.fitting_setup.yield_parameters

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_efficiencies_dict(self) -> EfficiencyNumbersDict:
        eff_path = self.input_dict[f"efficiency_{self.fitting_setup_name}"]
        efficiency_numbers_dict = EfficiencyNumbersDict()
        efficiency_numbers_dict.from_file(file_path=eff_path)
        return efficiency_numbers_dict

    def plot_bfs(
        self,
        bf_calculator: Union[BFCalculator, None],
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:

        output_key = f"{self.fitting_setup_name}_branching_fractions"
        output_path = os.path.join(self.base_dir, f"{output_key}.pdf")

        if dry_run:
            return output_key, output_path
        else:
            assert bf_calculator is not None

        plot_style.set_matplotlibrc_params()
        y_pos = np.arange(len(self.yield_parameters))
        fig, ax = plt.subplots(figsize=(6.5, 5))
        bfs = []
        lower_errors = []
        upper_errors = []
        for yield_parameter in self.yield_parameters:
            bf, upper_error, lower_error = bf_calculator.get_bf_for_parameter(parameter=yield_parameter)
            bfs.append(bf)
            upper_errors.append(upper_error)
            lower_errors.append(lower_error)

        if self.fitting_setup.nominal_bfs:
            bf_dict = self.fitting_setup.nominal_bfs
            plot_colors = self.fit_case.plot_colors
            assert isinstance(plot_colors, dict)
            ax.errorbar(
                x=np.array([bf_dict[yield_param] for yield_param in self.yield_parameters]) * 100,
                y=y_pos,
                yerr=0.4,
                xerr=None,
                ls="",
                elinewidth=4,
                marker="",
                color=[plot_colors[yield_param] for yield_param in self.yield_parameters],
                capsize=0,
            )

        ax.errorbar(
            x=np.array(bfs) * 100,
            y=y_pos,
            yerr=0,
            xerr=(np.array(lower_errors) * 100, np.array(upper_errors) * 100),
            ls="",
            marker=".",
            color="black",
            capsize=3,
            label=self.fit_case.plot_info_text,
        )

        plt.legend(loc="best")
        ax.set_xlabel(r"$\mathcal{B}(B\rightarrow D^{**} \ell \nu)$ in %", plot_style.xlabel_pos)
        ax.set_yticks(y_pos)
        plot_labels = self.fit_case.plot_labels
        assert isinstance(plot_labels, dict)
        ax.set_yticklabels([plot_labels[yield_param] for yield_param in self.yield_parameters])

        ax.set_xlim(0, 1)

        export(fig=fig, filename=f"{output_key}", target_dir=self.base_dir)

        return output_key, output_path

    def plot_bfs_with_scaled_efficiencies(
        self,
        bf_calculator: Union[BFCalculator, None],
        parameter: str,
        scalings: List[Tuple[float, float]],
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        output_key = f"{self.fitting_setup_name}_branching_fractions_scaled_eff_{parameter}"
        output_path = os.path.join(self.base_dir, f"{output_key}.pdf")

        if dry_run:
            return output_key, output_path
        else:
            assert bf_calculator is not None

        plot_style.set_matplotlibrc_params()
        y_pos = np.arange(len(scalings) + 1)
        fig, ax = plt.subplots(figsize=(6.5, 5))

        bf, upper_error, lower_error = bf_calculator.get_bf_for_parameter(parameter=parameter)
        ax.errorbar(
            x=bf * 100,
            y=y_pos[0],
            yerr=0,
            xerr=([lower_error * 100], [upper_error * 100]),
            ls="",
            marker=".",
            color="black",
            capsize=3,
            label=self.fit_case.plot_info_text,
        )
        for i in range(0, len(scalings)):
            scaling = scalings[i]

            scaled_bf_calculator = copy.copy(bf_calculator)
            scaled_bf_calculator.reset()
            scaled_bf_calculator.set_eff_scaling(scale_eff_statistics=scaling)
            scaled_bf_calculator.calculate()
            scaled_bf, scaled_upper_error, scaled_lower_error = scaled_bf_calculator.get_bf_for_parameter(
                parameter=parameter
            )

            ax.errorbar(
                x=scaled_bf * 100,
                y=y_pos[i + 1],
                yerr=0,
                xerr=([scaled_lower_error * 100], [scaled_upper_error * 100]),
                ls="",
                marker=".",
                capsize=3,
                label=f"{self.fit_case.plot_info_text} \n"
                + rf"sig eff $\times$ {scaling[0]}"
                + "\n"
                + rf"norm eff $\times$ {scaling[1]}",
            )

        plt.legend(loc="best")
        ax.set_xlabel(r"$\mathcal{B}(B\rightarrow D^{**} \ell \nu)$ in %", plot_style.xlabel_pos)
        ax.set_yticks(y_pos)
        plot_labels = self.fit_case.plot_labels
        assert isinstance(plot_labels, dict)
        ax.set_title(plot_labels[parameter], loc="right")

        ax.set_xlim(0, 1)

        export(fig=fig, filename=f"{output_key}", target_dir=self.base_dir)

        return output_key, output_path

    def run_bf_calculation(self, dry_run: bool = False) -> Dict[str, PathType]:

        output_dict = dict()  # type: Dict[str, PathType]

        yield_output_key = f"yields_{self.fitting_setup_name}"
        yield_output_path = os.path.join(self.base_dir, f"{yield_output_key}.json")
        output_dict.update({yield_output_key: yield_output_path})

        efficiency_output_key = f"efficiencies_{self.fitting_setup_name}"
        efficiency_output_path = os.path.join(self.base_dir, f"{efficiency_output_key}.json")
        output_dict.update({efficiency_output_key: efficiency_output_path})

        branching_fractions_output_key = f"branching_fractions_{self.fitting_setup_name}"
        branching_fractions_output_path = os.path.join(self.base_dir, f"{branching_fractions_output_key}.json")
        output_dict.update({branching_fractions_output_key: branching_fractions_output_path})

        if dry_run:
            bf_calculator = None
        else:
            efficiencies_dict = self.get_efficiencies_dict()
            fit_result_path = self.input_dict[f"{self.fit_output_tag}_FitResult_{self.fitting_setup_name}"]
            fit_result = MinimizeResult.load_from(path=fit_result_path)

            bf_calculator = BFCalculator(
                fitting_setup=self.fitting_setup,
                fit_result=fit_result,
                efficiency_dict=efficiencies_dict,
            )
            bf_calculator.calculate()
            bf_calculator.save_yields_to_file(file_path=yield_output_path)
            bf_calculator.save_efficiencies_to_file(file_path=efficiency_output_path)
            bf_calculator.save_branching_fractions_to_file(file_path=branching_fractions_output_path)
            print(bf_calculator.eff_dict())
            print(bf_calculator.bf_dict())

        output_key, output_path = self.plot_bfs(
            bf_calculator=bf_calculator,
            dry_run=dry_run,
        )
        output_dict.update({output_key: output_path})

        for parameter in self.yield_parameters:
            output_key, output_path = self.plot_bfs_with_scaled_efficiencies(
                bf_calculator=bf_calculator,
                parameter=parameter,
                scalings=[(1, 10), (10, 10), (35, 400)],
                dry_run=dry_run,
            )
            output_dict.update({output_key: output_path})

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_bf_calculation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_bf_calculation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(BFFittingTask)
