import copy
import os
import logging
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Dict, Generator, Tuple, List

from rdstar.offline_analysis.tasks import utilities
from rdstar.utilities import PathType

from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetupCollection
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import SimpleEfficiencyCalculator
from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.systematics.systematics import SysColsInfo
from rdstar.offline_analysis.dss_fitting.bf_measurement_systematics import pid_systematics

from templatefitter.fitter import TemplateFitter
from templatefitter.fitter import MinimizeResult

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "BFSystematicsTask",
]


@inherits(FittingPreparationTask)
class BFSystematicsTask(utilities.BaseOfflineTask):
    fitting_setup_name = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "bf_systematics", self.fitting_setup_name)

        self.fitting_setup = BFFittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = self.fitting_setup.fit_case

        self.fit_output_tag = self.fit_case.output_tag

        self.mc_key = "combined_mc_prepared"
        # self.mc_key = "asimov_mc_prepared"
        self.data_key = "data"

        self.original_weight_column = copy.copy(self.fit_case.weight_column_name)
        self.varied_weight_column = "varied_weight"  # type: str

        self.yield_parameters = self.fitting_setup.yield_parameters

        self.sys_fit_results = dict()  # type: Dict[str, List[MinimizeResult]]

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_data_path(self, data_key: str) -> PathType:
        input_key = f"{data_key}_{self.selection}"
        return self.input_dict[input_key]

    def get_fit_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}_{self.fitting_setup_name}"
        return key, f"{key}.json"

    def get_fit_plot_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}_{self.fitting_setup_name}"
        return key, f"{key}.pdf"

    def update_fit_case(self) -> None:
        original_fit_case = copy.copy(self.fit_case)
        self.fit_case = FitCase(
            name=original_fit_case.name,
            base_dir_name=original_fit_case.base_dir_name,
            sample_key=original_fit_case.sample_key,
            output_tag=original_fit_case.output_tag,
            plot_info_text=original_fit_case.plot_info_text,
            asimov_data=original_fit_case.asimov_data,
            round_asimov_data=original_fit_case.round_asimov_data,
            hist_variables=original_fit_case.hist_variables,
            binning_by_mode_dicts=original_fit_case.binning_by_mode_dicts,
            scope_by_mode_dicts=original_fit_case.scope_by_mode_dicts,
            weight_column_name=self.varied_weight_column,
            systematics_cols=original_fit_case.systematics_cols,
            reco_mode_infos=original_fit_case.reco_mode_infos,
            template_setup=original_fit_case.template_setup,
            plot_colors=original_fit_case.plot_colors,
            plot_labels=original_fit_case.plot_labels,
            sig_id_column_name=original_fit_case.sig_id_column_name,
            sig_queries=original_fit_case.sig_queries,
            minimizer_id=original_fit_case.minimizer_id,
        )

    def perform_pid_variation(
        self,
        df: pd.DataFrame,
        sys_col_info: SysColsInfo,
    ) -> Generator[Tuple[str, pd.DataFrame], None, None]:
        variations = [var for var in df.columns if var.startswith(SysColsInfo.var)]
        for variation in variations:
            df.loc[:, self.varied_weight_column] = (
                df[self.original_weight_column] / df[sys_col_info.ratio] * df[variation]
            )
            yield variation, df

    def initialize_model_creator(self) -> FitModelCreator:
        FitModelCreator.fit_cases = {self.fit_case.name: self.fit_case}

        model_creator = FitModelCreator(
            fit_case=self.fit_case.name,
        )

        return model_creator

    def run_fit(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]

        efficiency_key = f"efficiency_{self.fitting_setup_name}"
        efficiency_path = os.path.join(self.base_dir, f"{efficiency_key}.json")
        output_dict.update({efficiency_key: efficiency_path})

        if not dry_run:
            df_mc = pd.read_hdf(path_or_buf=self.get_data_path(data_key=self.mc_key))
            assert isinstance(df_mc, pd.DataFrame)
            df_mc.reset_index(drop=True, inplace=True)

            df_data = (
                pd.read_hdf(path_or_buf=self.get_data_path(data_key=self.data_key))
                if not self.fit_case.asimov_data
                else None
            )
            if df_data:
                assert isinstance(df_data, pd.DataFrame)
                df_data.reset_index(drop=True, inplace=True)

            self.update_fit_case()

            for systematic in pid_systematics:
                fit_results = []  # type: List[MinimizeResult]
                for variation, varied_df in self.perform_pid_variation(df=df_mc, sys_col_info=systematic):
                    logging.debug(f"Setting up Fit Model for variation {variation}...")
                    model_creator = self.initialize_model_creator()
                    model_creator.setup_model(
                        mc_df=varied_df,
                        data_df=df_data,
                        use_asimov_data=self.fit_case.asimov_data,
                        round_asimov_data_bin_counts=self.fit_case.round_asimov_data,
                        asimov_data_weight_column=self.original_weight_column,
                    )
                    fit_model = model_creator.fit_model
                    logging.debug("Fit Model set up!")

                    fitter = TemplateFitter(fit_model=fit_model, minimizer_id="iminuit")
                    fit_result = fitter.do_fit(
                        update_templates=True,
                        get_hesse=True,
                        verbose=True,
                        fix_nui_params=self.fitting_setup.fix_nui_params,
                    )
                    fit_results.append(fit_result)
                    logging.debug("Fit performed!")

                self.sys_fit_results.update({systematic.var: fit_results})

            efficiency_calculator = SimpleEfficiencyCalculator(
                df=df_mc,
                sig_id_data_type_mapping=self.fitting_setup.sig_id_data_type_mapping,
                sig_id_column_name=self.fit_case.sig_id_column_name,
            )
            efficiency_calculator.generate_efficiency_numbers_dict()
            efficiency_calculator.write_eff_dict_to_file(file_path=efficiency_path)

        return output_dict

    def run_systematics_calculation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]

        fit_output = self.run_fit(dry_run=dry_run)
        output_dict.update(fit_output)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_systematics_calculation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_fit(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(FittingPreparationTask)
