"""
Main Task for the D** analysis.

Patrick Ecker 2020
"""

import logging
import b2luigi as luigi
import os

from luigi.util import inherits

from typing import Generator

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.dss_fitting.tasks.ratio_plotter import RatioCalculationTask
from rdstar.offline_analysis.dss_fitting.tasks.significance_plotter import SignificancePlotterTask
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.tasks.migration_task import MigrationTask
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection
from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetupCollection
from rdstar.offline_analysis.dss_fitting.tasks.bf_calculation_task import BFCalculationTask

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

mpl_logger = logging.getLogger("matplotlib")
mpl_logger.setLevel(logging.WARNING)
luigi_logger = logging.getLogger("luigi")
luigi_logger.setLevel(logging.WARNING)

ch = logging.StreamHandler()
formatter = logging.Formatter(
    fmt="%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s", datefmt="%Y-%m-%d:%H:%M:%S"
)
ch.setFormatter(formatter)
logger.addHandler(ch)

__all__ = ["FittingMaster"]


@inherits(FittingPreparationTask)
class FittingMaster(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for fitting_setup in FittingSetupCollection.available_setups.keys():
            yield self.clone(RatioCalculationTask, fitting_setup_name=fitting_setup)
            yield self.clone(SignificancePlotterTask, fitting_setup_name=fitting_setup)
            if FittingSetupCollection.available_setups[fitting_setup].fit_in_bins_of.df_label == "qSquaredReco":
                yield self.clone(MigrationTask, fitting_setup_name=fitting_setup)

        for fitting_setup in BFFittingSetupCollection.available_setups.keys():
            yield self.clone(BFCalculationTask, fitting_setup_name=fitting_setup)


if __name__ == "__main__":

    selection = "dss_selection_dss_sec_sel_mva"
    # selection = "dss_selection_full_pl_dss_sec_sel_mva"
    # production_key = "data_set_prod_08thSep2020"
    production_key = "data_set_prod_13thJan2021"

    os.environ["NUMEXPR_MAX_THREADS"] = "40"

    num_workers = 15
    luigi.process(
        FittingMaster(
            selection=selection,
            production_key=production_key,
            working_dir=f"/ceph/pecker/rdstar/fitting/{production_key}/{selection}",
        ),
        workers=num_workers,
    )
