import os
import logging
import pandas as pd
import b2luigi as luigi
import numpy as np

from typing import Dict, List

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.rdstar_samples import main_sample_keys
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable

from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils

# from rdstar.offline_analysis.weight_management.weight_finalization import WeightFinalizer
from rdstar.utilities import PathType

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "FittingPreparationTask",
]


class FittingPreparationTask(utilities.BaseOfflineTask):
    selection = luigi.Parameter(positional=False)  # type: str
    production_key = luigi.Parameter(positional=False)  # type: str
    working_dir = luigi.Parameter(positional=False)  # type: str

    sig_b_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"  # type: str
    sig_b_tau_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
    advanced_sig_id_col = "advanced_sig_id"  # type: str

    decay_mode_col = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str

    dss_pdgs = [10411, 10421, 10413, 10423, 415, 425, 20413, 10423]
    dss_pdgs.extend(list(-np.array(dss_pdgs)))

    dss_sig_id_col = "dss_sig_id"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_input_path = f"/ceph/fmetzner/rdstar/{self.production_key}_stream0/CombinedSamples/{self.selection}"

        self.data_key = "DataY4S"

        self.dss_sig_mc_keys = [
            "DssPiZeroEff",
            "DssPiPlusEff",
            "DssTauPiZEff",
            "DssTauPiPEff",
        ]  # type: List[str]

        self.sample_keys_for_eff_mc = [
            "Charged",
            "Mixed",
            "Dss",
        ]

        self.base_dir = os.path.join(self.working_dir, "samples")

        self.dummy_input_dict = {
            f"pid_hadron_for_flag_dict_{self.selection}": os.path.join(
                f"/ceph/fmetzner/rdstar/data_set_prod_13thJan2021_stream0/Weight_Analysis/{self.selection}_with_CS_CS1/pid_hadron_for_flag_dict.json"
            )
        }  # type: Dict[str, PathType]

    def get_data_path(self, data_type: str) -> PathType:
        file_name = f"{data_type}_withFinalWeights_withMMSQResCorr_{self.selection}_with_CS_CS1.h5"
        return os.path.join(self.base_input_path, file_name)

    def get_eff_path(self, data_type: str) -> PathType:
        base_eff_path = f"/ceph/fmetzner/rdstar/{self.production_key}_stream0/{data_type}"
        file_name = f"{data_type}_Y4S_ntuple_{self.selection}_with_CS_CS1.h5"
        return os.path.join(base_eff_path, file_name)

    @staticmethod
    def apply_weights(df: pd.DataFrame) -> None:
        assert fitting_utils.final_weight_column not in df.columns

        df.loc[:, fitting_utils.final_weight_column] = 1
        for weight_info in WeightManager.weight_collection:
            # todo check which weights have to be used
            if weight_info.name != "b_bf_weight":  # and weight_info.name != "tag_weight":
                df[fitting_utils.final_weight_column] *= df[weight_info.column_name]

    @staticmethod
    def apply_qsquare_infos(df: pd.DataFrame) -> None:
        QSquaredResidualVariable.add_q_squared_residual_definitions(df=df)

    @staticmethod
    def get_cond_dict(df: pd.DataFrame):
        is_D_cond = df.sig_gen_D_pdg.isin([-411, -421, 411, 421])
        is_Ds_cond = df.sig_gen_D_pdg.isin([-413, -423, 413, 423])

        cond_dict = {
            101: (df.SIG_ID == 5) & is_D_cond,
            102: (df.SIG_ID == 5) & is_Ds_cond,
            103: (df.SIG_ID == 6) & is_D_cond,
            104: (df.SIG_ID == 6) & is_Ds_cond,
            105: (df.SIG_ID == 3),
            106: (df.SIG_ID == 4),
            107: (df.SIG_ID.isin([1, 2, 7, 8, 9])),
        }
        return cond_dict

    def create_advanced_sig_id(self, df: pd.DataFrame) -> None:
        test_cond = df[self.sig_b_mdst_col].isin([1, 2])

        df.loc[:, "b_mdst"] = np.nan
        df.loc[test_cond, "b_mdst"] = df[self.sig_b_mdst_col]

        b1_cond = df.b_mdst == 1
        b2_cond = df.b_mdst == 2

        df.loc[:, "sig_gen_D_pdg"] = np.nan
        df.loc[b1_cond, "sig_gen_D_pdg"] = df.loc[
            b1_cond, "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"
        ]
        df.loc[b2_cond, "sig_gen_D_pdg"] = df.loc[
            b2_cond, "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"
        ]

        df.loc[:, self.advanced_sig_id_col] = np.nan
        for advanced_id, cond in self.get_cond_dict(df=df).items():
            df.loc[cond, self.advanced_sig_id_col] = advanced_id

    @staticmethod
    def get_dss_cond_dict(df: pd.DataFrame):
        pi_b1_col = "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmmcDaughter__bo2__cmPDG__bc__bc__bc__bc"
        has_two_pi_b1 = df[pi_b1_col].isin([111, -111, 211, -211])
        pi_b2_col = "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmmcDaughter__bo2__cmPDG__bc__bc__bc__bc"
        has_two_pi_b2 = df[pi_b2_col].isin([111, -111, 211, -211])

        b1_cond = df.b_mdst == 1
        b2_cond = df.b_mdst == 2

        is_dss_b1 = df.genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc.isin(
            FittingPreparationTask.dss_pdgs
        )
        is_dss_b2 = df.genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc.isin(
            FittingPreparationTask.dss_pdgs
        )

        is_charged_B_cond = df.genUpsilon4S__bomcDaughter__bo0__cm__spPDG__bc__bc.isin([521, -521])

        cond_dict = {
            # B+
            201: (df.SIG_ID == 5) & df.sig_gen_id.isin([54]) & is_charged_B_cond,  # Dzerostar -> D(*) pi0
            202: (df.SIG_ID == 5) & df.sig_gen_id.isin([53]) & is_charged_B_cond,  # Doneprime -> D(*) pi0
            203: (df.SIG_ID == 5) & df.sig_gen_id.isin([51]) & is_charged_B_cond,  # Done -> D(*) pi0
            204: (df.SIG_ID == 5) & df.sig_gen_id.isin([52]) & is_charged_B_cond,  # Dtwostar -> D(*) pi0
            205: (df.SIG_ID == 6) & df.sig_gen_id.isin([54]) & is_charged_B_cond,  # Dzerostar -> D(*) pi+
            206: (df.SIG_ID == 6) & df.sig_gen_id.isin([53]) & is_charged_B_cond,  # Doneprime -> D(*) pi+
            207: (df.SIG_ID == 6) & df.sig_gen_id.isin([51]) & is_charged_B_cond,  # Done -> D(*) pi+
            208: (df.SIG_ID == 6) & df.sig_gen_id.isin([52]) & is_charged_B_cond,  # Dtwostar -> D(*) pi+
            209: (df.SIG_ID == 8) & b1_cond & has_two_pi_b1 & is_dss_b1 & is_charged_B_cond,  # Done -> D pi pi
            210: (df.SIG_ID == 8) & b2_cond & has_two_pi_b2 & is_dss_b2 & is_charged_B_cond,  # Done -> D pi pi
            211: (df.SIG_ID == 3) & is_charged_B_cond,
            212: (df.SIG_ID == 4) & is_charged_B_cond,
            213: (df.SIG_ID == 8)
            & ~(b1_cond & has_two_pi_b1 & is_dss_b1)
            & ~(b2_cond & has_two_pi_b2 & is_dss_b2)
            & is_charged_B_cond,
            214: (df.SIG_ID.isin([1, 2, 7, 9])) & is_charged_B_cond,
            # B0
            301: (df.SIG_ID == 5) & df.sig_gen_id.isin([54]) & ~is_charged_B_cond,  # Dzerostar -> D(*) pi0
            302: (df.SIG_ID == 5) & df.sig_gen_id.isin([53]) & ~is_charged_B_cond,  # Doneprime -> D(*) pi0
            303: (df.SIG_ID == 5) & df.sig_gen_id.isin([51]) & ~is_charged_B_cond,  # Done -> D(*) pi0
            304: (df.SIG_ID == 5) & df.sig_gen_id.isin([52]) & ~is_charged_B_cond,  # Dtwostar -> D(*) pi0
            305: (df.SIG_ID == 6) & df.sig_gen_id.isin([54]) & ~is_charged_B_cond,  # Dzerostar -> D(*) pi+
            306: (df.SIG_ID == 6) & df.sig_gen_id.isin([53]) & ~is_charged_B_cond,  # Doneprime -> D(*) pi+
            307: (df.SIG_ID == 6) & df.sig_gen_id.isin([51]) & ~is_charged_B_cond,  # Done -> D(*) pi+
            308: (df.SIG_ID == 6) & df.sig_gen_id.isin([52]) & ~is_charged_B_cond,  # Dtwostar -> D(*) pi+
            309: (df.SIG_ID == 8) & b1_cond & has_two_pi_b1 & is_dss_b1 & ~is_charged_B_cond,  # Done -> D pi pi
            310: (df.SIG_ID == 8) & b2_cond & has_two_pi_b2 & is_dss_b2 & ~is_charged_B_cond,  # Done -> D pi pi
            311: (df.SIG_ID == 3) & ~is_charged_B_cond,
            312: (df.SIG_ID == 4) & ~is_charged_B_cond,
            313: (df.SIG_ID == 8)
            & ~(b1_cond & has_two_pi_b1 & is_dss_b1)
            & ~(b2_cond & has_two_pi_b2 & is_dss_b2)
            & ~is_charged_B_cond,
            314: (df.SIG_ID.isin([1, 2, 7, 9])) & ~is_charged_B_cond,
        }
        return cond_dict

    def create_dss_sig_id(self, df: pd.DataFrame) -> None:
        test_cond = df[self.sig_b_mdst_col].isin([1, 2])
        test_cond_tau = df[self.sig_b_tau_mdst_col].isin([1, 2])

        df.loc[:, "b_mdst"] = np.nan
        df.loc[test_cond, "b_mdst"] = df[self.sig_b_mdst_col]
        df.loc[test_cond_tau, "b_mdst"] = df[self.sig_b_tau_mdst_col]

        b1_cond = df.b_mdst == 1
        b2_cond = df.b_mdst == 2

        df.loc[:, "sig_gen_id"] = np.nan
        df.loc[b1_cond, "sig_gen_id"] = df.loc[b1_cond, "GEN_B1_ID"]
        df.loc[b2_cond, "sig_gen_id"] = df.loc[b2_cond, "GEN_B2_ID"]

        df.loc[:, self.dss_sig_id_col] = np.nan
        for dss_id, cond in self.get_dss_cond_dict(df=df).items():
            df.loc[cond, self.dss_sig_id_col] = dss_id

    def run_fitting_preparation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]

        mc_output_key = f"combined_mc_prepared_{self.selection}"
        mc_output_path = os.path.join(self.base_dir, f"{mc_output_key}.h5")
        output_dict.update({mc_output_key: mc_output_path})

        signal_output_key = f"signal_mc_prepared_{self.selection}"
        signal_output_path = os.path.join(self.base_dir, f"{signal_output_key}.h5")
        output_dict.update({signal_output_key: signal_output_path})

        # asimov_output_key = f"asimov_mc_prepared_{self.selection}"
        # asimov_output_path = os.path.join(self.base_dir, f"{asimov_output_key}.h5")

        data_output_key = f"data_prepared_{self.selection}"
        data_output_path = os.path.join(self.base_dir, f"{data_output_key}.h5")
        output_dict.update({data_output_key: data_output_path})

        for data_type in self.sample_keys_for_eff_mc:
            key = f"{data_type}_prepared_{self.selection}"
            path = os.path.join(self.base_dir, f"{key}.h5")
            output_dict.update({key: path})

        if not dry_run:
            mc_data_types = main_sample_keys
            data_types = mc_data_types + [self.data_key] + self.dss_sig_mc_keys
            data_paths = {data_type: self.get_data_path(data_type) for data_type in data_types}

            # MC
            mc_dfs = [pd.read_hdf(data_paths[data_type]) for data_type in mc_data_types]
            assert all(isinstance(df, pd.DataFrame) for df in mc_dfs)
            df_mc = pd.concat(mc_dfs, ignore_index=True, sort=False)

            df_mc_reduced = df_mc.loc[df_mc[self.decay_mode_col].isin(fitting_utils.all_dss_modes)]

            # apply weights
            self.apply_weights(df=df_mc_reduced)
            # apply q^2 truth information
            self.apply_qsquare_infos(df=df_mc_reduced)
            # create advanced sig id
            self.create_advanced_sig_id(df=df_mc_reduced)
            # create dss sig id
            self.create_dss_sig_id(df=df_mc_reduced)

            # df_mc_reduced = df_mc[fitting_utils.get_required_columns(df=df_mc)]

            tmp_mc_output_path = os.path.join(self.base_dir, f"tmp_{mc_output_key}.h5")
            df_mc_reduced.to_hdf(path_or_buf=tmp_mc_output_path, key="fitting")
            os.rename(src=tmp_mc_output_path, dst=mc_output_path)

            # Signal MC
            df_sig_mcs = [pd.read_hdf(data_paths[data_type]) for data_type in self.dss_sig_mc_keys]
            assert all(isinstance(df, pd.DataFrame) for df in df_sig_mcs)
            df_sig = pd.concat(df_sig_mcs, ignore_index=True, sort=False)

            # apply weights
            self.apply_weights(df=df_sig)
            # apply q^2 truth information
            self.apply_qsquare_infos(df=df_sig)
            # create advanced sig id
            self.create_advanced_sig_id(df=df_sig)

            tmp_sig_output_path = os.path.join(self.base_dir, f"tmp_{signal_output_key}.h5")
            df_sig.to_hdf(path_or_buf=tmp_sig_output_path, key="fitting")
            os.rename(src=tmp_sig_output_path, dst=signal_output_path)

            for data_type in self.sample_keys_for_eff_mc:
                df = pd.read_hdf(self.get_data_path(data_type=data_type))
                assert isinstance(df, pd.DataFrame)
                df_reduced = df.loc[df[self.decay_mode_col].isin(fitting_utils.all_dss_modes)]

                # apply weights
                self.apply_weights(df=df_reduced)
                # apply q^2 truth information
                self.apply_qsquare_infos(df=df_reduced)
                # create advanced sig id
                self.create_advanced_sig_id(df=df_reduced)
                # create dss sig id
                self.create_dss_sig_id(df=df_reduced)

                key = f"{data_type}_prepared_{self.selection}"
                path = os.path.join(self.base_dir, f"{key}.h5")
                tmp_output_path = os.path.join(self.base_dir, f"tmp_{key}.h5")
                df_reduced.to_hdf(path_or_buf=tmp_output_path, key="fitting")
                os.rename(src=tmp_output_path, dst=path)

            # Eff samples for Asimov
            # df_asimov_mcs = []  # type: List[pd.DataFrame]
            # assert all(isinstance(df, pd.DataFrame) for df in df_asimov_mcs)
            #
            # for data_type in self.sample_keys_for_eff_mc:
            #    df = pd.read_hdf(self.get_eff_path(data_type=data_type))
            #    assert isinstance(df, pd.DataFrame)
            #
            #    if len(np.unique(df.index)) == 0:
            #        df = df.assign(**{"data_type": data_type})
            #    else:
            #        df.loc[:, "data_type"] = data_type
            #
            #    WeightFinalizer.apply_weight_finalization_to_df(
            #        df=df,
            #        flag_info_dict_path=WeightManager.get_flag_info_dict_path_from_task_input_dict(
            #            input_dict=self.dummy_input_dict,
            #            selection_name=self.selection,
            #        ),
            #        combo_name="Efficiency Calculation",
            #        key_hint=f"{data_type} {self.selection}_with_CS_CS1",
            #    )
            #    df_asimov_mcs.append(df)
            #
            # df_asimov_mc = pd.concat(df_asimov_mcs, ignore_index=True, sort=False)
            #
            # apply weights
            # self.apply_weights(df=df_asimov_mc)
            # apply q^2 truth information
            # self.apply_qsquare_infos(df=df_asimov_mc)
            # create advanced sig id
            # self.create_advanced_sig_id(df=df_asimov_mc)
            # create dss sig id
            # self.create_dss_sig_id(df=df_asimov_mc)
            #
            # tmp_asimov_output_path = os.path.join(self.base_dir, f"tmp_{asimov_output_key}.h5")
            # df_asimov_mc.to_hdf(path_or_buf=tmp_asimov_output_path, key="fitting")
            # os.rename(src=tmp_asimov_output_path, dst=asimov_output_path)

            # Data
            df_data = pd.read_hdf(data_paths[self.data_key])
            assert isinstance(df_data, pd.DataFrame)
            tmp_data_output_path = os.path.join(self.base_dir, f"tmp_{data_output_key}.h5")
            df_data.to_hdf(path_or_buf=tmp_data_output_path, key="fitting")
            os.rename(src=tmp_data_output_path, dst=data_output_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_fitting_preparation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_fitting_preparation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict
