import os
import logging
import b2luigi as luigi
import numpy as np
import matplotlib.pyplot as plt
import json

from luigi.util import inherits
from typing import Dict, List, Generator

from rdstar.offline_analysis.tasks import utilities
from rdstar.utilities import PathType
from rdstar.offline_analysis.mc_ids import rdstar_sig_queries

from rdstar.offline_analysis.dss_fitting.tasks.splitter_task import SampleSplitterTask
from rdstar.offline_analysis.dss_fitting.tasks.fitting_task import FittingTask
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection
from rdstar.offline_analysis.rdstar_plots import get_rdstar_color_dict

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "SignificancePlotterTask",
]


@inherits(SampleSplitterTask)
class SignificancePlotterTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "results", self.fitting_setup_name)

        self.fitting_setup = FittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = self.fitting_setup.fit_case_dict.pivot_fit_case
        self.fit_output_tag = self.fit_case.output_tag

        self.yield_parameters = self.fitting_setup.ratio_yield_parameters
        assert self.fit_case.plot_labels is not None
        self.yield_parameter_labels = self.fit_case.plot_labels

        self.variable = self.fitting_setup.fit_in_bins_of

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def run_significance_plotting(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_key = f"{self.fitting_setup_name}_significance"
        output_path = os.path.join(self.base_dir, f"{output_key}.pdf")

        if not dry_run:
            significance_dict = {yield_param: [] for yield_param in self.yield_parameters}  # type: Dict[str, List[float]]

            bins = self.variable.n_bins
            scopes = self.fitting_setup.fit_in_bins_of_scopes
            bin_centers = [scopes[i + 1] - ((scopes[i + 1] - scopes[i]) / 2) for i in range(0, bins)]
            bin_widths = [scopes[i + 1] - scopes[i] for i in range(0, bins)]
            assert len(bin_centers) == len(scopes) - 1
            assert len(bin_widths) == len(scopes) - 1

            for i in range(0, bins):
                with open(self.input_dict[f"significance_{self.fitting_setup_name}_bin_{i}"]) as f:
                    significance = json.load(f)

                    for yield_param in self.yield_parameters:
                        current_list = significance_dict[yield_param]
                        current_list.append(significance[yield_param])
                        significance_dict[yield_param] = current_list

            logging.debug(significance_dict)

            plot_style.set_matplotlibrc_params()
            fig, ax = plt.subplots(figsize=(6, 5))

            for yield_param in self.yield_parameters:
                if self.fit_case.plot_colors is None:
                    color = get_rdstar_color_dict()[rdstar_sig_queries[yield_param].id]
                else:
                    color = self.fit_case.plot_colors[yield_param]
                # ax.bar(
                #     x=bin_centers,
                #     height=0.1,
                #     width=bin_widths,
                #     bottom=np.array(significance_dict[yield_param]) - 0.05,
                #     color=color,
                #     alpha=1,
                #     label=self.yield_parameter_labels[yield_param]
                # )

                ax.errorbar(
                    x=bin_centers,
                    y=np.array(significance_dict[yield_param]),
                    yerr=None,
                    xerr=np.array(bin_widths) / 2,
                    ls="",
                    elinewidth=4,
                    marker="",
                    color=color,
                    capsize=0,
                    label=self.yield_parameter_labels[yield_param],
                )

            ax.set_ylim(bottom=0)
            ax.set_ylim(top=21)

            ax.set_xlabel(f"{self.variable.variable_name} in {self.variable.unit}", plot_style.xlabel_pos)
            ax.set_ylabel(r"Significance $ Z $ in $\sigma$", plot_style.ylabel_pos)
            ax.legend(loc=1)

            export(fig=fig, filename=output_key, target_dir=self.base_dir)

        return {output_key: output_path}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_significance_plotting()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_significance_plotting(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for bin_number in range(0, self.fitting_setup.fit_in_bins_of.n_bins):
            yield self.clone(FittingTask, bin_number=bin_number)

        yield self.clone(FittingPreparationTask)
