from typing import List, Dict, Tuple
import pandas as pd

from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetupManager,
)
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.systematics.systematics import SysColsInfo
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import BtoDststFFReweighting as BtoDss_FF
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import DststType as BtoDststFF_decay_types
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from templatefitter.plotter.histogram_variable import HistVariable

final_weight_column = "final_weights"  # type: str

base_required_columns = [
    "daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
    "qSquaredReco",
    MissingMassResolutionCorrector.get_corrected_col_name(
        correction_setup_name=ResolutionCorrectionSetupManager.setup_name_to_use
    ),
    "m2RecoilSignalSide",
    "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc",
    "SIG_ID",
    "__tracking_systematic__",
    "__tagging_fei_corr_uncert__",
    final_weight_column,
    "dss_sec_sel_mva_output",
    "e_miss_minus_p_miss",
    "daughter__bo1__cm__spformula__bodaughterInvM__bo0__cm__sp2__bc__sp__mi__spdaughter__bo0__cm__spInvM__bc__bc__bc",
    "pStarL",
]  # type: List[str]


def get_required_columns(df: pd.DataFrame) -> List[str]:
    required_cols = base_required_columns
    for item in df.columns:
        if item.startswith("__weight"):
            required_cols.append(item)
        if item.startswith("weight"):
            required_cols.append(item)
        if item in [info.column_name for info in WeightManager.weight_collection]:
            required_cols.append(item)
    required_cols.extend(QSquaredResidualVariable.get_new_columns_list())
    required_cols.append(QSquaredResidualVariable.q_squared_truth_col_name)

    return list(set(required_cols))


dss_missing_mass_fine_binning_dict = {
    "DZ": 30,
    "DM": 30,
    "DSZ": 30,
    "DSM": 30,
    "DZPiZ": 25,
    "DMPiZ": 20,
    "DSZPiZ": 20,
    "DSMPiZ": 20,
    "DZPiM": 30,
    "DPPiM": 30,
    "DSZPiM": 25,
    "DSPPiM": 20,
    "all_modes": 30,
}  # type: Dict[str, int]

dss_missing_mass_scope_dict_for_bin0 = {
    "DZ": (-2.0, 2.0),
    "DM": (-2.0, 2.0),
    "DSZ": (-2.0, 2.0),
    "DSM": (-2.0, 2.0),
    "DZPiZ": (-2.0, 2.0),
    "DMPiZ": (-2.0, 2.0),
    "DSZPiZ": (-2.0, 2.0),
    "DSMPiZ": (-2.0, 2.0),
    "DZPiM": (-2.0, 2.0),
    "DPPiM": (-2.0, 2.0),
    "DSZPiM": (-2.0, 2.0),
    "DSPPiM": (-2.0, 2.0),
    "all_modes": (-2.0, 2.0),
}  # type: Dict[str, Tuple[float, float]]

dss_missing_mass_scope_dict_for_advanced_bin0 = {
    "DZ": (-2.0, 1.0),
    "DM": (-2.0, 1.0),
    "DSZ": (-2.0, 1.0),
    "DSM": (-2.0, 1.0),
    "DZPiZ": (-2.0, 1.0),
    "DMPiZ": (-2.0, 1.0),
    "DSZPiZ": (-2.0, 1.0),
    "DSMPiZ": (-2.0, 1.0),
    "DZPiM": (-2.0, 1.0),
    "DPPiM": (-2.0, 1.0),
    "DSZPiM": (-2.0, 1.0),
    "DSPPiM": (-2.0, 1.0),
    "all_modes": (-2.0, 1.0),
}  # type: Dict[str, Tuple[float, float]]

dss_modes = {
    "DZPiZ": [63, 64],
    "DMPiZ": [61, 62],
    "DSZPiZ": [67, 68],
    "DSMPiZ": [65, 66],
    "DZPiM": [71, 72],
    "DPPiM": [73, 74],
    "DSZPiM": [75, 76],
    "DSPPiM": [77, 78],
}  # type: Dict[str, List[int]]

all_dss_modes = []  # type: List[int]
for modes in dss_modes.values():
    all_dss_modes.extend(modes)

dss_labels = {
    "DZPiZ": r"$B^- \rightarrow D^0 \pi^0 \ell^-$",
    "DMPiZ": r"$B^0 \rightarrow D^- \pi^0 \ell^+$",
    "DSZPiZ": r"$B^- \rightarrow D^{*0} \pi^0 \ell^-$",
    "DSMPiZ": r"$B^0 \rightarrow D^{*-} \pi^0 \ell^+$",
    "DZPiM": r"$B^0 \rightarrow \overline{D}^{0} \pi^- \ell^+$",
    "DPPiM": r"$B^- \rightarrow D^{+} \pi^- \ell^-$",
    "DSZPiM": r"$B^0 \rightarrow \overline{D}^{*0} \pi^- \ell^+$",
    "DSPPiM": r"$B^- \rightarrow D^{*+} \pi^- \ell^-$",
    "all_modes": "All Reconstruction Modes",
}  # type: Dict[str, str]

dss_charged_B_charged_pi_modes = {
    "DPPiM": [73, 74],
    "DSPPiM": [77, 78],
}  # type: Dict[str, List[int]]

dss_neutral_B_charged_pi_modes = {
    "DZPiM": [71, 72],
    "DSZPiM": [75, 76],
}  # type: Dict[str, List[int]]

dss_charged_B_neutral_pi_modes = {
    "DZPiZ": [63, 64],
    "DSZPiZ": [67, 68],
}  # type: Dict[str, List[int]]

dss_neutral_B_neutral_pi_modes = {
    "DMPiZ": [61, 62],
    "DSMPiZ": [65, 66],
}  # type: Dict[str, List[int]]

dss_modes_for_DPiZ = {
    "DZPiZ": [63, 64],
    "DMPiZ": [61, 62],
}  # type: Dict[str, List[int]]

dss_modes_for_DSPiZ = {
    "DSZPiZ": [67, 68],
    "DSMPiZ": [65, 66],
}  # type: Dict[str, List[int]]

dss_modes_for_DPiP = {
    "DZPiM": [71, 72],
    "DPPiM": [73, 74],
}  # type: Dict[str, List[int]]

dss_modes_for_DSPiP = {
    "DSZPiM": [75, 76],
    "DSPPiM": [77, 78],
}  # type: Dict[str, List[int]]

q_square_advanced_binning_for_fitting_var = HistVariable(
    df_label="qSquaredReco",
    n_bins=6,
    scope=(0, 10),
    var_name=LatexObsNames.fitting_q2,
    unit=LatexUnits.gev_sq,
)

pStarL_advanced_binning_for_fitting_var = HistVariable(
    df_label="pStarL",
    n_bins=8,
    scope=(0.1, 2.2),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)

q_square_advanced_binning_scopes = (0.0, 1.0, 2.0, 3.0, 4.5, 7.0, 10.0)

pStarL_advanced_binning_scopes = (0.1, 0.5, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.2)

dss_ff_systematics = []  # type: List[SysColsInfo]
for dstst_decay_type in BtoDststFF_decay_types:
    dss_ff_systematics.append(
        SysColsInfo(
            var=BtoDss_FF.get_var_dstst_weight_name(dstst_type=dstst_decay_type.value),
            ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type=dstst_decay_type.value),
            from_df=True,
        )
    )
