"""
Provides class for the calculation of reconstruction efficiencies.

Felix Metzner 2020
"""

import os
import stat
import json
import tempfile
import numpy as np
import pandas as pd

from typing import Union, Optional, Tuple, List, Iterable, Dict, Generator

from analysistools.statistics import limits as rdstar_limits

from rdstar.utilities import PathType
from rdstar.offline_analysis import mc_ids
from rdstar.offline_analysis import rdstar_plots
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis import rdstar_sample_weights

from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.weight_management.weight_finalization import WeightFinalizer

__all__ = [
    "EfficiencyDictType",
    "EfficiencyCalculator",
]


# EfficiencyDictType
# First key: Data type on which the efficiencies are evaluated on
# Dictionary containing
#   - the original number of events of this sample;
#   - the number of all events in this sample after the selection "after"
#   - the sum of weights of all events in this sample after the selection "sum_of_weights_after"
#   - the signal ID for this sample;
#   - and a dictionary containing the efficiencies for different reconstruction modes in this sample:
#       * Second key: Reconstruction channels that are evaluated on this sample
#           Dictionary containing the efficiency values for this sample.
EfficiencyDictEntry = Dict[str, Union[int, str, Dict[str, Union[float, int]]]]
EfficiencyDictType = Dict[str, EfficiencyDictEntry]


class EfficiencyCalculator:
    output_identifier = "efficiencies"  # type: str
    output_file_suffix = "_dict.json"  # type: str

    original_event_count_info_column = "prod"

    _key_sig_id = "sig_id"  # type: str
    _key_sample_combination_name = "sample_combination_name"  # type: str
    _key_before = "before"  # type: str
    _key_after = "after"  # type: str
    _key_evt_eff = "event_based_efficiency"  # type: str
    _key_evt_up = "event_based_up"  # type: str
    _key_evt_dn = "event_based_down"  # type: str
    _key_sow_after = "sum_of_weights_after"  # type: str
    _key_sow_eff = "weight_based_efficiency"  # type: str
    _key_sow_up = "weight_based_up"  # type: str
    _key_sow_dn = "weight_based_down"  # type: str

    sig_id_col = "SIG_ID"  # type: str
    decay_mode_id_col = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str

    default_weight_columns = WeightManager.efficiency_weight_column_names  # type: Tuple[str, ...]

    def __init__(
        self,
        selection_tag: str,
        output_dir_path: PathType,
        flag_info_dict_path: PathType,
        weight_column_names: Iterable[str],
    ) -> None:
        self.selection_tag = selection_tag  # type: str
        self.output_dir_path = output_dir_path  # type: PathType
        self.flag_info_dict_path = flag_info_dict_path  # type: PathType
        self.__weight_column_names = list(weight_column_names)  # type: List[str]
        self.__required_column_names = [
            *self.__weight_column_names,
            self.sig_id_col,
            self.decay_mode_id_col,
        ]  # type: List[str]

        self.eff_output_key = f"{self.output_identifier}_{self.selection_tag}"  # type: str
        self.output_file_name = f"{self.eff_output_key}{self.output_file_suffix}"  # type: str
        self.eff_target_path = os.path.join(self.output_dir_path, self.output_file_name)  # type: PathType

        dec_modes = {
            **rdstar_plots.rdstar_dec_modes,
            **rdstar_plots.dss_dec_modes,
            **rdstar_plots.dss_charged_dec_modes,
        }  # type: Dict[str, List[int]]

        self.all_decay_mode_ids = tuple([i for v in dec_modes.values() for i in v])  # type: Tuple[int, ...]

    def calculate_efficiency_values(
        self,
        df: pd.DataFrame,
        n_before: int,
        dictionary: EfficiencyDictEntry,
        mask: Optional[pd.Series] = None,
        name: Optional[str] = None,
    ) -> None:
        assert (mask is None) == (name is None)
        assert self._key_before in dictionary.keys()

        new_entry_name_key = self.output_identifier if name is None else f"{self.output_identifier}_{name}"
        assert new_entry_name_key not in dictionary.keys(), (new_entry_name_key, dictionary.keys())

        n_after_sow = None  # type: Optional[float]
        if mask is None:
            n_after = len(df.index)
            if len(self.__weight_column_names):
                n_after_sow = df.loc[:, list(self.__weight_column_names)].product(axis=1).sum()
        else:
            n_after = len(df.loc[mask].index)
            if len(self.__weight_column_names):
                n_after_sow = df.loc[mask, list(self.__weight_column_names)].product(axis=1).sum()

        efficiency = 1.0 * n_after / n_before
        upper, lower = rdstar_limits.binomial_limit(n=n_before, k=n_after, sigma=1)

        if len(self.__weight_column_names):
            assert n_after_sow is not None, n_after_sow
            assert n_after_sow >= 0.0, n_after_sow
            efficiency_sow = 1.0 * n_after_sow / n_before
            upper_sow, lower_sow = rdstar_limits.binomial_limit(n=n_before, k=n_after_sow, sigma=1)
        else:
            assert n_after_sow is None, n_after_sow
            efficiency_sow, upper_sow, lower_sow = (np.NaN, np.NaN, np.NaN)

        new_dict_entry_value = {
            self._key_before: n_before,
            self._key_after: n_after,
            self._key_evt_eff: efficiency,
            self._key_evt_up: upper,
            self._key_evt_dn: lower,
            self._key_sow_after: n_after_sow,
            self._key_sow_eff: efficiency_sow,
            self._key_sow_up: upper_sow,
            self._key_sow_dn: lower_sow,
        }

        dictionary.update({new_entry_name_key: new_dict_entry_value})

    @staticmethod
    def get_sig_id_and_sample_combination_key(data_type: str) -> Tuple[int, str]:
        dss_pi_zero_data_types = (
            rdstar_samples.data_type_combinations["DssPiZeroEff"] + rdstar_samples.data_type_combinations["DssTauPiZEff"]
        )
        dss_pi_plus_data_types = (
            rdstar_samples.data_type_combinations["DssPiPlusEff"] + rdstar_samples.data_type_combinations["DssTauPiPEff"]
        )

        if data_type in dss_pi_zero_data_types:
            sample_combination_key = "DssPiZeroEff"
            sig_type = "DSSl"
        elif data_type in dss_pi_plus_data_types:
            sample_combination_key = "DssPiPlusEff"
            sig_type = "DSSlbkg"
        elif data_type in rdstar_samples.data_type_combinations["SignalEff"]:
            sample_combination_key = "SignalEff"
            if data_type.startswith("DS"):
                if data_type.endswith("Tau"):
                    sig_type = "DStau"
                else:
                    sig_type = "DSl"
            else:
                if data_type.endswith("Tau"):
                    sig_type = "Dtau"
                else:
                    sig_type = "Dl"
        else:
            raise KeyError(f"Trying to calculate signal efficiency for unexpected data type {data_type}...")

        return mc_ids.rdstar_sig_queries[sig_type].id, sample_combination_key

    def get_sig_id_decay_mode_id_combinations(
        self,
        data_type: str,
    ) -> Generator[Tuple[str, Tuple[int, ...]], None, None]:

        sig_id, sample_combination_key = self.get_sig_id_and_sample_combination_key(data_type=data_type)

        sig_dec_ids = tuple([k for k, v in mc_ids.decay_mode_to_sig_id.items() if sig_id in v])  # type: Tuple[int, ...]

        yield "for_decay_mode_ids", tuple([])
        for sig_dec_id in sig_dec_ids:
            yield f"for_sig_dec_id_{sig_dec_id}", tuple([sig_dec_id])
        yield "for_all_matching_dec_ids", sig_dec_ids
        for dec_id in self.all_decay_mode_ids:
            yield f"for_dec_id_{dec_id}", tuple([dec_id])
        yield "for_all_dec_ids", self.all_decay_mode_ids

        signal_chs_decay_ids = tuple([i for v in rdstar_plots.rdstar_dec_modes.values() for i in v])
        yield "for_all_SignalEff_dec_ids", signal_chs_decay_ids
        signal_chs_charged_b_decay_ids = tuple([i for v in rdstar_plots.rdstar_charged_b_dec_modes.values() for i in v])
        yield "for_charged_b_SignalEff_dec_ids", signal_chs_charged_b_decay_ids
        signal_chs_neutral_b_decay_ids = tuple([i for v in rdstar_plots.rdstar_neutral_b_dec_modes.values() for i in v])
        yield "for_neutral_b_SignalEff_dec_ids", signal_chs_neutral_b_decay_ids

        dss_chs_decay_ids = tuple([i for v in rdstar_plots.dss_dec_modes.values() for i in v])
        yield "for_all_DssPiZeroEff_dec_ids", dss_chs_decay_ids
        dss_chs_charged_b_decay_ids = tuple([i for v in rdstar_plots.dss_charged_b_dec_modes.values() for i in v])
        yield "for_charged_b_DssPiZeroEff_dec_ids", dss_chs_charged_b_decay_ids
        dss_chs_neutral_b_decay_ids = tuple([i for v in rdstar_plots.dss_neutral_b_dec_modes.values() for i in v])
        yield "for_neutral_b_DssPiZeroEff_dec_ids", dss_chs_neutral_b_decay_ids

        dss_pip_chs_decay_ids = tuple([i for v in rdstar_plots.dss_charged_dec_modes.values() for i in v])
        yield "for_all_DssPiPlusEff_dec_ids", dss_pip_chs_decay_ids
        dss_pip_chs_bp_decay_ids = tuple([i for v in rdstar_plots.dss_charged_charged_b_dec_modes.values() for i in v])
        yield "for_charged_b_DssPiPlusEff_dec_ids", dss_pip_chs_bp_decay_ids
        dss_pip_chs_b0_decay_ids = tuple([i for v in rdstar_plots.dss_charged_neutral_b_dec_modes.values() for i in v])
        yield "for_neutral_b_DssPiPlusEff_dec_ids", dss_pip_chs_b0_decay_ids

    def evaluate_efficiencies(
        self,
        efficiency_sample_tuples: List[Tuple[str, PathType]],
    ) -> EfficiencyDictType:
        efficiency_dict = dict()  # type: EfficiencyDictType

        for data_type, df_path in efficiency_sample_tuples:
            assert os.path.exists(df_path), df_path

            df = pd.read_hdf(df_path)
            assert isinstance(df, pd.DataFrame)

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{"data_type": data_type})
            else:
                df.loc[:, "data_type"] = data_type

            WeightFinalizer.apply_weight_finalization_to_df(
                df=df,
                flag_info_dict_path=self.flag_info_dict_path,
                combo_name="Efficiency Calculation",
                key_hint=f"{data_type} {self.selection_tag}",
            )

            assert all(c in df.columns for c in self.__required_column_names), [
                c for c in self.__required_column_names if c not in df.columns
            ]

            if self.original_event_count_info_column == "prod":
                n_before = rdstar_sample_weights.signal_mc_evts[data_type].prod  # type: int
            elif self.original_event_count_info_column == "meta":
                n_before = rdstar_sample_weights.signal_mc_evts[data_type].meta
            elif self.original_event_count_info_column == "summary":
                n_before = rdstar_sample_weights.signal_mc_evts[data_type].summary
            else:
                raise ValueError(f"The option {self.original_event_count_info_column} is not valid!")

            if len(self.__weight_column_names):
                sow = df[list(self.__weight_column_names)].product(axis=1).sum()
            else:
                sow = np.nan

            sig_id, sample_combination_key = self.get_sig_id_and_sample_combination_key(data_type=data_type)

            efficiency_dict[data_type] = {
                self._key_sig_id: sig_id,
                self._key_sample_combination_name: sample_combination_key,
                self._key_before: n_before,
                self._key_after: len(df.index),
                self._key_sow_after: sow,
            }

            all_true_mask = pd.Series(True, index=df.index)

            for name, decay_mode_id_tuple in self.get_sig_id_decay_mode_id_combinations(data_type=data_type):
                mask = df[self.decay_mode_id_col].isin(decay_mode_id_tuple) if decay_mode_id_tuple else all_true_mask
                sig_mask = df[self.sig_id_col] == sig_id

                self.calculate_efficiency_values(
                    df=df,
                    n_before=n_before,
                    dictionary=efficiency_dict[data_type],
                    mask=mask,
                    name=name,
                )

                self.calculate_efficiency_values(
                    df=df,
                    n_before=n_before,
                    dictionary=efficiency_dict[data_type],
                    mask=(mask & sig_mask),
                    name=f"{name}_and_sig_id_match",
                )

        assert len(efficiency_dict) > 0, len(efficiency_dict)
        assert len(efficiency_dict) == len(efficiency_sample_tuples), (
            len(efficiency_dict),
            len(efficiency_sample_tuples),
        )
        return efficiency_dict

    def get_output_key_file_path_pair(self) -> Tuple[str, PathType]:
        return self.eff_output_key, self.eff_target_path

    def atomic_save(self, efficiency_dict: EfficiencyDictType) -> None:
        assert os.path.isdir(self.output_dir_path), self.output_dir_path
        output_data = json.dumps(efficiency_dict, indent=2)
        with tempfile.NamedTemporaryFile(mode="w", dir=self.output_dir_path) as out_file:
            out_file.write(output_data)
            out_file.flush()
            os.fsync(out_file.fileno())
            os.rename(out_file.name, self.eff_target_path)
            os.chmod(path=self.eff_target_path, mode=(stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH))
            open(out_file.name, "w").close()
