import pandas as pd
import os
import json
import numpy as np

from typing import Tuple, Dict, NamedTuple, Union

from rdstar.utilities import PathType
from rdstar.offline_analysis.rdstar_sample_weights import signal_mc_evts
from rdstar.offline_analysis.weight_management.weight_information import WeightTypeCollection

from analysistools.statistics import limits as rdstar_limits

__all__ = [
    "DataTypeFraction",
    "EfficiencyNumbersDict",
    "SimpleEfficiencyCalculator",
]


class DataTypeFraction(NamedTuple):
    data_types: Tuple[str, ...]
    fraction: float


class EfficiencyNumbersDict:
    def __init__(self) -> None:
        self._eff_numbers_dict = dict()  # type: Dict[str, Dict[str, Union[float, Dict[str, float]]]]

    def add_entry(
        self,
        sig_id: int,
        produced_events: float,
        selected_events: Dict[str, float],
    ) -> None:
        self._eff_numbers_dict.update(
            {
                f"SIG_ID_{sig_id}": {
                    "produced_events": produced_events,
                    "selected_events": selected_events,
                }
            }
        )

    def get_produced_events(
        self,
        sig_id: int,
    ) -> float:
        assert f"SIG_ID_{sig_id}" in self._eff_numbers_dict.keys()

        produced_events = self._eff_numbers_dict[f"SIG_ID_{sig_id}"]["produced_events"]
        assert isinstance(produced_events, float)

        return produced_events

    def get_selected_events(self, sig_id: int, decay_mode_id: int) -> float:
        assert f"SIG_ID_{sig_id}" in self._eff_numbers_dict.keys()

        selected_dict = self._eff_numbers_dict[f"SIG_ID_{sig_id}"]["selected_events"]
        assert isinstance(selected_dict, dict)
        assert f"decay_mode_id_{int(decay_mode_id)}" in selected_dict.keys()

        selected_events = selected_dict[f"decay_mode_id_{int(decay_mode_id)}"]
        assert isinstance(selected_events, float)

        return selected_events

    def get_efficiency(
        self,
        sig_ids: Tuple[int, ...],
        decay_mode_ids: Tuple[int, ...],
    ) -> Tuple[float, float, float]:
        n_produced = 0.0
        n_selected = 0.0
        for sig_id in sig_ids:
            n_produced += self.get_produced_events(sig_id=sig_id)
            for decay_id in decay_mode_ids:
                n_selected += self.get_selected_events(
                    sig_id=sig_id,
                    decay_mode_id=decay_id,
                )

        eff = n_selected / n_produced
        upper, lower = rdstar_limits.binomial_limit(n_produced, n_selected)
        return eff, upper - eff, eff - lower

    def to_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path, "w") as fp:
            json.dump(self._eff_numbers_dict, fp, indent=2)

    def from_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path) as f:
            self._eff_numbers_dict = json.load(f)


class SimpleEfficiencyCalculator:
    decay_mode_column = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str
    file_dict_base_path = "/home/fmetzner/Documents/file_dicts"  # type: PathType

    eff_weight_column = "efficiency_weight"  # type: str

    def __init__(
        self,
        df: pd.DataFrame,
        sig_id_data_type_mapping: Dict[int, DataTypeFraction],
        sig_id_column_name: str,
    ) -> None:
        self._df = df
        self._sig_id_data_type_mapping = sig_id_data_type_mapping
        self._sig_id_column_name = sig_id_column_name

        self._df.loc[:, self.eff_weight_column] = 1
        for weight in WeightTypeCollection.efficiency_weight_columns:
            self._df.loc[:, self.eff_weight_column] *= self._df[weight]

        self._eff_numbers_dict = EfficiencyNumbersDict()

    def get_decay_mode_ids(self) -> Tuple[int, ...]:
        return tuple(self._df[self.decay_mode_column].unique())

    def get_file_dict(
        self,
        data_type: str,
    ) -> Dict[str, Dict[str, int]]:
        file_dict_path = os.path.join(self.file_dict_base_path, f"evtgen_{data_type}/{data_type}.json")
        assert os.path.isfile(file_dict_path), f"{file_dict_path}:No such file!"
        with open(file_dict_path) as f:
            file_dict = json.load(f)
        return file_dict

    def get_produced_events(
        self,
        data_type: str,
    ) -> int:
        if data_type not in signal_mc_evts.keys():
            file_dict = self.get_file_dict(data_type=data_type)
            n_events = 0
            for file, meta_data in file_dict.items():
                n_events += meta_data["nevents"]
        else:
            n_events = signal_mc_evts[data_type].prod

        return n_events

    def get_produced_events_for_process(
        self,
        sig_id: int,
    ) -> float:
        assert sig_id in self._sig_id_data_type_mapping.keys()
        data_type_fraction = self._sig_id_data_type_mapping[sig_id]
        produced_events = 0
        for data_type in data_type_fraction.data_types:
            produced_events += self.get_produced_events(data_type=data_type)

        return produced_events * data_type_fraction.fraction

    def get_selected_events_for_process(
        self,
        sig_id: int,
        decay_mode_id: int,
    ) -> float:
        data_type_cond = self._df["data_type"].isin(self._sig_id_data_type_mapping[sig_id].data_types)
        decay_mode_cond = self._df[self.decay_mode_column] == decay_mode_id
        sig_id_cond = self._df[self._sig_id_column_name] == sig_id

        selected_df = self._df.loc[decay_mode_cond & sig_id_cond & data_type_cond]
        selected_events = float(np.sum(selected_df[self.eff_weight_column]))

        return selected_events

    def generate_efficiency_numbers_dict(self) -> None:
        for sig_id in self._sig_id_data_type_mapping.keys():
            produced_events = self.get_produced_events_for_process(sig_id=sig_id)
            selected_dict = dict()  # type: Dict[str, float]
            for decay_mode_id in self.get_decay_mode_ids():
                selected_events = self.get_selected_events_for_process(
                    sig_id=sig_id,
                    decay_mode_id=decay_mode_id,
                )
                selected_dict.update({f"decay_mode_id_{int(decay_mode_id)}": selected_events})
            self._eff_numbers_dict.add_entry(
                sig_id=sig_id,
                produced_events=produced_events,
                selected_events=selected_dict,
            )

    def write_eff_dict_to_file(self, file_path: PathType) -> None:
        self._eff_numbers_dict.to_file(file_path=file_path)
