"""
Provides fit models based on the TemplateFitter package and tailored to this analysis.

Felix Metzner 2019
"""

import copy
from typing import Union, Optional, Tuple, List, Dict, Sequence

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.binned_distributions.binning import BinsInputType
from templatefitter.minimizer import available_template_fitter_minimizer

from rdstar.offline_analysis.systematics.systematics import SysColsInfo
from rdstar.offline_analysis.mc_ids import rdstar_sig_queries, SigQueryDictType, default_sig_id_col_name


__all__ = [
    "BinningDictType",
    "ScopeDictType",
    "FitCase",
]

BinningDictType = Tuple[Optional[Dict[str, BinsInputType]], ...]
ScopeDictType = Tuple[Optional[Dict[str, Tuple[float, float]]], ...]


class FitCase:
    def __init__(
        self,
        name: str,
        base_dir_name: str,
        sample_key: str,
        output_tag: str,
        plot_info_text: str,
        # fix_nuisance_parameters: bool,  # TODO!!!!
        asimov_data: bool,
        round_asimov_data: bool,
        hist_variables: Tuple[HistVariable, ...],
        binning_by_mode_dicts: BinningDictType,
        scope_by_mode_dicts: ScopeDictType,
        weight_column_name: str,
        systematics_cols: List[SysColsInfo],
        reco_mode_infos: Sequence[Tuple[str, Sequence[int]]],
        template_setup: Union[Dict[str, List[str]], Dict[str, Tuple[str, ...]], List[str], Tuple[str, ...]],
        plot_colors: Dict[str, str],
        plot_labels: Dict[str, str],
        sig_id_column_name: Optional[str] = None,
        sig_queries: Optional[SigQueryDictType] = None,
        minimizer_id: Optional[str] = None,
    ) -> None:
        self.__name = name  # type: str
        self.__base_dir_name = base_dir_name  # type: str
        self.__sample_key = sample_key  # type: str
        self.__output_tag = output_tag  # type: str
        self.__plot_info_text = plot_info_text  # type: str
        self.__asimov_data = asimov_data  # type: bool
        self.__round_asimov_data = round_asimov_data  # type: bool

        self.__hist_variables = hist_variables  # type: Tuple[HistVariable, ...]
        self.__binning_by_mode_dicts = binning_by_mode_dicts  # type: BinningDictType
        self.__scope_by_mode_dicts = scope_by_mode_dicts  # type: ScopeDictType

        self.__weigth_column_name = weight_column_name  # type: str
        self.__systematics_cols = systematics_cols  # type: List[SysColsInfo]

        self.__reco_mode_infos = tuple(
            [(name, tuple(ids)) for name, ids in reco_mode_infos]
        )  # type: Tuple[Tuple[str, Tuple[int, ...]], ...]

        if isinstance(template_setup, dict):
            __all_values = [v for vs in template_setup.values() for v in vs]
            assert len(set(__all_values)) == len(__all_values), template_setup
        else:
            assert len(template_setup) == len(set(template_setup)), template_setup

        self.__template_setup = copy.copy(
            template_setup
        )  # type: Union[Dict[str, List[str]], Dict[str, Tuple[str, ...]], List[str], Tuple[str, ...]]

        assert all(key in plot_colors for key in self.template_setup)
        self.__plot_colors = copy.copy(plot_colors)  # type: Dict[str, str]

        assert all(key in plot_labels for key in self.template_setup)
        self.__plot_labels = copy.copy(plot_labels)  # type: Dict[str, str]

        if not isinstance(self.template_setup, tuple):
            if isinstance(self.template_setup, dict):
                self.__unpacked_plot_order = tuple(
                    [str(key) for key in self.template_setup.keys()]
                )  # type: Tuple[str, ...]
            else:
                self.__unpacked_plot_order = tuple([comp for comp in self.template_setup])
        else:
            self.__unpacked_plot_order = self.template_setup

        if sig_id_column_name is not None:
            self.__sig_id_column_name = sig_id_column_name  # type: str
        else:
            self.__sig_id_column_name = default_sig_id_col_name

        if sig_queries is not None:
            self.__sig_queries = sig_queries  # type: SigQueryDictType
        else:
            self.__sig_queries = rdstar_sig_queries

        if minimizer_id is not None:
            self.__minimizer_id = minimizer_id  # type: str
        else:
            self.__minimizer_id = "iminuit"
        assert self.__minimizer_id in available_template_fitter_minimizer.keys(), (
            self.__minimizer_id,
            available_template_fitter_minimizer.keys(),
        )

        if isinstance(self.template_setup, dict):
            assert all(
                comp in self.sig_queries for template_comp in self.template_setup.values() for comp in template_comp
            )
        else:
            assert all(comp in self.sig_queries for comp in self.template_setup)

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def base_dir_name(self) -> str:
        return self.__base_dir_name

    @base_dir_name.setter
    def base_dir_name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def sample_key(self) -> str:
        return self.__sample_key

    @sample_key.setter
    def sample_key(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def output_tag(self) -> str:
        return self.__output_tag

    @output_tag.setter
    def output_tag(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def plot_info_text(self) -> str:
        return self.__plot_info_text

    @plot_info_text.setter
    def plot_info_text(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def asimov_data(self) -> bool:
        return self.__asimov_data

    @asimov_data.setter
    def asimov_data(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def round_asimov_data(self) -> bool:
        return self.__round_asimov_data

    @round_asimov_data.setter
    def round_asimov_data(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def hist_variables(self) -> Tuple[HistVariable, ...]:
        return self.__hist_variables

    @hist_variables.setter
    def hist_variables(self, value: Tuple[HistVariable, ...]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def binning_by_mode_dicts(self) -> BinningDictType:
        return self.__binning_by_mode_dicts

    @binning_by_mode_dicts.setter
    def binning_by_mode_dicts(self, value: BinningDictType) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def scope_by_mode_dicts(self) -> ScopeDictType:
        return self.__scope_by_mode_dicts

    @scope_by_mode_dicts.setter
    def scope_by_mode_dicts(self, value: ScopeDictType) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def weight_column_name(self) -> str:
        return self.__weigth_column_name

    @weight_column_name.setter
    def weight_column_name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def systematics_cols(self) -> List[SysColsInfo]:
        return self.__systematics_cols

    @systematics_cols.setter
    def systematics_cols(self, value: List[SysColsInfo]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def reco_mode_infos(self) -> Tuple[Tuple[str, Tuple[int, ...]], ...]:
        return self.__reco_mode_infos

    @reco_mode_infos.setter
    def reco_mode_infos(self, value: Tuple[Tuple[str, Tuple[int, ...]], ...]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def template_setup(self) -> Union[Dict[str, List[str]], Dict[str, Tuple[str, ...]], List[str], Tuple[str, ...]]:
        return self.__template_setup

    @template_setup.setter
    def template_setup(
        self,
        value: Union[Dict[str, List[str]], Dict[str, Tuple[str, ...]], List[str], Tuple[str, ...]],
    ) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def plot_order(self) -> Tuple[str, ...]:
        return self.__unpacked_plot_order

    @property
    def plot_colors(self) -> Dict[str, str]:
        return self.__plot_colors

    @plot_colors.setter
    def plot_colors(self, value: Dict[str, str]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def plot_labels(self) -> Dict[str, str]:
        return self.__plot_labels

    @plot_labels.setter
    def plot_labels(self, value: Dict[str, str]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def sig_id_column_name(self) -> str:
        return self.__sig_id_column_name

    @sig_id_column_name.setter
    def sig_id_column_name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def sig_queries(self) -> SigQueryDictType:
        return self.__sig_queries

    @sig_queries.setter
    def sig_queries(self, value: SigQueryDictType) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def minimizer_id(self) -> str:
        return self.__minimizer_id

    @minimizer_id.setter
    def minimizer_id(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    def get_sig_id_for_component(self, component_name: str) -> Tuple[int, ...]:
        if isinstance(self.template_setup, dict):
            return tuple([self.sig_queries[comp].id for comp in self.template_setup[component_name]])
        else:
            return (self.sig_queries[component_name].id,)
