"""
Provides fit models based on the TemplateFitter package and tailored to this analysis.

Felix Metzner 2019
"""

import numpy as np
import pandas as pd

from collections import defaultdict
from typing import Optional, Tuple, List, Dict, Sequence

from templatefitter.fit_model.template import Template
from templatefitter.fit_model.model_builder import FitModel
from templatefitter.fit_model.parameter_handler import ParameterHandler

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.binned_distributions.binning import BinsInputType

from rdstar.offline_analysis.systematics.systematics import get_systematics

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.rdstar_fit_models import rdstar_fit_cases


__all__ = [
    "FitModelCreator",
]

BinningDictType = Tuple[Optional[Dict[str, BinsInputType]], ...]
ScopeDictType = Tuple[Optional[Dict[str, Tuple[float, float]]], ...]


class FitModelCreator:
    decay_mode_column_name = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str

    fit_cases = rdstar_fit_cases  # type: Dict[str, FitCase]

    def __init__(
        self,
        fit_case: str,
    ) -> None:
        if fit_case not in self.fit_cases.keys():
            raise ValueError(
                f"The parameter 'fit_case' must be one of {list(self.fit_cases.keys())}!\n"
                f"You provided '{fit_case}'..."
            )

        self._fit_case = fit_case  # type: str

        self.param_handler = ParameterHandler()
        self.model = FitModel(parameter_handler=self.param_handler)

        self.channel_templates = defaultdict(list)  # type: Dict[str, List[str]]
        self.channel_eff_parameters = defaultdict(list)  # type: Dict[str, List[str]]

        self.initial_yield_dict = {}  # type: Dict[str, float]
        self.initial_eff_dict = defaultdict(dict)  # type: Dict[str, Dict[str, float]]

    def setup_model(
        self,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame] = None,
        use_asimov_data: bool = False,
        round_asimov_data_bin_counts: bool = False,
        asimov_data_weight_column: Optional[str] = None,
    ) -> Tuple[ParameterHandler, FitModel]:
        if not use_asimov_data:
            assert data_df is not None
            assert isinstance(data_df, pd.DataFrame), type(data_df)
        else:
            assert data_df is None

        if len(mc_df.index) != len(np.unique(mc_df.index)):
            raise RuntimeError(
                f"The provided MC Data Frame must have a properly set index to use it as basis for the "
                f"fit model. The provided data frame has len(mc_df.index) = {len(mc_df.index)} "
                f"!= len(np.unique(mc_df.index)) = {len(np.unique(mc_df.index))}!"
            )

        if len(mc_df.index) != mc_df.index[-1] + 1:
            raise RuntimeError(
                f"The maximal index and the length of the index does not match: len(index) = {len(mc_df.index)},"
                f"max_index = {mc_df.index[-1]}. Try resetting the index first!"
            )

        self._get_initial_mc_yields(mc_df=mc_df)
        self._get_initial_mc_efficiencies(mc_df=mc_df)

        self._add_yield_parameters()
        self._add_templates_and_channels(mc_df=mc_df)

        if use_asimov_data and asimov_data_weight_column is None:
            self.model.add_asimov_data_from_templates(round_bin_counts=round_asimov_data_bin_counts)
        elif use_asimov_data and asimov_data_weight_column is not None:
            self._add_data(
                data_df=mc_df,
                weight_column_name=asimov_data_weight_column,
            )
        else:
            self._add_data(data_df=data_df)

        self.model.finalize_model()

        return self.param_handler, self.model

    def _get_initial_mc_yields(self, mc_df: pd.DataFrame) -> None:
        assert len(self.initial_yield_dict) == 0, self.initial_yield_dict

        for component in self.plot_order:
            sig_ids = self.fit_case.get_sig_id_for_component(component_name=component)  # type: Tuple[int, ...]
            sig_id_constraint = mc_df[self.fit_case.sig_id_column_name].isin(sig_ids)  # type: pd.Series

            initial_yield = 0.0  # type: float
            for mode_name, mode_ids in self.fit_case.reco_mode_infos:
                # TODO: Decide if one should really use the range constrained here for the yields, as one wants to
                #       determine the yield without this constraint! ATM it is not considered! See below!

                # variable_range_constraint = pd.Series(np.ones(len(mc_df.index), dtype=bool))
                # scope_for_mode = self.get_fit_variable_scopes_for_mode(mode=mode_name)
                # for col, scope in zip(self.fit_variable_column_names, scope_for_mode):
                #     variable_range_constraint &= ((mc_df[col] >= scope[0]) & (mc_df[col] <= scope[1]))

                mode_constraint = mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(
                    mode_ids
                )  # type: pd.Series
                all_constraints = sig_id_constraint & mode_constraint  # & variable_range_constraint  # TODO

                initial_yield += mc_df.loc[all_constraints, self.weight_column_name].sum()

            self.initial_yield_dict[component] = initial_yield

    def _get_initial_mc_efficiencies(self, mc_df: pd.DataFrame) -> None:
        assert len(self.initial_yield_dict) == len(self.plot_order), (
            self.initial_yield_dict,
            len(self.initial_yield_dict),
            len(self.plot_order),
        )
        assert len(self.initial_eff_dict) == 0, self.initial_eff_dict

        for component in self.plot_order:
            sig_ids = self.fit_case.get_sig_id_for_component(component_name=component)  # type: Tuple[int, ...]
            sig_id_constraint = mc_df[self.fit_case.sig_id_column_name].isin(sig_ids)  # type: pd.Series
            assert isinstance(sig_id_constraint, pd.Series)

            for mode_name, mode_ids in self.fit_case.reco_mode_infos:
                variable_range_constraint = pd.Series(np.ones(len(mc_df.index), dtype=bool))  # type: pd.Series
                scope_for_mode = self.get_fit_variable_scopes_for_mode(mode=mode_name)
                for col, scope in zip(self.fit_variable_column_names, scope_for_mode):
                    variable_range_constraint &= (mc_df[col] >= scope[0]) & (mc_df[col] <= scope[1])

                mode_constraint = mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(
                    mode_ids
                )  # type: pd.Series

                all_constraints = sig_id_constraint & mode_constraint & variable_range_constraint  # type: pd.Series
                count_in_mode = mc_df.loc[all_constraints, self.weight_column_name].sum()  # type: int

                self.initial_eff_dict[component][mode_name] = count_in_mode * 1.0 / self.initial_yield_dict[component]

    def _add_yield_parameters(self) -> None:
        for component in self.plot_order:
            self.model.add_model_parameter(
                name=f"yield_{component}",
                parameter_type=ParameterHandler.yield_parameter_type,
                floating=True,
                initial_value=self.initial_yield_dict[component],
                constrain_to_value=None,
                constraint_sigma=None,
            )

    def _add_templates_and_channels(self, mc_df: pd.DataFrame) -> None:
        for mode_name, mode_ids in self.fit_case.reco_mode_infos:
            decay_mode_condition = self._get_decay_mode_condition(df=mc_df, mode_ids=mode_ids)  # type: pd.Series

            for component in self.plot_order:
                sig_ids = self.fit_case.get_sig_id_for_component(component_name=component)  # type: Tuple[int, ...]
                component_condition = mc_df[self.fit_case.sig_id_column_name].isin(sig_ids)  # type: pd.Series
                template_condition = decay_mode_condition & component_condition  # type: pd.Series

                template = Template(
                    name=f"{mode_name}_{component}",
                    latex_label=self.fit_case.plot_labels[component],
                    color=self.fit_case.plot_colors[component],
                    process_name=component,
                    dimensions=self.fit_dimensions,
                    bins=self.get_fit_variable_binnings_for_mode(mode=mode_name),
                    scope=self.get_fit_variable_scopes_for_mode(mode=mode_name),
                    params=self.param_handler,
                    data_column_names=self.fit_variable_column_names,
                    data=mc_df.loc[template_condition],
                    weights=self.weight_column_name,
                    systematics=get_systematics(df=mc_df, sys_cols_infos=self.fit_case.systematics_cols),
                    log_scale_mask=False,
                )

                template_id = self.model.add_template(
                    template=template, yield_parameter=f"yield_{component}", use_other_systematics=True
                )
                self.channel_templates[mode_name].append(f"temp_{template_id}")

                eff_param_name = f"eff_{mode_name}_{component}"
                self.model.add_model_parameter(
                    name=eff_param_name,
                    parameter_type="efficiency",
                    floating=False,
                    initial_value=self.initial_eff_dict[component][mode_name],
                    constrain_to_value=self.initial_eff_dict[component][mode_name],
                    constraint_sigma=0.1,
                )
                self.channel_eff_parameters[mode_name].append(eff_param_name)

            self.model.add_channel(
                efficiency_parameters=self.channel_eff_parameters[mode_name],
                name=mode_name,
                components=self.channel_templates[mode_name],
            )

    def _add_data(
        self,
        data_df: pd.DataFrame,
        weight_column_name: Optional[str] = None,
    ) -> None:
        data_channel_infos = {}  # type: Dict[str, pd.DataFrame]
        data_channel_weight_infos = {}  # type: Dict[str, str]
        for mode_name, mode_ids in self.fit_case.reco_mode_infos:
            decay_mode_condition = self._get_decay_mode_condition(df=data_df, mode_ids=mode_ids)
            data_channel_infos.update({f"{mode_name}": data_df.loc[decay_mode_condition]})
            if weight_column_name is not None:
                assert weight_column_name in data_df.columns, weight_column_name
                data_channel_weight_infos.update({f"{mode_name}": weight_column_name})

        self.model.add_data(
            channels=data_channel_infos,
            channel_weights=data_channel_weight_infos if weight_column_name is not None else None,
        )

    def _get_decay_mode_condition(self, df: pd.DataFrame, mode_ids: Sequence[int]) -> pd.Series:
        return df[self.decay_mode_column_name].isin(mode_ids)

    @property
    def fit_model(self) -> FitModel:
        return self.model

    @property
    def fit_case_name(self) -> str:
        return self._fit_case

    @property
    def fit_case(self) -> FitCase:
        return self.fit_cases[self.fit_case_name]

    @property
    def plot_order(self) -> Tuple[str, ...]:
        return self.fit_case.plot_order

    @property
    def weight_column_name(self) -> str:
        return self.fit_case.weight_column_name

    @property
    def fit_variables(self) -> Tuple[HistVariable, ...]:
        return self.fit_case.hist_variables

    @property
    def fit_dimensions(self) -> int:
        return len(self.fit_variables)

    @property
    def fit_variable_column_names(self) -> List[str]:
        return [h_var.df_label for h_var in self.fit_variables]

    def get_fit_variable_binnings_for_mode(self, mode: str) -> BinsInputType:
        mode_dependent_bins = self.fit_case.binning_by_mode_dicts
        bins_in_mode = tuple(
            [
                binning[mode] if binning is not None else h_var.n_bins
                for binning, h_var in zip(mode_dependent_bins, self.fit_variables)
            ]
        )
        if 0 < sum([isinstance(binning, tuple) for binning in bins_in_mode]) < len(bins_in_mode):
            mode_dependent_scopes = self.get_fit_variable_scopes_for_mode(mode=mode)
            converted_binning = []  # type: List[Tuple[float, ...]]
            for binning, scope in zip(bins_in_mode, mode_dependent_scopes):
                if isinstance(binning, int):
                    updated_binning = np.linspace(start=scope[0], stop=scope[1], num=binning + 1, endpoint=True)
                    converted_binning.append(tuple(updated_binning))
                else:
                    converted_binning.append(binning)
            assert len(converted_binning) == len(bins_in_mode), (len(converted_binning), len(bins_in_mode))
            return tuple(converted_binning)
        else:
            return bins_in_mode

    def get_fit_variable_scopes_for_mode(self, mode: str) -> Tuple[Tuple[float, float], ...]:
        mode_dependent_scopes = self.fit_case.scope_by_mode_dicts
        return tuple(
            [
                scope[mode] if scope is not None else h_var.scope
                for scope, h_var in zip(mode_dependent_scopes, self.fit_variables)
            ]
        )
