"""
Provides fit models based on the TemplateFitter package and tailored to this analysis.

Felix Metzner 2019
"""


from typing import Dict, Tuple

from templatefitter.plotter.plot_style import KITColors
from templatefitter.plotter.histogram_variable import HistVariable

import rdstar.offline_analysis.histogram_variables as rdstar_hist_vars
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.systematics.systematics import full_systematics_cols
from rdstar.offline_analysis.mc_ids import rdstar_sig_queries, sig_id_to_label
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.rdstar_plots import rdstar_dec_modes, rdstar_plot_order, get_rdstar_color_dict
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector


__all__ = [
    "rdstar_fit_cases",
]

default_template_setup = tuple(rdstar_plot_order)  # type: Tuple[str, ...]
default_plot_colors = {
    comp: get_rdstar_color_dict()[rdstar_sig_queries[comp].id] for comp in default_template_setup
}  # type: Dict[str, str]
default_plot_labels = {
    comp: sig_id_to_label[rdstar_sig_queries[comp].id] for comp in default_template_setup
}  # type: Dict[str, str]


simple_signal_template_setup = {
    "continuum": ("CBKG",),
    "bbBkg": ("GBKG", "HBKG"),
    "myDSSl": ("DSSlbkg", "DSSl"),
    "myDSl": ("DSl",),
    "myDl": ("Dl",),
    "myDStau": ("DStau",),
    "myDtau": ("Dtau",),
}  # type: Dict[str, Tuple[str, ...]]
simple_signal_plot_colors = {
    "myDtau": KITColors.kit_green,
    "myDStau": KITColors.kit_maygreen,
    "myDl": KITColors.kit_blue,
    "myDSl": KITColors.kit_cyan,
    "myDSSl": KITColors.kit_red,
    "bbBkg": KITColors.grey,
    "continuum": KITColors.light_grey,
}  # type: Dict[str, str]
simple_signal_plot_labels = {
    "myDtau": LatexDecayProcesses.to_d_tau_nu,
    "myDStau": LatexDecayProcesses.to_d_st_tau_nu,
    "myDl": LatexDecayProcesses.to_d_l_nu,
    "myDSl": LatexDecayProcesses.to_d_st_l_nu,
    "myDSSl": LatexDecayProcesses.to_dstst_l_nu,
    "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
    "continuum": LatexDecayProcesses.to_continuum_txt,
}  # type: Dict[str, str]


sideband_template_setup = {
    "continuum": ("CBKG",),
    "bbBkg": ("GBKG", "HBKG", "DSSlbkg", "DSSl", "DStau", "Dtau"),
    "myDSl": ("DSl",),
    "myDl": ("Dl",),
}  # type: Dict[str, Tuple[str, ...]]
sideband_plot_colors = {
    "myDl": KITColors.kit_blue,
    "myDSl": KITColors.kit_cyan,
    "bbBkg": KITColors.kit_red,
    "continuum": KITColors.grey,
}  # type: Dict[str, str]
sideband_plot_labels = {
    "myDl": LatexDecayProcesses.to_d_l_nu,
    "myDSl": LatexDecayProcesses.to_d_st_l_nu,
    "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
    "continuum": LatexDecayProcesses.to_continuum_txt,
}  # type: Dict[str, str]


def replace_mm2_hv(
    fit_case: FitCase,
    new_mm2_col_name: str,
    tag_str: str,
) -> FitCase:
    _original_h_var = fit_case.hist_variables[0]
    _new_h_var = HistVariable(
        df_label=new_mm2_col_name,
        n_bins=_original_h_var.n_bins,
        scope=_original_h_var.scope,
        var_name=_original_h_var.variable_name,
        unit=_original_h_var.unit,
    )

    return FitCase(
        name=fit_case.name + "_" + tag_str.lower(),
        base_dir_name=fit_case.base_dir_name + "_" + tag_str,
        sample_key=fit_case.sample_key,
        output_tag=fit_case.output_tag + "_" + tag_str,
        plot_info_text=fit_case.plot_info_text,
        asimov_data=fit_case.asimov_data,
        round_asimov_data=fit_case.round_asimov_data,
        hist_variables=tuple([hv if _i != 0 else _new_h_var for _i, hv in enumerate(fit_case.hist_variables)]),
        binning_by_mode_dicts=fit_case.binning_by_mode_dicts,
        scope_by_mode_dicts=fit_case.scope_by_mode_dicts,
        weight_column_name=fit_case.weight_column_name,
        systematics_cols=fit_case.systematics_cols,
        reco_mode_infos=fit_case.reco_mode_infos,
        template_setup=fit_case.template_setup,
        plot_colors=fit_case.plot_colors,
        plot_labels=fit_case.plot_labels,
    )


simple_signal_fit_case = FitCase(
    name="simple_signal_fit",
    base_dir_name="Fitting_Signal_Simple",
    sample_key="",
    output_tag="Fit_Signal_Simple_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=simple_signal_template_setup,
    plot_colors=simple_signal_plot_colors,
    plot_labels=simple_signal_plot_labels,
)

simple_signal_asimov_fit_case = FitCase(
    name="simple_signal_asimov",
    base_dir_name="Fitting_AsimovSignal_Simple",
    sample_key="",
    output_tag="AsimovFit_Signal_Simple_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=simple_signal_template_setup,
    plot_colors=simple_signal_plot_colors,
    plot_labels=simple_signal_plot_labels,
)

signal_fit_case = FitCase(
    name="signal_fit",
    base_dir_name="Fitting_Signal",
    sample_key="",
    output_tag="Fit_Signal_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=default_template_setup,
    plot_colors=default_plot_colors,
    plot_labels=default_plot_labels,
)

signal_asimov_fit_case = FitCase(
    name="signal_asimov",
    base_dir_name="Fitting_AsimovSignal",
    sample_key="",
    output_tag="AsimovFit_Signal_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=default_template_setup,
    plot_colors=default_plot_colors,
    plot_labels=default_plot_labels,
)

signal_asimov_fit_case_scipy = FitCase(
    name="signal_asimov_scipy",
    base_dir_name="Fitting_AsimovSignal_Scipy",
    sample_key="",
    output_tag="AsimovFit_Signal_FixedNuisanceParams_Scipy",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=default_template_setup,
    plot_colors=default_plot_colors,
    plot_labels=default_plot_labels,
    minimizer_id="scipy",
)

q2_sideband_asimov_fit_case = FitCase(
    name="q2_sideband_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb, rdstar_hist_vars.pl_star_sideband_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=sideband_template_setup,
    plot_colors=sideband_plot_colors,
    plot_labels=sideband_plot_labels,
)

q2_sideband_fit_case = FitCase(
    name="q2_sideband_fit",
    base_dir_name="Fitting_Q2Sideband",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb, rdstar_hist_vars.pl_star_sideband_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=sideband_template_setup,
    plot_colors=sideband_plot_colors,
    plot_labels=sideband_plot_labels,
)

q2_sideband_asimov_fit_case_mm2_lep_ph_spc = replace_mm2_hv(
    fit_case=q2_sideband_asimov_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_corrected_col_name(
        correction_setup_name="lepton_phase_space_bin_id_based_mms_resolution_correction"
    ),
    tag_str="LepPhSpc",
)

q2_sideband_fit_case_mm2_lep_ph_spc = replace_mm2_hv(
    fit_case=q2_sideband_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_corrected_col_name(
        correction_setup_name="lepton_phase_space_bin_id_based_mms_resolution_correction"
    ),
    tag_str="LepPhSpc",
)

q2_sideband_fit_case_scipy = FitCase(
    name="q2_sideband_fit_scipy",
    base_dir_name="Fitting_Q2Sideband_Scipy",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_FixedNuisanceParams_Scipy",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb, rdstar_hist_vars.pl_star_sideband_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=sideband_template_setup,
    plot_colors=sideband_plot_colors,
    plot_labels=sideband_plot_labels,
    minimizer_id="scipy",
)

q2_sideband_fit_case_scipy_no_sys = FitCase(
    name="q2_sideband_fit_scipy_no_sys",
    base_dir_name="Fitting_Q2Sideband_Scipy_NoSys",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_Scipy_NoSys",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb, rdstar_hist_vars.pl_star_sideband_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=[],
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=sideband_template_setup,
    plot_colors=sideband_plot_colors,
    plot_labels=sideband_plot_labels,
    minimizer_id="scipy",
)

signal_wide_fit_case = FitCase(
    name="signal_wide_fit",
    base_dir_name="Fitting_Signal_Wide",
    sample_key="",
    output_tag="Fit_Signal_Wide_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_wide, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=default_template_setup,
    plot_colors=default_plot_colors,
    plot_labels=default_plot_labels,
)

signal_wide_asimov_fit_case = FitCase(
    name="signal_wide_asimov",
    base_dir_name="Fitting_AsimovSignal_Wide",
    sample_key="",
    output_tag="AsimovFit_Signal_Wide_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_wide, rdstar_hist_vars.pl_star_signal_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=default_template_setup,
    plot_colors=default_plot_colors,
    plot_labels=default_plot_labels,
)

q2_sideband_wide_asimov_fit_case = FitCase(
    name="q2_sideband_wide_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband_Wide",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    asimov_data=True,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb_wide, rdstar_hist_vars.pl_star_sideband_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=sideband_template_setup,
    plot_colors=sideband_plot_colors,
    plot_labels=sideband_plot_labels,
)

q2_sideband_wide_fit_case = FitCase(
    name="q2_sideband_wide_fit",
    base_dir_name="Fitting_Q2Sideband_Wide",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Data",
    asimov_data=False,
    round_asimov_data=False,
    hist_variables=(rdstar_hist_vars.missing_mass_res_corr_q2sb_wide, rdstar_hist_vars.pl_star_sideband_fit),
    binning_by_mode_dicts=(rdstar_hist_vars.missing_mass_binning_dict, None),
    scope_by_mode_dicts=(None, None),
    weight_column_name=WeightManager.final_weight_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=[(k, v) for k, v in rdstar_dec_modes.items()],
    template_setup=sideband_template_setup,
    plot_colors=sideband_plot_colors,
    plot_labels=sideband_plot_labels,
)

rdstar_fit_cases = {
    # Signal Window
    # simple_signal_fit_case.name: simple_signal_fit_case,  # TODO: Include for Box Opening
    simple_signal_asimov_fit_case.name: simple_signal_asimov_fit_case,
    # signal_fit_case.name: signal_fit_case,  # TODO: Include for Box Opening
    # signal_asimov_fit_case.name: signal_asimov_fit_case,  # TODO: Does not work!!!
    # signal_wide_fit_case.name: signal_wide_fit_case,  # TODO: Include for Box Opening
    # signal_wide_asimov_fit_case.name: signal_wide_asimov_fit_case,  # TODO: Does not work!!!
    #
    # q2 Sideband
    q2_sideband_asimov_fit_case.name: q2_sideband_asimov_fit_case,
    q2_sideband_fit_case.name: q2_sideband_fit_case,
    q2_sideband_asimov_fit_case_mm2_lep_ph_spc.name: q2_sideband_asimov_fit_case_mm2_lep_ph_spc,
    q2_sideband_fit_case_mm2_lep_ph_spc.name: q2_sideband_fit_case_mm2_lep_ph_spc,
    q2_sideband_wide_asimov_fit_case.name: q2_sideband_wide_asimov_fit_case,  # TODO: Does not work with first selection!
    # q2_sideband_wide_fit_case.name: q2_sideband_wide_fit_case,  # Removed because of Resolution issues.
    #
    # Signal Window --- SciPy
    # signal_asimov_fit_case_scipy.name: signal_asimov_fit_case_scipy,  # TODO: Does not work: Fix issues with SciPy
    #
    # q2 Sideband --- SciPy
    # q2_sideband_fit_case_scipy.name: q2_sideband_fit_case_scipy,  # TODO: Does not work: Fix issues with SciPy
    # q2_sideband_fit_case_scipy_no_sys.name: q2_sideband_fit_case_scipy_no_sys,  # TODO: Does not work: Fix issues with SciPy
}  # type: Dict[str, FitCase]
