"""
Histogram Variables

This file provides the histogram variables used in the analysis.

Felix Metzner 2020
"""

from typing import Dict

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable
from rdstar.offline_analysis.observables.e_miss_minus_p_miss import EMissMinusPMissVariable
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector

__all__ = [
    # For Fitting
    "missing_mass",
    "missing_mass_res_corr",
    "missing_mass_res_corr_wide",
    "missing_mass_res_corr_q2sb",
    "missing_mass_res_corr_q2sb_wide",
    "missing_mass_binning_dict",
    "pl_star_signal_fit",
    "pl_star_sideband_fit",
    "pl_star",
    "e_miss_minus_p_miss",
    # For Sideband Validation:
    "neextra_fine",
    "eextra",
    "cs_mva_output",
    "loose_cs_mva_output",
    "pl_star_SidebandValidation",
    "missing_mass_q2_SidebandValidation",
    "missing_mass_res_corr_q2sbSidebandValidation",
    "mbc_SidebandValidation",
    "tag_delta_e",
    "mbc_in_mbcsb_SidebandValidation",
    "mbc_in_mbcsb_SidebandValidation_rough_binning",
    "sig_vertex_fit_prob",
    # For FSR Validation
    "fsr_energy_cms",
    "fsr_energy_rest_frame",
    "fsr_energy_lab",
    "fsr_momentum_cms",
    "fsr_momentum_rest_frame",
    "fsr_momentum_lab",
    # Analysis Variables Definitions:
    "q2_residual_with_q2reco",
    "q2_residual_with_q2reco2",
    "q2_residual_with_q2reco3",
    "q2_residual_with_q2reco4",
    "q2_resolution_with_q2reco",
    "q2_resolution_with_q2reco2",
    "q2_resolution_with_q2reco3",
    "q2_resolution_with_q2reco4",
]

# Missing Mass Squared Definitions

missing_mass = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=30,
    scope=(-2.0, 6.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_res_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=42,
    scope=(-0.5, 10.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_res_corr_wide = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=48,
    scope=(-2.0, 10.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_res_corr_q2sb = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=24,
    scope=(-0.5, 4.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_res_corr_q2sb_wide = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=24,
    scope=(-2.0, 4.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_binning_dict = {"DZ": 22, "DM": 20, "DSZ": 22, "DSM": 17, "all_modes": 22}  # type: Dict[str, int]

# Lepton Momentum


pl_star_signal_fit = HistVariable(
    df_label="pStarL",
    n_bins=15,
    scope=(0.1, 2.2),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)

pl_star_sideband_fit = HistVariable(
    df_label="pStarL",
    n_bins=15,
    scope=(0.2, 2.3),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)

pl_star = HistVariable(
    df_label="pStarL",
    n_bins=20,
    scope=(0.1, 2.3),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)

# E_Miss - p_Miss Definition

e_miss_minus_p_miss = HistVariable(
    df_label=EMissMinusPMissVariable.e_miss_minus_p_miss_col_name,
    n_bins=33,
    scope=(-5.125, 3.125),
    var_name=LatexObsNames.fitting_e_miss_minus_p_miss,
    unit=LatexUnits.gev,
)

e_miss_minus_p_miss_cms = HistVariable(
    df_label=EMissMinusPMissVariable.e_miss_minus_p_miss_cms_col_name,
    n_bins=60,
    scope=(-1.0, 2.0),
    var_name=LatexObsNames.fitting_e_miss_minus_p_miss_cms,
    unit=LatexUnits.gev,
)  # type: HistVariable

# Sideband Validation Variables

neextra_fine = HistVariable(
    df_label="ROE_neextra__boCleanROE__bc",
    n_bins=40,
    scope=(0.0, 2.5),
    var_name=LatexObsNames.roe_n_e_ecl_extra,
    unit=LatexUnits.gev,
)

eextra = HistVariable(
    df_label="ROE_eextra__boCleanROE__bc",
    n_bins=20,
    scope=(0.0, 3.0),
    var_name=LatexObsNames.roe_e_ecl_extra,
    unit=LatexUnits.gev,
)

cs_mva_output = HistVariable(
    df_label="cs_mva_output",
    n_bins=40,
    scope=(0.5, 1.0),
    var_name=LatexObsNames.mva_cs_output,
    unit=LatexUnits.no_unit,
)

loose_cs_mva_output = HistVariable(
    df_label="cs_mva_output",
    n_bins=40,
    scope=(0.2, 1.0),
    var_name=LatexObsNames.mva_cs_output,
    unit=LatexUnits.no_unit,
)

sig_vertex_fit_prob = HistVariable(
    df_label="daughter__bo1__cm__spchiProb__bc",
    n_bins=40,
    scope=(0.0, 1.0),
    var_name=LatexObsNames.b_sig_vertex_prob,
    unit=LatexUnits.no_unit,
)

pl_star_SidebandValidation = HistVariable(
    df_label="pStarL",
    n_bins=25,
    scope=(0.1, 2.2),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)

missing_mass_q2_SidebandValidation = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=25,
    scope=(-2.0, 4.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_res_corr_q2sbSidebandValidation = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=25,
    scope=(-2.0, 4.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

mbc_SidebandValidation = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=25,
    scope=(5.27, 5.29),
    var_name=LatexObsNames.b_tag_mbc,
    unit=LatexUnits.gev,
)

tag_delta_e = HistVariable(
    df_label="daughter__bo0__cm__spdeltaE__bc",
    n_bins=30,
    scope=(-0.05, 0.1),
    var_name=LatexObsNames.b_tag_delta_e,
    unit=LatexUnits.gev,
)

mbc_in_mbcsb_SidebandValidation = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=25,
    scope=(5.22, 5.27),
    var_name=LatexObsNames.b_tag_mbc,
    unit=LatexUnits.gev,
)

mbc_in_mbcsb_SidebandValidation_rough_binning = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=15,
    scope=(5.22, 5.27),
    var_name=LatexObsNames.b_tag_mbc,
    unit=LatexUnits.gev,
)

# FSR Correction Validation Variables
fsr_energy_cms = HistVariable(
    df_label="useCMSFrame__bofsrE__bc",
    n_bins=30,
    scope=(0, 1.0),
    var_name=LatexObsNames.fsr_cms_energy,
    unit=LatexUnits.gev,
)

fsr_energy_rest_frame = HistVariable(
    df_label="useRestFrame__bofsrE__bc",
    n_bins=30,
    scope=(0, 1.0),
    var_name=LatexObsNames.fsr_rest_energy,
    unit=LatexUnits.gev,
)

fsr_energy_lab = HistVariable(
    df_label="useRestFrame__bofsrE__bc",
    n_bins=30,
    scope=(0, 1.0),
    var_name=LatexObsNames.fsr_lab_energy,
    unit=LatexUnits.gev,
)

fsr_momentum_cms = HistVariable(
    df_label="useCMSFrame__bofsrMomentum__bc",
    n_bins=30,
    scope=(0, 1.0),
    var_name=LatexObsNames.fsr_cms_p,
    unit=LatexUnits.gev,
)

fsr_momentum_rest_frame = HistVariable(
    df_label="useRestFrame__bofsrMomentum__bc",
    n_bins=30,
    scope=(0, 1.0),
    var_name=LatexObsNames.fsr_rest_p,
    unit=LatexUnits.gev,
)

fsr_momentum_lab = HistVariable(
    df_label="useRestFrame__bofsrMomentum__bc",
    n_bins=30,
    scope=(0, 1.0),
    var_name=LatexObsNames.fsr_lab_p,
    unit=LatexUnits.gev,
)

# Unused Variables
# Dstarstar Variables
missing_mass_dstarstar = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=25,
    scope=(-2.5, 8.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_dstarstar_rough_binning = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=15,
    scope=(-2.5, 8.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_dstarstar_mass_ge_1p5 = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=25,
    scope=(1.5, 8.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_dstarstar_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=25,
    scope=(-2.5, 8.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_mass_dstarstar_rough_binning_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=15,
    scope=(-2.5, 8.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

missing_dstarstar_mass_ge_1p5_corr = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_correction_column_name(),
    n_bins=25,
    scope=(1.5, 8.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

# Other
diff_dstar_d_mass = HistVariable(
    df_label="daughter__bo1__cm__spformula__bodaughterInvM__bo0__cm__sp2__bc__sp__mi__spdaughter__bo0__cm__spInvM__bc__bc__bc",
    n_bins=30,
    scope=(0.1, 1.0),
    var_name=LatexObsNames.d_st_st_delta_m,
    unit=LatexUnits.gev,
)

missing_mom_cms = HistVariable(
    df_label="missingMomentumInCMSFrame",
    n_bins=30,
    scope=(0.0, 1.5),
    var_name=LatexObsNames.fitting_p_miss_star,
    unit=LatexUnits.gev,
)

# Analysis Plot Variables

q2_residual_with_q2reco = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_residual_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_resi,
    unit=LatexUnits.gev_sq,
)
q2_residual_with_q2reco2 = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_residual_with_q2reco_2_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_resi_v2,
    unit=LatexUnits.gev_sq,
)
q2_residual_with_q2reco3 = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_residual_with_q2reco_3_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_resi_v3,
    unit=LatexUnits.gev_sq,
)
q2_residual_with_q2reco4 = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_residual_with_q2reco_4_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_resi_v4,
    unit=LatexUnits.gev_sq,
)

q2_resolution_with_q2reco = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_resolution_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_reso,
    unit=LatexUnits.no_unit,
)
q2_resolution_with_q2reco2 = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_resolution_with_q2reco_2_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_reso_v2,
    unit=LatexUnits.no_unit,
)
q2_resolution_with_q2reco3 = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_resolution_with_q2reco_3_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_reso_v3,
    unit=LatexUnits.no_unit,
)
q2_resolution_with_q2reco4 = HistVariable(
    df_label=QSquaredResidualVariable.q_squared_resolution_with_q2reco_4_col_name,
    n_bins=100,
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.q2_reso_v4,
    unit=LatexUnits.no_unit,
)
