"""
Simple script to run manual merge of Data n-tuple files, which is necessary because of their large size.

Felix Metzner 2021
"""

import os
import numpy as np
import pandas as pd

from typing import List

from analysistools.dataframe_tools.hdf5_utils import load_root_file

from rdstar.utilities import PathType

__all__ = [
    "ManualDataNTupleMerger",
]


class ManualDataNTupleMerger:
    file_tag = "base"  # type: str
    base_file_name = "DataY4S_Y4S_ntuple"  # type: str

    def __init__(self, dir_base_path: PathType, parts: int) -> None:
        self.base_path = dir_base_path  # type: PathType
        self.parts = parts  # type: int

    def do_merge(self) -> None:
        dfs = []  # type: List[pd.DataFrame]
        for part in range(self.parts):
            print(f"Loading data frame {part + 1} of {self.parts}...")
            dfs.append(self.load_part(part=int(part + 1)))

        list_of_events = [len(df.index) for df in dfs]
        list_of_events_u = [len(np.unique(df.index)) for df in dfs]

        print(f"Merging {self.parts} data frames...")
        merged_df = pd.concat(dfs, ignore_index=True, sort=False)

        for df in dfs:
            if isinstance(df, pd.DataFrame):
                del df

        merged_df.reset_index(drop=True, inplace=True)

        print("Checking merged data frame...")
        assert list_of_events == list_of_events_u, (list_of_events, list_of_events_u)
        assert len(np.unique(merged_df.index)) == sum(list_of_events), (
            len(np.unique(merged_df.index)),
            sum(list_of_events),
            list_of_events,
        )

        hdf5_key = self.base_file_name  # type: str
        file_name = f"{hdf5_key}_{self.file_tag}.h5"
        hdf5_file_path = os.path.join(self.base_path, file_name)  # type: PathType
        tmp_hdf5_file_path = os.path.join(self.base_path, f"tmp_{file_name}")  # type: PathType
        print(f"Writing merged data frame to {hdf5_file_path}")
        assert not os.path.exists(hdf5_file_path), hdf5_file_path
        assert not os.path.exists(tmp_hdf5_file_path), tmp_hdf5_file_path
        merged_df.to_hdf(path_or_buf=tmp_hdf5_file_path, key=hdf5_key)
        os.rename(src=tmp_hdf5_file_path, dst=hdf5_file_path)
        print(f"Finished writing merged data frame to {hdf5_file_path}")

    def load_part(self, part: int) -> pd.DataFrame:
        root_part_path = self.get_root_part_path(part=part)  # type: PathType
        print(f"Loading part {part}: {root_part_path}")
        assert os.path.exists(root_part_path), root_part_path

        hdf5_key = f"{self.base_file_name}_part{part}"  # type: str
        file_name = f"{hdf5_key}.h5"  # type: str
        hdf5_file_path = os.path.join(self.base_path, file_name)  # type: PathType
        tmp_hdf5_file_path = os.path.join(self.base_path, f"tmp_{file_name}")  # type: PathType
        assert not os.path.exists(hdf5_file_path), hdf5_file_path
        assert not os.path.exists(tmp_hdf5_file_path), tmp_hdf5_file_path

        df = load_root_file(file_path=root_part_path)

        print(f"Writing part {part} to hdf5 file: {hdf5_file_path}")
        df.to_hdf(path_or_buf=tmp_hdf5_file_path, key=hdf5_key)
        os.rename(src=tmp_hdf5_file_path, dst=hdf5_file_path)
        print(f"Saved part {part}: {hdf5_file_path}")
        return df

    def get_root_part_path(self, part: int) -> PathType:
        return os.path.join(self.base_path, f"{self.base_file_name}_part{part}.root")

    @staticmethod
    def get_base_dir_path(production_tag: str) -> PathType:
        return f"/ceph/fmetzner/rdstar/data_set_{production_tag}_stream0/DataY4S/"


if __name__ == "__main__":
    my_production_tag = "prod_02ndMay2021"
    path = ManualDataNTupleMerger.get_base_dir_path(production_tag=my_production_tag)

    merger = ManualDataNTupleMerger(dir_base_path=path, parts=3)
    merger.do_merge()
