"""
This file contains definitions and functions for adding signal and
generator ids to monte carlo samples.
"""
import copy
import logging

import numpy as np
import pandas as pd

from typing import Optional, Dict, List, Callable, NamedTuple

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "MatchQueryInfo",
    "SigQueryDictType",
    "default_sig_id_col_name",
    "lepton_pdg_ids",
    "tau_pdg_id",
    "d_pdg_ids",
    "dst_pdg_ids",
    "all_d_pdg_ids",
    "valid_forth_daughter",
    "pizero_pdg_id",
    "picharged_pdg_id",
    "rdstar_sig_queries",
    "sig_id_to_label",
    "decay_mode_to_sig_id",
    "decay_mode_to_latex_label",
    "add_sig_id_to_df",
    "gen_b_ids",
    "gen_id_to_label",
    "add_gen_b_id_to_df",
]


class MatchQueryInfo(NamedTuple):
    query: Callable
    id: int


SigQueryDictType = Dict[str, MatchQueryInfo]
GenQueryDictType = SigQueryDictType


# general stuff
# =============

default_sig_id_col_name = "SIG_ID"  # type: str

lepton_pdg_ids = [11, 13]  # type: List[int]
tau_pdg_id = 15  # type: int

d_pdg_ids = [411, 421]  # type: List[int]
dst_pdg_ids = [413, 423]  # type: List[int]

done_pdg_ids = [10413, 10423]  # type: List[int]
dtwostar_pdg_ids = [415, 425]  # type: List[int]
dzerostar_pdg_ids = [10411, 10421]  # type: List[int]
dprimeone_pdg_ids = [20413, 20423]  # type: List[int]

dstst_pdg_ids = done_pdg_ids + dtwostar_pdg_ids + dzerostar_pdg_ids + dprimeone_pdg_ids  # type: List[int]
all_d_pdg_ids = d_pdg_ids + dst_pdg_ids + dstst_pdg_ids  # type: List[int]

b_meson_pdg_ids = [511, 521]  # type: List[int]
valid_forth_daughter = [22, -999]  # type: List[int]

pizero_pdg_id = 111  # type: int
picharged_pdg_id = 211  # type: int


# signal ids
# ==========
def is_lepton_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the mcMatched(PDG) code of daughter 1 of the signal B meson
    matches a lepton pdg id.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    fourth_pdg_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo3__cmPDG__bc__bc__bc__bc"

    is_lepton = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc).isin(lepton_pdg_ids)
    is_from_b = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc).isin(b_meson_pdg_ids)
    has_no_forth_daughter = sample[fourth_pdg_var].isin(valid_forth_daughter)

    return is_lepton & is_from_b & has_no_forth_daughter


def is_tau_to_lepton_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the mcMatched(PDG) code of daughter 1 of the signal B meson
    matches a lepton pdg id and if the mcMother(PDG) is a tau lepton.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    b_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__boPDG__bc__bc__bc__bc"
    fourth_pdg_var = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomcDaughter__bo3__cmPDG__bc__bc__bc__bc__bc"
    )

    is_lepton = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc).isin(lepton_pdg_ids)
    is_from_tau = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc) == tau_pdg_id
    is_from_b = abs(sample[b_pdg_var_name]).isin(b_meson_pdg_ids)
    has_no_forth_daughter = sample[fourth_pdg_var].isin(valid_forth_daughter)

    return is_lepton & is_from_tau & is_from_b & has_no_forth_daughter


def is_tau_to_hadron_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the reconstructed first daughter of the B meson originates from a tau,
    but does not match a lepton PDG.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    lepton_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc"
    b_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__boPDG__bc__bc__bc__bc"
    fourth_pdg_var = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomcDaughter__bo3__cmPDG__bc__bc__bc__bc__bc"
    )

    is_not_lepton = ~abs(sample[lepton_pdg_var_name]).isin(lepton_pdg_ids)
    is_from_tau = abs(sample.daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc) == tau_pdg_id
    is_from_b = abs(sample[b_pdg_var_name]).isin(b_meson_pdg_ids)
    has_no_forth_daughter = sample[fourth_pdg_var].isin(valid_forth_daughter)

    return is_not_lepton & is_from_tau & is_from_b & has_no_forth_daughter


def b_to_dstst_l_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D**(->D(*)pi0)lv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"
    d_daughter_pdg_var_name = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp"
        "mcDaughter__bo1__cm__spPDG__bc__bc__bc__bc__bc"
    )
    dstst2_pdg_var = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp"
        "mcDaughter__bo2__cm__spPDG__bc__bc__bc__bc__bc"
    )

    is_lepton = is_lepton_signal(sample)
    is_b_to_dstst = np.abs(sample[d_pdg_var_name]).isin(dstst_pdg_ids)
    is_dstst_with_pi_zero = np.abs(sample[d_daughter_pdg_var_name]) == pizero_pdg_id
    is_not_continuum = sample.isNotContinuumEvent == 1.0

    if dstst2_pdg_var in sample.columns:
        dstst_has_no_second_daughter = sample[dstst2_pdg_var].isin(valid_forth_daughter)
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_zero & is_not_continuum & dstst_has_no_second_daughter
    else:
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_zero & is_not_continuum


def b_to_dstst_l_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D**(->D(*)pi+)lv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"
    d_daughter_pdg_var_name = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp"
        "mcDaughter__bo1__cm__spPDG__bc__bc__bc__bc__bc"
    )
    dstst2_pdg_var = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__sp"
        "mcDaughter__bo2__cm__spPDG__bc__bc__bc__bc__bc"
    )

    is_lepton = is_lepton_signal(sample)
    is_b_to_dstst = np.abs(sample[d_pdg_var_name]).isin(dstst_pdg_ids)
    is_dstst_with_pi_charged = np.abs(sample[d_daughter_pdg_var_name]) == picharged_pdg_id
    is_not_continuum = sample.isNotContinuumEvent == 1.0
    if dstst2_pdg_var in sample.columns:
        dstst_has_no_second_daughter = sample[dstst2_pdg_var].isin(valid_forth_daughter)
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_charged & is_not_continuum & dstst_has_no_second_daughter
    else:
        return is_lepton & is_b_to_dstst & is_dstst_with_pi_charged & is_not_continuum


def b_to_d_l_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->Dlv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"

    is_lepton = is_lepton_signal(sample)
    is_b_to_d = np.abs(sample[d_pdg_var_name]).isin(d_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.0

    return is_lepton & is_b_to_d & is_not_continuum


def b_to_dst_l_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D*lv decay.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcDaughter__bo0__cm__spPDG__bc__bc__bc__bc"

    is_lepton = is_lepton_signal(sample)
    is_b_to_dst = np.abs(sample[d_pdg_var_name]).isin(dst_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.0

    return is_lepton & is_b_to_dst & is_not_continuum


def b_to_d_tau_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->Dtv.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bo"
        "mcDaughter__bo0__cm__spPDG__bc__bc__bc__bc__bc"
    )

    is_tau_to_lepton = is_tau_to_lepton_signal(sample)
    is_b_to_d = np.abs(sample[d_pdg_var_name]).isin(d_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.0

    return is_tau_to_lepton & is_b_to_d & is_not_continuum


def b_to_dst_tau_signal(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is a lepton
    originating from a B->D*tv.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    d_pdg_var_name = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bo"
        "mcDaughter__bo0__cm__spPDG__bc__bc__bc__bc__bc"
    )

    is_tau_to_lepton = is_tau_to_lepton_signal(sample)
    is_b_to_dst = np.abs(sample[d_pdg_var_name]).isin(dst_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.0

    return is_tau_to_lepton & is_b_to_dst & is_not_continuum


def hadronic_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the first daughter of the signal B meson is not
    a lepton.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    l_pdg_var_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc"

    fake_lepton = ~np.abs(sample[l_pdg_var_name]).isin(lepton_pdg_ids)
    is_not_continuum = sample.isNotContinuumEvent == 1.0

    return fake_lepton & is_not_continuum


def continuum_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if the event is a continuum event.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    return sample.isContinuumEvent == 1.0


def bbar_bkg(sample: pd.DataFrame) -> pd.Series:
    """
    Checks if an event does not match any other checks above.
    This would be classified as BBbar background.

    :param sample: pandas.DataFrame
    :return: pandas.Series with boolean entries.
    """
    not_dl = ~b_to_d_l_signal(sample)
    not_dstl = ~b_to_dst_l_signal(sample)
    not_dststl = ~b_to_dstst_l_signal(sample)
    not_dststl_bkg = ~b_to_dstst_l_bkg(sample)
    not_dtau = ~b_to_d_tau_signal(sample)
    not_dsttau = ~b_to_dst_tau_signal(sample)
    not_hadronic_bkg = ~hadronic_bkg(sample)
    not_continuum = ~continuum_bkg(sample)
    return not_dl & not_dstl & not_dststl & not_dststl_bkg & not_dtau & not_dsttau & not_hadronic_bkg & not_continuum


rdstar_sig_queries = {
    "Dtau": MatchQueryInfo(query=b_to_d_tau_signal, id=1),
    "DStau": MatchQueryInfo(query=b_to_dst_tau_signal, id=2),
    "Dl": MatchQueryInfo(query=b_to_d_l_signal, id=3),
    "DSl": MatchQueryInfo(query=b_to_dst_l_signal, id=4),
    "DSSl": MatchQueryInfo(query=b_to_dstst_l_signal, id=5),
    "DSSlbkg": MatchQueryInfo(query=b_to_dstst_l_bkg, id=6),
    "HBKG": MatchQueryInfo(query=hadronic_bkg, id=7),
    "GBKG": MatchQueryInfo(query=bbar_bkg, id=8),
    "CBKG": MatchQueryInfo(query=continuum_bkg, id=9),
}  # type: SigQueryDictType

sig_id_to_label = {
    1: LatexDecayProcesses.to_d_tau_nu,
    2: LatexDecayProcesses.to_d_st_tau_nu,
    3: LatexDecayProcesses.to_d_l_nu,
    4: LatexDecayProcesses.to_d_st_l_nu,
    5: LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero,
    6: LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm,
    7: LatexDecayProcesses.to_hadronic_bkg_txt,
    8: LatexDecayProcesses.to_b_bbar_bkg,
    9: LatexDecayProcesses.to_continuum_txt,
}  # type: Dict[int, str]

decay_mode_to_sig_id = {
    11: [1, 3],  #  B^0 to D- e+                # noqa: E262
    12: [1, 3],  #  B^0 to D- mu+               # noqa: E262
    13: [1, 3],  #  B^- to D0 e-                # noqa: E262
    14: [1, 3],  #  B^- to D0 mu-               # noqa: E262
    15: [2, 4],  #  B^0 to D*- e+               # noqa: E262
    16: [2, 4],  #  B^0 to D*- mu+              # noqa: E262
    17: [2, 4],  #  B^- to D*0 e-               # noqa: E262
    18: [2, 4],  #  B^- to D*0 mu-              # noqa: E262
    61: [5],  #     B0 to D- e+ pi0             # noqa: E262
    62: [5],  #     B0 to D- mu+ pi0            # noqa: E262
    63: [5],  #     B- to D0 e- pi0             # noqa: E262
    64: [5],  #     B- to D0 mu- pi0            # noqa: E262
    65: [5],  #     B0 to D*- e+ pi0            # noqa: E262
    66: [5],  #     B0 to D*- mu+ pi0           # noqa: E262
    67: [5],  #     B- to D*0 e- pi0            # noqa: E262
    68: [5],  #     B- to D*0 mu- pi0           # noqa: E262
    71: [6],  #     B0 to anti-D0 e+ pi-        # noqa: E262
    72: [6],  #     B0 to anti-D0 mu+ pi-       # noqa: E262
    73: [6],  #     B- to D+ e- pi-             # noqa: E262
    74: [6],  #     B- to D+ mu- pi-            # noqa: E262
    75: [6],  #     B0 to anti-D*0 e+ pi-       # noqa: E262
    76: [6],  #     B0 to anti-D*0 mu+ pi-      # noqa: E262
    77: [6],  #     B- to D*+ e- pi-            # noqa: E262
    78: [6],  #     B- to D*+ mu- pi-           # noqa: E262
}  # type: Dict[int, List[int]]

decay_mode_to_latex_label = {
    11: LatexDecayProcesses.reco_b0_to_dm_ep,  #                B0  to  D- e+                 # noqa: E262
    12: LatexDecayProcesses.reco_b0_to_dm_mup,  #               B0  to  D- mu+                # noqa: E262
    13: LatexDecayProcesses.reco_bm_to_d0_em,  #                B-  to  D0 e-                 # noqa: E262
    14: LatexDecayProcesses.reco_bm_to_d0_mum,  #               B-  to  D0 mu-                # noqa: E262
    15: LatexDecayProcesses.reco_b0_to_d_stm_ep,  #             B0  to  D*- e+                # noqa: E262
    16: LatexDecayProcesses.reco_b0_to_d_stm_mup,  #            B0  to  D*- mu+               # noqa: E262
    17: LatexDecayProcesses.reco_bm_to_d_st0_em,  #             B-  to  D*0 e-                # noqa: E262
    18: LatexDecayProcesses.reco_bm_to_d_st0_mum,  #            B-  to  D*0 mu-               # noqa: E262
    61: LatexDecayProcesses.reco_b0_to_dm_ep_pi0,  #            B0  to  D- e+ pi0             # noqa: E262
    62: LatexDecayProcesses.reco_b0_to_dm_mup_pi0,  #           B0  to  D- mu+ pi0            # noqa: E262
    63: LatexDecayProcesses.reco_bm_to_d0_em_pi0,  #            B-  to  D0 e- pi0             # noqa: E262
    64: LatexDecayProcesses.reco_bm_to_d0_mum_pi0,  #           B-  to  D0 mu- pi0            # noqa: E262
    65: LatexDecayProcesses.reco_b0_to_d_stm_ep_pi0,  #         B0  to  D*- e+ pi0            # noqa: E262
    66: LatexDecayProcesses.reco_b0_to_d_stm_mup_pi0,  #        B0  to  D*- mu+ pi0           # noqa: E262
    67: LatexDecayProcesses.reco_bm_to_d_st0_em_pi0,  #         B-  to  D*0 e- pi0            # noqa: E262
    68: LatexDecayProcesses.reco_bm_to_d_st0_mum_pi0,  #        B-  to  D*0 mu- pi0           # noqa: E262
    71: LatexDecayProcesses.reco_b0_to_anti_d0_ep_pim,  #       B0  to  anti-D0 e+ pi-        # noqa: E262
    72: LatexDecayProcesses.reco_b0_to_anti_d0_mup_pim,  #      B0  to  anti-D0 mu+ pi-       # noqa: E262
    73: LatexDecayProcesses.reco_bm_to_dp_em_pim,  #            B-  to  D+ e- pi-             # noqa: E262
    74: LatexDecayProcesses.reco_bm_to_dp_mum_pim,  #           B-  to  D+ mu- pi-            # noqa: E262
    75: LatexDecayProcesses.reco_b0_to_anti_d_st0_ep_pim,  #    B0  to  anti-D*0 e+ pi-       # noqa: E262
    76: LatexDecayProcesses.reco_b0_to_anti_d_st0_mup_pim,  #   B0  to  anti-D*0 mu+ pi-      # noqa: E262
    77: LatexDecayProcesses.reco_bm_to_d_stp_em_pim,  #         B-  to  D*+ e- pi-            # noqa: E262
    78: LatexDecayProcesses.reco_bm_to_d_stp_mum_pim,  #        B-  to  D*+ mu- pi-           # noqa: E262
}  # type: Dict[int, str]


def add_sig_id_to_df_dict(
    df_dict: Dict[str, Dict[str, pd.DataFrame]],
    sig_query_dict: Optional[SigQueryDictType] = None,
) -> None:
    """Adds signal ids to a dictionary of dictionary pandas.DataFrames.

    :param df_dict: Standard Dict of Dict of pd.DataFrames used in this analysis.
    :param sig_query_dict: rdstar_sig_queries Dict returned from get_rdstar_sig_queries.
        Default is mc_ids.rdstar_sig_queries.
    :return: None
    """
    if sig_query_dict is None:
        sig_query_dict = copy.deepcopy(rdstar_sig_queries)

    for data_type in df_dict.keys():
        logging.info(f"Adding signal ids for data type: {data_type}")

        add_sig_id_to_df(df=df_dict[data_type]["Y4S_ntuple"], sig_query_dict=sig_query_dict)


def add_sig_id_to_df(
    df: pd.DataFrame,
    sig_id_column_name: Optional[str] = None,
    sig_query_dict: Optional[SigQueryDictType] = None,
) -> None:
    if sig_query_dict is None:
        _sig_query_dict = copy.deepcopy(rdstar_sig_queries)  # type: SigQueryDictType
    else:
        _sig_query_dict = sig_query_dict

    if sig_id_column_name is None:
        _sig_id_column_name = default_sig_id_col_name  # type: str
    else:
        _sig_id_column_name = sig_id_column_name

    logging.info(f"Number of candidates: {len(df.index)}")

    assert _sig_id_column_name not in df.columns, _sig_id_column_name
    df.loc[:, _sig_id_column_name] = np.nan

    for key, query_info in _sig_query_dict.items():
        query_fcn = query_info.query  # type: Callable
        query_result = query_fcn(df)

        logging.info(f"Processing query {key}...")
        logging.info(f"Number of matching candidates: {len(df.loc[query_result].index)}")

        if not np.all(pd.isnull(df.loc[query_result, _sig_id_column_name])):
            raise RuntimeWarning("Careful, you are overwriting signal ids")

        df.loc[query_result, _sig_id_column_name] = query_info.id

    check = df[_sig_id_column_name].isnull().sum()
    assert check == 0, f"Sig ID queries do not cover all events. Number of Rows w/o ID: {check}"


# generator ids
# =============
def is_continuum_event(df: pd.DataFrame) -> pd.Series:
    return df["isContinuumEvent"] == 1.0


def second_daughter_is_lepton(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmPDG__bc__bc__bc"]).isin(
        lepton_pdg_ids
    )


def second_daughter_is_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmPDG__bc__bc__bc"]) == tau_pdg_id


def second_daughter_is_tau_to_l(df: pd.DataFrame, b_index: int) -> pd.Series:
    l_pdg = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"

    return second_daughter_is_tau(df, b_index) & abs(df[l_pdg]).isin(lepton_pdg_ids)


def second_daughter_is_tau_to_had(df: pd.DataFrame, b_index: int) -> pd.Series:
    l_pdg = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"

    return second_daughter_is_tau(df, b_index) & ~(abs(df[l_pdg]).isin(lepton_pdg_ids))


def first_daughter_is_D(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(d_pdg_ids)


def first_daughter_is_Dst(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(dst_pdg_ids)


def first_daughter_is_Done(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        done_pdg_ids
    )


def first_daughter_is_Dtwostar(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dtwostar_pdg_ids
    )


def first_daughter_is_Dprimeone(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dprimeone_pdg_ids
    )


def first_daughter_is_Dzerostar(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        dzerostar_pdg_ids
    )


def first_daughter_is_any_D(df: pd.DataFrame, b_index: int) -> pd.Series:
    return abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        all_d_pdg_ids
    )


def first_daughter_is_not_any_D(df: pd.DataFrame, b_index: int) -> pd.Series:
    return ~abs(df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"]).isin(
        all_d_pdg_ids
    )


def b_has_three_daughters(df: pd.DataFrame, b_index: int) -> pd.Series:
    # allow radiative photons also fourth daughter
    return df[f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo3__cmPDG__bc__bc__bc"].isin([22, -999])


def gen_BtoDtau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_d = first_daughter_is_D(df, b_index)
    has_tau_to_l = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_d & has_tau_to_l & has_three_daughters


def gen_BtoDsttau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dst = first_daughter_is_Dst(df, b_index)
    has_tau_to_l = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dst & has_tau_to_l & has_three_daughters


def gen_BtoDl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_d = first_daughter_is_D(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_d & has_l & has_three_daughters


def gen_BtoDstl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dst = first_daughter_is_Dst(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dst & has_l & has_three_daughters


def gen_BtoDonel(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_done = first_daughter_is_Done(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_done & has_l & has_three_daughters


def gen_BtoDtwostarl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dtwostar = first_daughter_is_Dtwostar(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dtwostar & has_l & has_three_daughters


def gen_BtoDprimeonel(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dprimeone = first_daughter_is_Dprimeone(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dprimeone & has_l & has_three_daughters


def gen_BtoDzerostarl(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dzerostar = first_daughter_is_Dzerostar(df, b_index)
    has_l = second_daughter_is_lepton(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dzerostar & has_l & has_three_daughters


def gen_BtoDonetau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_done = first_daughter_is_Done(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_done & has_tau & has_three_daughters


def gen_BtoDtwostartau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dtwostar = first_daughter_is_Dtwostar(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dtwostar & has_tau & has_three_daughters


def gen_BtoDprimeonetau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dprimeone = first_daughter_is_Dprimeone(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dprimeone & has_tau & has_three_daughters


def gen_BtoDzerostartau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_dzerostar = first_daughter_is_Dzerostar(df, b_index)
    has_tau = second_daughter_is_tau_to_l(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_dzerostar & has_tau & has_three_daughters


def gen_hadronic_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    has_any_d = first_daughter_is_any_D(df, b_index)
    has_tau_to_had = second_daughter_is_tau_to_had(df, b_index)
    has_three_daughters = b_has_three_daughters(df, b_index)

    return not_continuum & has_any_d & has_tau_to_had & has_three_daughters


def gen_other(df: pd.DataFrame, b_index: int) -> pd.Series:
    not_continuum = ~is_continuum_event(df)
    not_gen_BtoDl = ~gen_BtoDl(df, b_index)
    not_gen_BtoDstl = ~gen_BtoDstl(df, b_index)
    not_gen_BtoDtau = ~gen_BtoDtau(df, b_index)
    not_gen_BtoDsttau = ~gen_BtoDsttau(df, b_index)
    not_gen_BtoDonel = ~gen_BtoDonel(df, b_index)
    not_gen_BtoDonetau = ~gen_BtoDonetau(df, b_index)
    not_gen_BtoDtwostarl = ~gen_BtoDtwostarl(df, b_index)
    not_gen_BtoDtwostartau = ~gen_BtoDtwostartau(df, b_index)
    not_gen_BtoDprimeonel = ~gen_BtoDprimeonel(df, b_index)
    not_gen_BtoDprimeonetau = ~gen_BtoDprimeonetau(df, b_index)
    not_gen_BtoDzerostarl = ~gen_BtoDzerostarl(df, b_index)
    not_gen_BtoDzerostartau = ~gen_BtoDzerostartau(df, b_index)
    not_gen_hadronic_tau = ~gen_hadronic_tau(df, b_index)

    return (
        not_continuum
        & not_gen_hadronic_tau
        & not_gen_BtoDl
        & not_gen_BtoDstl
        & not_gen_BtoDtau
        & not_gen_BtoDsttau
        & not_gen_BtoDonel
        & not_gen_BtoDonetau
        & not_gen_BtoDtwostarl
        & not_gen_BtoDtwostartau
        & not_gen_BtoDprimeonel
        & not_gen_BtoDprimeonetau
        & not_gen_BtoDzerostarl
        & not_gen_BtoDzerostartau
    )


def gen_continuum(df: pd.DataFrame, b_index: int) -> pd.Series:
    return is_continuum_event(df)


gen_b_ids = {
    "Dtau": MatchQueryInfo(query=gen_BtoDtau, id=1),
    "Dsttau": MatchQueryInfo(query=gen_BtoDsttau, id=2),
    "Dl": MatchQueryInfo(query=gen_BtoDl, id=3),
    "Dstl": MatchQueryInfo(query=gen_BtoDstl, id=4),
    "Donel": MatchQueryInfo(query=gen_BtoDonel, id=51),
    "Donetau": MatchQueryInfo(query=gen_BtoDonetau, id=55),
    "Dtwostarl": MatchQueryInfo(query=gen_BtoDtwostarl, id=52),
    "Dtwostartau": MatchQueryInfo(query=gen_BtoDtwostartau, id=56),
    "Dprimeonel": MatchQueryInfo(query=gen_BtoDprimeonel, id=53),
    "Dprimeonetau": MatchQueryInfo(query=gen_BtoDprimeonetau, id=57),
    "Dzerostarl": MatchQueryInfo(query=gen_BtoDzerostarl, id=54),
    "Dzerostartau": MatchQueryInfo(query=gen_BtoDzerostartau, id=58),
    "hadronic_tau": MatchQueryInfo(query=gen_hadronic_tau, id=6),
    "other": MatchQueryInfo(query=gen_other, id=7),
    "continuum": MatchQueryInfo(query=gen_continuum, id=8),
}  # type: GenQueryDictType

gen_id_to_label = {
    1: LatexDecayProcesses.to_d_tau_nu,
    2: LatexDecayProcesses.to_d_st_tau_nu,
    3: LatexDecayProcesses.to_d_l_nu,
    4: LatexDecayProcesses.to_d_st_l_nu,
    51: LatexDecayProcesses.to_d_1_l_nu,
    52: LatexDecayProcesses.to_dst_2_l_nu,
    53: LatexDecayProcesses.to_dprime1_l_nu,
    54: LatexDecayProcesses.to_d0st_l_nu,
    55: LatexDecayProcesses.to_d_1_tau_nu,
    56: LatexDecayProcesses.to_dst_2_tau_nu,
    57: LatexDecayProcesses.to_dprime1_tau_nu,
    58: LatexDecayProcesses.to_d0st_tau_nu,
    6: LatexDecayProcesses.to_hadronic_tau_bkg_txt,
    7: LatexDecayProcesses.to_other_bkg_txt,
    8: LatexDecayProcesses.to_continuum_txt,
    -999: LatexDecayProcesses.to_inconclusive_txt,
}  # type: Dict[int, str]


def add_gen_b_id_to_df(
    df: pd.DataFrame,
    gen_b_mdst_index: int,
    gen_query_dict: Optional[GenQueryDictType] = None,
) -> None:
    """Adds the given generator B meson ids for the generator B
    with mdst index `gen_b_mdst_index` to a given data frame.

    :param df: A pandas.DataFrame instance (has to contain
    genUpsilon4S variables.)
    :param gen_b_mdst_index: Integer specifying the generator B
    mdst index. Has to be in {1, 2}.
    :param gen_query_dict: Dictionary with True/False values for
    the index. Default is mc_ids.gen_b_ids
    """
    assert gen_b_mdst_index in [1, 2], gen_b_mdst_index

    if gen_query_dict is None:
        gen_query_dict = copy.deepcopy(gen_b_ids)

    logging.debug(f"Number of candidates: {len(df.index)}")
    gen_b_column_name = f"GEN_B{gen_b_mdst_index}_ID"
    df.loc[:, gen_b_column_name] = np.nan

    for key, query_info in gen_query_dict.items():
        logging.debug(f"Processing query {key}...")
        query_fcn = query_info.query  # type: Callable
        query_result = query_fcn(df, gen_b_mdst_index - 1)

        logging.debug(f"Number of matching candidates: {len(df.loc[query_result].index)}")

        if not np.all(pd.isnull(df.loc[query_result, gen_b_column_name])):
            raise RuntimeError("Careful, you are overwriting signal ids")

        df.loc[query_result, gen_b_column_name] = query_info.id

    check = df[gen_b_column_name].isnull().sum()

    if check > 0:
        raise ValueError(f"Gen ID queries do not cover all events.\n" f"Number of Rows w/o ID: {check}")


# generic functions
def apply_fcn_to_df_dict(
    df_dict: Dict[str, Dict[str, pd.DataFrame]],
    fcn: Callable,
    ntuple_key: str = "Y4S_ntuple",
    **kwargs,
) -> None:
    for key, val in df_dict.items():
        logging.info(f"Applying function {fcn.__name__} to data type {key}..")
        fcn(val[ntuple_key], **kwargs)
