"""
Missing Mass Squared Resolution Correction

Felix Metzner 2020
"""

import os
import json
import logging
import numpy as np
import pandas as pd

from typing import Optional, Tuple, List, Dict, NamedTuple, Callable, Any

from rdstar.utilities import PathType
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetup,
    ResolutionCorrectionSetupManager,
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "ResolutionInfo",
    "MissingMassResolutionCorrector",
]


class ResolutionInfo(NamedTuple):
    resolution_correction_setup_name: str
    number_of_bins: int
    sigmas: Tuple[float, ...]
    bin_edges: Optional[Tuple[Tuple[float, float], ...]] = None
    secondary_var: Optional[str] = None
    additional_info: Optional[Dict[str, Any]] = None

    @staticmethod
    def from_file(input_file_path: PathType) -> "ResolutionInfo":
        assert os.path.isfile(input_file_path), input_file_path
        with open(input_file_path, "r") as in_file:
            input_dict = json.load(in_file)

        new_res_info = ResolutionInfo(
            resolution_correction_setup_name=input_dict["resolution_correction_setup_name"],
            number_of_bins=int(input_dict["number_of_bins"]),
            sigmas=tuple(input_dict["sigmas"]),
            bin_edges=tuple([tuple(bin_edge_pair) for bin_edge_pair in input_dict["bin_edges"]]),  # type: ignore
            secondary_var=input_dict["secondary_var"],
            additional_info=input_dict["additional_info"],
        )

        assert len(new_res_info.sigmas) == new_res_info.number_of_bins, (
            new_res_info.number_of_bins,
            len(new_res_info.sigmas),
            new_res_info.sigmas,
        )
        assert all(sig >= 0.0 for sig in new_res_info.sigmas), new_res_info.sigmas
        assert isinstance(new_res_info.bin_edges, tuple), new_res_info.bin_edges
        assert len(new_res_info.bin_edges) == new_res_info.number_of_bins, (
            new_res_info.number_of_bins,
            len(new_res_info.bin_edges),
            new_res_info.bin_edges,
        )
        assert all(isinstance(edges, tuple) for edges in new_res_info.bin_edges), new_res_info.bin_edges
        assert all(len(edges) == 2 for edges in new_res_info.bin_edges), new_res_info.bin_edges
        assert all(edges[0] < edges[1] for edges in new_res_info.bin_edges), new_res_info.bin_edges

        return new_res_info

    @staticmethod
    def create_from_fit_result(
        resolution_correction_setup_name: str,
        secondary_column_name: str,
        number_of_secondary_bins: int,
        secondary_bin_edge_pairs: Tuple[Tuple[float, float], ...],
        mc_results: Dict[int, float],
        data_results: Dict[int, float],
    ) -> "ResolutionInfo":
        assert len(secondary_bin_edge_pairs) == number_of_secondary_bins, (
            number_of_secondary_bins,
            len(secondary_bin_edge_pairs),
            secondary_bin_edge_pairs,
        )

        sigmas = []  # type: List[float]
        additional_info_dict = {}  # type: Dict[str, Any]
        for bin_index, secondary_edges in enumerate(secondary_bin_edge_pairs):
            assert bin_index in mc_results.keys(), bin_index
            assert bin_index in data_results.keys(), bin_index

            mc = mc_results[bin_index]  # type: float
            data = data_results[bin_index]  # type: float
            plain_sigma_diff = abs(data - mc)  # type: float

            if data <= mc:
                diff = 0.0  # type: float
            else:
                diff = np.sqrt(np.abs(data ** 2 - mc ** 2))

            additional_info_dict.update(
                {
                    f"bin_{bin_index}": {
                        "index": bin_index,
                        "lower_bound": secondary_edges[0],
                        "upper_bound": secondary_edges[1],
                        "mc_sigma": mc,
                        "data_sigma": data,
                        "sigma_diff": diff,
                        "plain_sigma_diff": plain_sigma_diff,
                    }
                }
            )

            sigmas.append(diff)

            logging.info(f"Bin {bin_index} ({secondary_edges}): " f"mc={mc}, " f"data={data} " f"diff={diff}")

        assert len(additional_info_dict.keys()) == number_of_secondary_bins
        return ResolutionInfo(
            resolution_correction_setup_name=resolution_correction_setup_name,
            number_of_bins=number_of_secondary_bins,
            sigmas=tuple(sigmas),
            bin_edges=secondary_bin_edge_pairs,
            secondary_var=secondary_column_name,
            additional_info=additional_info_dict,
        )

    def write_to_file(self, file_path: PathType) -> None:
        with open(file_path, "w") as output_file:
            json.dump(self._asdict(), output_file, indent=2)


class MissingMassResolutionCorrector:
    default_correction_col_tag = "corrected_resolution"

    def __init__(
        self,
        resolution_correction_setup_name: str,
        resolution_info_path: PathType,
    ) -> None:

        self._res_corr_setup_name = resolution_correction_setup_name  # type: str
        self._resolution_info = ResolutionInfo.from_file(input_file_path=resolution_info_path)  # type: ResolutionInfo

        assert self._resolution_info.resolution_correction_setup_name == self._res_corr_setup_name, (
            self._resolution_info.resolution_correction_setup_name,
            self._res_corr_setup_name,
        )

        self._res_corr_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=self._res_corr_setup_name
        )  # type: ResolutionCorrectionSetup

        self._original_col = self._get_original_col_name()  # type: str
        self._new_col = self._get_corrected_col_name()  # type: str

    def add_required_columns(self, sample: pd.DataFrame) -> None:
        variable_definition_function = self._res_corr_setup.variable_definition  # type: Optional[Callable]
        if variable_definition_function is not None:
            variable_definition_function(df=sample, check_range=False)

    def apply_resolution_correction(
        self,
        sample: pd.DataFrame,
        random_seed: int = 13374747,
        correct_overflows: bool = True,
    ) -> None:
        assert self.original_column in sample.columns, self.original_column
        assert self.secondary_var_col in sample.columns, self.secondary_var_col

        assert self.new_column not in sample.columns, self.new_column

        np.random.RandomState(seed=random_seed)

        sample.loc[:, self.new_column] = np.nan

        _applied_conditions = []  # type: List[pd.Series]
        for bin_index, (sigma, bin_edges) in enumerate(zip(self.sigmas, self.bin_edge_pairs)):
            if correct_overflows and (bin_index == 0 or bin_index == self.number_of_bins - 1):
                if self.number_of_bins == 1:
                    condition = pd.Series([True] * len(sample.index))  # type: pd.Series
                    assert len(condition.index) == len(sample.index)
                elif bin_index == 0:
                    condition = sample[self.secondary_var_col] < bin_edges[1]
                elif bin_index == self.number_of_bins - 1:
                    condition = sample[self.secondary_var_col] >= bin_edges[0]
                else:
                    raise RuntimeError("This case should not be reached...")
            else:
                left_condition = sample[self.secondary_var_col] >= bin_edges[0]

                if bin_index == self.number_of_bins - 1:
                    # Special case for last bin
                    right_condition = sample[self.secondary_var_col] <= bin_edges[1]
                else:
                    right_condition = sample[self.secondary_var_col] < bin_edges[1]

                condition = left_condition & right_condition

            assert all(not np.any(condition & c) for c in _applied_conditions), (
                self.new_column,
                self._res_corr_setup_name,
                bin_index,
                bin_edges,
            )
            _applied_conditions.append(condition)

            assert pd.isnull(sample.loc[condition, self.new_column]).all(), (self.new_column, self._res_corr_setup_name)

            if sigma == 0.0:
                sample.loc[condition, self.new_column] = sample.loc[condition, self.original_column].values
            else:
                sample.loc[condition, self.new_column] = np.random.normal(
                    loc=sample.loc[condition, self.original_column].values,
                    scale=sigma,
                )

        if not correct_overflows:
            left_condition = sample[self.secondary_var_col] < self.bin_edge_pairs[0][0]
            right_condition = sample[self.secondary_var_col] > self.bin_edge_pairs[-1][1]
            condition = left_condition & right_condition
            assert pd.isnull(sample.loc[condition, self.new_column]).all()
            sample.loc[condition, self.new_column] = sample.loc[condition, self.original_column]

        assert not pd.isnull(sample[self.new_column]).any()

    @staticmethod
    def add_corrected_column_for_data(
        sample: pd.DataFrame,
        correction_setup_name: str,
    ) -> None:
        orig_column = MissingMassResolutionCorrector.get_original_col_name(correction_setup_name=correction_setup_name)
        new_column = MissingMassResolutionCorrector.get_corrected_col_name(correction_setup_name=correction_setup_name)

        assert orig_column in sample.columns, orig_column
        assert new_column not in sample.columns, new_column

        sample.loc[:, new_column] = sample[orig_column].values
        assert new_column in sample.columns, new_column

    @staticmethod
    def add_corrected_column_to_empty_data_frame(
        sample: pd.DataFrame,
        correction_setup_name: str,
    ) -> None:
        orig_column = MissingMassResolutionCorrector.get_original_col_name(correction_setup_name=correction_setup_name)
        new_column = MissingMassResolutionCorrector.get_corrected_col_name(correction_setup_name=correction_setup_name)
        assert orig_column in sample.columns, orig_column
        assert new_column not in sample.columns, new_column
        sample = sample.assign(**{new_column: sample[orig_column]})
        assert new_column in sample.columns, new_column

    @property
    def original_column(self) -> str:
        return self._original_col

    @property
    def new_column(self) -> str:
        return self._new_col

    @property
    def secondary_var_col(self) -> Optional[str]:
        return self._resolution_info.secondary_var

    @property
    def bin_edge_pairs(self) -> Tuple[Tuple[float, float], ...]:
        assert isinstance(self._resolution_info.bin_edges, tuple), self._resolution_info.bin_edges
        return self._resolution_info.bin_edges

    @property
    def sigmas(self) -> Tuple[float, ...]:
        return self._resolution_info.sigmas

    @property
    def number_of_bins(self) -> int:
        return self._resolution_info.number_of_bins

    @staticmethod
    def get_corrected_col_name(correction_setup_name: str) -> str:
        cor_setup = ResolutionCorrectionSetupManager.get_setup_definition(name=correction_setup_name)
        cor_appendix = cor_setup.corrected_variable_appendix  # type: str
        original_column_name = cor_setup.primary_hist_var.df_label  # type: str
        return f"{original_column_name}_{cor_appendix}_{MissingMassResolutionCorrector.default_correction_col_tag}"

    def _get_corrected_col_name(self) -> str:
        return MissingMassResolutionCorrector.get_corrected_col_name(
            correction_setup_name=self._resolution_info.resolution_correction_setup_name
        )

    @staticmethod
    def get_original_col_name(correction_setup_name: str) -> str:
        cor_setup = ResolutionCorrectionSetupManager.get_setup_definition(name=correction_setup_name)
        return cor_setup.primary_hist_var.df_label

    def _get_original_col_name(self) -> str:
        return MissingMassResolutionCorrector.get_original_col_name(
            correction_setup_name=self._resolution_info.resolution_correction_setup_name
        )

    @staticmethod
    def get_main_correction_column_name() -> str:
        main_res_corr_setup_name = ResolutionCorrectionSetupManager.setup_name_to_use  # type: str
        return MissingMassResolutionCorrector.get_corrected_col_name(correction_setup_name=main_res_corr_setup_name)
