"""
Definition of Missing Mass Squared resolution correction setups

Felix Metzner 2021
"""

import logging

from typing import Optional, Tuple, NamedTuple, Dict, Callable

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.observables.e_miss_minus_p_miss import EMissMinusPMissVariable
from rdstar.offline_analysis.observables.reco_modes_observables import (
    RecoIDObservables,
    SimpleDMesonRecoIDDefinition,
    DAndDsMesonRecoIDDefinition,
    BSigMesonRecoIDDefinition,
    BTagMesonRecoIDDefinition,
    U4SRecoIDDefinition,
)
from rdstar.offline_analysis.observables.decay_process_information import (
    DecayProcessVariables,
    BSigTrackCountVariable,
    BSigNeutralCountVariable,
    BTagTrackCountVariable,
    BTagNeutralCountVariable,
)
from rdstar.offline_analysis.observables.kinematic_observables import LeptonPhaseSpaceBinID, KinematicObservables


logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "missing_mass_hist_var",
    "ResolutionCorrectionSetup",
    "ResolutionCorrectionSetupManager",
]

missing_mass_hist_var = HistVariable(
    df_label="m2RecoilSignalSide",
    n_bins=250,
    scope=(-0.5, 2.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
    use_log_scale=False,
)

default_mbc_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=100,  # Only for visualization, as associated binning will be provided by the dedicated Binning instance.
    scope=(5.270, 5.290),
    var_name=LatexObsNames.b_tag_mbc,
    unit=LatexUnits.gev,
    use_log_scale=False,
)

default_mbc_binning = Binning(
    bins=(5.2700, 5.2750, 5.2780, 5.2805, 5.2835, 5.290),  # Binning by eye.
    dimensions=1,
    scope=default_mbc_hist_var.scope,
    log_scale=False,
)

w_reco_hist_var = HistVariable(
    df_label="wReco",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(1.0, 2.0),
    var_name=LatexObsNames.ff_w,
    unit=LatexUnits.gev,
    use_log_scale=False,
)

w_reco_binning = Binning(
    bins=(1.0, 1.25204834, 1.33214282, 1.39304942, 1.42476468, 1.46135439, 1.50742468, 2.0),  # Eq. fr. binning from mc.
    dimensions=1,
    scope=w_reco_hist_var.scope,
    log_scale=False,
)

q2_reco_hist_var = HistVariable(
    df_label="qSquaredReco",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(-20.0, 4.0),
    var_name=LatexObsNames.fitting_q2,
    unit=LatexUnits.gev_sq,
    use_log_scale=False,
)

q2_reco_binning = Binning(
    bins=(-20.0, 0.67099225, 1.4066534, 2.01698814, 2.56469315, 3.0785986, 3.57038442, 4.0),  # Eq. fr. binning from mc.
    dimensions=1,
    scope=q2_reco_hist_var.scope,
    log_scale=False,
)

lepton_p_hist_var = HistVariable(
    df_label="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 2.0),
    var_name=LatexObsNames.lepton_p,
    unit=LatexUnits.gev,
    use_log_scale=False,
)

lepton_p_binning = Binning(
    bins=LeptonPhaseSpaceBinID.lepton_p_5_bins_equal_frequency_binning,  # Equal frequency binning from mc.
    dimensions=1,
    scope=lepton_p_hist_var.scope,
    log_scale=False,
)

lepton_theta_hist_var = HistVariable(
    df_label="daughter__bo1__cm__spdaughter__bo1__cm__sptheta__bc__bc",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 3.0),
    var_name=LatexObsNames.lepton_theta,
    unit=LatexUnits.rad,
    use_log_scale=False,
)

lepton_theta_binning = Binning(
    bins=LeptonPhaseSpaceBinID.lepton_theta_5_bins_equal_frequency_binning,  # Equal frequency binning from mc.
    dimensions=1,
    scope=lepton_theta_hist_var.scope,
    log_scale=False,
)

event_track_count_hist_var = HistVariable(
    df_label="nCleanedTracks__bodr__sp__st__sp2__spand__spabs__bodz__bc__sp__st__sp4__bc",
    n_bins=21,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.5, 20.5),
    var_name=LatexObsNames.decay_info_number_of_tracks_in_event,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

event_track_count_binning = Binning(
    bins=(0.5, 7.5, 8.5, 9.5, 10.5, 20.5),  # Binning by eye.
    dimensions=1,
    scope=event_track_count_hist_var.scope,
    log_scale=False,
)

event_neutral_count_hist_var = HistVariable(
    df_label="nCleanedECLClusters__bogoodBelleGamma__sp__eq__eq__sp1__spand__spclusterTrackMatch__sp__eq__eq__sp0__bc",
    n_bins=27,  # Only for visualization; is overwritten by binning instance below.
    scope=(-0.5, 26.5),
    var_name=LatexObsNames.decay_info_number_of_neutrals_in_event,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

event_neutral_count_binning = Binning(
    bins=(-0.5, 2.5, 4.5, 6.5, 8.5, 26.5),  # Binning by eye.
    dimensions=1,
    scope=event_neutral_count_hist_var.scope,
    log_scale=False,
)

fei_probability_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 1.0),
    var_name=LatexObsNames.fei_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

fei_probability_binning = Binning(
    bins=(0.0, 0.00214756, 0.00561211, 0.019896, 0.1106317, 1.0),  # Equal frequency binning determined on mc sample.
    dimensions=1,
    scope=fei_probability_hist_var.scope,
    log_scale=False,
)

b_sig_vertex_fit_hist_var = HistVariable(
    df_label="daughter__bo1__cm__spchiProb__bc",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 1.0),
    var_name=LatexObsNames.b_sig_vertex_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

b_sig_vertex_fit_binning = Binning(
    bins=(0.0, 4.29092033e-04, 6.24699183e-02, 2.86736459e-01, 6.16876781e-01, 1.0),  # Equal frequency binning from mc.
    dimensions=1,
    scope=b_sig_vertex_fit_hist_var.scope,
    log_scale=False,
)

b_tag_daughter_vertex_fit_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spdaughter__bo0__cm__spchiProb__bc__bc",
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.b_tag_daughter_vertex_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

b_tag_daughter_vertex_fit_binning = Binning(
    bins=(-1.0, 0.0, 0.4258008, 1.0),  # Equal frequency binning (as good as possible) determined on mc sample.
    dimensions=1,
    scope=b_tag_daughter_vertex_fit_hist_var.scope,
    log_scale=False,
)


class ResolutionCorrectionSetup(NamedTuple):
    name: str
    latex_title_appendix: str
    corrected_variable_appendix: str
    primary_hist_var: HistVariable
    secondary_hist_var: HistVariable
    secondary_var_binning: Optional[Binning]
    variable_definition: Optional[Callable] = None
    variable_definition_class: Optional[object] = None
    tick_label_mapping: Optional[Dict[int, str]] = None


default_res_cor_setup = ResolutionCorrectionSetup(
    name="mbc_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.b_tag_mbc,
    corrected_variable_appendix="mbc_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=default_mbc_hist_var,
    secondary_var_binning=default_mbc_binning,
)

w_reco_res_cor_setup = ResolutionCorrectionSetup(
    name="w_reco_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.ff_w,
    corrected_variable_appendix="w_reco_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=w_reco_hist_var,
    secondary_var_binning=w_reco_binning,
)

q2_reco_res_cor_setup = ResolutionCorrectionSetup(
    name="q2_reco_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.fitting_q2,
    corrected_variable_appendix="q2_reco_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=q2_reco_hist_var,
    secondary_var_binning=q2_reco_binning,
)

lepton_p_res_cor_setup = ResolutionCorrectionSetup(
    name="lepton_p_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.lepton_p,
    corrected_variable_appendix="lepton_p_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=lepton_p_hist_var,
    secondary_var_binning=lepton_p_binning,
)

lepton_theta_res_cor_setup = ResolutionCorrectionSetup(
    name="lepton_theta_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.lepton_theta,
    corrected_variable_appendix="lepton_theta_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=lepton_theta_hist_var,
    secondary_var_binning=lepton_theta_binning,
)

lepton_phase_space_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{LeptonPhaseSpaceBinID.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.lepton_p_theta_bin,
    corrected_variable_appendix="lepton_phase_space_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=LeptonPhaseSpaceBinID.hist_var,
    secondary_var_binning=LeptonPhaseSpaceBinID.binning,
    variable_definition=KinematicObservables.get_definition_function(name=LeptonPhaseSpaceBinID.new_column_name),
    variable_definition_class=KinematicObservables.get_kinematic_definition(name=LeptonPhaseSpaceBinID.new_column_name),
    tick_label_mapping=LeptonPhaseSpaceBinID.get_tick_label_mapping(),
)

d_meson_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{SimpleDMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.reco_d_decay_mode_label,
    corrected_variable_appendix="simple_d_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=SimpleDMesonRecoIDDefinition.hist_var,
    secondary_var_binning=SimpleDMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=SimpleDMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=SimpleDMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=SimpleDMesonRecoIDDefinition.get_tick_label_mapping(),
)

d_and_dst_meson_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{DAndDsMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.reco_d_or_dst_decay_mode_label,
    corrected_variable_appendix="d_and_dst_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=DAndDsMesonRecoIDDefinition.hist_var,
    secondary_var_binning=DAndDsMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=DAndDsMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=DAndDsMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=DAndDsMesonRecoIDDefinition.get_tick_label_mapping(),
)

b_sig_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BSigMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.reco_b_sig_decay_mode_label,
    corrected_variable_appendix="b_sig_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BSigMesonRecoIDDefinition.hist_var,
    secondary_var_binning=BSigMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=BSigMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=BSigMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=BSigMesonRecoIDDefinition.get_tick_label_mapping(),
)

signal_neutrals_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BSigNeutralCountVariable.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.decay_info_number_of_neutrals_in_b_sig_short,
    corrected_variable_appendix="sig_neutrals_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BSigNeutralCountVariable.hist_var,
    secondary_var_binning=BSigNeutralCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BSigNeutralCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BSigNeutralCountVariable.new_column_name
    ),
)

signal_track_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BSigTrackCountVariable.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.decay_info_number_of_tracks_in_b_sig_short,
    corrected_variable_appendix="sig_track_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BSigTrackCountVariable.hist_var,
    secondary_var_binning=BSigTrackCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BSigTrackCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BSigTrackCountVariable.new_column_name
    ),
)

b_tag_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BTagMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.reco_b_tag_decay_mode_label,
    corrected_variable_appendix="b_tag_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagMesonRecoIDDefinition.hist_var,
    secondary_var_binning=BTagMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=BTagMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=BTagMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=BTagMesonRecoIDDefinition.get_tick_label_mapping(),
)

tag_neutrals_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BTagNeutralCountVariable.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.decay_info_number_of_neutrals_in_b_tag_short,
    corrected_variable_appendix="tag_neutrals_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagNeutralCountVariable.hist_var,
    secondary_var_binning=BTagNeutralCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BTagNeutralCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BTagNeutralCountVariable.new_column_name
    ),
)

tag_track_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BTagTrackCountVariable.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.decay_info_number_of_tracks_in_b_tag_short,
    corrected_variable_appendix="tag_track_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagTrackCountVariable.hist_var,
    secondary_var_binning=BTagTrackCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BTagTrackCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BTagTrackCountVariable.new_column_name
    ),
)

u4s_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{U4SRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.reco_y4s_decay_mode_label,
    corrected_variable_appendix="u4s_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=U4SRecoIDDefinition.hist_var,
    secondary_var_binning=U4SRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=U4SRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=U4SRecoIDDefinition.new_column_name),
    tick_label_mapping=U4SRecoIDDefinition.get_tick_label_mapping(),
)

event_track_count_res_cor_setup = ResolutionCorrectionSetup(
    name="event_track_count_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.decay_info_number_of_tracks_in_event_short,
    corrected_variable_appendix="event_track_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=event_track_count_hist_var,
    secondary_var_binning=event_track_count_binning,
)

event_neutral_count_res_cor_setup = ResolutionCorrectionSetup(
    name="event_neutral_count_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.decay_info_number_of_neutrals_in_event_short,
    corrected_variable_appendix="event_neutral_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=event_neutral_count_hist_var,
    secondary_var_binning=event_neutral_count_binning,
)

fei_probability_res_cor_setup = ResolutionCorrectionSetup(
    name="fei_probability_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.fei_prob,
    corrected_variable_appendix="fei_probability_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=fei_probability_hist_var,
    secondary_var_binning=fei_probability_binning,
)

b_sig_vertex_fit_res_cor_setup = ResolutionCorrectionSetup(
    name="b_sig_vertex_fit_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.b_sig_vertex_prob,
    corrected_variable_appendix="b_sig_vertex_fit_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=b_sig_vertex_fit_hist_var,
    secondary_var_binning=b_sig_vertex_fit_binning,
)

b_tag_daughter_vertex_fit_res_cor_setup = ResolutionCorrectionSetup(
    name="b_tag_daughter_vertex_fit_based_mms_resolution_correction",
    latex_title_appendix=LatexObsNames.b_tag_daughter_vertex_prob,
    corrected_variable_appendix="b_tag_daughter_vertex_fit_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=b_tag_daughter_vertex_fit_hist_var,
    secondary_var_binning=b_tag_daughter_vertex_fit_binning,
)


# E_Miss - P_Miss resolution correction setups
e_miss_minus_p_miss_cms = HistVariable(
    df_label=EMissMinusPMissVariable.e_miss_minus_p_miss_cms_col_name,
    n_bins=60,
    scope=(-1.0, 2.0),
    var_name=LatexObsNames.fitting_e_miss_minus_p_miss_cms,
    unit=LatexUnits.gev,
)  # type: HistVariable

e_miss_minus_p_miss_res_corr_setup = ResolutionCorrectionSetup(
    name="mbc_based_e_miss_minus_p_miss_resolution_correction",
    latex_title_appendix=LatexObsNames.b_tag_mbc,
    corrected_variable_appendix="mbc_based",
    primary_hist_var=e_miss_minus_p_miss_cms,
    secondary_hist_var=default_mbc_hist_var,
    secondary_var_binning=default_mbc_binning,
)


class ResolutionCorrectionSetupManager:
    setup_name_to_use = default_res_cor_setup.name  # type: str

    # Further setups to be considered:
    #  - D Meson channel with ids boiled down to similar groups of decay processes
    #  - D and Dst Meson channels with ids boiled down to similar groups of decay processes
    #  - B Sig channel with ids boiled down to similar groups of decay processes
    #  - B Tag channel with ids boiled down to similar groups of decay processes
    #  - B Sig + Y(4S) (using the latter to differentiate between normal and wrong combinations) channels ?
    #       -> Only for boiled down B_sig modes, because otherwise we get to many bins!
    #  - Multiplicity variable, which combines tracks and clusters...

    setup_definitions = {
        # Default Setup: Mbc
        default_res_cor_setup.name: default_res_cor_setup,
        #
        # Event kinematic based setups:
        w_reco_res_cor_setup.name: w_reco_res_cor_setup,
        q2_reco_res_cor_setup.name: q2_reco_res_cor_setup,
        #
        # Lepton based setups:
        lepton_p_res_cor_setup.name: lepton_p_res_cor_setup,
        lepton_theta_res_cor_setup.name: lepton_theta_res_cor_setup,
        lepton_phase_space_based_res_cor_setup.name: lepton_phase_space_based_res_cor_setup,
        #
        # D Meson based setups:
        d_meson_decay_id_based_res_cor_setup.name: d_meson_decay_id_based_res_cor_setup,
        d_and_dst_meson_decay_id_based_res_cor_setup.name: d_and_dst_meson_decay_id_based_res_cor_setup,
        #
        # B Sig Meson based setups:
        b_sig_decay_id_based_res_cor_setup.name: b_sig_decay_id_based_res_cor_setup,
        signal_neutrals_count_based_res_cor_setup.name: signal_neutrals_count_based_res_cor_setup,
        signal_track_count_based_res_cor_setup.name: signal_track_count_based_res_cor_setup,
        #
        # B Sig Meson based setups:
        # b_tag_decay_id_based_res_cor_setup.name: b_tag_decay_id_based_res_cor_setup,  # Rejected, because too many channels -> too small statistics!  # noqa: E501, E950
        tag_neutrals_count_based_res_cor_setup.name: tag_neutrals_count_based_res_cor_setup,
        tag_track_count_based_res_cor_setup.name: tag_track_count_based_res_cor_setup,
        #
        # Y(4S) Meson based setups:
        u4s_decay_id_based_res_cor_setup.name: u4s_decay_id_based_res_cor_setup,
        #
        # Event based setups:
        event_track_count_res_cor_setup.name: event_track_count_res_cor_setup,
        event_neutral_count_res_cor_setup.name: event_neutral_count_res_cor_setup,
        #
        # Other setups:
        fei_probability_res_cor_setup.name: fei_probability_res_cor_setup,
        b_sig_vertex_fit_res_cor_setup.name: b_sig_vertex_fit_res_cor_setup,
        b_tag_daughter_vertex_fit_res_cor_setup.name: b_tag_daughter_vertex_fit_res_cor_setup,
        #
        #
        # E_Miss - P_Miss resolution correction setups
        e_miss_minus_p_miss_res_corr_setup.name: e_miss_minus_p_miss_res_corr_setup,
    }  # type: Dict[str, ResolutionCorrectionSetup]

    @staticmethod
    def get_setup_definition_names() -> Tuple[str, ...]:
        return tuple(ResolutionCorrectionSetupManager.setup_definitions.keys())

    @staticmethod
    def get_setup_definition(name: str) -> ResolutionCorrectionSetup:
        return ResolutionCorrectionSetupManager.setup_definitions[name]

    @staticmethod
    def get_n_bin_options_for_setup(setup_name: str) -> Tuple[int, ...]:
        n_bins_options = (
            20,
            # 25,
            50,
            # 75,
            # 125,
            250,
        )  # type: Tuple[int, ...]
        return n_bins_options

    @staticmethod
    def get_scope_options_for_setup(setup_name: str) -> Dict[str, Tuple[float, float]]:
        corr_setup = ResolutionCorrectionSetupManager.setup_definitions[setup_name]  # type: ResolutionCorrectionSetup
        v_label = corr_setup.primary_hist_var.df_label  # type: str

        v1 = missing_mass_hist_var.df_label  # type: str
        v2 = e_miss_minus_p_miss_cms.df_label  # type: str

        valid_vs = (v1, v2)  # type: Tuple[str, ...]

        if v_label not in valid_vs:
            raise KeyError(
                f"No scope settings defined for setup '{setup_name}' with primary variable column name '{v_label}'!\n"
                f"Valid variable column names are {valid_vs}."
            )

        hist_var_scope_settings = {
            "wide1": {v1: (-0.5, 2.0), v2: (-1.0, 1.0)},
            # "wide2": {v1: (-0.5, 1.0), v2: (-1.0, 0.8)},
            # "wide3": {v1: (0.0, 2.0), v2: (-1.2, 1.2)},
            # "wide4": {v1: (-0.1, 2.0), v2: (-1.5, 1.5)},
            # "wide5": {v1: (0.0, 1.0), v2: (-2.0, 2.0)},
            "wide6": {v1: (-0.1, 1.0), v2: (-0.5, 1.0)},
            # "tight1": {v1: (-0.5, 0.5), v2: (-0.5, 0.5)},
            "tight2": {v1: (-0.3, 0.3), v2: (-0.3, 0.3)},
            # "tight3": {v1: (-0.3, 0.5), v2: (-0.3, 0.5)},
            # "tight4": {v1: (-0.1, 0.5), v2: (-0.1, 0.5)},
        }  # type: Dict[str, Dict[str, Tuple[float, float]]]

        return {k: v[v_label] for k, v in hist_var_scope_settings.items()}

    @staticmethod
    def get_special_options_for_setup(setup_name: str) -> Dict[Tuple[str, int], Tuple[float, float]]:
        corr_setup = ResolutionCorrectionSetupManager.setup_definitions[setup_name]  # type: ResolutionCorrectionSetup
        v_label = corr_setup.primary_hist_var.df_label  # type: str

        v1 = missing_mass_hist_var.df_label  # type: str
        v2 = e_miss_minus_p_miss_cms.df_label  # type: str

        valid_vs = (v1, v2)  # type: Tuple[str, ...]

        if v_label not in valid_vs:
            return {}

        special_option_cases = {
            ("wide1", 20): {v2: (-0.5, 1.0)},
            ("wide6", 50): {v1: (-0.2, 1.8), v2: (-0.2, 1.8)},
        }  # type: Dict[Tuple[str, int], Dict[str, Tuple[float, float]]]

        return {k: v[v_label] for k, v in special_option_cases.items() if v_label in v.keys()}
