"""
Missing Mass Squared Resolution Correction

Felix Metzner 2020
"""

import os
import logging
import probfit
import iminuit
import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from matplotlib import ticker as mpl_ticker
from matplotlib import colors as mpl_colors
from matplotlib.offsetbox import AnchoredText
from matplotlib import font_manager as mpl_font_mgr

from recordclass import RecordClass
from typing import Union, Optional, Tuple, List, Dict, NamedTuple

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export
from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot

from rdstar.utilities import PathType
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import ResolutionInfo
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetup,
    ResolutionCorrectionSetupManager,
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "MissingMassResolutionEvaluator",
]


class FitInputData(NamedTuple):
    mc_miss_mass: np.ndarray
    mc_weights: np.ndarray
    mc_sig_miss_mass: np.ndarray
    mc_sig_weights: np.ndarray
    data_miss_mass: np.ndarray
    data_weights: Optional[np.ndarray] = None


class FitParameters(RecordClass):
    gauss_n_guess: float
    gauss_mean_guess: float
    gauss_sigma_guess: float
    cruijff_n_guess: float
    cruijff_mean_guess: float
    cruijff_sigma_l_guess: float
    cruijff_sigma_r_guess: float
    cruijff_alpha_l_guess: float
    cruijff_alpha_r_guess: float
    sig_cruijff_n_guess: float
    sig_cruijff_mean_guess: float
    sig_cruijff_sigma_l_guess: float
    sig_cruijff_sigma_r_guess: float
    sig_cruijff_alpha_l_guess: float
    sig_cruijff_alpha_r_guess: float

    @property
    def parameter_names(self) -> Tuple[str, ...]:
        return tuple(self.__annotations__.keys())

    @property
    def as_dict(self) -> Dict[str, float]:
        return {name: getattr(self, name) for name in self.parameter_names}


class CruijffFitResult(NamedTuple):
    n: float
    mean: float
    sigma_l: float
    sigma_r: float
    alpha_l: float
    alpha_r: float


class GaussFitResult(NamedTuple):
    n: float
    mean: float
    sigma: float


class FitOutput(NamedTuple):
    fit_model: probfit.BinnedLH
    minuit_instance_before_fit: iminuit.Minuit
    minuit_instance_after_fit: iminuit.Minuit
    gauss: Optional[GaussFitResult] = None
    cruijff: Optional[CruijffFitResult] = None


class MissingMassResolutionEvaluator:
    default_is_signal_col = "isSignalAcceptMissingNeutrino"
    default_result_file = "missing_mass_squared_resolution_eval_result"  # type: str

    fit_plot_base_name = "missing_mass_resolution_fit_plot"
    default_fig_size = (6, 4)  # type: Tuple[float, float]

    parameter_name_dict = {
        "N_g": r"$N_\mathrm{Gauss}$",
        "m_g": r"$m_\mathrm{Gauss}$",
        "sigma_g": r"$\sigma_\mathrm{Gauss}$",
        "N_c": r"$N_\mathrm{Cruijff}$",
        "m_c": r"$m_\mathrm{Cruijff}$",
        "sigma_c_L": r"$\sigma^\mathrm{L}_\mathrm{Cruijff}$",
        "sigma_c_R": r"$\sigma^\mathrm{R}_\mathrm{Cruijff}$",
        "alpha_c_L": r"$\alpha^\mathrm{L}_\mathrm{Cruijff}$",
        "alpha_c_R": r"$\alpha^\mathrm{R}_\mathrm{Cruijff}$",
    }  # type: Dict[str, str]

    type_string_mapping = {
        "mc": "on MC",
        "data": "on Data",
        "signal": "on Signal MC",
    }  # type: Dict[str, str]

    def __init__(
        self,
        resolution_correction_setup_name: str,
        weight_column: str,
        results_path: PathType,
        is_signal_column: Optional[str] = None,
        number_of_bins: int = 50,
        normalize: bool = False,
        produce_plots: bool = True,
        fig_size: Optional[Tuple[Union[int, float], Union[int, float]]] = None,
    ) -> None:
        if normalize:
            raise RuntimeError("You did set the option normalize to True, but this does not work properly atm...")
        self._res_corr_setup_name = resolution_correction_setup_name  # type: str
        self._res_corr_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=self._res_corr_setup_name
        )  # type: ResolutionCorrectionSetup
        self._secondary_variable, self._secondary_binning = self._init_secondary_var()  # type: HistVariable, Binning

        self._weight_column = weight_column  # type: str
        self._missing_mass = self._res_corr_setup.primary_hist_var  # type: HistVariable
        self._number_of_bins = number_of_bins  # type: int

        self._is_signal_column = self.default_is_signal_col if is_signal_column is None else is_signal_column  # type: str

        self._normalize = normalize  # type: bool

        self._results_path = results_path  # type: PathType

        self._produce_plots = produce_plots  # type: bool
        self._fig_size = self.default_fig_size if fig_size is None else fig_size  # type: Tuple[float, float]

    def evaluate_resolution(self, mc_df: pd.DataFrame, data_df: pd.DataFrame) -> Dict[str, PathType]:
        self._check_columns(mc_cols=list(mc_df.columns), data_cols=list(data_df.columns))
        os.makedirs(self._results_path, exist_ok=True)

        mc_results = {}  # type: Dict[int, float]
        data_results = {}  # type: Dict[int, float]
        output_path_dict = {}  # type: Dict[str, PathType]

        new_out = self.produce_secondary_variable_plot(df=mc_df, dry_run=False)
        assert all(new_k not in output_path_dict for new_k in new_out.keys())
        output_path_dict.update(new_out)

        for bin_index, secondary_edges in enumerate(self.secondary_bin_edge_pairs):
            assert bin_index not in mc_results
            assert bin_index not in data_results

            miss_mass_mc_cond = self.get_missing_mass_condition(df=mc_df)
            miss_mass_data_cond = self.get_missing_mass_condition(df=data_df)
            is_sig_mc_cond = self.get_signal_condition(df=mc_df)
            secondary_mc_cond = self.get_secondary_condition(df=mc_df, secondary_edges=secondary_edges)
            secondary_data_cond = self.get_secondary_condition(df=data_df, secondary_edges=secondary_edges)

            mc_cond = miss_mass_mc_cond & secondary_mc_cond
            data_cond = miss_mass_data_cond & secondary_data_cond

            mc_miss_mass = mc_df.loc[mc_cond, self.missing_mass_col].values  # type: np.ndarray
            mc_weights = mc_df.loc[mc_cond, self.weight_col].values  # type: np.ndarray

            mc_sig_miss_mass = mc_df.loc[mc_cond & is_sig_mc_cond, self.missing_mass_col].values  # type: np.ndarray
            mc_sig_weights = mc_df.loc[mc_cond & is_sig_mc_cond, self.weight_col].values  # type: np.ndarray

            data_miss_mass = data_df.loc[data_cond, self.missing_mass_col].values  # type: np.ndarray
            data_weights = None  # type: Optional[np.ndarray]

            if self.normalize:
                mc_weights = mc_weights / np.sum(mc_weights)
                mc_sig_weights = mc_sig_weights / np.sum(mc_sig_weights)
                data_weights = np.ones_like(data_miss_mass) * 1.0 / data_miss_mass.size

            fit_data = FitInputData(
                mc_miss_mass=mc_miss_mass,
                mc_weights=mc_weights,
                mc_sig_miss_mass=mc_sig_miss_mass,
                mc_sig_weights=mc_sig_weights,
                data_miss_mass=data_miss_mass,
                data_weights=data_weights,
            )

            mc_and_data_sigma, new_out = self.do_fit_routine(
                fit_data=fit_data,
                bin_index=bin_index if self.multiple_secondary_bins else None,
                dry_run=False,
            )

            mc_results.update({bin_index: mc_and_data_sigma[0]})
            data_results.update({bin_index: mc_and_data_sigma[1]})
            assert all(new_k not in output_path_dict for new_k in new_out.keys())
            output_path_dict.update(new_out)

        new_out = self.write_result(mc_results=mc_results, data_results=data_results)
        assert all(new_k not in output_path_dict for new_k in new_out.keys())
        output_path_dict.update(new_out)
        return output_path_dict

    def get_output_dict(self) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]
        for bin_index, _ in enumerate(self.secondary_bin_edge_pairs):
            _, new_out = self.do_fit_routine(bin_index=bin_index, dry_run=True)
            assert all(new_k not in output_dict for new_k in new_out.keys())
            output_dict.update(new_out)

        new_out = self.produce_secondary_variable_plot(df=pd.DataFrame(), dry_run=True)
        assert all(new_k not in output_dict for new_k in new_out.keys())
        output_dict.update(new_out)

        res_key, res_path = self.result_file_key_path_pair
        output_dict.update({res_key: res_path})

        return output_dict

    def write_result(
        self,
        mc_results: Dict[int, float],
        data_results: Dict[int, float],
    ) -> Dict[str, PathType]:
        res_corr_info = ResolutionInfo.create_from_fit_result(
            resolution_correction_setup_name=self._res_corr_setup.name,
            secondary_column_name=self.secondary_col,
            number_of_secondary_bins=self.number_of_secondary_bins,
            secondary_bin_edge_pairs=tuple(self.secondary_bin_edge_pairs),
            mc_results=mc_results,
            data_results=data_results,
        )

        res_key, res_path = self.result_file_key_path_pair
        res_corr_info.write_to_file(file_path=res_path)
        return {res_key: res_path}

    def do_fit_routine(
        self,
        fit_data: Optional[FitInputData] = None,
        bin_index: Optional[int] = None,
        dry_run: bool = False,
    ) -> Tuple[Tuple[float, float], Dict[str, PathType]]:
        output_dict = {}  # type: Dict[str, PathType]
        if dry_run:
            sig_out = self.produce_before_after_plots(fit_type="signal", bin_index=bin_index, dry_run=dry_run)
            assert all(new_key not in output_dict.keys() for new_key in sig_out)
            output_dict.update(sig_out)
            mc_out = self.produce_before_after_plots(fit_type="mc", bin_index=bin_index, dry_run=dry_run)
            assert all(new_key not in output_dict.keys() for new_key in mc_out)
            output_dict.update(mc_out)
            d_out = self.produce_before_after_plots(fit_type="data", bin_index=bin_index, dry_run=dry_run)
            assert all(new_key not in output_dict.keys() for new_key in d_out)
            output_dict.update(d_out)
            return (0.0, 0.0), output_dict
        else:
            assert fit_data is not None
            assert isinstance(fit_data, FitInputData)

        fit_parameters = FitParameters(
            gauss_n_guess=float(np.sum(fit_data.mc_sig_weights)),
            gauss_mean_guess=0.0,
            gauss_sigma_guess=0.05,
            cruijff_n_guess=float(np.sum(fit_data.mc_weights) - np.sum(fit_data.mc_sig_weights)),
            cruijff_mean_guess=-0.1,
            cruijff_sigma_l_guess=0.1,
            cruijff_sigma_r_guess=0.7,
            cruijff_alpha_l_guess=0.7,
            cruijff_alpha_r_guess=0.1,
            sig_cruijff_n_guess=float(np.sum(fit_data.mc_sig_weights)),
            sig_cruijff_mean_guess=0.0,
            sig_cruijff_sigma_l_guess=0.03,
            sig_cruijff_sigma_r_guess=0.06,
            sig_cruijff_alpha_l_guess=0.15,
            sig_cruijff_alpha_r_guess=0.15,
        )  # type: FitParameters

        sig_fit_res = self.do_cruijff_fit(
            **{k.split("_", 1)[1]: v for k, v in fit_parameters.as_dict.items() if k.startswith("sig")},
            events=fit_data.mc_sig_miss_mass,
            event_weights=fit_data.mc_sig_weights,
        )  # type: FitOutput
        sig_out = self.produce_before_after_plots(fit_type="signal", fit_output=sig_fit_res, bin_index=bin_index)
        assert all(new_key not in output_dict.keys() for new_key in sig_out)
        output_dict.update(sig_out)

        assert sig_fit_res.cruijff is not None
        sig_sigma_guess = abs(sig_fit_res.cruijff.sigma_l + sig_fit_res.cruijff.sigma_r) / 2.0
        fit_parameters.gauss_sigma_guess = sig_sigma_guess

        fit_parameters.sig_cruijff_n_guess = sig_fit_res.cruijff.n
        fit_parameters.sig_cruijff_mean_guess = sig_fit_res.cruijff.mean
        fit_parameters.sig_cruijff_sigma_l_guess = sig_fit_res.cruijff.sigma_l
        fit_parameters.sig_cruijff_sigma_r_guess = sig_fit_res.cruijff.sigma_r
        fit_parameters.sig_cruijff_alpha_l_guess = sig_fit_res.cruijff.alpha_l
        fit_parameters.sig_cruijff_alpha_r_guess = sig_fit_res.cruijff.alpha_r

        mc_fit_res = self.do_gauss_cruijff_fit(
            **{k: v for k, v in fit_parameters.as_dict.items() if not k.startswith("sig")},
            events=fit_data.mc_miss_mass,
            event_weights=fit_data.mc_weights,
        )  # type: FitOutput
        mc_out = self.produce_before_after_plots(fit_type="mc", fit_output=mc_fit_res, bin_index=bin_index)
        assert all(new_key not in output_dict.keys() for new_key in mc_out)
        output_dict.update(mc_out)

        data_fit_res = self.do_gauss_cruijff_fit(
            **{k: v for k, v in fit_parameters.as_dict.items() if not k.startswith("sig")},
            events=fit_data.data_miss_mass,
            event_weights=fit_data.data_weights,
        )  # type: FitOutput
        d_out = self.produce_before_after_plots(fit_type="data", fit_output=data_fit_res, bin_index=bin_index)
        assert all(new_key not in output_dict.keys() for new_key in d_out)
        output_dict.update(d_out)

        assert mc_fit_res.gauss is not None
        assert data_fit_res.gauss is not None
        return (mc_fit_res.gauss.sigma, data_fit_res.gauss.sigma), output_dict

    def do_gauss_cruijff_fit(
        self,
        gauss_n_guess: Union[float, int],
        gauss_mean_guess: float,
        gauss_sigma_guess: float,
        cruijff_n_guess: Union[float, int],
        cruijff_mean_guess: float,
        cruijff_sigma_l_guess: float,
        cruijff_sigma_r_guess: float,
        cruijff_alpha_l_guess: float,
        cruijff_alpha_r_guess: float,
        events: np.ndarray,
        event_weights: Optional[np.ndarray] = None,
    ) -> FitOutput:
        if event_weights is not None:
            assert len(event_weights) == len(events), (len(event_weights), len(events))

        pdf1 = probfit.rename(probfit.gaussian, ("x", "m_g", "sigma_g"))
        pdf2 = probfit.rename(probfit.cruijff, ("x", "m_c", "sigma_c_L", "sigma_c_R", "alpha_c_L", "alpha_c_R"))

        ext_pdf1 = probfit.Extended(pdf1, extname="N_g")
        ext_pdf2 = probfit.Extended(pdf2, extname="N_c")

        composite_pdf = probfit.AddPdf(ext_pdf1, ext_pdf2)

        fit_model = probfit.BinnedLH(f=composite_pdf, data=events, bins=self.bins, weights=event_weights, extended=True)
        before_fit, minuit_instance = self.create_minuit_instances(
            fit_model=fit_model,
            m_g=gauss_mean_guess,
            m_c=cruijff_mean_guess,
            sigma_g=gauss_sigma_guess,
            sigma_c_L=cruijff_sigma_l_guess,
            sigma_c_R=cruijff_sigma_r_guess,
            alpha_c_L=cruijff_alpha_l_guess,
            alpha_c_R=cruijff_alpha_r_guess,
            N_g=gauss_n_guess,
            N_c=cruijff_n_guess,
        )

        minuit_instance.migrad()

        gauss_res = GaussFitResult(
            n=minuit_instance.values["N_g"],
            mean=minuit_instance.values["m_g"],
            sigma=minuit_instance.values["sigma_g"],
        )
        cruijff_res = CruijffFitResult(
            n=minuit_instance.values["N_c"],
            mean=minuit_instance.values["m_c"],
            sigma_l=minuit_instance.values["sigma_c_L"],
            sigma_r=minuit_instance.values["sigma_c_R"],
            alpha_l=minuit_instance.values["alpha_c_L"],
            alpha_r=minuit_instance.values["alpha_c_R"],
        )

        return FitOutput(
            fit_model=fit_model,
            minuit_instance_before_fit=before_fit,
            minuit_instance_after_fit=minuit_instance,
            gauss=gauss_res,
            cruijff=cruijff_res,
        )

    def do_cruijff_fit(
        self,
        cruijff_n_guess: Union[float, int],
        cruijff_mean_guess: float,
        cruijff_sigma_l_guess: float,
        cruijff_sigma_r_guess: float,
        cruijff_alpha_l_guess: float,
        cruijff_alpha_r_guess: float,
        events: np.ndarray,
        event_weights: Optional[np.ndarray] = None,
    ) -> FitOutput:
        if event_weights is not None:
            assert len(event_weights) == len(events), (len(event_weights), len(events))

        pdf = probfit.rename(probfit.cruijff, ("x", "m_c", "sigma_c_L", "sigma_c_R", "alpha_c_L", "alpha_c_R"))

        ext_pdf = probfit.Extended(pdf, extname="N_c")

        fit_model = probfit.BinnedLH(f=ext_pdf, data=events, bins=self.bins, weights=event_weights, extended=True)
        before_fit, minuit_instance = self.create_minuit_instances(
            fit_model=fit_model,
            m_c=cruijff_mean_guess,
            sigma_c_L=cruijff_sigma_l_guess,
            sigma_c_R=cruijff_sigma_r_guess,
            alpha_c_L=cruijff_alpha_l_guess,
            alpha_c_R=cruijff_alpha_r_guess,
            N_c=cruijff_n_guess,
        )

        minuit_instance.migrad()

        cruijff_res = CruijffFitResult(
            n=minuit_instance.values["N_c"],
            mean=minuit_instance.values["m_c"],
            sigma_l=minuit_instance.values["sigma_c_L"],
            sigma_r=minuit_instance.values["sigma_c_R"],
            alpha_l=minuit_instance.values["alpha_c_L"],
            alpha_r=minuit_instance.values["alpha_c_R"],
        )

        return FitOutput(
            fit_model=fit_model,
            minuit_instance_before_fit=before_fit,
            minuit_instance_after_fit=minuit_instance,
            cruijff=cruijff_res,
        )

    @staticmethod
    def create_minuit_instances(fit_model: probfit.BinnedLH, **kwargs) -> Tuple[iminuit.Minuit, iminuit.Minuit]:
        instance_1 = iminuit.Minuit(fit_model, **kwargs)
        instance_2 = iminuit.Minuit(fit_model, **kwargs)
        return instance_1, instance_2

    def produce_secondary_variable_plot(
        self,
        df: pd.DataFrame,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        plot_style.set_matplotlibrc_params()

        file_name = "missing_mass_resolution_correction_secondary_variable_binning"
        path = os.path.join(self.results_path, f"{file_name}.pdf")  # type: PathType
        output_dict = {file_name: path}
        if dry_run:
            return output_dict

        shp = SimpleHistogramPlot(variable=self._secondary_variable)
        shp.add_component(
            label=self._secondary_variable.variable_name,
            data=df,
            weights=self.weight_col,
            color=plot_style.KITColors.kit_blue,
        )

        fig = plt.figure(figsize=self.fig_size, dpi=200)
        ax = plt.gca()

        if self._res_corr_setup.tick_label_mapping is not None:

            def format_fn(tick_val: float, tick_pos: float) -> str:
                assert isinstance(self._res_corr_setup.tick_label_mapping, dict)
                if int(tick_val) in self._res_corr_setup.tick_label_mapping.keys():
                    return self._res_corr_setup.tick_label_mapping[int(tick_val)]
                else:
                    return ""

            ax.xaxis.set_major_formatter(mpl_ticker.FuncFormatter(format_fn))
            min_num_of_ticks = self.number_of_secondary_bins if self.number_of_secondary_bins < 16 else 2  # type: int
            ax.xaxis.set_major_locator(mpl_ticker.MaxNLocator(integer=True, min_n_ticks=min_num_of_ticks))

            tick_font_size = (
                "medium"
                if self.number_of_secondary_bins < 10
                else mpl_font_mgr.FontProperties(size=plt.rcParams["xtick.labelsize"]).get_size_in_points() * 0.8
            )
            plt.xticks(rotation=45, horizontalalignment="right", fontsize=tick_font_size)

        shp.plot_on(ax=ax, draw_legend=False, y_axis_scale=1.1)

        for a_bin_edge in self.secondary_bin_edges:
            ax.axvline(x=a_bin_edge, lw=2, color="black")

        plt.title(
            label=f"Binning in {self._secondary_variable.variable_name} "
            f"for {self._missing_mass.variable_name} Resolution Correction",
            loc="right",
            fontsize=mpl_font_mgr.FontProperties(size=plt.rcParams["axes.titlesize"]).get_size_in_points() * 0.9,
        )

        export(fig=fig, filename=file_name, target_dir=self.results_path, close_figure=True)

        return output_dict

    def produce_before_after_plots(
        self,
        fit_type: str,
        fit_output: Optional[FitOutput] = None,
        bin_index: Optional[int] = None,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        if not self.produce_plots:
            return {}

        assert fit_type in self.type_string_mapping.keys()
        title_info = self.map_plot_title_info(type_string=fit_type)
        bin_info = "" if bin_index is None else f"_bin{bin_index}"
        bin_title_info = ""
        if bin_index is not None:
            sec_bin_edges = self.secondary_bin_edge_pairs[bin_index]
            edges_info = f"[{sec_bin_edges[0]:.3f}, {sec_bin_edges[1]:.3f}] {self._secondary_variable.unit}"
            bin_title_info = (
                f" in Bin {bin_index + 1}/{self.number_of_secondary_bins} "
                rf" in {self._secondary_variable.variable_name} $\in$ {edges_info}"
            )
        output_dict = {}  # type: Dict[str, PathType]

        b_fit_file_name = self.get_fit_plot_file_name(fitted=False, bin_info=bin_info, type_string=fit_type)
        output_dict.update({b_fit_file_name: os.path.join(self.results_path, f"{b_fit_file_name}.pdf")})

        a_fit_file_name = self.get_fit_plot_file_name(fitted=True, bin_info=bin_info, type_string=fit_type)
        output_dict.update({a_fit_file_name: os.path.join(self.results_path, f"{b_fit_file_name}.pdf")})

        residuals_file_name = self.get_residual_plot_file_name(bin_info=bin_info, type_string=fit_type)
        output_dict.update({residuals_file_name: os.path.join(self.results_path, f"{residuals_file_name}.pdf")})

        if dry_run:
            return output_dict
        else:
            assert fit_output is not None
            assert isinstance(fit_output, FitOutput)

        self.plot_probfit_status(
            fit_model=fit_output.fit_model,
            status=fit_output.minuit_instance_before_fit,
            file_name=b_fit_file_name,
            title=f"{self._missing_mass.variable_name} Resolution Fit {title_info} Before Fit\n{bin_title_info}",
        )

        self.plot_probfit_status(
            fit_model=fit_output.fit_model,
            status=fit_output.minuit_instance_after_fit,
            file_name=a_fit_file_name,
            title=f"{self._missing_mass.variable_name} Resolution Fit {title_info} After Fit\n{bin_title_info}",
        )

        self.plot_probfit_residuals(
            fit_model=fit_output.fit_model,
            status=fit_output.minuit_instance_after_fit,
            file_name=residuals_file_name,
            title=f"{self._missing_mass.variable_name} Resolution Fit {title_info} Residuals\n{bin_title_info}",
        )

        return output_dict

    def plot_probfit_status(
        self,
        fit_model: probfit.BinnedLH,
        status: iminuit.Minuit,
        file_name: str,
        title: Optional[str] = None,
    ) -> None:
        fig = plt.figure(figsize=self.fig_size, dpi=200)
        ax = plt.gca()
        fit_model.draw(minuit=status, ax=ax, parts=True, print_par=False)
        info_text = self.get_fit_info_text(fit_model=fit_model, status=status)
        anchored_text = AnchoredText(info_text, loc="upper right", frameon=False)
        plt.gca().add_artist(anchored_text)

        ax.set_xlabel(self._missing_mass.x_label, plot_style.xlabel_pos)
        ax.set_ylabel(self.get_y_axis_label(), plot_style.ylabel_pos)

        if title is not None:
            plt.title(title)

        export(fig=fig, filename=file_name, target_dir=self.results_path, close_figure=True)

    def plot_probfit_residuals(
        self,
        fit_model: probfit.BinnedLH,
        status: iminuit.Minuit,
        file_name: str,
        title: Optional[str] = None,
    ) -> None:
        fig = plt.figure(figsize=self.fig_size, dpi=200)
        ax = plt.gca()

        # Resetting color codes used by draw_residual method implemented in probfit
        mpl_colors.colorConverter.colors["b"] = plot_style.KITColors.kit_blue
        mpl_colors.colorConverter.colors["r"] = plot_style.KITColors.kit_red

        fit_model.draw_residual(minuit=status, ax=ax, norm=True)

        ax.set_xlabel(self._missing_mass.x_label, plot_style.xlabel_pos)
        ax.set_ylabel(r"(Fit$-$Data Points) / $\sigma_\mathrm{Fit}$", plot_style.ylabel_pos)

        if title is not None:
            plt.title(title)

        export(fig=fig, filename=file_name, target_dir=self.results_path, close_figure=True)

    def get_fit_info_text(self, fit_model: probfit.BinnedLH, status: iminuit.Minuit) -> str:
        parameters = iminuit.util.describe(fit_model)
        assert all(p in self.parameter_name_dict.keys() for p in parameters), (parameters, self.parameter_name_dict)
        txt = ""
        for (param, value) in zip(parameters, status.values):
            txt += rf"{self.parameter_name_dict[param]} = {value:.4g}"
            if status.errors is not None:
                txt += rf"$\pm${status.errors[param]:.4g}"
            txt += "\n"
        return txt

    def get_y_axis_label(self) -> str:
        if self.normalize:
            return "Normalized in arb. units"
        else:
            return "Events / ({b:.2g}{v})".format(
                b=self.bin_width, v=" " + self._missing_mass.unit if self._missing_mass.unit else ""
            )

    def get_fit_plot_file_name(self, fitted: bool, bin_info: str, type_string: str) -> str:
        b_or_a = "after" if fitted else "before"
        return f"{self.fit_plot_base_name}{bin_info}_{type_string}_{b_or_a}_fit"

    def get_residual_plot_file_name(self, bin_info: str, type_string: str) -> str:
        return f"{self.fit_plot_base_name}{bin_info}_{type_string}_residuals"

    def map_plot_title_info(self, type_string: str) -> str:
        assert type_string in self.type_string_mapping.keys()
        return self.type_string_mapping[type_string]

    def get_missing_mass_condition(self, df: pd.DataFrame) -> pd.Series:
        return (df[self.missing_mass_col] >= self.missing_mass_bounds[0]) & (
            df[self.missing_mass_col] <= self.missing_mass_bounds[1]
        )

    def get_signal_condition(self, df: pd.DataFrame) -> pd.Series:
        return df[self.is_signal_col] == 1.0

    def get_secondary_condition(self, df: pd.DataFrame, secondary_edges: Tuple[float, float]) -> pd.Series:
        assert isinstance(secondary_edges, tuple), type(secondary_edges)
        assert len(secondary_edges) == 2, (len(secondary_edges), secondary_edges)
        return (df[self.secondary_col] >= secondary_edges[0]) & (df[self.secondary_col] <= secondary_edges[1])

    @property
    def missing_mass_bounds(self) -> Tuple[float, float]:
        return self._missing_mass.scope

    @property
    def secondary_bin_edge_pairs(self) -> List[Tuple[float, float]]:
        bin_edges = self._secondary_binning.bin_edges[0]
        return [(left, right) for left, right in zip(bin_edges[:-1], bin_edges[1:])]

    @property
    def secondary_bin_edges(self) -> List[float]:
        return list(self._secondary_binning.bin_edges[0])

    @property
    def missing_mass_col(self) -> str:
        return self._missing_mass.df_label

    @property
    def secondary_col(self) -> str:
        return self._secondary_variable.df_label

    @property
    def weight_col(self) -> str:
        return self._weight_column

    @property
    def is_signal_col(self) -> str:
        return self._is_signal_column

    @property
    def bins(self) -> int:
        return self._number_of_bins

    @property
    def bin_width(self) -> float:
        window_width = self.missing_mass_bounds[1] - self.missing_mass_bounds[0]
        return window_width / self.bins

    @property
    def number_of_secondary_bins(self) -> int:
        return self._secondary_binning.num_bins[0]

    @property
    def multiple_secondary_bins(self) -> bool:
        return self.number_of_secondary_bins > 1

    @property
    def normalize(self) -> bool:
        return self._normalize

    @property
    def results_path(self) -> PathType:
        return self._results_path

    @property
    def result_file_key_path_pair(self) -> Tuple[str, PathType]:
        key = self.default_result_file
        path = os.path.join(self.results_path, f"{self.default_result_file}.json")
        return key, path

    @property
    def fig_size(self) -> Tuple[Union[int, float], Union[int, float]]:
        return self._fig_size

    @property
    def produce_plots(self) -> bool:
        return self._produce_plots

    def _check_columns(self, mc_cols: List[str], data_cols: List[str]):
        assert self.missing_mass_col in mc_cols
        assert self.missing_mass_col in data_cols

        assert self.secondary_col in mc_cols
        assert self.secondary_col in data_cols

        assert self.weight_col in mc_cols
        assert self.is_signal_col in mc_cols

    def _init_secondary_var(self) -> Tuple[HistVariable, Binning]:
        hist_var = self._res_corr_setup.secondary_hist_var  # type: HistVariable
        binning = self._res_corr_setup.secondary_var_binning  # type: Optional[Binning]
        if binning is None:
            binning = Binning(
                bins=hist_var.n_bins,
                dimensions=1,
                scope=hist_var.scope,
                log_scale=False,
            )

        assert binning.dimensions == 1, binning.dimensions
        assert isinstance(binning.range, tuple), type(binning.range)
        assert len(binning.range) == 1, (binning.range, len(binning.range))
        assert isinstance(binning.range[0], tuple), (type(binning.range[0]), binning.range[0])

        assert hist_var.scope == binning.range[0], (hist_var.scope, binning.range, hist_var.df_label)
        return hist_var, binning
