"""
Definition of E_miss - p_miss

Felix Metzner 2020
"""

import logging
import pandas as pd

from typing import Tuple, List, Dict

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "EMissMinusPMissVariable",
]


class EMissMinusPMissVariable:
    e_miss_minus_p_miss_col_name = "e_miss_minus_p_miss"  # type: str
    e_miss_minus_p_miss_e_and_p_rescaled_col_name = "e_miss_minus_p_miss_e_p_rescaled"  # type: str
    e_miss_minus_p_miss_cms_col_name = "e_miss_minus_p_miss_in_cms"  # type: str
    e_miss_minus_p_miss_cms_ep_col_name = "e_miss_minus_p_miss_in_cms_p_miss_e_p_rescaled"  # type: str

    # Missing Energy
    mELabTagERescaled = "mELabTagERescaled"  # type: str
    mELabTagEAndPRescaled = "mELabTagEAndPRescaled"  # type: str
    mELabNoScaling = "mELabNoScaling"  # type: str
    mECMSTagERescaled = "mECMSTagERescaled"  # type: str
    mECMSNoScaling = "mECMSNoScaling"  # type: str
    # Missing Momentum
    mPLabTagERescaled = "mPLabTagERescaled"  # type: str
    mPLabTagEAndPRescaled = "mPLabTagEAndPRescaled"  # type: str
    mPCMSTagERescaled = "mPCMSTagERescaled"  # type: str
    mPCMSTagEAndPRescaled = "mPCMSTagEAndPRescaled"  # type: str

    required_variables = {
        # Missing Energy Definitions
        mELabTagERescaled: "missingEnergyLab",
        mELabTagEAndPRescaled: "missingEnergyRescaledTagLab",  # Contains Nan values for off-resonance data!
        mELabNoScaling: "missingEnergyNoScalingLab",
        mECMSTagERescaled: "missingEnergyCMS",
        mECMSNoScaling: "missingEnergyNoScalingCMS",
        # Missing Momentum Definitions
        mPLabTagERescaled: "missingMomentum",
        mPLabTagEAndPRescaled: "missingMomentumRescaledTag",  # Contains NaN values for off-resonance case!
        mPCMSTagERescaled: "missingMomentumInCMSFrame",
        mPCMSTagEAndPRescaled: "missingMomentumRescaledTagInCMSFrame",  # Contains NaN values for off-resonance case!
    }  # type: Dict[str, str]

    definitions = {
        e_miss_minus_p_miss_col_name: (mELabTagERescaled, mPLabTagERescaled),
        e_miss_minus_p_miss_e_and_p_rescaled_col_name: (mELabTagEAndPRescaled, mPLabTagEAndPRescaled),
        e_miss_minus_p_miss_cms_col_name: (mECMSTagERescaled, mPCMSTagERescaled),
        e_miss_minus_p_miss_cms_ep_col_name: (mECMSTagERescaled, mPCMSTagEAndPRescaled),
    }  # type: Dict[str, Tuple[str, str]]

    @staticmethod
    def add_e_miss_minus_p_miss_definitions(df: pd.DataFrame, is_off_resonance: bool) -> None:
        assert all(c in df.columns for c in EMissMinusPMissVariable.get_required_columns_list()), [
            c for c in EMissMinusPMissVariable.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in EMissMinusPMissVariable.definitions.keys()), [
            c for c in EMissMinusPMissVariable.definitions.keys() if c in df.columns
        ]

        for new_col, (e_key, p_key) in EMissMinusPMissVariable.definitions.items():
            e_col = EMissMinusPMissVariable.required_variables[e_key]
            p_col = EMissMinusPMissVariable.required_variables[p_key]
            df.loc[:, new_col] = df[e_col] - df[p_col]
            exception = new_col in EMissMinusPMissVariable.get_bad_definitions_for_off_resonance() and is_off_resonance
            if exception:
                logging.warning(
                    f"Check for NaNs will be omitted for off resonance sample for the definition {new_col}"
                    + " which would have contained NaNs"
                    if pd.isnull(df[new_col]).any()
                    else "."
                )
            else:
                assert not pd.isnull(df[new_col]).any(), new_col

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(EMissMinusPMissVariable.required_variables.values())

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(EMissMinusPMissVariable.definitions.keys())

    @staticmethod
    def get_bad_definitions_for_off_resonance() -> List[str]:
        # These definitions will lead to NaN values for off-resonance data and MC,
        # because the values they are based one already do.
        return [
            EMissMinusPMissVariable.e_miss_minus_p_miss_e_and_p_rescaled_col_name,
            EMissMinusPMissVariable.e_miss_minus_p_miss_cms_ep_col_name,
        ]
