"""
Definition of additional kinematic variables.

Felix Metzner 2021
"""

import logging
import numpy as np
import pandas as pd

from scipy.stats import binned_statistic_2d
from typing import Union, Tuple, List, Dict, Callable, Type, Any

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.observables.offline_observable_definition_base import BinnedOfflineObservableBase


logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "LeptonPhaseSpaceBinID",
    "KinematicObservables",
]


class LeptonPhaseSpaceBinID(BinnedOfflineObservableBase):
    new_column_name = "lepton_phase_space_bin_id"  # type: str

    _use_simple_tick_labels = False  # type: bool

    lepton_p_5_bins_equal_frequency_binning = (
        0.0,
        0.87832912,
        1.14442348,
        1.3894603,
        1.65222468,
        2.0,
    )  # type: Tuple[float, ...]  # Equal frequency binning from mc.
    lepton_theta_5_bins_equal_frequency_binning = (
        0.0,
        0.7990633,
        1.09147632,
        1.37113117,
        1.71175465,
        3.0,
    )  # type: Tuple[float, ...]  # Equal frequency binning from mc.

    lepton_p_3_bins_equal_frequency_binning = (
        0.0,
        1.06035184,
        1.47254149,
        12.0,  # Most values are below 2.0 GeV, but in continuum much larger values can occur which must be covered for the 2D approach.
    )  # type: Tuple[float, ...]  # Equal frequency binning from mc.
    lepton_theta_3_bins_equal_frequency_binning = (
        0.0,
        0.99992182,
        1.47253153,
        3.15,
    )  # type: Tuple[float, ...]  # Equal frequency binning from mc.

    use_binning = (
        lepton_p_3_bins_equal_frequency_binning,
        lepton_theta_3_bins_equal_frequency_binning,
    )  # type: Tuple[Tuple[float, ...], Tuple[float, ...]]

    _number_of_phase_space_bins = 9  # type: int  # Results from 2-dimensional binning...

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=_number_of_phase_space_bins,
        scope=tuple(np.arange(_number_of_phase_space_bins + 1) - 0.5)[::_number_of_phase_space_bins],  # from 0 to 8
        var_name=LatexObsNames.lepton_p_theta_bin,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    lepton_momentum = "lepton_momentum_p"  # type: str
    lepton_theta = "lepton_momentum_theta"  # type: str

    required_variables = {
        lepton_momentum: "daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
        lepton_theta: "daughter__bo1__cm__spdaughter__bo1__cm__sptheta__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        dimensions = (len(cls.use_binning[0]) - 1, len(cls.use_binning[1]) - 1)  # type: Tuple[int, int]
        assert cls._number_of_phase_space_bins == dimensions[0] * dimensions[1], (
            cls._number_of_phase_space_bins,
            dimensions,
        )

        lepton_momenta = df[cls.required_variables[cls.lepton_momentum]].values  # type: np.array
        lepton_thetas = df[cls.required_variables[cls.lepton_theta]].values  # type: np.array

        _, __, ___, _bin_indices = binned_statistic_2d(
            x=lepton_momenta,
            y=lepton_thetas,
            values=lepton_momenta,  # Arbitrary choice, because we are not interested in the statistic...
            statistic="min",  # Arbitrary choice, because we are not interested in the statistic...
            bins=cls.use_binning,
            expand_binnumbers=True,
        )  # type: Any, Any, Any, np.ndarray

        reduced_bin_indices = _bin_indices - 1  # type: np.ndarray
        assert np.all(reduced_bin_indices[0] >= 0), np.min(
            reduced_bin_indices[0], np.unique(lepton_momenta[reduced_bin_indices[0] < 0])
        )
        assert np.all(reduced_bin_indices[1] >= 0), np.min(
            reduced_bin_indices[1], np.unique(lepton_thetas[reduced_bin_indices[1] < 0])
        )
        assert np.all(reduced_bin_indices[0] < dimensions[0]), (
            np.max(reduced_bin_indices[0]),
            np.unique(lepton_momenta[reduced_bin_indices[0] >= dimensions[0]]),
            dimensions,
        )
        assert np.all(reduced_bin_indices[1] < dimensions[1]), (
            np.max(reduced_bin_indices[1]),
            np.unique(lepton_thetas[reduced_bin_indices[1] >= dimensions[1]]),
            dimensions,
        )

        bin_indices = np.ravel_multi_index(multi_index=reduced_bin_indices, dims=dimensions)  # type: np.array

        assert len(np.unique(bin_indices)) <= cls._number_of_phase_space_bins, (
            len(np.unique(bin_indices)),
            dimensions,
            cls._number_of_phase_space_bins,
            np.unique(bin_indices),
        )
        assert np.all(np.unique(bin_indices) > cls.hist_var.scope[0]), (np.unique(bin_indices), cls.hist_var.scope)
        assert np.all(np.unique(bin_indices) < cls.hist_var.scope[-1]), (np.unique(bin_indices), cls.hist_var.scope)

        return bin_indices

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        p_edges = LeptonPhaseSpaceBinID.use_binning[0]  # type: Tuple[float, ...]
        t_edges = LeptonPhaseSpaceBinID.use_binning[1]  # type: Tuple[float, ...]
        dimensions = (len(p_edges) - 1, len(t_edges) - 1)  # type: Tuple[int, int]
        bin_a, bin_b = np.unravel_index(
            indices=np.array(list(range(LeptonPhaseSpaceBinID._number_of_phase_space_bins))), dims=dimensions
        )  # type: np.array, np.array

        indices_mapping = {i: ids for i, ids in enumerate(zip(bin_a, bin_b))}  # type: Dict[int, Tuple[int, int]]

        if LeptonPhaseSpaceBinID._use_simple_tick_labels:
            pb_l = LatexObsNames.lepton_p_bin  # type: str
            tb_l = LatexObsNames.lepton_theta_bin  # type: str
            return {i: pb_l[:-1] + rf" {ids[0]},\; " + tb_l[1:-1] + f"{ids[1]}$" for i, ids in indices_mapping.items()}

        p_l = LatexObsNames.lepton_p  # type: str
        t_l = LatexObsNames.lepton_theta  # type: str

        return {
            i: p_l[:-1]
            + r"\in"
            + rf"[{p_edges[ids[0]]:.2f}, {p_edges[ids[0] + 1]:.2f}),\; "
            + t_l[1:-1]
            + r"\in"
            + f"[{t_edges[ids[1]]:.2f}, {t_edges[ids[1] + 1]:.2f})$"
            for i, ids in indices_mapping.items()
        }


class KinematicObservables:

    definitions = {
        LeptonPhaseSpaceBinID.get_new_column_name(): LeptonPhaseSpaceBinID,
    }  # type: Dict[str, Type[BinnedOfflineObservableBase]]

    required_variables = tuple(
        [c for d in definitions.values() for c in d.get_required_columns()]
    )  # type: Tuple[str, ...]

    @staticmethod
    def add_kinematic_observables_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in KinematicObservables.get_required_columns_list()), [
            c for c in KinematicObservables.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in KinematicObservables.get_new_columns_list()), [
            c for c in KinematicObservables.get_new_columns_list() if c in df.columns
        ]

        for new_col, definition_class in KinematicObservables.definitions.items():
            new_values = definition_class.calculate_new_column_values(df=df)
            assert len(new_values) == len(df), (len(new_values), len(df), new_col)
            df.loc[:, new_col] = new_values

            assert not pd.isnull(df[new_col]).any(), new_col

    @staticmethod
    def get_definition_function(name: str) -> Callable:
        def definition_function(df: pd.DataFrame, check_range: bool = True) -> None:
            definition = KinematicObservables.definitions[name]  # type: Type[BinnedOfflineObservableBase]
            new_col_name = definition.get_new_column_name()
            assert new_col_name not in df.columns, new_col_name
            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(KinematicObservables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(KinematicObservables.definitions.keys())

    @staticmethod
    def get_kinematic_definition(name: str) -> Type[BinnedOfflineObservableBase]:
        return KinematicObservables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(KinematicObservables.definitions.keys())
