"""
Definition of Q Squared residuals (q1_truth - q2_reco)

Felix Metzner 2020
"""

import logging
import numpy as np
import pandas as pd

from typing import Tuple, List, Dict

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "QSquaredResidualVariable",
]


class QSquaredResidualVariable:
    may_contain_nan_values = True  # type: bool

    q_squared_residual_col_name = "qSquaredRecoMinusTruthResidual"  # type: str
    q_squared_residual_with_q2reco_2_col_name = "qSquaredRecoMinusTruthResidual2"  # type: str
    q_squared_residual_with_q2reco_3_col_name = "qSquaredRecoMinusTruthResidual3"  # type: str
    q_squared_residual_with_q2reco_4_col_name = "qSquaredRecoMinusTruthResidual4"  # type: str
    q_squared_resolution_col_name = "qSquaredRecoMinusTruthResolution"  # type: str
    q_squared_resolution_with_q2reco_2_col_name = "qSquaredRecoMinusTruthResolution2"  # type: str
    q_squared_resolution_with_q2reco_3_col_name = "qSquaredRecoMinusTruthResolution3"  # type: str
    q_squared_resolution_with_q2reco_4_col_name = "qSquaredRecoMinusTruthResolution4"  # type: str

    # q_squared truth column
    q_squared_truth_col_name = "qSquaredTruth"  # type: str
    # q_squared reco definitions
    q_squared_reco_1_col_name = "qSquaredReco"  # type: str
    q_squared_reco_2_col_name = "qSquaredReco2"  # type: str
    q_squared_reco_3_col_name = "qSquaredReco3"  # type: str
    q_squared_reco_4_col_name = "qSquaredReco4"  # type: str

    required_variables = {
        q_squared_reco_1_col_name: q_squared_reco_1_col_name,
        q_squared_reco_2_col_name: q_squared_reco_2_col_name,
        q_squared_reco_3_col_name: q_squared_reco_3_col_name,
        q_squared_reco_4_col_name: q_squared_reco_4_col_name,
    }  # type: Dict[str, str]

    definitions = {
        q_squared_residual_col_name: (q_squared_reco_1_col_name, q_squared_truth_col_name, False),
        q_squared_residual_with_q2reco_2_col_name: (q_squared_reco_2_col_name, q_squared_truth_col_name, False),
        q_squared_residual_with_q2reco_3_col_name: (q_squared_reco_3_col_name, q_squared_truth_col_name, False),
        q_squared_residual_with_q2reco_4_col_name: (q_squared_reco_4_col_name, q_squared_truth_col_name, False),
        q_squared_resolution_col_name: (q_squared_reco_1_col_name, q_squared_truth_col_name, True),
        q_squared_resolution_with_q2reco_2_col_name: (q_squared_reco_2_col_name, q_squared_truth_col_name, True),
        q_squared_resolution_with_q2reco_3_col_name: (q_squared_reco_3_col_name, q_squared_truth_col_name, True),
        q_squared_resolution_with_q2reco_4_col_name: (q_squared_reco_4_col_name, q_squared_truth_col_name, True),
    }  # type: Dict[str, Tuple[str, str, bool]]

    mc_q_squared_for_mdst_1_b_col_name = "genUpsilon4S__bomcDaughter__bo0__cm__spqSquaredMC__bc__bc"  # type: str
    mc_q_squared_for_mdst_2_b_col_name = "genUpsilon4S__bomcDaughter__bo1__cm__spqSquaredMC__bc__bc"  # type: str
    b_mdst_col_name = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"  # type: str
    b_mdst_tau_col_name = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
    )  # type: str

    additional_required_columns = (
        mc_q_squared_for_mdst_1_b_col_name,
        mc_q_squared_for_mdst_2_b_col_name,
        b_mdst_col_name,
        b_mdst_tau_col_name,
    )  # type: Tuple[str, ...]

    valid_b_mdst_indices = (1, 2)  # type: Tuple[int, int]

    mdst_info_pairs = {
        1: mc_q_squared_for_mdst_1_b_col_name,
        2: mc_q_squared_for_mdst_2_b_col_name,
    }  # type: Dict[int, str]

    @staticmethod
    def add_q_squared_residual_definitions(df: pd.DataFrame) -> None:
        assert (
            QSquaredResidualVariable.q_squared_truth_col_name not in df.columns
        ), QSquaredResidualVariable.q_squared_truth_col_name

        assert all(c in df.columns for c in QSquaredResidualVariable.get_required_columns_list()), [
            c for c in QSquaredResidualVariable.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in QSquaredResidualVariable.get_new_columns_list()), [
            c for c in QSquaredResidualVariable.get_new_columns_list() if c in df.columns
        ]

        kc = set(QSquaredResidualVariable.valid_b_mdst_indices) == set(QSquaredResidualVariable.mdst_info_pairs.keys())
        assert kc, (QSquaredResidualVariable.valid_b_mdst_indices, QSquaredResidualVariable.mdst_info_pairs.keys())

        decay_via_lep_mask = df[QSquaredResidualVariable.b_mdst_col_name].isin(
            QSquaredResidualVariable.valid_b_mdst_indices
        )  # type: pd.Series
        decay_via_tau_mask = df[QSquaredResidualVariable.b_mdst_tau_col_name].isin(
            QSquaredResidualVariable.valid_b_mdst_indices
        )  # type: pd.Series

        b_mdst_info_cols = {
            QSquaredResidualVariable.b_mdst_col_name: decay_via_lep_mask,
            QSquaredResidualVariable.b_mdst_tau_col_name: decay_via_tau_mask,
        }  # type: Dict[str, pd.Series]

        df.loc[:, QSquaredResidualVariable.q_squared_truth_col_name] = np.nan
        for b_mdst_info_col, b_mdst_mask in b_mdst_info_cols.items():
            for mc_mdst_index, mc_q_squared_col_name in QSquaredResidualVariable.mdst_info_pairs.items():
                get_matching_b_mdst_index_mask = df[b_mdst_info_col] == mc_mdst_index  # type: pd.Series
                mask = b_mdst_mask & get_matching_b_mdst_index_mask  # type: pd.Series
                df.loc[mask, QSquaredResidualVariable.q_squared_truth_col_name] = df[mc_q_squared_col_name]

        all_matching_mask = decay_via_lep_mask | decay_via_tau_mask  # type: pd.Series
        assert not df.loc[all_matching_mask, QSquaredResidualVariable.q_squared_truth_col_name].isnull().any()

        for new_col, (reco_col, truth_col, is_resolution) in QSquaredResidualVariable.definitions.items():
            df.loc[:, new_col] = np.nan
            if is_resolution:
                none_zero_mask = df[truth_col] != 0.0
                assert isinstance(none_zero_mask, pd.Series)
                df.loc[all_matching_mask & none_zero_mask, new_col] = (df[reco_col] - df[truth_col]) / df[truth_col]
                df.loc[all_matching_mask & (~none_zero_mask), new_col] = (df[reco_col] - df[truth_col] - 0.01) / (
                    df[truth_col] + 0.01
                )
            else:
                df.loc[all_matching_mask, new_col] = df[reco_col] - df[truth_col]
            assert not pd.isnull(df.loc[all_matching_mask, new_col]).any(), new_col

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(QSquaredResidualVariable.required_variables.values()) + list(
            QSquaredResidualVariable.additional_required_columns
        )

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(QSquaredResidualVariable.definitions.keys())
