"""
Definition of decay modes for resolution correction.

Felix Metzner 2020
"""

import copy
import logging
import numpy as np
import pandas as pd

from typing import Union, Tuple, List, Dict, Callable, Type

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.offline_analysis.fei_b_tagging.fei_reco_mode_interpreter import FEIDecayModeIDInterpreter
from rdstar.online_analysis.decay_channels import B_decay_label_to_decay_id, U4S_decay_label_to_decay_id
from rdstar.online_analysis.decay_channels import Dstar_decay_label_to_decay_id, D_decay_label_to_decay_id
from rdstar.offline_analysis.observables.offline_observable_definition_base import BinnedOfflineObservableBase
from rdstar.offline_analysis.rdstar_plots import d_meson_decay_mode_id_to_latex_label_dict, decay_mode_id_dict
from rdstar.online_analysis.decay_channels import B_Dlpi0_decay_label_to_decay_id, B_Dlpim_decay_label_to_decay_id


logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "SimpleDMesonRecoIDDefinition",
    "DAndDsMesonRecoIDDefinition",
    "BSigMesonRecoIDDefinition",
    "BSigMesonWithDststRecoIDDefinition",
    "BTagMesonRecoIDDefinition",
    "U4SRecoIDDefinition",
    "RecoIDObservables",
]


class SimpleDMesonRecoIDDefinition(BinnedOfflineObservableBase):
    new_column_name = "generic_d_meson_reco_mode_id"  # type: str

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(D_decay_label_to_decay_id.values()),
        scope=tuple(np.arange(len(D_decay_label_to_decay_id) + 1) + 0.5)[:: len(D_decay_label_to_decay_id)],
        var_name=LatexObsNames.reco_d_decay_mode_label,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    d_or_d_st_meson_reco_mode_id = "d_or_d_st_meson_reco_mode_id"  # type: str
    d_from_d_st_meson_reco_mode_id = "d_from_d_st_meson_reco_mode_id"  # type: str

    required_variables = {
        d_or_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc",
        d_from_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        d_reco_ids = tuple(D_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        d_st_reco_ids = tuple(Dstar_decay_label_to_decay_id.values())  # type: Tuple[int, ...]

        reco_id_mapping = {r_id: i + 1 for i, r_id in zip(range(len(d_reco_ids)), d_reco_ids)}  # type: Dict[int, int]

        is_d = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].isin(d_reco_ids)
        is_from_dst = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].isin(d_st_reco_ids)
        is_d_from_dst = is_from_dst & (df[cls.required_variables[cls.d_from_d_st_meson_reco_mode_id]].isin(d_reco_ids))

        assert np.sum(is_d & is_d_from_dst) == 0, np.sum(is_d & is_d_from_dst)
        assert np.sum(is_d | is_d_from_dst) == len(np.unique(df.index)), (
            np.sum(is_d | is_d_from_dst),
            len(np.unique(df.index)),
        )

        d_reco_ids_array = np.select(
            condlist=[is_d.values, is_d_from_dst.values],
            choicelist=[
                df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].values,
                df[cls.required_variables[cls.d_from_d_st_meson_reco_mode_id]].values,
            ],
            default=-1,
        )

        assert np.sum(d_reco_ids_array == -1) == 0, np.sum(d_reco_ids_array == -1)
        assert np.all(np.isin(element=np.unique(d_reco_ids_array), test_elements=list(reco_id_mapping.keys())))

        return np.vectorize(reco_id_mapping.get)(d_reco_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        d_reco_ids = tuple(D_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        return {
            i + 1: d_meson_decay_mode_id_to_latex_label_dict()[r_id]
            for i, r_id in zip(range(len(d_reco_ids)), d_reco_ids)
        }


class DAndDsMesonRecoIDDefinition(BinnedOfflineObservableBase):
    new_column_name = "d_and_ds_meson_reco_mode_id"  # type: str

    n_bins = len(D_decay_label_to_decay_id) + len(Dstar_decay_label_to_decay_id)  # type: int

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=n_bins,
        scope=tuple(np.arange(n_bins + 1) + 0.5)[::n_bins],
        var_name=LatexObsNames.reco_d_or_dst_decay_mode_label,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    d_or_d_st_meson_reco_mode_id = "d_or_d_st_meson_reco_mode_id"  # type: str

    required_variables = {
        d_or_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        d_reco_ids = tuple(D_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        d_st_reco_ids = tuple(Dstar_decay_label_to_decay_id.values())  # type: Tuple[int, ...]

        d_and_dst_reco_ids = d_reco_ids + d_st_reco_ids  # type: Tuple[int, ...]

        reco_id_mapping = {
            r_id: i + 1 for i, r_id in zip(range(len(d_and_dst_reco_ids)), d_and_dst_reco_ids)
        }  # type: Dict[int, int]

        d_reco_ids_array = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].values
        assert np.all(np.isin(element=np.unique(d_reco_ids_array), test_elements=list(reco_id_mapping.keys())))

        return np.vectorize(reco_id_mapping.get)(d_reco_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        d_reco_ids = tuple(D_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        d_st_reco_ids = tuple(Dstar_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        d_and_dst_reco_ids = d_reco_ids + d_st_reco_ids  # type: Tuple[int, ...]
        return {
            i + 1: d_meson_decay_mode_id_to_latex_label_dict()[r_id]
            for i, r_id in zip(range(len(d_and_dst_reco_ids)), d_and_dst_reco_ids)
        }


class BSigMesonRecoIDDefinition(BinnedOfflineObservableBase):
    new_column_name = "b_sig_meson_reco_mode_id"  # type: str

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(B_decay_label_to_decay_id.values()),
        scope=tuple(np.arange(len(B_decay_label_to_decay_id) + 1) + 0.5)[:: len(B_decay_label_to_decay_id)],
        var_name=LatexObsNames.reco_b_sig_decay_mode_label,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    __b_charge_based_mapping = False  # type: bool
    b_dss_charged_pi_to_regular_b_mapping = {
        71: 11 if __b_charge_based_mapping else 13,  # "B0:D[0]_e_pi-": 71  ->  "B0:D[+]_e": 11
        72: 12 if __b_charge_based_mapping else 14,  # "B0:D[0]_mu_pi-": 72  ->  "B0:D[+]_mu": 12
        73: 13 if __b_charge_based_mapping else 11,  # "B-:D+_e_pi-": 73  ->  "B-:D0_e": 13
        74: 14 if __b_charge_based_mapping else 12,  # "B-:D+_mu_pi-": 74  ->  "B-:D0_mu": 14
        75: 15 if __b_charge_based_mapping else 17,  # "B0:D*[0]_e_pi-": 75  ->  "B0:D*[+]_e": 15
        76: 16 if __b_charge_based_mapping else 18,  # "B0:D*[0]_mu_pi-": 76  ->  "B0:D*[+]_mu": 16
        77: 17 if __b_charge_based_mapping else 15,  # "B-:D*+_e_pi-": 77  ->  "B-:D*0_e": 17
        78: 18 if __b_charge_based_mapping else 16,  # "B-:D*+_mu_pi-": 78  ->  "B-:D*0_mu": 18
    }  # type: Dict[int, int]

    b_sig_reco_mode_id = "b_sig_reco_mode_id"  # type: str

    required_variables = {
        b_sig_reco_mode_id: "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        b_sig_reco_ids = tuple(B_decay_label_to_decay_id.values())  # type: Tuple[int, ...]

        reco_id_mapping = {
            r_id: i + 1 for i, r_id in zip(range(len(b_sig_reco_ids)), b_sig_reco_ids)
        }  # type: Dict[int, int]

        b_sig_mapping_charged_pi_update = {
            r_id: reco_id_mapping[cls.b_dss_charged_pi_to_regular_b_mapping[r_id]]
            for r_id in B_Dlpim_decay_label_to_decay_id.values()
        }  # type: Dict[int, int]
        assert all(k not in reco_id_mapping.keys() for k in b_sig_mapping_charged_pi_update.keys())
        assert all(v in reco_id_mapping.values() for v in b_sig_mapping_charged_pi_update.values())
        reco_id_mapping.update(b_sig_mapping_charged_pi_update)

        b_sig_mapping_neutral_pi_update = {
            r_id: reco_id_mapping[B_decay_label_to_decay_id[decay_str.rsplit(sep="_pi", maxsplit=1)[0]]]
            for decay_str, r_id in B_Dlpi0_decay_label_to_decay_id.items()
        }  # type: Dict[int, int]
        assert all(k not in reco_id_mapping.keys() for k in b_sig_mapping_neutral_pi_update.keys())
        assert all(v in reco_id_mapping.values() for v in b_sig_mapping_neutral_pi_update.values())
        reco_id_mapping.update(b_sig_mapping_neutral_pi_update)

        b_sig_reco_ids_array = df[cls.required_variables[cls.b_sig_reco_mode_id]].values
        assert np.all(np.isin(element=np.unique(b_sig_reco_ids_array), test_elements=list(reco_id_mapping.keys()))), (
            np.unique(b_sig_reco_ids_array),
            list(reco_id_mapping.keys()),
        )

        return np.vectorize(reco_id_mapping.get)(b_sig_reco_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        b_sig_reco_ids = tuple(B_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        return {i + 1: decay_mode_id_dict()[r_id] for i, r_id in zip(range(len(b_sig_reco_ids)), b_sig_reco_ids)}


class BSigMesonWithDststRecoIDDefinition(BinnedOfflineObservableBase):
    new_column_name = "b_sig_meson_with_Dstst_reco_mode_id"  # type: str

    all_b_sig_reco_ids = tuple(
        [
            decay_id
            for decay_map in [B_decay_label_to_decay_id, B_Dlpi0_decay_label_to_decay_id, B_Dlpim_decay_label_to_decay_id]
            for decay_id in decay_map.values()
        ]
    )  # type: Tuple[int, ...]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(all_b_sig_reco_ids),
        scope=tuple(np.arange(len(all_b_sig_reco_ids) + 1) + 0.5)[:: len(all_b_sig_reco_ids)],
        var_name=LatexObsNames.reco_b_sig_decay_mode_label,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    b_sig_reco_mode_id = "b_sig_reco_mode_id"  # type: str

    required_variables = {
        b_sig_reco_mode_id: "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        reco_id_mapping = {
            r_id: i + 1 for i, r_id in zip(range(len(cls.all_b_sig_reco_ids)), cls.all_b_sig_reco_ids)
        }  # type: Dict[int, int]

        b_sig_reco_ids_array = df[cls.required_variables[cls.b_sig_reco_mode_id]].values  # type: np.ndarray
        assert np.all(np.isin(element=np.unique(b_sig_reco_ids_array), test_elements=list(reco_id_mapping.keys()))), (
            np.unique(b_sig_reco_ids_array),
            list(reco_id_mapping.keys()),
        )

        return np.vectorize(reco_id_mapping.get)(b_sig_reco_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        all_b_sig_id_values = BSigMesonWithDststRecoIDDefinition.all_b_sig_reco_ids  # type: Tuple[int, ...]
        return {
            i + 1: decay_mode_id_dict()[r_id] for i, r_id in zip(range(len(all_b_sig_id_values)), all_b_sig_id_values)
        }


class BTagMesonRecoIDDefinition(BinnedOfflineObservableBase):
    new_column_name = "b_tag_reco_mode_id"  # type: str

    fei_tag_label_mapping = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_to_latex_label_dict(
        include_b_to=False
    )  # type: Dict[int, str]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(fei_tag_label_mapping),
        scope=(min(fei_tag_label_mapping.keys()) - 0.5, max(fei_tag_label_mapping.keys()) + 0.5),
        var_name=LatexObsNames.reco_b_tag_decay_mode_label,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    b_tag_pdg = "b_tag_pdg_code"  # type: str
    b_tag_reco_mode_id = "b_tag_reco_mode_id"  # type: str

    required_variables = {
        b_tag_pdg: "daughter__bo0__cm__spPDG__bc",
        b_tag_reco_mode_id: "daughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        original_b_tag_reco_ids_array = df[cls.required_variables[cls.b_tag_reco_mode_id]].values
        assert np.all(
            np.isin(
                element=np.unique(original_b_tag_reco_ids_array),
                test_elements=[
                    k for d in FEIDecayModeIDInterpreter.get_fei_decay_mode_id_dict().values() for k in d.keys()
                ],
            )
        )

        b_tag_reco_ids_array = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_ids(
            df=df,
            b_meson_pdg_column_name=cls.required_variables[cls.b_tag_pdg],
            fei_decay_mode_id_column_name=cls.required_variables[cls.b_tag_reco_mode_id],
        )

        assert np.all(
            np.isin(element=np.unique(b_tag_reco_ids_array), test_elements=list(cls.fei_tag_label_mapping.keys()))
        )

        return b_tag_reco_ids_array

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        return BTagMesonRecoIDDefinition.fei_tag_label_mapping


class U4SRecoIDDefinition(BinnedOfflineObservableBase):
    new_column_name = "u4s_reco_mode_id"  # type: str

    u4s_id_mapping = {
        decay_id: 1 + 2 * int("mixed" in ch_str) + int("_wt" in ch_str)
        for ch_str, decay_id in U4S_decay_label_to_decay_id.items()
    }  # type: Dict[int, int]

    _u_to = LatexParticleNames.y4s_str[:-1] + r" \rightarrow "  # type: str
    u4s_label_mapping = {
        1: _u_to + LatexParticleNames.b_sig_minus_str[1:-1] + LatexParticleNames.b_tag_plus_str[1:],  # charged
        2: _u_to + LatexParticleNames.b_sig_minus_str[1:-1] + LatexParticleNames.b_tag_neutral_str[1:],  # ch + wrong tag
        3: _u_to + LatexParticleNames.b_sig_neutral_str[1:-1] + LatexParticleNames.b_tag_anti_neutral_str[1:],  # mixed
        4: _u_to + LatexParticleNames.b_sig_neutral_str[1:-1] + LatexParticleNames.b_tag_plus_str[1:],  # mix + wrong tag
    }  # type: Dict[int, str]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(u4s_label_mapping),
        scope=tuple(np.arange(len(u4s_label_mapping) + 1) + 0.5)[:: len(u4s_label_mapping)],
        var_name=LatexObsNames.reco_y4s_decay_mode_label,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    binning = Binning(
        bins=(0.5, 2.5, 4.5),
        dimensions=1,
        scope=hist_var.scope,
        log_scale=False,
    )  # type: Binning

    u4s_reco_mode_id = "u4s_reco_mode_id"  # type: str

    required_variables = {
        u4s_reco_mode_id: "extraInfo__bodecayModeID__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        reco_id_mapping = U4SRecoIDDefinition.u4s_id_mapping  # type: Dict[int, int]

        u4s_reco_ids_array = df[cls.required_variables[cls.u4s_reco_mode_id]].values
        assert np.all(np.isin(element=np.unique(u4s_reco_ids_array), test_elements=list(reco_id_mapping.keys())))

        return np.vectorize(reco_id_mapping.get)(u4s_reco_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        keys_set = set(U4SRecoIDDefinition.u4s_label_mapping.keys())
        values_set = set(U4SRecoIDDefinition.u4s_id_mapping.values())
        assert keys_set == values_set, (keys_set.symmetric_difference(values_set), keys_set, values_set)

        return copy.copy(U4SRecoIDDefinition.u4s_label_mapping)


class RecoIDObservables:

    definitions = {
        SimpleDMesonRecoIDDefinition.get_new_column_name(): SimpleDMesonRecoIDDefinition,
        DAndDsMesonRecoIDDefinition.get_new_column_name(): DAndDsMesonRecoIDDefinition,
        BSigMesonRecoIDDefinition.get_new_column_name(): BSigMesonRecoIDDefinition,
        BSigMesonWithDststRecoIDDefinition.get_new_column_name(): BSigMesonWithDststRecoIDDefinition,
        BTagMesonRecoIDDefinition.get_new_column_name(): BTagMesonRecoIDDefinition,
        U4SRecoIDDefinition.get_new_column_name(): U4SRecoIDDefinition,
    }  # type: Dict[str, Type[BinnedOfflineObservableBase]]

    required_variables = tuple(
        [c for d in definitions.values() for c in d.get_required_columns()]
    )  # type: Tuple[str, ...]

    @staticmethod
    def add_reco_id_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in RecoIDObservables.get_required_columns_list()), [
            c for c in RecoIDObservables.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in RecoIDObservables.get_new_columns_list()), [
            c for c in RecoIDObservables.get_new_columns_list() if c in df.columns
        ]

        for new_col, definition_class in RecoIDObservables.definitions.items():
            new_values = definition_class.calculate_new_column_values(df=df)
            assert len(new_values) == len(df), (len(new_values), len(df), new_col)
            df.loc[:, new_col] = new_values

            assert not pd.isnull(df[new_col]).any(), new_col

    @staticmethod
    def get_definition_function(name: str) -> Callable:
        def definition_function(df: pd.DataFrame, check_range: bool = True) -> None:
            definition = RecoIDObservables.definitions[name]  # type: Type[BinnedOfflineObservableBase]
            new_col_name = definition.get_new_column_name()
            assert new_col_name not in df.columns, new_col_name
            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(RecoIDObservables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(RecoIDObservables.definitions.keys())

    @staticmethod
    def get_reco_id_definition(name: str) -> Type[BinnedOfflineObservableBase]:
        return RecoIDObservables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(RecoIDObservables.definitions.keys())
