"""
This module provides the plotting methods for the rdstar analysis.

Felix Metzner 2019
"""

import os
import copy
import numpy as np
import pandas as pd
from scipy.stats import chi2
import matplotlib.pyplot as plt

from typing import Any, Union, Optional, Tuple, List, Dict, Generator

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import StackedHistogramPlot
from templatefitter.plotter.plot_utilities import FigureType, AxesType

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses
from rdstar.offline_analysis.mc_ids import sig_id_to_label, rdstar_sig_queries, decay_mode_to_latex_label
from rdstar.online_analysis.decay_channels import Dstar_decay_label_to_decay_id, D_decay_label_to_decay_id

plot_style.set_matplotlibrc_params()

__all__ = [
    "rdstar_color_map",
    "rdstar_plot_order",
    "rdstar_dec_modes",
    "rdstar_neutral_b_dec_modes",
    "rdstar_charged_b_dec_modes",
    "dss_dec_modes",
    "dss_neutral_b_dec_modes",
    "dss_charged_b_dec_modes",
    "dss_charged_dec_modes",
    "dss_charged_neutral_b_dec_modes",
    "dss_charged_charged_b_dec_modes",
    "rdstar_dec_titles",
    "get_short_dec_titles",
    "dss_dec_titles",
    "dss_charged_dec_titles",
    "dss_mc_titles",
    "decay_mode_id_dict",
    "u4s_decay_mode_id_dict",
    "d_meson_decay_name_to_id_dict",
    "d_meson_decay_mode_id_to_latex_label_dict",
    "get_rdstar_color_dict",
    "plot_chi_squared_toys",
    "plot_stacked_for_all_modes",
    "DecayModeDictType",
    "DecayTitleDictType",
]

DecayModeDictType = Dict[str, List[int]]
DecayTitleDictType = Dict[str, str]

rdstar_color_map = "viridis"  # type: str

rdstar_plot_order = ["CBKG", "GBKG", "HBKG", "DSSlbkg", "DSSl", "DSl", "Dl", "DStau", "Dtau"]  # type: List[str]


rdstar_neutral_b_dec_modes = {
    "DM": [11, 12],
    "DSM": [15, 16],
}  # type: DecayModeDictType

rdstar_charged_b_dec_modes = {
    "DZ": [13, 14],
    "DSZ": [17, 18],
}  # type: DecayModeDictType

rdstar_dec_modes = {
    **rdstar_neutral_b_dec_modes,
    **rdstar_charged_b_dec_modes,
}  # type: DecayModeDictType

dss_neutral_b_dec_modes = {
    "DMPi": [61, 62],
    "DSMPi": [65, 66],
}  # type: DecayModeDictType

dss_charged_b_dec_modes = {
    "DZPi": [63, 64],
    "DSZPi": [67, 68],
}  # type: DecayModeDictType

dss_dec_modes = {
    **dss_neutral_b_dec_modes,
    **dss_charged_b_dec_modes,
}  # type: DecayModeDictType

dss_charged_neutral_b_dec_modes = {
    "DZPiPlus": [71, 72],
    "DSZPiPlus": [75, 76],
}  # type: DecayModeDictType

dss_charged_charged_b_dec_modes = {
    "DMPiPlus": [73, 74],
    "DSMPiPlus": [77, 78],
}  # type: DecayModeDictType

dss_charged_dec_modes = {
    **dss_charged_neutral_b_dec_modes,
    **dss_charged_charged_b_dec_modes,
}  # type: DecayModeDictType

rdstar_dec_titles = {
    "DZ": LatexDecayProcesses.reco_bm_to_d0_lm,
    "DM": LatexDecayProcesses.reco_b0_to_dm_lp,
    "DSZ": LatexDecayProcesses.reco_bm_to_d_st0_lm,
    "DSM": LatexDecayProcesses.reco_b0_to_d_stm_lp,
    "all_modes": LatexDecayProcesses.all_modes_label,
}  # type: DecayTitleDictType

dss_dec_titles = {
    "DZPi": LatexDecayProcesses.reco_bm_to_d0_pi0_lm,
    "DMPi": LatexDecayProcesses.reco_b0_to_dm_pi0_lp,
    "DSZPi": LatexDecayProcesses.reco_bm_to_d_st0_pi0_lm,
    "DSMPi": LatexDecayProcesses.reco_b0_to_d_stm_pi0_lp,
}  # type: DecayTitleDictType

dss_charged_dec_titles = {
    "DZPiPlus": LatexDecayProcesses.reco_b0_to_anti_d0_pim_lp,
    "DMPiPlus": LatexDecayProcesses.reco_bm_to_dp_pim_lm,
    "DSZPiPlus": LatexDecayProcesses.reco_b0_to_anti_d_st0_pim_lp,
    "DSMPiPlus": LatexDecayProcesses.reco_bm_to_d_stp_pim_lm,
}  # type: DecayTitleDictType

dss_mc_titles = {
    "Done": LatexDecayProcesses.b_to_d_1_l_nu,
    "Dtwostar": LatexDecayProcesses.b_to_dst_2_l_nu,
    "Dprimeone": LatexDecayProcesses.b_to_dprime1_l_nu,
    "Dzerostar": LatexDecayProcesses.b_to_d0st_l_nu,
}  # type: DecayTitleDictType


def get_short_dec_titles(dec_title_dict: DecayTitleDictType) -> DecayTitleDictType:
    return {k: "All Reco. Modes" if k == "all_modes" else v for k, v in dec_title_dict.items()}


def decay_mode_id_dict() -> Dict[int, str]:
    return decay_mode_to_latex_label


def u4s_decay_mode_id_dict() -> Dict[int, str]:
    return {
        1001: LatexDecayProcesses.u4s_to_bp_tag_bm_sig,
        1000: LatexDecayProcesses.u4s_to_b0_tag_anti_b0_sig,
        1011: LatexDecayProcesses.u4s_to_b0_tag_bm_sig,
        1021: LatexDecayProcesses.u4s_to_b0_tag_bp_sig,
        1010: LatexDecayProcesses.u4s_to_bp_tag_b0_sig,
        1020: LatexDecayProcesses.u4s_to_bm_tag_b0_sig,
    }


def d_meson_decay_name_to_id_dict() -> Dict[str, int]:
    mapping = copy.copy(Dstar_decay_label_to_decay_id)  # type: Dict[str, int]
    mapping.update(copy.copy(D_decay_label_to_decay_id))
    return mapping


def d_meson_decay_mode_id_to_latex_label_dict() -> Dict[int, str]:
    return {
        # D*0 decay Channels
        21: LatexDecayProcesses.dst0_to_d0_pi0,
        22: LatexDecayProcesses.dst0_to_d0_gamma,
        # D*+ decay Channels
        23: LatexDecayProcesses.dstp_to_d0_pip,
        24: LatexDecayProcesses.dstp_to_dp_pi0,
        # D+ decay Channels
        31: LatexDecayProcesses.dp_to_kaon_m_pip_pip,
        32: LatexDecayProcesses.dp_to_kaon_m_pip_pip_pi0,
        33: LatexDecayProcesses.dp_to_kaon_m_pip_pip_pip_pim,
        34: LatexDecayProcesses.dp_to_kaon_s0_pip,
        35: LatexDecayProcesses.dp_to_kaon_s0_pip_pi0,
        36: LatexDecayProcesses.dp_to_kaon_s0_pip_pip_pim,
        37: LatexDecayProcesses.dp_to_kaon_s0_kaon_p,
        38: LatexDecayProcesses.dp_to_kaon_p_kaon_m_pip,
        # D0 decay Channels
        41: LatexDecayProcesses.d0_to_kaon_m_pip,
        42: LatexDecayProcesses.d0_to_kaon_m_pip_pi0,
        43: LatexDecayProcesses.d0_to_kaon_m_pip_pip_pim,
        44: LatexDecayProcesses.d0_to_kaon_m_pip_pip_pim_pi0,
        45: LatexDecayProcesses.d0_to_kaon_s0_pi0,
        46: LatexDecayProcesses.d0_to_kaon_s0_pip_pim,
        47: LatexDecayProcesses.d0_to_kaon_s0_pip_pim_pi0,
        48: LatexDecayProcesses.d0_to_kaon_m_kaon_p,
    }


assert all(k in d_meson_decay_name_to_id_dict().values() for k in d_meson_decay_mode_id_to_latex_label_dict().keys())
assert len(d_meson_decay_name_to_id_dict().keys()) == len(d_meson_decay_mode_id_to_latex_label_dict().keys())


# Define reconstructed decay modes to be combined:
def get_mode_dict() -> Tuple[DecayModeDictType, DecayTitleDictType]:
    rdstar_mode_dict = {
        **rdstar_dec_modes,
        **dss_dec_modes,
        **dss_charged_dec_modes,
    }  # type: DecayModeDictType

    mode_to_title = {
        **rdstar_dec_titles,
        **dss_dec_titles,
        **dss_charged_dec_titles,
    }  # type: DecayTitleDictType

    return rdstar_mode_dict, mode_to_title


def get_rdstar_color_dict() -> Dict[int, str]:
    color_dict = dict()
    color_dict[1] = plot_style.KITColors.kit_green
    color_dict[2] = plot_style.KITColors.kit_maygreen
    color_dict[3] = plot_style.KITColors.kit_blue
    color_dict[4] = plot_style.KITColors.kit_cyan
    color_dict[5] = plot_style.KITColors.kit_red
    color_dict[6] = plot_style.KITColors.kit_orange
    color_dict[7] = plot_style.KITColors.kit_brown
    color_dict[8] = plot_style.KITColors.grey
    color_dict[9] = plot_style.KITColors.light_grey

    assert len(color_dict.keys()) == len(sig_id_to_label.keys()), (len(color_dict.keys()), len(sig_id_to_label.keys()))
    assert set(color_dict.keys()) == set(sig_id_to_label.keys()), (color_dict.keys(), sig_id_to_label.keys())

    return color_dict


def plot_stacked(
    df: pd.DataFrame,
    variable: HistVariable,
    weight_column: Optional[str] = None,
    cut_string: Optional[str] = None,
    y_lim_percentage: Optional[float] = None,
    output_name: Optional[str] = None,
    output_path: Optional[PathType] = None,
) -> Tuple[FigureType, AxesType, Dict[str, Any]]:
    if weight_column is not None:
        assert isinstance(weight_column, str)
        assert weight_column in df.columns
    if (output_name is not None) and (output_path is None):
        raise ValueError("If a output file name is provided, the output_path must also be set!")

    color_dict = get_rdstar_color_dict()

    shp = StackedHistogramPlot(variable)
    for ID in sorted(list(df.SIG_ID.unique()), reverse=True):
        if cut_string:
            shp.add_component(
                sig_id_to_label[ID],
                df.query(f"(SIG_ID == {ID}) and {cut_string}"),
                weights=weight_column,
                color=color_dict[ID],
            )
        else:
            shp.add_component(
                sig_id_to_label[ID],
                df.query(f"SIG_ID == {ID}"),
                weights=weight_column,
                color=color_dict[ID],
            )

    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=100, facecolor="w", edgecolor="k")
    ax = plt.gca()

    shp.plot_on(ax=ax)

    if y_lim_percentage is not None:
        assert 0.0 < y_lim_percentage < 1.0
        y_min, y_max = ax.get_ylim()
        ax.set_ylim(bottom=y_min, top=y_lim_percentage * y_max)

    if output_name:
        export(fig, output_name, output_path)

    return fig, ax, {"filename": output_name, "target_dir": output_path}


def plot_stacked_for_all_modes(
    df: pd.DataFrame,
    variable: HistVariable,
    weight_column: Optional[str] = None,
    cut_string: Optional[str] = None,
    dec_modes: Optional[DecayModeDictType] = None,
    dec_titles: Optional[DecayTitleDictType] = None,
    plot_order: Optional[List] = None,
    y_lim_percentage: Optional[float] = None,
    figsize: Tuple[float, float] = (6, 4.5),
    additional_text: Optional[str] = None,
    output_name: Optional[str] = None,
    base_output_path: Optional[PathType] = None,
    close_figure: bool = False,
    dry_run: bool = False,
) -> Generator[Tuple[Optional[FigureType], Optional[AxesType], str, Optional[PathType]], None, None]:

    if (output_name is not None) and (base_output_path is None):
        raise ValueError("If a output file name is provided, the output_path must also be set!")

    if dec_modes is None:
        dec_modes = rdstar_dec_modes
    if dec_titles is None:
        dec_titles = rdstar_dec_titles
    if plot_order is None:
        plot_order = rdstar_plot_order

    all_mode_ids = [d_id for d_ids in dec_modes.values() for d_id in d_ids]

    mode_infos = [(k, v) for k, v in dec_modes.items()]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        output_key = f"{output_name}_{mode_name}" if output_name else ""
        output_path = (
            os.path.join(base_output_path, f"{output_key}.pdf") if base_output_path else None
        )  # type: Optional[PathType]

        if dry_run:
            yield None, None, output_key, output_path
        else:
            if weight_column is not None:
                assert isinstance(weight_column, str)
                assert weight_column in df.columns

            shp = StackedHistogramPlot(variable)
            for comp in plot_order:
                reco_id = rdstar_sig_queries[comp].id  # type: int
                if cut_string:
                    shp.add_component(
                        sig_id_to_label[reco_id],
                        df[df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)].query(
                            f"(SIG_ID == {reco_id}) and {cut_string}"
                        ),
                        weights=weight_column,
                        color=get_rdstar_color_dict()[reco_id],
                    )
                else:
                    shp.add_component(
                        sig_id_to_label[reco_id],
                        df[df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)].query(
                            f"SIG_ID == {reco_id}"
                        ),
                        weights=weight_column,
                        color=get_rdstar_color_dict()[reco_id],
                    )

            fig = plt.figure(num=None, figsize=figsize, dpi=100, facecolor="w", edgecolor="k")
            ax = plt.gca()

            ax.set_title(dec_titles[mode_name], loc="right")

            if additional_text:
                ax.text(
                    0.05,
                    0.95,
                    additional_text,
                    horizontalalignment="left",
                    verticalalignment="top",
                    transform=ax.transAxes,
                )

            shp.plot_on(ax=ax)

            if y_lim_percentage is not None:
                assert 0.0 < y_lim_percentage < 1.0
                y_min, y_max = ax.get_ylim()
                ax.set_ylim(bottom=y_min, top=y_lim_percentage * y_max)

            if output_name:
                export(fig=fig, filename=output_key, target_dir=base_output_path, close_figure=close_figure)

            yield fig, ax, output_key, output_path


def plot_chi_squared_toys(
    toys: np.ndarray,
    bin_edges: np.ndarray,
    ndf: int,
    observed_chi2: float,
    fig_size: Tuple[Union[int, float], Union[int, float]] = (5.5, 5.0),
    full_output_path: Optional[PathType] = None,
    tag: Optional[str] = None,
) -> Tuple[FigureType, AxesType]:
    fig, ax = plt.subplots(1, 1, figsize=fig_size, dpi=200)

    ax.hist(
        toys,
        bins=bin_edges,
        color=plot_style.KITColors.kit_blue,
        label=r"Toy $\chi^2$ Distribution",
        density=True,
    )

    x_vals = np.linspace(start=bin_edges[0], stop=np.ceil(bin_edges[-1]), num=100)
    ax.plot(
        x_vals,
        chi2.pdf(x_vals, ndf),
        c=plot_style.KITColors.kit_red,
        ls="-",
        lw=2.0,
        label=r"$\chi^2$ Distribution",
    )

    ax.axvline(
        x=observed_chi2,
        c=plot_style.KITColors.kit_orange,
        ls="--",
        lw=2.0,
        label=rf"Observed $\chi^2$ = {observed_chi2:0.2f}",
    )

    ax.set_xlabel(r"$\chi^2$")
    ax.legend(frameon=False)

    if full_output_path is not None:
        output_path, base_name = os.path.split(full_output_path)
        output_name, _ = os.path.splitext(base_name)
        if tag is not None:
            output_name = f"{output_name}_{tag}"
        export(fig, output_name, output_path)

    return fig, ax
