"""
Collection of data samples used in the RDStar analysis

Felix Metzner 2019
"""

from typing import Dict, Tuple, List

from rdstar.offline_analysis import rdstar_sample_weights
from analysistools.batch_tools.file_handler import Belle_list_types


__all__ = [
    "private_data_types",
    "data_type_combinations",
    "recorded_data_samples",
    "main_sample_keys",
    "unused_sample_keys",
    "main_samples",
    "alternative_gap_samples",
    "efficiency_sample_combo_keys",
    "efficiency_samples",
    "mva_samples",
    "off_resonance_mc_samples",
    "unused_samples",
    "is_off_resonance",
]

private_data_types = list(rdstar_sample_weights.signal_mc_bfs.keys())

data_type_combinations = {
    "Charged": ["Charged"],
    "Mixed": ["Mixed"],
    "UDS": ["UDS"],
    "Charm": ["Charm"],
    "Special": [c for c in Belle_list_types if "psi" not in c],
    "SpecialPsi": [c for c in Belle_list_types if "psi" in c],
    "Dss": [c for c in private_data_types if c.startswith("B") and not ("MVA" in c or "Eff" in c)],
    "DssTau": [c for c in private_data_types if c.startswith("TB") and not ("MVA" in c or "Eff" in c)],
    "GapD1": [c for c in private_data_types if c.startswith("GapD1") and "ViaD" not in c],
    "GapDEtaOpt1": [c for c in private_data_types if c.startswith("GapDEta") and "ViaD" not in c],
    "GapDsEtaOpt1": [c for c in private_data_types if c.startswith("GapDstEta") and "ViaD" not in c],
    "GapDPPOpt1": [c for c in private_data_types if c.startswith("GapDPiPi") and "ViaD" not in c],
    "GapDsPPOpt1": [c for c in private_data_types if c.startswith("GapDstPiPi") and "ViaD" not in c],
    "GapDEtaOpt2": [c for c in private_data_types if c.startswith("GapDEta") and "ViaD" in c],
    "GapDsEtaOpt2": [c for c in private_data_types if c.startswith("GapDstEta") and "ViaD" in c],
    "GapDPPOpt2": [c for c in private_data_types if c.startswith("GapDPiPi") and "ViaD" in c],
    "GapDsPPOpt2": [c for c in private_data_types if c.startswith("GapDstPiPi") and "ViaD" in c],
    "SignalEff": [c for c in private_data_types if c.startswith("D") and not c.endswith("MVA")],
    "SignalMVA": [c for c in private_data_types if c.startswith("D") and c.endswith("MVA")],
    "DssPiZeroEff": [c for c in private_data_types if c.startswith("B") and c.endswith("Eff") and "Charged" not in c],
    "DssPiPlusEff": [c for c in private_data_types if c.startswith("B") and c.endswith("ChargedPiEff")],
    "DssPiZeroMVA": [c for c in private_data_types if c.startswith("B") and c.endswith("MVA") and "Charged" not in c],
    "DssPiPlusMVA": [c for c in private_data_types if c.startswith("B") and c.endswith("ChargedPiMVA")],
    "DssTauPiZEff": [c for c in private_data_types if c.startswith("TB") and c.endswith("TauEff")],
    "DssTauPiPEff": [c for c in private_data_types if c.startswith("TB") and c.endswith("TauChargedPiEff")],
    "OffResonanceUDS": ["OffResonanceUDS"],
    "OffResonanceCharm": ["OffResonanceCharm"],
    "DataY4S": ["DataY4S"],
    "DataContinuum": ["DataContinuum"],
}  # type: Dict[str, List[str]]

# Some checks of data sample combinations:
assert all(len(v) > 0 for v in data_type_combinations.values()), [
    k for k, v in data_type_combinations.items() if len(v) == 0
]
assert all(any(k in v for v in data_type_combinations.values()) for k in private_data_types), [
    k for k in private_data_types if not any(k in v for v in data_type_combinations.values())
]
for combo_key, combo_content in data_type_combinations.items():
    for dt_key in combo_content:
        if any(dt_key in other_c and other_k != combo_key for other_k, other_c in data_type_combinations.items()):
            all_combos_with_dt = [k for k, v in data_type_combinations.items() if dt_key in v]
            raise RuntimeError(f"Key {dt_key} appears in multiple combos: {all_combos_with_dt}")

recorded_data_samples = [
    *data_type_combinations["DataY4S"],
    *data_type_combinations["DataContinuum"],
]  # type: List[str]

main_sample_keys = [
    "Charged",
    "Mixed",
    "UDS",
    "Charm",
    "Special",
    "Dss",
    "DssTau",
    "GapD1",
    "GapDEtaOpt1",
    "GapDsEtaOpt1",
    "GapDPPOpt1",
    "GapDsPPOpt1",
]  # type: List[str]

unused_sample_keys = tuple(
    [
        "SpecialPsi",
    ]
)  # type: Tuple[str, ...]

main_samples = [
    *data_type_combinations["Charged"],
    *data_type_combinations["Mixed"],
    *data_type_combinations["UDS"],
    *data_type_combinations["Charm"],
    *data_type_combinations["Special"],
    *data_type_combinations["Dss"],
    *data_type_combinations["DssTau"],
    *data_type_combinations["GapD1"],
    *data_type_combinations["GapDEtaOpt1"],
    *data_type_combinations["GapDsEtaOpt1"],
    *data_type_combinations["GapDPPOpt1"],
    *data_type_combinations["GapDsPPOpt1"],
]  # type: List[str]

alternative_gap_samples = [
    *data_type_combinations["GapDEtaOpt2"],
    *data_type_combinations["GapDsEtaOpt2"],
    *data_type_combinations["GapDPPOpt2"],
    *data_type_combinations["GapDsPPOpt2"],
]  # type: List[str]

efficiency_sample_combo_keys = [
    "SignalEff",
    "DssPiZeroEff",
    "DssPiPlusEff",
    "DssTauPiZEff",
    "DssTauPiPEff",
]  # type: List[str]

efficiency_samples = [
    *data_type_combinations["SignalEff"],
    *data_type_combinations["DssPiZeroEff"],
    *data_type_combinations["DssPiPlusEff"],
    *data_type_combinations["DssTauPiZEff"],
    *data_type_combinations["DssTauPiPEff"],
]  # type: List[str]

mva_samples = [
    *data_type_combinations["SignalMVA"],
    *data_type_combinations["DssPiZeroMVA"],
    *data_type_combinations["DssPiPlusMVA"],
]  # type: List[str]

off_resonance_mc_samples = [
    *data_type_combinations["OffResonanceUDS"],
    *data_type_combinations["OffResonanceCharm"],
]  # type: List[str]

unused_samples = [
    *data_type_combinations["SpecialPsi"],
]  # type: List[str]

assert all(dc in main_samples for dtk in main_sample_keys for dc in data_type_combinations[dtk])
assert all(dc in unused_samples for dtk in unused_sample_keys for dc in data_type_combinations[dtk])
assert all(dc in efficiency_samples for dtk in efficiency_sample_combo_keys for dc in data_type_combinations[dtk])

all_samples = (
    recorded_data_samples
    + main_samples
    + alternative_gap_samples
    + efficiency_samples
    + mva_samples
    + off_resonance_mc_samples
    + unused_samples
)  # type: List[str]

assert len([s for v in data_type_combinations.values() for s in v]) == len(all_samples)
assert len(set([s for v in data_type_combinations.values() for s in v])) == len(set(all_samples))


def is_off_resonance(data_type_key: str) -> bool:
    return data_type_key in ["DataContinuum", "OffResonanceCharm", "OffResonanceUDS"]
