"""
Class collecting and evaluating information on Signal Selection

Felix Metzner 2020
"""

import copy
import logging
import warnings
import numpy as np
import pandas as pd

from collections import defaultdict
from typing import Optional, List, Tuple, Dict, Any
from pandas.core.common import SettingWithCopyWarning

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.rdstar_plots import decay_mode_id_dict, u4s_decay_mode_id_dict
from rdstar.offline_analysis.mc_ids import rdstar_sig_queries, gen_b_ids, decay_mode_to_sig_id

from rdstar.offline_analysis.selection.selection_info_containers import BCSVariableInfo

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "SelectionEvaluator",
]


class SelectionEvaluator:
    decay_mode_id_col = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str
    u_eid_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()  # type: str

    sig_id_col = "SIG_ID"  # type: str
    is_sig_col = "isSignalAcceptMissingNeutrino"  # type: str
    is_b_sig_col = "daughter__bo1__cm__spisSignalAcceptMissingNeutrino__bc"  # type: str

    gen_b1_id_col = "GEN_B1_ID"  # type: str
    gen_b2_id_col = "GEN_B2_ID"  # type: str

    is_c_bkg_col = "isContinuumEvent"  # type: str

    l_mother_pdg_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__boPDG__bc__bc__bc"  # type: str
    mdst_index_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"  # type: str
    mdst_index_tau_col = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__" "bc__bc__bc"
    )  # type: str

    def __init__(self, df: pd.DataFrame, data_type: str) -> None:
        self.data_type = data_type  # type: str

        self._evaluation_info = {}  # type: Dict[str, Any]
        self._applied_cuts = []  # type: List[str]

        self._evaluation_info.update({"0_initial_no_cuts": self._get_evaluation_info(df=df, bcs=None)})

    def update_after_cut(self, df: pd.DataFrame, cut: str) -> None:
        self._applied_cuts.append(cut)
        number_of_applied_cuts = len(self._applied_cuts)
        info_key = f"{number_of_applied_cuts}_after_cut_{number_of_applied_cuts}"
        self._evaluation_info.update({info_key: self._get_evaluation_info(df=df)})

    def finalize_after_best_candidate_selection(
        self, df: pd.DataFrame, bcs: Tuple[BCSVariableInfo, BCSVariableInfo]
    ) -> Dict[str, Any]:
        index = len(self._applied_cuts) + 1
        self._evaluation_info.update({f"{index}_after_bcs": self._get_evaluation_info(df=df, bcs=bcs)})
        return self._evaluation_info

    def _add_vs_comparisons(
        self, info_dict: Dict[str, Any], df: pd.DataFrame, info_name: str, col: str, cond_col: str, cond_val: int
    ) -> None:
        cond = df[cond_col] == cond_val
        assert f"{info_name}_{cond_val}" not in info_dict, f"{info_name}_{cond_val}"
        assert f"u_{info_name}_{cond_val}" not in info_dict, f"u_{info_name}_{cond_val}"

        vs_info = df.loc[cond, col].value_counts()
        info_dict.update({f"{info_name}_{cond_val}": {k: v for k, v in zip(vs_info.index, vs_info)}})
        u_vs_info = df.loc[cond].groupby(self.u_eid_col)[col].first().value_counts()
        info_dict.update({f"u_{info_name}_{cond_val}": {k: v for k, v in zip(u_vs_info.index, u_vs_info)}})

    def _add_cut_and_bcs_validation_info(self, info_dict: Dict[str, Any], df: pd.DataFrame) -> None:
        assert isinstance(df, pd.DataFrame)
        res = defaultdict(dict)  # type: Dict[str, Any]
        is_sig_strict = df[self.is_sig_col] == 1
        is_b_sig_strict = df[self.is_b_sig_col] == 1

        for decay_id in decay_mode_id_dict():
            is_rec = df[self.decay_mode_id_col] == decay_id
            is_sig = df[self.sig_id_col].isin(decay_mode_to_sig_id[decay_id])

            if len(decay_mode_to_sig_id[decay_id]) == 2:
                is_sig_wn = df[self.sig_id_col].isin(decay_mode_to_sig_id[decay_id][:1])
                is_norm = df[self.sig_id_col].isin(decay_mode_to_sig_id[decay_id][1:])
            else:
                is_sig_wn = is_sig
                is_norm = False

            sig = len(df.loc[is_rec & is_sig].index)
            res["cut_info_sig"].update({decay_id: sig})
            sig_wn = len(df.loc[is_rec & is_sig_wn].index)
            res["cut_info_wn_sig"].update({decay_id: sig_wn})
            norm = len(df.loc[is_rec & is_norm].index)
            res["cut_info_wn_norm"].update({decay_id: norm})
            sig_u = len(df.loc[is_rec & is_sig].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_sig"].update({decay_id: sig_u})
            sig_wn_u = len(df.loc[is_rec & is_sig_wn].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_sig"].update({decay_id: sig_wn_u})
            norm_u = len(df.loc[is_rec & is_norm].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_norm"].update({decay_id: norm_u})

            strict_sig = len(df.loc[is_rec & is_sig_strict].index)
            res["cut_info_strict_sig"].update({decay_id: strict_sig})
            strict_b_sig = len(df.loc[is_rec & is_b_sig_strict].index)
            res["cut_info_strict_b_sig"].update({decay_id: strict_b_sig})
            strict_sig_u = len(df.loc[is_rec & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_strict_sig"].update({decay_id: strict_sig_u})
            strict_b_sig_u = len(df.loc[is_rec & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_strict_b_sig"].update({decay_id: strict_b_sig_u})

            v_strict_sig = len(df.loc[is_rec & is_sig & is_sig_strict].index)
            res["cut_info_v_strict_sig"].update({decay_id: v_strict_sig})
            v_strict_b_sig = len(df.loc[is_rec & is_sig & is_b_sig_strict].index)
            res["cut_info_v_strict_b_sig"].update({decay_id: v_strict_b_sig})
            v_strict_sig_wn = len(df.loc[is_rec & is_sig_wn & is_sig_strict].index)
            res["cut_info_wn_v_strict_sig"].update({decay_id: v_strict_sig_wn})
            v_strict_b_sig_wn = len(df.loc[is_rec & is_sig_wn & is_b_sig_strict].index)
            res["cut_info_wn_v_strict_b_sig"].update({decay_id: v_strict_b_sig_wn})
            v_strict_norm = len(df.loc[is_rec & is_norm & is_sig_strict].index)
            res["cut_info_wn_v_strict_norm"].update({decay_id: v_strict_norm})
            v_strict_b_norm = len(df.loc[is_rec & is_norm & is_b_sig_strict].index)
            res["cut_info_wn_v_strict_b_norm"].update({decay_id: v_strict_b_norm})
            v_strict_sig_u = len(df.loc[is_rec & is_sig & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_v_strict_sig"].update({decay_id: v_strict_sig_u})
            v_stri_b_sig_u = len(df.loc[is_rec & is_sig & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_v_strict_b_sig"].update({decay_id: v_stri_b_sig_u})
            v_strict_sig_wn_u = len(
                df.loc[is_rec & is_sig_wn & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index
            )
            res["bcs_info_wn_v_strict_sig"].update({decay_id: v_strict_sig_wn_u})
            v_strict_b_sig_wn_u = len(
                df.loc[is_rec & is_sig_wn & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index
            )
            res["bcs_info_wn_v_strict_b_sig"].update({decay_id: v_strict_b_sig_wn_u})
            v_strict_norm_u = len(df.loc[is_rec & is_norm & is_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_v_strict_norm"].update({decay_id: v_strict_norm_u})
            v_st_b_norm_u = len(df.loc[is_rec & is_norm & is_b_sig_strict].drop_duplicates(subset=self.u_eid_col).index)
            res["bcs_info_wn_v_strict_b_norm"].update({decay_id: v_st_b_norm_u})

            bkg = len(df.loc[is_rec & ~is_sig].index)
            res["bcs_info_bkg"].update({decay_id: bkg})
            bkg_wn = len(df.loc[is_rec & ~is_sig_wn & ~is_norm].index)
            res["bcs_info_wn_bkg"].update({decay_id: bkg_wn})

            strict_bkg = len(df.loc[is_rec & ~is_sig_strict].index)
            res["bcs_info_strict_bkg"].update({decay_id: strict_bkg})
            strict_b_bkg = len(df.loc[is_rec & ~is_b_sig_strict].index)
            res["bcs_info_strict_b_bkg"].update({decay_id: strict_b_bkg})

            v_strict_bkg = len(df.loc[is_rec & ~(is_sig & is_sig_strict)].index)
            res["bcs_info_v_strict_bkg"].update({decay_id: v_strict_bkg})
            v_strict_b_bkg = len(df.loc[is_rec & ~(is_sig & is_b_sig_strict)].index)
            res["bcs_info_v_strict_b_bkg"].update({decay_id: v_strict_b_bkg})
            v_strict_bkg_wn = len(df.loc[is_rec & ~(is_sig_wn & is_sig_strict) & ~(is_norm & is_sig_strict)].index)
            res["bcs_info_wn_v_strict_bkg"].update({decay_id: v_strict_bkg_wn})
            v_stric_b_bkg_wn = len(df.loc[is_rec & ~(is_sig_wn & is_b_sig_strict) & ~(is_norm & is_b_sig_strict)].index)
            res["bcs_info_wn_v_strict_b_bkg"].update({decay_id: v_stric_b_bkg_wn})

        assert all(k not in info_dict for k in res)
        info_dict.update(res)

    def _add_event_multiplicity_info(self, info_dict: Dict[str, Any], df: pd.DataFrame) -> None:
        em_name = "event_multiplicities"
        em_per_dt_name = "event_multiplicities_per_data_type"
        assert em_name not in info_dict.keys(), em_name
        assert em_per_dt_name not in info_dict.keys(), em_per_dt_name

        # Event multiplicity; get histogram by applying value_counts twice:
        multiplicity_hist = df[self.u_eid_col].value_counts().value_counts()
        info_dict.update({em_name: {k: v for k, v in zip(multiplicity_hist.index, multiplicity_hist)}})
        info_dict.update({em_per_dt_name: {str(self.data_type): info_dict[em_name]}})

        em_per_decay_id_name = "event_multiplicities_per_decay_id"
        assert em_per_decay_id_name not in info_dict.keys(), em_per_decay_id_name
        em_per_decay_id_is_sig_name = "event_multiplicities_is_sig_per_decay_id"
        assert em_per_decay_id_is_sig_name not in info_dict.keys(), em_per_decay_id_is_sig_name
        em_per_decay_id_is_sig_strict_name = "event_multiplicities_is_sig_strict_per_decay_id"
        assert em_per_decay_id_is_sig_strict_name not in info_dict.keys(), em_per_decay_id_is_sig_strict_name
        em_per_decay_id_is_b_sig_strict_name = "event_multiplicities_is_sig_strict_b_per_decay_id"
        assert em_per_decay_id_is_b_sig_strict_name not in info_dict.keys(), em_per_decay_id_is_b_sig_strict_name
        em_per_decay_id_dict = {}
        em_p_did_is_sig_dict = {}
        em_p_did_is_sig_strict_dict = {}
        em_p_did_is_b_sig_stri_dict = {}
        for decay_id in decay_mode_id_dict():
            is_decay = df[self.decay_mode_id_col] == decay_id  # type: pd.Series
            d_id_em_hist = df.loc[is_decay, self.u_eid_col].value_counts().value_counts()
            em_per_decay_id_dict.update({decay_id: {k: v for k, v in zip(d_id_em_hist.index, d_id_em_hist)}})

            if self.data_type not in rdstar_samples.recorded_data_samples:
                is_sig = df[self.sig_id_col].isin(decay_mode_to_sig_id[decay_id])  # type: pd.Series
                is_sig_strict = df[self.is_sig_col] == 1  # type: pd.Series
                is_b_sig_strict = df[self.is_b_sig_col] == 1  # type: pd.Series

                sig_em_h = df.loc[is_decay & is_sig, self.u_eid_col].value_counts().value_counts()
                em_p_did_is_sig_dict.update({decay_id: {k: v for k, v in zip(sig_em_h.index, sig_em_h)}})
                strict_em_h = df.loc[is_decay & is_sig & is_sig_strict, self.u_eid_col].value_counts().value_counts()
                em_p_did_is_sig_strict_dict.update({decay_id: {k: v for k, v in zip(strict_em_h.index, strict_em_h)}})
                stric_b_em_h = df.loc[is_decay & is_sig & is_b_sig_strict, self.u_eid_col].value_counts().value_counts()
                em_p_did_is_b_sig_stri_dict.update({decay_id: {k: v for k, v in zip(stric_b_em_h.index, stric_b_em_h)}})

        info_dict.update({em_per_decay_id_name: em_per_decay_id_dict})
        info_dict.update({em_per_decay_id_is_sig_name: em_p_did_is_sig_dict})
        info_dict.update({em_per_decay_id_is_sig_strict_name: em_p_did_is_sig_strict_dict})
        info_dict.update({em_per_decay_id_is_b_sig_strict_name: em_p_did_is_b_sig_stri_dict})

        if self.data_type not in rdstar_samples.recorded_data_samples:
            em_per_sig_id_name = "event_multiplicities_per_sig_id"
            assert em_per_sig_id_name not in info_dict.keys(), em_per_sig_id_name
            em_per_sig_id_dict = {}
            for sig_id_info in rdstar_sig_queries.values():
                sig_id = sig_id_info.id  # type: int
                is_sig = df[self.sig_id_col] == sig_id
                sig_id_em_hist = df.loc[is_sig, self.u_eid_col].value_counts().value_counts()
                em_per_sig_id_dict.update({sig_id: {k: v for k, v in zip(sig_id_em_hist.index, sig_id_em_hist)}})
            info_dict.update({em_per_sig_id_name: em_per_sig_id_dict})

    def _get_evaluation_info(
        self,
        df: pd.DataFrame,
        bcs: Optional[Tuple[BCSVariableInfo, BCSVariableInfo]] = None,
    ) -> Dict[str, Any]:
        info = {
            "bcs": copy.copy(bcs),
            "cuts": copy.copy(self._applied_cuts) if len(self._applied_cuts) > 0 else None,
        }  # type: Dict[str, Any]

        tag_decay_mode_id = "daughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc"
        u4s_decay_mode_id = "extraInfo__bodecayModeID__bc"

        # Event multiplicity
        self._add_event_multiplicity_info(info_dict=info, df=df)

        info.update({"n_events": len(df[self.u_eid_col].unique())})
        info.update({"n_candidates": len(df.index)})
        # Number of candidates per data type
        info.update({"data_type_counts": {str(self.data_type): len(df.index)}})
        # Number of events per data type
        info.update({"data_type_counts_unique": {str(self.data_type): len(df[self.u_eid_col].unique())}})

        # Reconstruction modes
        decay_modes = df[self.decay_mode_id_col].value_counts()
        info.update({"events_per_decay_mode_id": {k: v for k, v in zip(decay_modes.index, decay_modes)}})
        info.update({"decay_id_vs_data_type": {str(self.data_type): info["events_per_decay_mode_id"]}})
        u_decay_modes = df.groupby(self.u_eid_col)[self.decay_mode_id_col].first().value_counts()
        info.update({"unique_events_per_decay_mode_id": {k: v for k, v in zip(u_decay_modes.index, u_decay_modes)}})
        info.update({"decay_id_vs_data_type_unique": {str(self.data_type): info["unique_events_per_decay_mode_id"]}})
        tag_modes = df[tag_decay_mode_id].value_counts()
        info.update({"events_per_tag_mode_id": {k: v for k, v in zip(tag_modes.index, tag_modes)}})
        info.update({"tag_id_vs_data_type": {str(self.data_type): info["events_per_tag_mode_id"]}})
        u_tag_modes = df.groupby(self.u_eid_col)[tag_decay_mode_id].first().value_counts()
        info.update({"unique_events_per_tag_mode_id": {k: v for k, v in zip(u_tag_modes.index, u_tag_modes)}})
        info.update({"tag_id_vs_data_type_unique": {str(self.data_type): info["unique_events_per_tag_mode_id"]}})
        u4s_modes = df[u4s_decay_mode_id].value_counts()
        info.update({"events_per_u4s_mode_id": {k: v for k, v in zip(u4s_modes.index, u4s_modes)}})
        info.update({"u4s_id_vs_data_type": {str(self.data_type): info["events_per_u4s_mode_id"]}})
        u_u4s_modes = df.groupby(self.u_eid_col)[u4s_decay_mode_id].first().value_counts()
        info.update({"unique_events_per_u4s_mode_id": {k: v for k, v in zip(u_u4s_modes.index, u_u4s_modes)}})
        info.update({"u4s_id_vs_data_type_unique": {str(self.data_type): info["unique_events_per_u4s_mode_id"]}})

        for decay_id in decay_mode_id_dict():
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="tag_id_vs_decay_id",
                col=tag_decay_mode_id,
                cond_col=self.decay_mode_id_col,
                cond_val=decay_id,
            )

            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="u4s_id_vs_decay_id",
                col=u4s_decay_mode_id,
                cond_col=self.decay_mode_id_col,
                cond_val=decay_id,
            )

        for u4s_id in u4s_decay_mode_id_dict():
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="tag_id_vs_u4s_id",
                col=tag_decay_mode_id,
                cond_col=u4s_decay_mode_id,
                cond_val=u4s_id,
            )

        if self.data_type in rdstar_samples.recorded_data_samples:
            return info

        # MC-Dependent information from here on!

        self._add_cut_and_bcs_validation_info(info_dict=info, df=df)

        is_signal_strict = df.isSignalAcceptMissingNeutrino == 1
        info.update({"n_signal_strict": len(df.loc[is_signal_strict].index)})
        info.update({"n_signal_unique_strict": len(df.loc[is_signal_strict, self.u_eid_col].unique())})

        decay_modes = df.loc[is_signal_strict, self.decay_mode_id_col].value_counts()
        info.update({"is_signal_events_per_decay_mode_id": {k: v for k, v in zip(decay_modes.index, decay_modes)}})
        info.update(
            {"is_signal_decay_id_vs_data_type": {str(self.data_type): info["is_signal_events_per_decay_mode_id"]}}
        )
        u_decay_modes = df.loc[is_signal_strict].groupby(self.u_eid_col)[self.decay_mode_id_col].first().value_counts()
        info.update(
            {"is_signal_unique_events_per_decay_mode_id": {k: v for k, v in zip(u_decay_modes.index, u_decay_modes)}}
        )
        info.update(
            {
                "is_signal_decay_id_vs_data_type_unique": {
                    str(self.data_type): info["is_signal_unique_events_per_decay_mode_id"]
                }
            }
        )
        tag_modes = df.loc[is_signal_strict, tag_decay_mode_id].value_counts()
        info.update({"is_signal_events_per_tag_mode_id": {k: v for k, v in zip(tag_modes.index, tag_modes)}})
        info.update({"is_signal_tag_id_vs_data_type": {str(self.data_type): info["is_signal_events_per_tag_mode_id"]}})
        u_tag_modes = df.loc[is_signal_strict].groupby(self.u_eid_col)[tag_decay_mode_id].first().value_counts()
        info.update({"is_signal_unique_events_per_tag_mode_id": {k: v for k, v in zip(u_tag_modes.index, u_tag_modes)}})
        info.update(
            {
                "is_signal_tag_id_vs_data_type_unique": {
                    str(self.data_type): info["is_signal_unique_events_per_tag_mode_id"]
                }
            }
        )
        u4s_modes = df.loc[is_signal_strict, u4s_decay_mode_id].value_counts()
        info.update({"is_signal_events_per_u4s_mode_id": {k: v for k, v in zip(u4s_modes.index, u4s_modes)}})
        info.update({"is_signal_u4s_id_vs_data_type": {str(self.data_type): info["is_signal_events_per_u4s_mode_id"]}})
        u_u4s_modes = df.loc[is_signal_strict].groupby(self.u_eid_col)[u4s_decay_mode_id].first().value_counts()
        info.update({"is_signal_unique_events_per_u4s_mode_id": {k: v for k, v in zip(u_u4s_modes.index, u_u4s_modes)}})
        info.update(
            {
                "is_signal_u4s_id_vs_data_type_unique": {
                    str(self.data_type): info["is_signal_unique_events_per_u4s_mode_id"]
                }
            }
        )

        # Number of candidates per sig_id
        sig_id_counts = df[self.sig_id_col].value_counts()
        info.update({"sig_id_counts": {k: v for k, v in zip(sig_id_counts.index, sig_id_counts)}})
        u_sig_id_counts = df.groupby(self.u_eid_col)[self.sig_id_col].first().value_counts()
        info.update({"sig_id_counts_unique": {k: v for k, v in zip(u_sig_id_counts.index, u_sig_id_counts)}})

        is_continuum = df[self.is_c_bkg_col] == 1

        tmp_gen_id_col = "tmp_gen_id"
        assert tmp_gen_id_col not in df.columns

        conditions = [
            (df[self.mdst_index_col] == 1),
            (df[self.mdst_index_col] == 2),
            ((df[self.mdst_index_tau_col] == 1) & (abs(df[self.l_mother_pdg_col]) == 15)),
            ((df[self.mdst_index_tau_col] == 2) & (abs(df[self.l_mother_pdg_col]) == 15)),
            (is_continuum & (df[self.gen_b1_id_col] == df[self.gen_b2_id_col])),
        ]

        continuum_check_series = df.loc[is_continuum, self.gen_b1_id_col] == df.loc[is_continuum, self.gen_b2_id_col]
        assert isinstance(continuum_check_series, pd.Series)
        assert continuum_check_series.all()
        assert all(pd.DataFrame({c: d for c, d in enumerate(conditions)}).sum(axis=1) <= 1), "Conditions are not unique"

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, tmp_gen_id_col] = np.select(
                condlist=conditions,
                choicelist=[
                    df[self.gen_b1_id_col],
                    df[self.gen_b2_id_col],
                    df[self.gen_b1_id_col],
                    df[self.gen_b2_id_col],
                    df[self.gen_b1_id_col],
                ],
                default=-999,
            )
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

        incon_gen_1_id = df.loc[df[tmp_gen_id_col] == -999, self.gen_b1_id_col].value_counts()
        incon_gen_2_id = df.loc[df[tmp_gen_id_col] == -999, self.gen_b2_id_col].value_counts()
        info.update(
            {
                "gen_b1_id_for_inconclusive": {k: v for k, v in zip(incon_gen_1_id.index, incon_gen_1_id)},
                "gen_b2_id_for_inconclusive": {k: v for k, v in zip(incon_gen_2_id.index, incon_gen_2_id)},
            }
        )

        gen_id_counts = df[tmp_gen_id_col].value_counts()
        info.update({"gen_id_counts": {k: v for k, v in zip(gen_id_counts.index, gen_id_counts)}})
        info.update({"gen_id_per_data_type": {str(self.data_type): info["gen_id_counts"]}})
        u_gen_id_counts = df.groupby(self.u_eid_col)[tmp_gen_id_col].first().value_counts()
        info.update({"gen_id_counts_unique": {k: v for k, v in zip(u_gen_id_counts.index, u_gen_id_counts)}})
        info.update({"gen_id_per_data_type_unique": {str(self.data_type): info["gen_id_counts_unique"]}})
        s_gen_id_counts = df.loc[is_signal_strict, tmp_gen_id_col].value_counts()
        info.update({"gen_id_counts_strict": {k: v for k, v in zip(s_gen_id_counts.index, s_gen_id_counts)}})
        us_gen_id_counts = df.loc[is_signal_strict].groupby(self.u_eid_col)[tmp_gen_id_col].first().value_counts()
        info.update({"gen_id_counts_unique_strict": {k: v for k, v in zip(us_gen_id_counts.index, us_gen_id_counts)}})

        for sig_id, sig_id_info in rdstar_sig_queries.items():
            is_sig = df[self.sig_id_col] == sig_id_info.id  # type: pd.Series

            info.update({f"n_signal_{sig_id}": len(df.loc[is_sig].index)})
            info.update({f"n_signal_unique_{sig_id}": len(df.loc[is_sig, self.u_eid_col].unique())})

            info.update({f"data_type_per_sig_id_{sig_id_info.id}": {str(self.data_type): len(df.loc[is_sig].index)}})
            num_unique_per_sig_id = len(df.loc[is_sig, self.u_eid_col].unique())
            info.update({f"u_data_type_per_sig_id_{sig_id_info.id}": {str(self.data_type): num_unique_per_sig_id}})

            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="gen_id_per_sig_id",
                col=tmp_gen_id_col,
                cond_col=self.sig_id_col,
                cond_val=sig_id_info.id,
            )

            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="decay_id_vs_sig_id",
                col=self.decay_mode_id_col,
                cond_col=self.sig_id_col,
                cond_val=sig_id_info.id,
            )
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="tag_id_vs_sig_id",
                col=tag_decay_mode_id,
                cond_col=self.sig_id_col,
                cond_val=sig_id_info.id,
            )
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="u4s_id_vs_sig_id",
                col=u4s_decay_mode_id,
                cond_col=self.sig_id_col,
                cond_val=sig_id_info.id,
            )

        for gen_id, gen_id_info in gen_b_ids.items():
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="decay_id_vs_gen_id",
                col=self.decay_mode_id_col,
                cond_col=tmp_gen_id_col,
                cond_val=gen_id_info.id,
            )
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="tag_id_vs_gen_id",
                col=tag_decay_mode_id,
                cond_col=tmp_gen_id_col,
                cond_val=gen_id_info.id,
            )
            self._add_vs_comparisons(
                info_dict=info,
                df=df,
                info_name="u4s_id_vs_gen_id",
                col=u4s_decay_mode_id,
                cond_col=tmp_gen_id_col,
                cond_val=gen_id_info.id,
            )

        del df[tmp_gen_id_col]

        return info
