"""
Definition of Classes implementing functionalities related to Event Selections.

Felix Metzner 2019
"""

from abc import ABC, abstractmethod
from typing import Optional, List, Tuple, NamedTuple, Dict, Generator

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.mva_selection.mva_selector_setups import MVASelectorSetupBase, MVASelectorSetups

__all__ = [
    "BCSVariableInfo",
    "MVAInfo",
    "SelectionSpecifics",
    "SubSelectionInfo",
    "SelectionInfo",
    "SelectionSetupBase",
]


class BCSVariableInfo(NamedTuple):
    """
    Definition of a information set used for the BCS:
    Contains:
      - bcs_variable: variable name of variable to be used for bcs
                      or "random" if selection shall be done random
                      or None e.g. if now secondary bcs variable shall be used or no BCS shall be applied at all.
      - absolute_value: Boolean indicating if absolute value of variable should be taken for the bcs
      - max: Boolean whether to select the candidate based on the maximum (True) or minimum (False)
    """

    bcs_variable: Optional[str]
    absolute_value: bool = False
    max: bool = True


class MVAInfo(NamedTuple):
    mva_setup_name: str
    mva_post_cut: float
    mva_setup: MVASelectorSetupBase


class SelectionSpecifics(NamedTuple):
    name: str
    cuts: Tuple[CutInfo, ...]
    primary_bcs_var: BCSVariableInfo
    secondary_bcs_var: BCSVariableInfo
    check_cuts: Tuple[CutInfo, ...]
    check_bcs: bool
    mva_setup_name: Optional[str] = None
    post_mva_cut_info: Optional[CutInfo] = None


class SubSelectionInfo(NamedTuple):
    selection_tag: str
    cut_list: Tuple[CutInfo, ...]
    check_cuts: Tuple[CutInfo, ...]
    check_bcs: bool


class SelectionInfo(NamedTuple):
    selection_tag: str
    selection: SelectionSpecifics
    target_path: PathType
    output_key: str
    info_target_path: Optional[PathType] = None
    info_output_key: Optional[str] = None


class SelectionSetupBase(ABC):
    selection_name = ""  # type: str
    _mva_setup_name = None  # type: Optional[str]
    _post_mva_cut = None  # type: Optional[float]
    all_cut_infos = []  # type: List[CutInfo]

    @staticmethod
    @abstractmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_selection(cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        raise NotImplementedError()

    @classmethod
    def get_cut_label_dict(cls) -> Dict[str, str]:
        cut_label_dict = {cut.cut_str: cut.cut_label for cut in cls.all_cut_infos}
        if cls.has_mva_setup_definition():
            mva_cut_info = cls.get_post_mva_cut_info()
            assert mva_cut_info is not None
            assert mva_cut_info.cut_str not in cut_label_dict, (mva_cut_info.cut_str, cut_label_dict)
            cut_label_dict.update({mva_cut_info.cut_str: mva_cut_info.cut_label})
        return cut_label_dict

    @classmethod
    def get_cut(cls, cut_name: str) -> str:
        cuts = [cut.cut_str for cut in cls.all_cut_infos if cut.cut_name == cut_name]
        if len(cuts) == 1:
            return cuts[0]
        elif len(cuts) == 0:
            raise ValueError(f"No cut named {cut_name} available in the Selection class {cls.__name__}!")
        else:
            raise ValueError(f"Name {cut_name} not unique in the Selection class {cls.__name__}!")

    @classmethod
    def has_mva_setup_definition(cls) -> bool:
        return cls._mva_setup_name is not None

    @classmethod
    def get_mva_setup_name(cls) -> str:
        assert cls.has_mva_setup_definition()
        assert cls._mva_setup_name is not None
        return cls._mva_setup_name

    @classmethod
    def get_mva_setup(cls) -> MVASelectorSetupBase:
        assert cls.has_mva_setup_definition()
        return MVASelectorSetups.get_selection_mva_setup(setup_name=cls.get_mva_setup_name())

    @classmethod
    def get_post_mva_cut(cls) -> float:
        assert cls.has_mva_setup_definition()

        assert cls._post_mva_cut is not None
        assert isinstance(cls._post_mva_cut, float), type(cls._post_mva_cut)
        assert 0 <= cls._post_mva_cut <= 1.0, cls._post_mva_cut
        return cls._post_mva_cut

    @classmethod
    def get_post_mva_cut_info(cls) -> Optional[CutInfo]:
        if not cls.has_mva_setup_definition():
            return None

        mva_setup = cls.get_mva_setup()
        assert cls._post_mva_cut is not None

        mva_output_column_name = mva_setup.get_selection_mva_output_column_name()
        return CutInfo(
            cut_str=f"{mva_output_column_name} > {cls._post_mva_cut}",
            cut_label=LatexObsNames.mva_sel_output[:-1] + r" > $" + f"{cls._post_mva_cut}",
            cut_name=f"{mva_output_column_name}_cut",
            cut_var=mva_output_column_name,
            cut_var_latex_label=LatexObsNames.mva_sel_output,
            cut_type=">",
            cut_limits=cls._post_mva_cut,
            latex_unit=None,
        )

    @classmethod
    def get_mva_info(cls) -> Optional[MVAInfo]:
        if not cls.has_mva_setup_definition():
            return None

        return MVAInfo(
            mva_setup_name=cls.get_mva_setup_name(),
            mva_post_cut=cls.get_post_mva_cut(),
            mva_setup=cls.get_mva_setup(),
        )

    @classmethod
    def get_selection_tags(cls, cs_tag: str) -> Tuple[str, ...]:
        return tuple([info.selection_tag for info in cls.get_selection(cs_tag=cs_tag)])

    @classmethod
    def get_main_selection_tag(cls, cs_tag: str) -> str:
        # Getting first element returned by the generator get_selection.
        # This first item is supposed to be the main selection definition of the analysis.
        return next(cls.get_selection(cs_tag=cs_tag)).selection_tag

    @classmethod
    def get_selection_specifics(cls, cs_tag: str) -> Dict[str, SelectionSpecifics]:
        return {
            infos.selection_tag: SelectionSpecifics(
                name=infos.selection_tag,
                cuts=infos.cut_list,
                primary_bcs_var=cls.get_bcs_info()[0],
                secondary_bcs_var=cls.get_bcs_info()[1],
                check_cuts=infos.check_cuts,
                check_bcs=infos.check_bcs,
                mva_setup_name=cls.get_mva_setup_name() if cls.has_mva_setup_definition() else None,
                post_mva_cut_info=cls.get_post_mva_cut_info(),
            )
            for infos in cls.get_selection(cs_tag=cs_tag)
        }

    @classmethod
    def print_selection_info(cls, cs_tag: str) -> None:
        print(f"Selection name: {cls.selection_name}\n")

        print("\tAll defined selections:\n\t\t- " + "\n\t\t- ".join(cls.get_selection_tags(cs_tag=cs_tag)) + "\n")

        print(f"\tMain selection tag: {cls.get_main_selection_tag(cs_tag=cs_tag)}\n")

        if cls.has_mva_setup_definition():
            mva_cut_info = cls.get_post_mva_cut_info()
            assert mva_cut_info is not None
            print(
                f"\tMVA Selection Info:\n\t\t"
                f"MVA Selection Name: {cls.get_mva_setup_name()}\n\t\t"
                f"MVA Selection Cut: {mva_cut_info.cut_str}\n"
            )

        for key, value in cls.get_selection_specifics(cs_tag=cs_tag).items():
            cut_string_list = [f"C {c.cut_str}" if c in value.check_cuts else f"  {c.cut_str}" for c in value.cuts]
            print(f"\t{key}:\n\t\t" + "\n\t\t".join(cut_string_list) + "\n")


if __name__ == "__main__":
    my_cut_info = CutInfo(
        cut_str="3. <= useCMSFrame(pt) < 4.",
        cut_label="test",
        cut_name="test",
        cut_type=("<=", "<"),
        cut_var="useCMSFrame(pt)",
        cut_var_latex_label=LatexObsNames.momentum_pt_cms,
        latex_unit=LatexUnits.gev,
        cut_abs=False,
        cut_limits=(3, 04.0),
        cut_online=True,
    )
    print(my_cut_info)
