"""
Definition of Signal and Sideband Selection Setups

This file covers the definitions of
    - Selection cuts
    - Best candidate selection
for
    - the main selection
    - the q2 sideband
    - the mbc sideband
    - the selection for the form factor study
    - the dedicated D** selection

Felix Metzner 2019
"""

from typing import Optional, List, Tuple, Dict, Generator, Type

from rdstar.mva_selection import mva_selector_setups
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.selection.selection_info_containers import (
    SelectionSetupBase,
    BCSVariableInfo,
    MVAInfo,
    SubSelectionInfo,
    SelectionSpecifics,
)


__all__ = [
    "SelectionSetups",
]


# Definition of my first version of the selection:
class FirstSelection(SelectionSetupBase):
    selection_name = "first_selection"  # type: str

    # Default Cuts
    fei_prob_cut = CutInfo(
        cut_str="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc > 0.001",
        cut_label=LatexObsNames.fei_prob[:-1] + r" > 0.001$",
        cut_name="fei_prob_cut",
        cut_var="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
        cut_var_latex_label=LatexObsNames.fei_prob,
        cut_type=">",
        cut_limits=0.001,
        latex_unit=LatexUnits.no_unit,
    )
    q_square_cut = CutInfo(
        cut_str="qSquaredReco > 4.",
        cut_label=LatexObsNames.fitting_q2[:-1] + r" > 4.0\," + LatexUnits.gev_sq[1:],
        cut_name="q_square_cut",
        cut_var="qSquaredReco",
        cut_var_latex_label=LatexObsNames.fitting_q2,
        cut_type=">",
        cut_limits=4.0,
        latex_unit=LatexUnits.gev_sq,
    )
    e_extra_cut = CutInfo(
        cut_str="ROE_eextra__boCleanROE__bc < 3.",
        cut_label=LatexObsNames.roe_e_ecl_extra[:-1] + r" < 3.0\," + LatexUnits.gev[1:],
        cut_name="e_extra_cut",
        cut_var="ROE_eextra__boCleanROE__bc",
        cut_var_latex_label=LatexObsNames.roe_e_ecl_extra,
        cut_type="<",
        cut_limits=3.0,
        latex_unit=LatexUnits.gev,
    )
    d_mass_cut = CutInfo(
        cut_str="abs(daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc) < 0.035",
        cut_label=LatexObsNames.d_w_wo_st_abs_delta_m[:-1] + r" < 35\," + LatexUnits.mev[1:],
        cut_name="d_mass_cut",
        cut_var="daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc",
        cut_var_latex_label=LatexObsNames.d_w_wo_st_abs_delta_m,
        cut_type="<",
        cut_limits=0.035,
        cut_abs=True,
        latex_unit=LatexUnits.gev,
    )
    tag_mbc_cut = CutInfo(
        cut_str="daughter__bo0__cm__spMbc__bc > 5.27",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" > 5.27\," + LatexUnits.gev[1:],
        cut_name="tag_mbc_cut",
        cut_var="daughter__bo0__cm__spMbc__bc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type=">",
        cut_limits=5.27,
        latex_unit=LatexUnits.gev,
    )
    tag_delta_e_cut = CutInfo(
        cut_str="-0.05 < daughter__bo0__cm__spdeltaE__bc < 0.1",
        cut_label=LatexObsNames.b_tag_delta_e[:-1] + r"\;\in\;(-0.05, 0.1)\," + LatexUnits.gev[1:],
        cut_name="tag_delta_e_cut",
        cut_var="daughter__bo0__cm__spdeltaE__bc",
        cut_var_latex_label=LatexObsNames.b_tag_delta_e,
        cut_type=("<", "<"),
        cut_limits=(-0.05, 0.1),
        latex_unit=LatexUnits.gev,
    )
    cs_mva_cut = CutInfo(
        cut_str="cs_mva_output > 0.5",
        cut_label=LatexObsNames.mva_cs_output[:-1] + r" > 0.5$",
        cut_name="cs_mva_cut",
        cut_var="cs_mva_output",
        cut_var_latex_label=LatexObsNames.mva_cs_output,
        cut_type=">",
        cut_limits=0.5,
        latex_unit=LatexUnits.no_unit,
    )
    lepton_p_cut = CutInfo(
        cut_str="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc < 2.",
        cut_label=LatexObsNames.lepton_p[:-1] + r" < 2\," + LatexUnits.gev[1:],
        cut_name="lepton_p_cut",
        cut_var="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
        cut_var_latex_label=LatexObsNames.lepton_p,
        cut_type="<",
        cut_limits=2.0,
        latex_unit=LatexUnits.gev,
    )
    b_vertex_fit_cut = CutInfo(
        cut_str="daughter__bo1__cm__spchiProb__bc > 0.0",
        cut_label=LatexObsNames.b_sig_vertex_prob[:-1] + r" > 0.0$",
        cut_name="b_vertex_fit_cut",
        cut_var="daughter__bo1__cm__spchiProb__bc",
        cut_var_latex_label=LatexObsNames.b_sig_vertex_prob,
        cut_type=">",
        cut_limits=0.0,
        latex_unit=LatexUnits.no_unit,
    )

    # Sideband Cuts
    q_square_anti_cut = CutInfo(
        cut_str="qSquaredReco <= 4.",
        cut_label=LatexObsNames.fitting_q2[:-1] + r" \leq 4.0\," + LatexUnits.gev_sq[1:],
        cut_name="q_square_anti_cut",
        cut_var="qSquaredReco",
        cut_var_latex_label=LatexObsNames.fitting_q2,
        cut_type="<=",
        cut_limits=4.0,
        latex_unit=LatexUnits.gev_sq,
    )
    tag_mbc_anti_cut = CutInfo(
        cut_str="daughter__bo0__cm__spMbc__bc <= 5.27",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" \leq 5.27\," + LatexUnits.gev[1:],
        cut_name="tag_mbc_anti_cut",
        cut_var="daughter__bo0__cm__spMbc__bc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type="<=",
        cut_limits=5.27,
        latex_unit=LatexUnits.gev,
    )

    all_cut_infos = [
        fei_prob_cut,
        q_square_cut,
        e_extra_cut,
        d_mass_cut,
        tag_mbc_cut,
        tag_delta_e_cut,
        cs_mva_cut,
        lepton_p_cut,
        b_vertex_fit_cut,
        q_square_anti_cut,
        tag_mbc_anti_cut,
    ]  # type: List[CutInfo]

    # Definition of Best Candidate Selection
    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        primary_bcs_var = BCSVariableInfo(
            bcs_variable="ROE_neextra__boCleanROE__bc",
            absolute_value=False,
            max=False,
        )
        secondary_bcs_var = BCSVariableInfo(
            bcs_variable="daughter__bo0__cm__spdeltaE__bc",
            absolute_value=True,
            max=False,
        )
        return primary_bcs_var, secondary_bcs_var

    # Selection Definitions
    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag = f"{cls.selection_name}_with_CS_{cs_tag}"  # type: str
        cut_list = (
            FirstSelection.fei_prob_cut,
            FirstSelection.q_square_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.tag_delta_e_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
        )  # type: Tuple[CutInfo, ...]
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains not well defined cuts!"

        yield SubSelectionInfo(selection_tag=selection_tag, cut_list=cut_list, check_cuts=(), check_bcs=False)

        sb_check_cut_list = (
            FirstSelection.e_extra_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
        )  # type: Tuple[CutInfo, ...]

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"  # type: str
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains not well defined cuts!"
        mbc_sb_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.tag_mbc_cut])
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            check_cuts=mbc_sb_check_cuts,
            check_bcs=True,
        )

        # q squared Sideband Selection
        q2_sb_tag = f"{selection_tag}_q2_sideband"  # type: str
        q2_sb_cut_list = [FirstSelection.q_square_anti_cut if c == FirstSelection.q_square_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in q2_sb_cut_list), f"{q2_sb_tag} contains not well defined cuts!"
        q2_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.q_square_cut])
        yield SubSelectionInfo(
            selection_tag=q2_sb_tag,
            cut_list=tuple(q2_sb_cut_list),
            check_cuts=q2_sbc_check_cuts,
            check_bcs=True,
        )


class MVASelection(SelectionSetupBase):
    _selection_name = "mva_selection"  # type: str
    _mva_setup_name = mva_selector_setups.SecondSelectionMVASetup.selection_mva_name  # type: Optional[str]
    selection_name = f"{_selection_name}_{_mva_setup_name}"  # type: str
    _post_mva_cut = 0.5  # type: Optional[float]

    all_cut_infos = [
        FirstSelection.fei_prob_cut,
        FirstSelection.q_square_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
        FirstSelection.q_square_anti_cut,
        FirstSelection.tag_mbc_anti_cut,
    ]  # type: List[CutInfo]

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag = f"{cls.selection_name}_with_CS_{cs_tag}"  # type: str
        cut_list = (
            FirstSelection.fei_prob_cut,
            FirstSelection.q_square_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
        )  # type: Tuple[CutInfo, ...]
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains not well defined cuts!"

        yield SubSelectionInfo(selection_tag=selection_tag, cut_list=cut_list, check_cuts=(), check_bcs=False)

        # q squared Sideband Selection
        q2_sb_tag = f"{selection_tag}_q2_sideband"  # type: str
        q2_sb_cut_list = [FirstSelection.q_square_anti_cut if c == FirstSelection.q_square_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in q2_sb_cut_list), f"{q2_sb_tag} contains not well defined cuts!"
        yield SubSelectionInfo(selection_tag=q2_sb_tag, cut_list=tuple(q2_sb_cut_list), check_cuts=(), check_bcs=False)

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"  # type: str
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains not well defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            check_cuts=(),
            check_bcs=False,
        )


class FullQ2Selection(SelectionSetupBase):
    selection_name = "full_q2_selection"  # type: str

    all_cut_infos = [
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.b_vertex_fit_cut,
        FirstSelection.tag_mbc_anti_cut,
    ]  # type: List[CutInfo]

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag = f"{cls.selection_name}_with_CS_{cs_tag}"  # type: str
        cut_list = (
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.b_vertex_fit_cut,
        )  # type: Tuple[CutInfo, ...]
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains not well defined cuts!"

        yield SubSelectionInfo(selection_tag=selection_tag, cut_list=cut_list, check_cuts=(), check_bcs=False)

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains not well defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            check_cuts=(),
            check_bcs=False,
        )


class DSSSelection(SelectionSetupBase):
    _selection_name = "dss_selection"  # type: str
    _mva_setup_name = mva_selector_setups.SecondSelectionDSSMVASetup.selection_mva_name  # type: Optional[str]
    selection_name = f"{_selection_name}_{_mva_setup_name}"  # type: str
    _post_mva_cut = 0.5  # type: Optional[float]

    all_cut_infos = [
        FirstSelection.fei_prob_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
        FirstSelection.tag_mbc_anti_cut,
    ]  # type: List[CutInfo]

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag = f"{cls.selection_name}_with_CS_{cs_tag}"  # type: str
        cut_list = (
            FirstSelection.fei_prob_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
        )  # type: Tuple[CutInfo, ...]
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains not well defined cuts!"

        yield SubSelectionInfo(selection_tag=selection_tag, cut_list=cut_list, check_cuts=(), check_bcs=False)

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"  # type: str
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains not well defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            check_cuts=(),
            check_bcs=False,
        )


class DSSSelectionFullPL(SelectionSetupBase):
    _selection_name = "dss_selection_full_pl"  # type: str
    _mva_setup_name = mva_selector_setups.SecondSelectionDSSMVASetup.selection_mva_name  # type: Optional[str]
    selection_name = f"{_selection_name}_{_mva_setup_name}"  # type: str
    _post_mva_cut = 0.5  # type: Optional[float]

    all_cut_infos = [
        FirstSelection.fei_prob_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.b_vertex_fit_cut,
        FirstSelection.tag_mbc_anti_cut,
    ]  # type: List[CutInfo]

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag = f"{cls.selection_name}_with_CS_{cs_tag}"  # type: str
        cut_list = (
            FirstSelection.fei_prob_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.b_vertex_fit_cut,
        )  # type: Tuple[CutInfo, ...]
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains not well defined cuts!"

        yield SubSelectionInfo(selection_tag=selection_tag, cut_list=cut_list, check_cuts=(), check_bcs=False)

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"  # type: str
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains not well defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            check_cuts=(),
            check_bcs=False,
        )


def check_setups_and_provide_correction_mapping(selections_dict, mapping_dict) -> Dict[str, str]:
    assert all(k in selections_dict for k in mapping_dict.keys()), [
        k for k in mapping_dict.keys() if k not in selections_dict
    ]
    assert all(k in selections_dict for k in mapping_dict.values()), [
        k for k in mapping_dict.values() if k not in selections_dict
    ]

    full_resolution_correction_mapping = {
        k: k if k not in mapping_dict else mapping_dict[k] for k in selections_dict.keys()
    }  # type: Dict[str, str]

    assert set(selections_dict.keys()) == set(full_resolution_correction_mapping.keys()), (
        selections_dict.keys(),
        full_resolution_correction_mapping.keys(),
    )

    return full_resolution_correction_mapping


# This interface class should be used to access all defined selections...
class SelectionSetups:
    available_selections = {
        FirstSelection.selection_name: FirstSelection,
        MVASelection.selection_name: MVASelection,
        # FullQ2Selection.selection_name: FullQ2Selection,
        # DSSSelection.selection_name: DSSSelection,
        # DSSSelectionFullPL.selection_name: DSSSelectionFullPL,
    }  # type: Dict[str, Type[SelectionSetupBase]]

    __alternative_res_corr_mapping = {
        # FullQ2Selection.selection_name: FirstSelection.selection_name,
        # DSSSelection.selection_name: FirstSelection.selection_name,
        # DSSSelectionFullPL.selection_name: FirstSelection.selection_name,
    }  # type: Dict[str, str]

    my_selections = (
        FirstSelection,
        MVASelection,
        DSSSelection,
        DSSSelectionFullPL,
    )  # type: Tuple[Type[SelectionSetupBase], ...]

    resolution_correction_mapping = check_setups_and_provide_correction_mapping(
        selections_dict=available_selections,
        mapping_dict=__alternative_res_corr_mapping,
    )  # type: Dict[str, str]

    @staticmethod
    def _get_selection_class(selection_name: str):
        if selection_name in SelectionSetups.available_selections:
            return SelectionSetups.available_selections[selection_name]
        else:
            raise ValueError(f"Selection with name {selection_name} is not available!")

    @staticmethod
    def get_available_selection_names() -> Tuple[str, ...]:
        return tuple([sel_name for sel_name in SelectionSetups.available_selections.keys()])

    @staticmethod
    def get_selection_infos(selection_name: str, cs_tag: str) -> Dict[str, SelectionSpecifics]:
        return SelectionSetups._get_selection_class(selection_name=selection_name).get_selection_specifics(cs_tag=cs_tag)

    @staticmethod
    def get_selection_cuts(
        selection_name: str,
        selection_tag: str,
        cs_tag: str,
    ) -> Tuple[Tuple[str, ...], Tuple[str, ...], Tuple[str, ...]]:
        specs = SelectionSetups.get_selection_infos(selection_name=selection_name, cs_tag=cs_tag)[selection_tag]

        cuts = tuple([c.cut_str for c in specs.cuts])
        check_cuts = tuple([c.cut_str for c in specs.check_cuts])
        base_cuts = tuple([c for c in cuts if c not in check_cuts])

        assert len(base_cuts) + len(check_cuts) == len(cuts), f"{len(base_cuts)} + {len(check_cuts)} == {len(cuts)}"
        return cuts, base_cuts, check_cuts

    @staticmethod
    def get_selection_bcs(
        selection_name: str,
        selection_tag: str,
        cs_tag: str,
    ) -> Tuple[BCSVariableInfo, BCSVariableInfo, bool]:
        specs = SelectionSetups.get_selection_infos(selection_name=selection_name, cs_tag=cs_tag)[selection_tag]
        return specs.primary_bcs_var, specs.secondary_bcs_var, specs.check_bcs

    @staticmethod
    def get_selection_mva_info(selection_name: str) -> Optional[MVAInfo]:
        return SelectionSetups._get_selection_class(selection_name=selection_name).get_mva_info()

    @staticmethod
    def get_selection_tags(selection_name: str, cs_tag: str) -> Tuple[str, ...]:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_selection_tags(cs_tag=cs_tag)

    @staticmethod
    def get_main_selection_tag(selection_name: str, cs_tag: str) -> str:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_main_selection_tag(cs_tag=cs_tag)

    @staticmethod
    def get_selection_cut_label_dict(selection_name: str) -> Dict[str, str]:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_cut_label_dict()


if __name__ == "__main__":
    for selection_setup in SelectionSetups.available_selections.values():
        selection_setup.print_selection_info(cs_tag="test")
