"""
Definition of Systematics
    via collection of weight variations...

Felix Metzner 2019
"""
import pandas as pd

from typing import NamedTuple, Optional, List

from templatefitter.binned_distributions.systematics import SystematicsInputType

from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection
from rdstar.offline_analysis.corrections.k_short_corrections import KShortEffCorrector as KSCorr
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector as TrkCorr
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection as PIDCorr

# from rdstar.offline_analysis.corrections.fei_tag_correction import InclusiveTagCorrection as InclTagCorr

from rdstar.mc_reweighting.BtoDlnu.BtoDlnu_weights import BtoDFFReweighting as BtoD_FF
from rdstar.mc_reweighting.BtoDstlnu.BtoDstlnu_weights import BtoDstarFFReweighting as BtoDs_FF
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import BtoDststFFReweighting as BtoDss_FF
from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import DststType as BtoDststFF_decay_types


__all__ = [
    "SysColsInfo",
    "full_systematics_name",
    "full_systematics_cols",
    "get_systematics",
]


class SysColsInfo(NamedTuple):
    var: str
    ratio: Optional[str] = None
    from_df: bool = False


# class SysColsInfo(NamedTuple):  # TODO: Use this in the future!
#     var: str
#     name: str = ""
#     ratio: Optional[str] = None
#     from_df: bool = False
#     latex_name: Optional[str] = None
#
#     @property
#     def name(self) -> str:
#         if self.name:
#             return self.name
#         else:
#             return self.var
#
#     @property
#     def latex_name(self) -> str:
#         if self.latex_name:
#             return self.latex_name
#         else:
#             return self.name


full_systematics_name = "full_sys"
_full_systematics_cols = [
    SysColsInfo(var=PIDCorr.pid_l_corr_var_prefix, ratio=PIDCorr.pid_l_corr_col_name, from_df=True),
    SysColsInfo(var=PIDCorr.pid_pi_corr_var_prefix, ratio=PIDCorr.pid_pi_corr_col_name, from_df=True),
    SysColsInfo(var=PIDCorr.pid_piFromK_corr_var_prefix, ratio=PIDCorr.pid_piFromK_corr_col_name, from_df=True),
    SysColsInfo(var=PIDCorr.pid_k_corr_var_prefix, ratio=PIDCorr.pid_k_corr_col_name, from_df=True),
    SysColsInfo(var=PIDCorr.pid_kFromPi_corr_var_prefix, ratio=PIDCorr.pid_kFromPi_corr_col_name, from_df=True),
    SysColsInfo(var=TrkCorr.slow_pi_corr_var_prefix, ratio=TrkCorr.slow_pi_corr_col_name, from_df=True),
    SysColsInfo(var=KSCorr.k_short_corr_var_prefix, ratio=KSCorr.k_short_corr_col_name, from_df=True),
    SysColsInfo(var=TrkCorr.tracking_systematic_name),
    SysColsInfo(var=TagCorrection.tagging_corr_uncert_name, ratio=TagCorrection.tagging_correction_name),
    # SysColsInfo(var=InclTagCorr.tagging_corr_uncert_prefix, ratio=InclTagCorr.tagging_correction_name, from_df=True),  # TODO: USE!
    # FF Systematics (Cov Matrix is calculated differently than Max did it!):
    SysColsInfo(var=BtoD_FF.base_var_weight_name, ratio=BtoD_FF.nom_weight_name, from_df=True),
    SysColsInfo(var=BtoDs_FF.base_var_weight_name, ratio=BtoDs_FF.nom_weight_name, from_df=True),
]

for dstst_decay_type in BtoDststFF_decay_types:
    _full_systematics_cols.append(
        SysColsInfo(
            var=BtoDss_FF.get_var_dstst_weight_name(dstst_type=dstst_decay_type.value),
            ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type=dstst_decay_type.value),
            from_df=True,
        )
    )

full_systematics_cols = _full_systematics_cols


def get_systematics(df: pd.DataFrame, sys_cols_infos: List[SysColsInfo]) -> SystematicsInputType:
    if not sys_cols_infos or sys_cols_infos is None:
        return None

    assert all(isinstance(info, SysColsInfo) for info in sys_cols_infos), [type(info) for info in sys_cols_infos]
    assert all(info.ratio in df.columns for info in sys_cols_infos if info.ratio is not None), [
        info.ratio for info in sys_cols_infos if info.ratio is not None and info.ratio not in df.columns
    ]
    assert all(info.var in df.columns for info in sys_cols_infos if not info.from_df), [
        info.var for info in sys_cols_infos if not info.from_df and info.var not in df.columns
    ]
    assert all(any(c.startswith(info.var) for c in df.columns) for info in sys_cols_infos if info.from_df), [
        info.var for info in sys_cols_infos if info.from_df and not any(c.startswith(info.var) for c in df.columns)
    ]

    infos_from_cols = [info.ratio for info in sys_cols_infos if info.from_df]
    infos_from_cols.extend([c for info in sys_cols_infos if info.from_df for c in df.columns if c.startswith(info.var)])
    assert all(c in df.columns for c in infos_from_cols)
    for col in infos_from_cols:
        assert not pd.isnull(df[col]).any(), (col, df.loc[pd.isnull(df[col]), "data_type"])

    return [
        (info.ratio, [c for c in df.columns if c.startswith(info.var)]) if info.from_df else (info.ratio, info.var)
        for info in sys_cols_infos
    ]
