import numpy as np
import json
from typing import Dict, Tuple, List

from abc import ABC, abstractmethod

from rdstar.utilities import PathType

from templatefitter.fitter import MinimizeResult

__all__ = [
    "SystematicsCalculatorBase",
]


class SystematicsCalculatorBase(ABC):
    def __init__(
        self,
        nominal_fit_result_path: PathType,
        fit_results_variations_path: PathType,
    ) -> None:
        self._nominal_fit_result_path = nominal_fit_result_path
        self._fit_results_variations_path = fit_results_variations_path

        self._nominal_fit_result = self.load_nominal_fit_result()
        self._variation_fit_results = self.get_variation_fit_results()

    def load_nominal_fit_result(self) -> MinimizeResult:
        return MinimizeResult.load_from(self._nominal_fit_result_path)

    def get_variation_fit_results(self) -> Dict[str, MinimizeResult]:
        with open(self._fit_results_variations_path) as f:
            results = json.load(f)
        return {variation: MinimizeResult.from_dict(result_dict) for variation, result_dict in results.items()}

    def get_nominal_yields_for_parameter(self) -> Tuple[np.ndarray, np.ndarray]:
        yields = self.nominal_fit_result.params.values
        errors = self._nominal_fit_result.params.errors
        return yields[: self.number_of_yield_parameters], errors[: self.number_of_yield_parameters]

    def get_variation_yields_for_parameter(self) -> Tuple[Tuple[np.ndarray, np.ndarray], ...]:
        yield_error_list = []  # type: List[Tuple[np.ndarray, np.ndarray]]
        for _, fit_result in self.variation_fit_results.items():
            yields = fit_result.params.values
            errors = fit_result.params.errors
            yield_error_list.append(
                (yields[: self.number_of_yield_parameters], errors[: self.number_of_yield_parameters])
            )

        assert len(yield_error_list) == self.number_of_variations, (len(yield_error_list), self.number_of_variations)
        return tuple(yield_error_list)

    @abstractmethod
    def generate_covariance_matrix(self) -> np.ndarray:
        raise NotImplementedError()

    @abstractmethod
    def generate_jacobi_matrix(self) -> np.ndarray:
        raise NotImplementedError()

    def transform_covariance(self) -> np.ndarray:
        covariance = self.generate_covariance_matrix()
        jacobi = self.generate_jacobi_matrix()
        assert covariance.shape == jacobi.shape, (covariance.shape, jacobi.shape)
        dov = np.matmul(np.matmul(jacobi, covariance), np.transpose(jacobi))
        return dov

    @property
    def nominal_fit_result(self) -> MinimizeResult:
        return self._nominal_fit_result

    @property
    def variation_fit_results(self) -> Dict[str, MinimizeResult]:
        return self._variation_fit_results

    @property
    def number_of_variations(self) -> int:
        return len(self.variation_fit_results.keys())

    @property
    def number_of_yield_parameters(self) -> int:
        return len([name for name in self.nominal_fit_result.params.names if name.startswith("yield")])
