import pandas as pd
import copy
import os
import json
from typing import Tuple, Generator, Optional, Any, Dict

from rdstar.utilities import PathType
from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator

from rdstar.offline_analysis.systematics.systematics import SysColsInfo

from templatefitter.fitter import TemplateFitter

__all__ = [
    "SystematicsFitter",
]


class SystematicsFitter:

    varied_weight_column = "varied_weight"  # type: str

    def __init__(
        self,
        df_mc: pd.DataFrame,
        df_data: Optional[pd.DataFrame],
        fit_case: FitCase,
        sys_col_infos: Tuple[SysColsInfo, ...],
        base_output_path: PathType,
    ) -> None:
        self._df_mc = df_mc
        self._df_data = df_data
        self._fit_case = fit_case
        self._original_weight_column = copy.copy(fit_case.weight_column_name)
        self._sys_col_infos = sys_col_infos
        self._base_output_path = base_output_path

    def update_fit_case(self) -> None:
        original_fit_case = copy.copy(self._fit_case)
        self._fit_case = FitCase(
            name=original_fit_case.name,
            base_dir_name=original_fit_case.base_dir_name,
            sample_key=original_fit_case.sample_key,
            output_tag=original_fit_case.output_tag,
            plot_info_text=original_fit_case.plot_info_text,
            asimov_data=original_fit_case.asimov_data,
            round_asimov_data=original_fit_case.round_asimov_data,
            hist_variables=original_fit_case.hist_variables,
            binning_by_mode_dicts=original_fit_case.binning_by_mode_dicts,
            scope_by_mode_dicts=original_fit_case.scope_by_mode_dicts,
            weight_column_name=self.varied_weight_column,
            systematics_cols=original_fit_case.systematics_cols,
            reco_mode_infos=original_fit_case.reco_mode_infos,
            template_setup=original_fit_case.template_setup,
            plot_colors=original_fit_case.plot_colors,
            plot_labels=original_fit_case.plot_labels,
            sig_id_column_name=original_fit_case.sig_id_column_name,
            sig_queries=original_fit_case.sig_queries,
            minimizer_id=original_fit_case.minimizer_id,
        )

    def perform_variation(
        self,
        sys_col_info: SysColsInfo,
    ) -> Generator[Tuple[str, pd.DataFrame], None, None]:
        variations = [var for var in self._df_mc.columns if var.startswith(SysColsInfo.var)]
        for variation in variations:
            self._df_mc.loc[:, self.varied_weight_column] = (
                self._df_mc[self._original_weight_column] / self._df_mc[sys_col_info.ratio] * self._df_mc[variation]
            )
            yield variation, self._df_mc

    def initialize_model_creator(self) -> FitModelCreator:
        FitModelCreator.fit_cases = {self._fit_case.name: self._fit_case}
        model_creator = FitModelCreator(
            fit_case=self._fit_case.name,
        )
        return model_creator

    def perform_fits(self) -> None:
        for systematic in self._sys_col_infos:
            fit_results = dict()  # type: Dict[str, Dict[Any, Any]]
            for variation, varied_df in self.perform_variation(sys_col_info=systematic):
                model_creator = self.initialize_model_creator()
                model_creator.setup_model(
                    mc_df=varied_df,
                    data_df=self._df_data,
                    use_asimov_data=self._fit_case.asimov_data,
                    round_asimov_data_bin_counts=self._fit_case.round_asimov_data,
                    asimov_data_weight_column=self._original_weight_column,
                )
                fit_model = model_creator.fit_model

                fitter = TemplateFitter(fit_model=fit_model, minimizer_id="iminuit")
                fit_result = fitter.do_fit(
                    update_templates=True,
                    get_hesse=True,
                    verbose=True,
                    fix_nui_params=True,  # todo Should be fixed in this kind of systematics handling
                )
                fit_results.update({variation: fit_result.as_dict})

            output_path = os.path.join(self._base_output_path, f"sys_{systematic.var}_all_fit_results.json")
            tmp_output_path = os.path.join(self._base_output_path, f"tmp_sys_{systematic.var}_all_fit_results.json")
            with open(tmp_output_path, "w") as fp:
                json.dump(fit_results, fp, indent=2)
            os.rename(src=tmp_output_path, dst=output_path)
