"""
Data Reduction Tasks

    These tasks have the purpose of reducing the size of the data frames by
        - Application of Continuum Suppression and a light cut on the output
        - Removing unneeded columns from the data frames

Felix Metzner 2020
"""

import os
import logging
import numpy as np
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Dict, Generator, Optional
from more_itertools import split_into, random_permutation

from analysistools.dataframe_tools import hdf5_utils

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar import continuum_suppression as cs_tools
from rdstar.offline_analysis import offline_observables
from rdstar.offline_analysis import rdstar_sample_weights
from rdstar.offline_analysis.observables.offline_observables_manager import OfflineObservablesManager

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.cs_tasks import CSTestTask
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "DataReductionTask",
]


@inherits(CSTestTask)
@inherits(DataPreparationMergerTask)
class DataReductionTask(utilities.BaseOfflineTask):
    evaluation_sample_fraction = 0.05  # type: float
    evaluation_sample_dir_name = "EvaluationSamples"  # type: str
    evaluation_sample_tag = "evaluation_subset"  # type: str

    @property
    def resources(self) -> Dict[str, int]:
        if self.data_type in ["Charged", "Charm", "UDS", "DataY4S"]:
            return {utilities.offline_memory_resource_name: 6}
        elif self.data_type in ["Mixed"]:
            return {utilities.offline_memory_resource_name: 4}
        else:
            return {utilities.offline_memory_resource_name: 1}

    @property
    def priority(self) -> int:
        return int(max(super().priority / 10.0, self.default_priority))

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType
        self.base_dir = os.path.join(self.working_dir, self.data_type)
        self.eval_sample_base_dir = os.path.join(self.working_dir, self.evaluation_sample_dir_name, self.data_type)

        if isinstance(offline_config["cs_config"], dict):
            self.cs_offline_config_path = self.offline_config_path
            cs_config = offline_config["cs_config"]
        else:
            configs_dir = utilities.get_config_directory()
            self.cs_offline_config_path = os.path.join(configs_dir, offline_config["cs_config"])

            if not os.path.isfile(self.cs_offline_config_path):
                raise RuntimeError(
                    f"Encountered invalid offline_config path for cs training:\n" f"\t{self.cs_offline_config_path}"
                )

            cs_offline_config = utilities.load_yaml_config(yaml_config_path=self.cs_offline_config_path)
            cs_config = cs_offline_config["cs_config"]

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f"{self.data_type}{self.stream_tag}_{self.tuple_type}"
        self.input_key = f"{self.file_name}_{DataPreparationTask.base_file_tag}"

        full_file_name = f"{self.file_name}_with_CS_{self.cs_tag}"  # type: str
        self.target_path = os.path.join(self.base_dir, f"{full_file_name}.h5")
        self.tmp_target_path = os.path.join(self.base_dir, f"tmp_{full_file_name}.h5")
        self.output_key = full_file_name

        self.eval_sample_target_path = os.path.join(self.eval_sample_base_dir, f"{full_file_name}.h5")
        self.tmp_eval_sample_target_path = os.path.join(self.eval_sample_base_dir, f"tmp_{full_file_name}.h5")
        self.eval_sample_key = f"{full_file_name}_{self.evaluation_sample_tag}"

        self.cs_column_name = cs_config["cs_column_name"]
        self.cs_cut_value = cs_config["cs_cut_value"] if "cs_cut_value" in cs_config else None

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def apply_cs_mva(self, df: pd.DataFrame) -> None:
        cs_classifier = cs_tools.ContinuumSuppression.load_classifier(
            weight_file_path=self.input_dict["cs_output_weight_file"],
            var_file_path=self.input_dict["cs_var_list_file"],
        )
        logging.info(f"\tApplying CS classifier to data frame {self.file_name}...")

        cs_classifier.apply(data=df, output_column=self.cs_column_name)

    def apply_soft_cs_mva_selection_cut(self, df: pd.DataFrame) -> None:
        if self.cs_cut_value is not None:
            logging.info(f"\tApplying cut on CS output to data frame {self.file_name}...")
            n_events_before = len(df.index)  # type: int
            df.query(f"{self.cs_column_name} > {self.cs_cut_value}", inplace=True)
            n_events_after = len(df.index)  # type: int
            red_factor = 1.0 * n_events_after / n_events_before  # type: float
            logging.info(
                f"\tApplied cut on CS output to data frame {self.file_name} and "
                f"reduced number of data rows by {red_factor} from {n_events_before} to {n_events_after}..."
            )

    @staticmethod
    def discard_unneeded_columns(df: pd.DataFrame) -> None:
        discard_cols = [c for c in offline_observables.observables_to_discard if c in df.columns]
        df.drop(discard_cols, axis=1, inplace=True)

    def add_scaling_weights(self, df: pd.DataFrame) -> None:
        rdstar_sample_weights.add_scaling_weights(
            df=df,
            data_type=self.data_type,
            global_file_dict_path=DataCopyTask.get_local_file_dict_path(offline_config_path=self.offline_config_path),
            generic_stream_dict=None,
        )
        rdstar_sample_weights.add_off_to_on_resonance_scaling_weights(
            df=df,
            data_type=self.data_type,
        )

    def data_reduction(self) -> None:
        df = hdf5_utils.load_hdf5(hdf5_file_path=self.input_dict[self.input_key], hdf5_key=self.file_name)

        self.apply_cs_mva(df=df)

        if self.data_type.startswith("Data"):
            df.reset_index(drop=True, inplace=True)
            utilities.add_unique_event_id(df=df, data_type=f"{self.data_type}{self.stream_tag}")

        evaluation_sample = None  # type: Optional[pd.DataFrame]
        if self.stream == 0 or self.stream == -999:
            evaluation_sample = self.extract_evaluation_sub_sample(df=df, sample_fraction=self.evaluation_sample_fraction)

        self.apply_soft_cs_mva_selection_cut(df=df)

        if self.data_type.startswith("Data"):
            # Resetting index for recorded Data, because events have been discarded in apply_cs_mva
            df.reset_index(drop=True, inplace=True)
            OfflineObservablesManager.add_offline_variables(
                df=df,
                is_off_resonance=rdstar_samples.is_off_resonance(data_type_key=self.data_type),
            )

        self.discard_unneeded_columns(df=df)
        self.add_scaling_weights(df=df)

        df.to_hdf(path_or_buf=self.tmp_target_path, key=self.file_name)
        os.rename(src=self.tmp_target_path, dst=self.target_path)

        del df

        if self.stream == 0 or self.stream == -999:
            assert evaluation_sample is not None
            if self.data_type.startswith("Data"):
                OfflineObservablesManager.add_offline_variables(
                    df=evaluation_sample,
                    is_off_resonance=rdstar_samples.is_off_resonance(data_type_key=self.data_type),
                )

            self.discard_unneeded_columns(df=evaluation_sample)
            self.add_scaling_weights(df=evaluation_sample)

            evaluation_sample.to_hdf(path_or_buf=self.tmp_eval_sample_target_path, key=self.eval_sample_key)
            os.rename(src=self.tmp_eval_sample_target_path, dst=self.eval_sample_target_path)

    @staticmethod
    def extract_evaluation_sub_sample(
        df: pd.DataFrame,
        sample_fraction: float,
    ) -> pd.DataFrame:
        assert 0.0 < sample_fraction < 1.0, sample_fraction
        unique_event_col_name = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()  # type: str

        sample_size = int(np.ceil(len(df.unique_event_col_name.unique()) * sample_fraction))  # type: int

        unique_event_groups = random_permutation(df.groupby(unique_event_col_name, group_keys=False))
        sub_sample_groups, _ = split_into(unique_event_groups, [sample_size, None])

        sub_sample_ids = pd.concat(
            [group[unique_event_col_name] for _, group in sub_sample_groups],
            axis=0,
            sort=False,
        ).index

        sub_sample_df = df.query(df.index.isin(sub_sample_ids), inplace=False)
        sub_sample_df.reset_index(drop=True, inplace=True)
        return sub_sample_df

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.data_reduction()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {self.output_key: luigi.LocalTarget(self.target_path)}
        if self.stream == 0 or self.stream == -999:
            output_dict.update({self.eval_sample_key: luigi.LocalTarget(self.eval_sample_target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CSTestTask, offline_config_path=self.cs_offline_config_path)
        yield self.clone(DataPreparationMergerTask)
