"""
Fitting Tasks

These tasks will
    - perform the actual fit of the analysis (TODO)
    - perform sideband fits on data to validate the model and the data-MC agreement (TODO)
    - run asimov fits on sideband samples (TODO)
    - run asimov fits on signal region samples (TODO)
    - perform toy studies (TODO)
    - perform profile scans (TODO)
    - plot the template shapes (TODO)
    - plot the distributions before and after the fits (TODO)

Felix Metzner 2020
"""

import os
import json
import logging
import pandas as pd
import b2luigi as luigi

from collections import abc
from luigi.util import inherits
from PyPDF2 import PdfFileMerger
from typing import Tuple, List, Dict, Generator

from templatefitter.fitter import TemplateFitter
from templatefitter.plotter.fit_result_plots import FitResultPlotter

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.rdstar_plots import rdstar_dec_titles
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.tasks.resolution_correction_task import (
    ResolutionCorrectionTask,
    ResolutionCorrectionManagerTask,
)

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.fitting.rdstar_fit_models import rdstar_fit_cases


logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "FitTask",
    "FittingTasksWrapperTask",
]


# TODO: Maybe combine sample for fit beforehand.
# TODO: Maybe move fit samples to Deepthought SSDs


@inherits(ResolutionCorrectionManagerTask)
class FitTask(utilities.BaseOfflineTask):
    fit_case_key = luigi.Parameter(positional=False)  # type: str

    fit_base_dir_name = "FitResults"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.fit_case_name = self.fit_case_key  # type: str
        self.fit_case = FitModelCreator.fit_cases[self.fit_case_key]  # type: FitCase
        self.base_dir_name = self.fit_case.base_dir_name  # type: PathType
        self.plot_info_text = self.fit_case.plot_info_text  # type: str
        self.fit_output_tag = self.fit_case.output_tag  # type: str

        input_sample_key = FitModelCreator.fit_cases[self.fit_case_key].sample_key  # type: str
        if input_sample_key != "":
            input_sample_key = f"_{input_sample_key}"

        self.base_path = os.path.join(
            self.working_dir, self.fit_base_dir_name, self.main_sel_tag, self.base_dir_name
        )  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.mc_sample_keys = rdstar_samples.main_sample_keys  # type: List[str]

        self.key = f"_{self.selection_name}_with_CS_{self.cs_tag}{input_sample_key}"  # type: str
        self.input_appendix = (
            f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"
        )  # type: str

        self.input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

    def load_mc_sample(self) -> pd.DataFrame:
        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{mc_key}_{self.input_appendix}_")
                and k.endswith(self.key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]  # type: List[PathType]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, self.key)

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        WeightManager.add_additional_weights(df=mc_sample)

        return mc_sample

    def load_recorded_data_sample(self) -> pd.DataFrame:
        sample_key = "DataY4S"
        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{sample_key}_{self.input_appendix}_")
                and k.endswith(self.key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]  # type: List[PathType]
        assert len(data_paths) == 1, (len(data_paths), data_paths, self.key)
        data_path = data_paths[0]
        assert os.path.exists(data_path), data_path
        data_df = pd.read_hdf(data_path)
        assert isinstance(data_df, pd.DataFrame)
        return data_df

    def get_fit_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}"
        return key, f"{key}.json"

    def get_fit_plot_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}"
        return key, f"{key}.pdf"

    @staticmethod
    def merge_plot_pdf_files(
        input_paths: List[PathType],
        output_path: PathType,
        overwrite: bool = False,
    ) -> None:
        pdf_merger = PdfFileMerger()
        assert all(os.path.exists(path) for path in input_paths), [p for p in input_paths if not os.path.exists(p)]
        assert all(os.path.splitext(path)[1] == ".pdf" for path in input_paths), [
            p for p in input_paths if not os.path.splitext(p)[1] == ".pdf"
        ]
        assert os.path.splitext(output_path)[1] == ".pdf", output_path
        if not overwrite:
            assert not os.path.exists(output_path), output_path

        for pdf_path in input_paths:
            pdf_merger.append(pdf_path)

        pdf_merger.write(fileobj=output_path)
        pdf_merger.close()

    def do_fit(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]
        fit_result_key, fit_result_file_name = self.get_fit_output_key_file_name_pair(file_tag="FitResult")
        temp_fit_result_path = os.path.join(self.base_path, f"temp_{fit_result_file_name}")
        fit_result_path = os.path.join(self.base_path, fit_result_file_name)
        output_dict.update({fit_result_key: fit_result_path})

        fit_sig_key, fit_sig_file_name = self.get_fit_output_key_file_name_pair(file_tag="FitSignificance")
        temp_fit_sig_path = os.path.join(self.base_path, f"temp_{fit_sig_file_name}")
        fit_sig_path = os.path.join(self.base_path, fit_sig_file_name)
        output_dict.update({fit_sig_key: fit_sig_path})

        pre_fit_plot_key, pre_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PreFitPlot")
        temp_pre_fit_plot_path = os.path.join(self.base_path, f"temp_{pre_fit_plot_file_name}")
        pre_fit_plot_path = os.path.join(self.base_path, pre_fit_plot_file_name)
        output_dict.update({pre_fit_plot_key: pre_fit_plot_path})

        post_fit_plot_key, post_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PostFitPlot")
        temp_post_fit_plot_path = os.path.join(self.base_path, f"temp_{post_fit_plot_file_name}")
        post_fit_plot_path = os.path.join(self.base_path, post_fit_plot_file_name)
        output_dict.update({post_fit_plot_key: post_fit_plot_path})

        if not dry_run:
            mc_sample = self.load_mc_sample()

            fit_model_creator = FitModelCreator(fit_case=self.fit_case_name)

            fit_model_creator.setup_model(
                mc_df=mc_sample,
                data_df=None if self.fit_case.asimov_data else self.load_recorded_data_sample(),
                use_asimov_data=self.fit_case.asimov_data,
                round_asimov_data_bin_counts=self.fit_case.round_asimov_data,
            )
            fit_model = fit_model_creator.fit_model

            fit_result_plotter = FitResultPlotter(
                variables=fit_model_creator.fit_variables,
                fit_model=fit_model,
                reference_dimension=0,
                channel_label_dict=rdstar_dec_titles,
                data_label=self.plot_info_text,
            )

            pre_fit_plot_paths = fit_result_plotter.plot_fit_result(
                use_initial_values=True,
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag,
            )
            assert len(pre_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["png"]), (
                len(pre_fit_plot_paths["pdf"]),
                len(pre_fit_plot_paths["png"]),
            )
            self.merge_plot_pdf_files(input_paths=pre_fit_plot_paths["pdf"], output_path=temp_pre_fit_plot_path)

            for fit_dim in range(len(self.fit_case.hist_variables)):
                fit_result_plotter.plot_fit_result_projections(
                    project_to=fit_dim,
                    use_initial_values=True,
                    output_dir_path=self.base_path,
                    output_name_tag=self.fit_output_tag,
                )
            os.rename(src=temp_pre_fit_plot_path, dst=pre_fit_plot_path)

            fitter = TemplateFitter(
                fit_model=fit_model,
                minimizer_id=self.fit_case.minimizer_id,
            )
            fit_result = fitter.do_fit(update_templates=True, get_hesse=True, verbose=True, fix_nui_params=True)

            fit_result.save_to(path=temp_fit_result_path, overwrite=True)
            os.rename(src=temp_fit_result_path, dst=fit_result_path)

            post_fit_plot_paths = fit_result_plotter.plot_fit_result(
                output_dir_path=self.base_path,
                output_name_tag=self.fit_output_tag,
            )
            assert len(post_fit_plot_paths["pdf"]) == len(post_fit_plot_paths["png"]), (
                len(post_fit_plot_paths["pdf"]),
                len(post_fit_plot_paths["png"]),
            )
            assert len(post_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["pdf"]), (
                len(post_fit_plot_paths["pdf"]),
                len(pre_fit_plot_paths["pdf"]),
            )
            self.merge_plot_pdf_files(input_paths=post_fit_plot_paths["pdf"], output_path=temp_post_fit_plot_path)

            for fit_dim in range(len(self.fit_case.hist_variables)):
                fit_result_plotter.plot_fit_result_projections(
                    project_to=fit_dim,
                    use_initial_values=False,
                    output_dir_path=self.base_path,
                    output_name_tag=self.fit_output_tag,
                )

            os.rename(src=temp_post_fit_plot_path, dst=post_fit_plot_path)

            significance_dict = dict()  # type: Dict[str, float]
            for yield_parameter in fit_model.get_yield_parameter_names():
                significance = fitter.get_significance(yield_parameter=yield_parameter, fix_nui_params=True)
                significance_dict.update({yield_parameter: significance})

            with open(temp_fit_sig_path, "w") as fp_sig:
                json.dump(obj=significance_dict, fp=fp_sig, indent=2)
            os.rename(src=temp_fit_sig_path, dst=fit_sig_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_fit()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_fit(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )


@inherits(ResolutionCorrectionManagerTask)
class FittingTasksWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def complete(self):
        # Catching special case of a FittingTasksWrapperTask for selection with which no output can be produced.
        if len(list(self.requires())) == 0:
            logging.warning(f"FittingTasksWrapperTask cannot produce output for selection name {self.selection_name}!")
            return True
        return super().complete()

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for fit_case_key, fit_case in rdstar_fit_cases.items():
            selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
            sample_tag = f"_{fit_case.sample_key}" if fit_case.sample_key else fit_case.sample_key
            required_tag = f"{self.selection_name}_with_CS_{self.cs_tag}{sample_tag}"  # type: str
            if any(st == required_tag for st in selection_tags):
                yield self.clone(FitTask, fit_case_key=fit_case_key)
