"""
Tasks performing the Flow Analysis of the Analysis

Felix Metzner 2019
"""

import os
import copy
import multiprocessing
import b2luigi as luigi

from collections import abc
from functools import partial
from luigi.util import requires
from matplotlib import pyplot as plt

from typing import Any, Dict, List, Callable

from templatefitter.plotter.plot_utilities import export

from analysistools.plotting.latex_tables import df_to_split_latex_table

from rdstar.utilities import PathType
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.validation import analysis_flow_validation, confusion_validation, multiplicity_validation

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.tasks.combiner_tasks import SelectionInfoCombinerTask

__all__ = [
    "AnalysisEvaluationTask",
]


@requires(SelectionInfoCombinerTask)
class AnalysisEvaluationTask(utilities.BaseOfflineTask):
    cs_tag = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "AnalysisEvaluation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.confusion_matrix_infos = copy.copy(confusion_validation.confusion_matrix_info_collection)

        self.multiplicity_infos = copy.copy(multiplicity_validation.multiplicity_info_collection)

        self.input_dict = {k: v.path for k, v in self.input().items()}  # type: Dict[str, PathType]
        assert all(k.startswith(DataSelectionTask.evaluation_info_prefix) for k in self.input_dict)
        assert all(
            os.path.basename(p).startswith(DataSelectionTask.evaluation_info_prefix) for p in self.input_dict.values()
        )
        assert all(os.path.splitext(p)[1] == ".json" for p in self.input_dict.values())

    def add_table_fraction_output_info(
        self,
        output_dict: Dict[str, PathType],
        name_prefix: str,
        table_fractions: List[int],
    ) -> None:
        analysis_flow_validation.update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=self.base_path,
            table_fractions=table_fractions,
        )

    def create_confusion_matrices(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        sub_info_dict = {}  # type: Dict[str, Any]
        if not dry_run:
            assert len(info_dict) > 0, len(info_dict)
            sub_info_dict = analysis_flow_validation.get_sub_dict_by_index(
                info_dict=info_dict,
                index=len(info_dict) - 1,
            )

        for conf_mat_info in self.confusion_matrix_infos:
            assert isinstance(conf_mat_info, confusion_validation.ConfusionMatrixInfo)
            file_name = f"{selection_tag}_cm_{conf_mat_info.name}_confusion_matrix"
            if not dry_run:
                matrix = confusion_validation.get_confusion_matrix(
                    sub_info_dict=sub_info_dict,
                    confusion_matrix_info=conf_mat_info,
                )
                fig, _ = confusion_validation.create_confusion_matrix_plot(
                    confusion_matrix=matrix,
                    confusion_matrix_info=conf_mat_info,
                )
                export(fig=fig, filename=file_name, target_dir=self.base_path)
                plt.close(fig=fig)
            assert file_name not in output_dict
            output_dict.update({file_name: os.path.join(self.base_path, f"{file_name}.pdf")})

        return output_dict

    def do_cut_flow_analysis(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        info_keys = copy.copy(analysis_flow_validation.cut_flow_info_keys)  # type: List[str]
        table_fractions = [2, 4]  # type: List[int]

        column_formatter = [
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s :0.2f}",
        ]

        for info_key in info_keys:
            name_prefix = f"{selection_tag}_cft_{info_key}_cut_flow"

            self.add_table_fraction_output_info(
                output_dict=output_dict,
                name_prefix=name_prefix,
                table_fractions=table_fractions,
            )

            if not dry_run:
                df = analysis_flow_validation.create_cut_flow_table(
                    info_dict=info_dict,
                    info_key=info_key,
                    selection_name=self.selection_name,
                )

                df_to_split_latex_table(
                    df=df,
                    table_fractions=table_fractions,
                    name_prefix=name_prefix,
                    output_lookup_dict=output_dict,
                    placeholder_element="---",
                    remove_column_name_index=True,
                    split_col_formatters=column_formatter,
                    create_pdf=True,
                    multicolumn_format="c",
                )

        return output_dict

    def do_cut_flow_analysis_for_data(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        table_fractions = [2, 4]  # type: List[int]

        column_formatter = [
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s :0.2f}",
        ]  # type: List[Callable]

        name_prefix = f"{selection_tag}_cft_data_cut_flow"

        self.add_table_fraction_output_info(
            output_dict=output_dict,
            name_prefix=name_prefix,
            table_fractions=table_fractions,
        )

        if not dry_run:
            df = analysis_flow_validation.create_cut_flow_table_for_data(
                info_dict=info_dict,
                selection_name=self.selection_name,
            )

            df_to_split_latex_table(
                df=df,
                table_fractions=table_fractions,
                name_prefix=name_prefix,
                output_lookup_dict=output_dict,
                placeholder_element="---",
                remove_column_name_index=True,
                split_col_formatters=column_formatter,
                create_pdf=True,
                multicolumn_format="c",
                allow_smaller_sub_dicts=True,
            )

        return output_dict

    def create_multiplicity_plots(
        self,
        info_dict: Dict[str, Any],
        multiplicity_info: multiplicity_validation.MultiplicityInfo,
        selection_tag: str,
        dry_run: bool = False,
        max_cands: int = 10,
        log_y_scale: bool = False,
        for_data: bool = False,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]
        em_key = f"event_multiplicities{multiplicity_info.key}"  # type: str

        if for_data:
            assert not multiplicity_info.key, multiplicity_info.key
            file_name_tag = "data"  # type: str
        else:
            file_name_tag = em_key

        file_name = f"{selection_tag}_emp_{file_name_tag}_plot"  # type: str
        assert file_name not in output_dict, file_name
        output_dict.update({file_name: os.path.join(self.base_path, f"{file_name}.pdf")})

        if not dry_run:
            cut_list = info_dict[list(info_dict.keys())[len(info_dict) - 1]]["cuts"]

            if multiplicity_info.key:
                tmp_dict = {k: copy.copy(info_dict[k][em_key]) for k in info_dict.keys()}
                assert all(all(isinstance(v, dict) for v in sub_di.values()) for sub_di in tmp_dict.values())
                emp_dict = {k: {} for k in info_dict.keys()}  # type: Dict[str, Any]
                for k in info_dict.keys():
                    for sub_dict in tmp_dict[k].values():
                        emp_dict[k].update(
                            SelectionInfoCombinerTask.update_selection_dicts(dict1=emp_dict[k], dict2=sub_dict)
                        )
            else:
                emp_dict = {k: copy.copy(info_dict[k][em_key]) for k in info_dict.keys()}

            assert all(all(isinstance(v, int) for v in sub_di.values()) for sub_di in emp_dict.values())

            fig, _ = multiplicity_validation.create_multiplicity_plot(
                multiplicity_dict=emp_dict,
                cut_list=cut_list,
                selection_name=self.selection_name,
                max_cands=max_cands,
                log_y_scale=log_y_scale,
            )
            export(fig=fig, filename=file_name, target_dir=self.base_path)
            plt.close(fig=fig)

        return output_dict

    def do_event_multiplicity_evaluation(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        for multiplicity_info in self.multiplicity_infos:
            if multiplicity_info.key:
                em_table_output_dict = multiplicity_validation.create_multiplicity_table(
                    info_dict=info_dict,
                    multiplicity_info=multiplicity_info,
                    selection_name=self.selection_name,
                    selection_tag=selection_tag,
                    base_path=self.base_path,
                    dry_run=dry_run,
                )
                assert all(k not in output_dict for k in em_table_output_dict), (em_table_output_dict, output_dict)
                output_dict.update(em_table_output_dict)

            em_plot_output_dict = self.create_multiplicity_plots(
                info_dict=info_dict,
                multiplicity_info=multiplicity_info,
                selection_tag=selection_tag,
                dry_run=dry_run,
            )
            assert all(k not in output_dict for k in em_plot_output_dict), (em_plot_output_dict, output_dict)
            output_dict.update(em_plot_output_dict)

        return output_dict

    def do_event_multiplicity_evaluation_for_data(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        em_table_output_dict = multiplicity_validation.create_multiplicity_table_for_data(
            info_dict=info_dict,
            selection_name=self.selection_name,
            selection_tag=selection_tag,
            base_path=self.base_path,
            dry_run=dry_run,
        )
        assert all(k not in output_dict for k in em_table_output_dict), (em_table_output_dict, output_dict)
        output_dict.update(em_table_output_dict)

        em_plot_output_dict = self.create_multiplicity_plots(
            info_dict=info_dict,
            multiplicity_info=multiplicity_validation.MultiplicityInfo(key="", column_name_dict=None),
            selection_tag=selection_tag,
            dry_run=dry_run,
            for_data=True,
        )
        assert all(k not in output_dict for k in em_plot_output_dict), (em_plot_output_dict, output_dict)
        output_dict.update(em_plot_output_dict)

        return output_dict

    def do_bcs_validation(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        info_key_parts = copy.copy(analysis_flow_validation.bcs_info_key_parts)  # type: List[str]
        table_fractions = [2, 4]  # type: List[int]

        for key_part in info_key_parts:
            name_prefix = f"{selection_tag}_bcst_{key_part}bcs"
            self.add_table_fraction_output_info(
                output_dict=output_dict,
                name_prefix=name_prefix,
                table_fractions=table_fractions,
            )

        if not dry_run:
            assert len(info_dict) > 0, len(info_dict)
            bcs_df_dict = analysis_flow_validation.create_bcs_table(
                info_dict=info_dict,
                info_key_parts=info_key_parts,
            )

            row_formatters = {}  # type: Dict[str, Callable]
            for row_key in ["Sig", "Bkg", "Norm"]:
                row_formatters.update(
                    {
                        f"{row_key} before": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                        f"{row_key} after": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                        f"{row_key} Eff": lambda s: r"\multicolumn{1}{c}{---}"
                        if s == "---"
                        else f"{s * 100:0.2f}" + r"$\,\si{\percent}$",
                    }
                )

            for key_part, bcs_df in bcs_df_dict.items():
                name_prefix = f"{selection_tag}_bcst_{key_part}bcs"

                df_to_split_latex_table(
                    df=bcs_df,
                    table_fractions=table_fractions,
                    name_prefix=name_prefix,
                    output_lookup_dict=output_dict,
                    placeholder_element="---",
                    remove_column_name_index=True,
                    create_pdf=True,
                    row_formatters=row_formatters,
                )

        return output_dict

    def do_bcs_validation_for_data(
        self,
        info_dict: Dict[str, Any],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        table_fractions = [2, 5]  # type: List[int]

        name_prefix = f"{selection_tag}_bcst_data_bcs"
        self.add_table_fraction_output_info(
            output_dict=output_dict,
            name_prefix=name_prefix,
            table_fractions=table_fractions,
        )

        if not dry_run:
            assert len(info_dict) > 0, len(info_dict)
            bcs_df = analysis_flow_validation.create_bcs_table_for_data(info_dict=info_dict)

            row_formatters = {
                "Before": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                "After": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                "Eff": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s * 100:0.2f}" + r"$\,\si{\percent}$",
            }

            df_to_split_latex_table(
                df=bcs_df,
                table_fractions=table_fractions,
                name_prefix=name_prefix,
                output_lookup_dict=output_dict,
                placeholder_element="---",
                remove_column_name_index=True,
                create_pdf=True,
                row_formatters=row_formatters,
                allow_smaller_sub_dicts=True,
            )

        return output_dict

    @staticmethod
    def do_validation_sub_task(
        sub_task_function: Callable,
        path: PathType,
        selection_tag: str,
        output_dict: Dict[str, PathType],
        dry_run: bool,
    ) -> Dict[str, PathType]:
        info_dict = utilities.load_dict_from_json(path=path) if not dry_run else None
        outputs = sub_task_function(
            info_dict=info_dict,
            selection_tag=selection_tag,
            dry_run=dry_run,
        )
        assert all(k not in output_dict for k in outputs)
        return outputs

    def do_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        mc_validation_sub_task_functions = [
            self.create_confusion_matrices,
            self.do_cut_flow_analysis,
            self.do_event_multiplicity_evaluation,
            self.do_bcs_validation,
        ]  # type: List[Callable]

        data_validation_sub_task_functions = [
            self.do_cut_flow_analysis_for_data,
            self.do_event_multiplicity_evaluation_for_data,
            self.do_bcs_validation_for_data,
        ]  # type: List[Callable]

        output_dict = {}  # type: Dict[str, PathType]
        for key, path in self.input_dict.items():
            combo_key_and_sel_tag = key.split(DataSelectionTask.evaluation_info_prefix + "_")[1]
            combo_key, selection_tag = combo_key_and_sel_tag.split("_", 1)
            assert combo_key in SelectionInfoCombinerTask.info_names, combo_key
            selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
            assert selection_tag in selection_tags, (selection_tag, selection_tags)

            if combo_key == SelectionInfoCombinerTask.main_info_name:
                validation_sub_task_functions = mc_validation_sub_task_functions
            elif combo_key == SelectionInfoCombinerTask.data_info_name:
                validation_sub_task_functions = data_validation_sub_task_functions
            else:
                continue

            with multiprocessing.Pool(processes=1 if dry_run else len(validation_sub_task_functions)) as pool:
                eval_outputs = pool.map(
                    partial(
                        AnalysisEvaluationTask.do_validation_sub_task,
                        path=path,
                        selection_tag=selection_tag,
                        output_dict=output_dict,
                        dry_run=dry_run,
                    ),
                    validation_sub_task_functions,
                )

            assert all(k not in output_dict for sub_dict in eval_outputs for k in sub_dict)
            assert all(v not in output_dict.values() for sub_dict in eval_outputs for v in sub_dict.values())
            for sub_dict in eval_outputs:
                output_dict.update(sub_dict)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}
