"""
Tasks performing the Validation of the FSR Correction

Patrick Ecker 2020
"""

import os
import json
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Dict, List, Generator, Iterable

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.combiner_tasks import SampleCombinerTask
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.tasks.data_reduction_tasks import DataReductionTask
from rdstar.offline_analysis.rdstar_samples import main_sample_keys, efficiency_samples
from rdstar.offline_analysis.histogram_variables import (
    fsr_energy_cms,
    fsr_energy_lab,
    fsr_energy_rest_frame,
    fsr_momentum_cms,
    fsr_momentum_lab,
    fsr_momentum_rest_frame,
)

from rdstar.offline_analysis.validation.fsr_correction_validation import (
    FSRCorrectionValidator,
    fsr_pl_star,
    fsr_missing_mass,
    FSRVariable,
)

__all__ = [
    "FSRCorrectionEfficiencyTask",
    "FSRCorrectionValidationTask",
]


@inherits(DataSelectionTask)
class FSRCorrectionEfficiencyTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: str
    selection_name = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "FSRCorrectionEfficiency"  # type: str

    fsr_corrected_col = "wasFSRCorrected"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)  # type: PathType

        self.input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

    def generate_efficiency_dict(
        self,
        unselected_df: pd.DataFrame,
        selected_df: pd.DataFrame,
    ) -> Dict[str, float]:
        unselected_df.query("abs(daughter__bo1__cm__spdaughter__bo1__cm__spPDG__bc__bc) == 11", inplace=True)
        total_unselected = len(unselected_df)
        fsr_corr_unselected = len(unselected_df.query(f"{self.fsr_corrected_col} == 1"))

        selected_df.query("abs(daughter__bo1__cm__spdaughter__bo1__cm__spPDG__bc__bc) == 11", inplace=True)
        total_selected = len(selected_df)
        fsr_corr_selected = len(selected_df.query(f"{self.fsr_corrected_col} == 1"))

        eff_dict = {
            "total_number_unselected": total_unselected,
            "fsr_corrected_number_unselected": fsr_corr_unselected,
            "ratio_unselected": fsr_corr_unselected / total_unselected,
            "total_number_selected": total_selected,
            "fsr_corrected_number_selected": fsr_corr_selected,
            "ratio_selected": fsr_corr_selected / total_selected,
        }
        return eff_dict

    def run_fsr_correction_efficiency_calculation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_key = f"{self.data_type}_{self.selection_name}_fsr_efficiency"
        output_path = os.path.join(self.base_path, f"{output_key}.json")

        if not dry_run:
            unselected_key = f"{self.data_type}_{self.tuple_type}_with_CS_{self.cs_tag}"
            df_unselected = pd.read_hdf(self.input_dict[unselected_key])
            assert isinstance(df_unselected, pd.DataFrame)

            selected_key = f"{self.data_type}_{self.tuple_type}_{self.selection_name}_with_CS_{self.cs_tag}"
            df_selected = pd.read_hdf(self.input_dict[selected_key])
            assert isinstance(df_selected, pd.DataFrame)

            efficiency_dict = self.generate_efficiency_dict(unselected_df=df_unselected, selected_df=df_selected)
            with open(output_path, "w") as output_fp:
                json.dump(efficiency_dict, output_fp, indent=2)

        return {output_key: output_path}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.run_fsr_correction_efficiency_calculation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.run_fsr_correction_efficiency_calculation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(DataSelectionTask)
        yield self.clone(DataReductionTask)


class FSRCorrectionValidationTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType
    selection_name = luigi.Parameter(positional=False)  # type: str
    cs_tag = luigi.Parameter(positional=False)  # type: str
    config_combinations = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name = "FSRCorrectionValidation"  # type: str

    fsr_comparison_vars = [fsr_missing_mass, fsr_pl_star]  # type: List[FSRVariable]
    fsr_vars = [
        fsr_energy_cms,
        fsr_energy_lab,
        fsr_energy_rest_frame,
        fsr_momentum_cms,
        fsr_momentum_lab,
        fsr_momentum_rest_frame,
    ]  # type: List[HistVariable]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def merge_efficiency_dict(self, dry_run: bool = False) -> Dict[str, Dict[str, float]]:
        eff_dict = {}  # type: Dict[str, Dict[str, float]]
        if dry_run:
            return eff_dict

        for data_type in efficiency_samples:
            eff_key = f"{data_type}_{self.selection_name}_fsr_efficiency"
            tmp_eff_dict = utilities.load_dict_from_json(self.input_dict[eff_key])
            eff_dict.update({data_type: tmp_eff_dict})

        return eff_dict

    def run_fsr_correction_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        mc_samples_paths = [
            self.input_dict[f"{key}_{self.selection_name}_with_CS_{self.cs_tag}"] for key in main_sample_keys
        ]  # type: List[PathType]

        efficiency_dict = self.merge_efficiency_dict(dry_run=dry_run)

        fsr_corr_validator = FSRCorrectionValidator(
            list_of_df_paths=mc_samples_paths,
            base_output_path=self.base_path,
            fsr_comparison_vars=self.fsr_comparison_vars,
            fsr_vars=self.fsr_vars,
            efficiency_dict=efficiency_dict,
            selection_name=self.selection_name,
        )
        output_info_list = fsr_corr_validator.do_validation(dry_run=dry_run)

        return {k: v for k, v in output_info_list}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.run_fsr_correction_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.run_fsr_correction_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, Iterable)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield SampleCombinerTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                config_params=relevant_configs,
            )
            for params in relevant_configs:
                if params["data_type"] in efficiency_samples:
                    yield self.clone(FSRCorrectionEfficiencyTask, selection_name=self.selection_name, **params)
