"""
Tasks producing the Validation Results for the Analysis

Felix Metzner 2019
"""

import b2luigi as luigi

from luigi.util import inherits
from typing import Tuple, List, Generator, Any

from rdstar.offline_analysis.selection.selection_setups import SelectionSetups

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.result_tasks import EfficiencyEvaluationTask
from rdstar.offline_analysis.tasks.fitting_tasks import FittingTasksWrapperTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightValidationWrapperTask
from rdstar.offline_analysis.tasks.flow_validation_tasks import AnalysisEvaluationTask
from rdstar.offline_analysis.tasks.analysis_plotter_tasks import AnalysisPlotManagerTask
from rdstar.offline_analysis.tasks.sideband_validation_tasks import SidebandEvaluationWrapperTask
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionValidationManagerTask
from rdstar.offline_analysis.tasks.fsr_correction_validation_tasks import FSRCorrectionValidationTask


__all__ = [
    "ValidationWrapperTask",
    "SelectionManagerTask",
]


@inherits(
    AnalysisEvaluationTask,
    SidebandEvaluationWrapperTask,
    ResolutionCorrectionValidationManagerTask,
)
class ValidationWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(AnalysisEvaluationTask)
        yield self.clone(SidebandEvaluationWrapperTask)
        if self.selection_name in SelectionSetups.resolution_correction_mapping.values():
            yield self.clone(ResolutionCorrectionValidationManagerTask)
        yield self.clone(WeightValidationWrapperTask)
        yield self.clone(FSRCorrectionValidationTask)
        yield self.clone(AnalysisPlotManagerTask)


class SelectionManagerTask(luigi.WrapperTask):
    selection_setup_names = luigi.TupleParameter(positional=False, hashed=True)  # type: Tuple[str, ...]
    config_combinations = luigi.ListParameter(positional=False, hashed=True)  # type: List[Any]
    working_dir = luigi.Parameter(positional=False)  # type: PathType
    cs_tag = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for selection_name in self.selection_setup_names:
            yield self.clone(FittingTasksWrapperTask, selection_name=selection_name)
            yield self.clone(ValidationWrapperTask, selection_name=selection_name)
            yield self.clone(EfficiencyEvaluationTask, selection_name=selection_name)
