"""
The objective of the tasks provided here are:
    - the determination of the difference in resolution of the observable missing mass squared on data and MC
    - and the application of a smearing based on this difference on the MC data sets.
The Resolution difference is obtained from the q^2 sideband.

Felix Metzner 2019
"""

import os
import itertools
import numpy as np
import pandas as pd
import multiprocessing
import b2luigi as luigi

from collections import abc
from functools import partial
from luigi.util import inherits
from typing import Union, Optional, Tuple, List, Dict, Generator, Any

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.rdstar_plots import rdstar_dec_modes, rdstar_dec_titles, get_short_dec_titles

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.utilities import InputDictType as IDType
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.validation.observable_validation import plot_bin_composition, plot_bin_migration

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.systematics.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.validation.sideband_validation import (
    plot_data_mc_comparison,
    SidebandInfo,
    reset_systematics_in_sideband_info,
)

from rdstar.offline_analysis.mms_resolution_correction.resolution_evaluation import MissingMassResolutionEvaluator
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetup,
    ResolutionCorrectionSetupManager,
)


__all__ = [
    "ResolutionEvaluationTask",
    "ResolutionCorrectionTask",
    "ResolutionCorrectionManagerTask",
    "ResolutionCorrectionValidationManagerTask",
]


class ResolutionEvaluationTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType
    selection_name = luigi.Parameter(positional=False)  # type: str
    cs_tag = luigi.Parameter(positional=False)  # type: str
    config_combinations = luigi.ListParameter(positional=False, hashed=True)
    resolution_correction_setup_name = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "ResolutionEvaluation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.main_sel_tag,
            self.resolution_correction_setup_name,
        )  # type: PathType

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"  # type: str
        self.sideband_key = "q2_sideband"  # type: str

        self.mc_sample_keys = rdstar_samples.main_sample_keys  # type: List[str]
        self.data_sample_key = "DataY4S"  # type: str

        self.res_corr_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=self.resolution_correction_setup_name
        )  # type: ResolutionCorrectionSetup

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}  # type: IDType

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"
        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{mc_key}_{WeightManager.finalized_weights_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]  # type: List[PathType]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{self.data_sample_key}_{WeightManager.finalized_weights_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        if self.res_corr_setup.variable_definition is not None:
            required_col = self.res_corr_setup.secondary_hist_var.df_label
            if required_col not in mc_sample.columns:
                self.res_corr_setup.variable_definition(df=mc_sample)
                self.res_corr_setup.variable_definition(df=data_sample)

        return mc_sample, data_sample

    def evaluate_resolution(self, dry_run: bool = False) -> Dict[str, PathType]:
        resolution_evaluator = MissingMassResolutionEvaluator(
            resolution_correction_setup_name=self.resolution_correction_setup_name,
            weight_column=WeightManager.final_weight_with_additions_col,
            results_path=self.base_path,
        )

        if dry_run:
            return resolution_evaluator.get_output_dict()

        mc_sample, data_sample = self.get_mc_and_data_sample()
        WeightManager.add_additional_weights(df=mc_sample)
        resolution_evaluator.evaluate_resolution(mc_df=mc_sample, data_df=data_sample)

        return resolution_evaluator.get_output_dict()

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.evaluate_resolution(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.evaluate_resolution(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield WeightFinalizationTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )


class ResolutionCorrectionManagerTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType
    selection_name = luigi.Parameter(positional=False)  # type: str
    cs_tag = luigi.Parameter(positional=False)  # type: str
    config_combinations = luigi.ListParameter(positional=False, hashed=True)

    results_dir_name = "Results"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(
            self.working_dir,
            ResolutionEvaluationTask.base_dir_name,
            self.main_sel_tag,
            self.results_dir_name,
        )  # type: PathType

        assert isinstance(self.input(), dict), self.input()
        self.input_dict = {
            k: v.path
            for k, d in self.input().items()
            if isinstance(d, dict)
            for kk, v in d.items()
            if kk == MissingMassResolutionEvaluator.default_result_file
        }  # type: IDType

    def run(self) -> None:
        for setup_name, path in self.input_dict.items():
            os.symlink(src=path, dst=self.create_target_path(correction_setup_name=setup_name))

    def create_target_path(self, correction_setup_name: str) -> PathType:
        file_name = f"{MissingMassResolutionEvaluator.default_result_file}_{correction_setup_name}.json"
        return os.path.join(self.base_path, file_name)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {}  # type: Dict[str, luigi.LocalTarget]
        for res_corr_setup_name in ResolutionCorrectionSetupManager.get_setup_definition_names():
            target_path = self.create_target_path(correction_setup_name=res_corr_setup_name)  # type: PathType
            output_dict.update({res_corr_setup_name: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Dict[str, utilities.TaskType]:
        required_tasks_dict = {}  # type: Dict[str, utilities.TaskType]
        for res_corr_setup_name in ResolutionCorrectionSetupManager.get_setup_definition_names():
            required_tasks_dict.update(
                {
                    res_corr_setup_name: ResolutionEvaluationTask(
                        working_dir=self.working_dir,
                        selection_name=SelectionSetups.resolution_correction_mapping[self.selection_name],
                        cs_tag=self.cs_tag,
                        config_combinations=self.config_combinations,
                        resolution_correction_setup_name=res_corr_setup_name,
                    )
                }
            )
        return required_tasks_dict


@inherits(WeightFinalizationTask)
class ResolutionCorrectionTask(utilities.BaseOfflineTask):
    appendix = "withMMSQResCorr"  # type: str

    input_data_paths_key = "input_from_weight_finalization_task"  # type: str
    input_res_corr_infos_key = "input_from_resolution_correction_manager_task"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.data_input_dict = {
            k: v.path
            for kk, d in self.input().items()
            if isinstance(d, dict) and kk == self.input_data_paths_key
            for k, v in d.items()
        }  # type: IDType

        self.res_corr_info_input_dict = {
            k: v.path
            for kk, d in self.input().items()
            if isinstance(d, dict) and kk == self.input_res_corr_infos_key
            for k, v in d.items()
        }  # type: IDType

    def get_input_files(self) -> Dict[str, PathType]:
        key_path_dict = {
            key: path
            for key, path in self.data_input_dict.items()
            if WeightManager.finalized_weights_appendix in key and os.path.splitext(path)[1] == ".h5"
        }
        assert all(key.startswith(self.combo_name) for key in key_path_dict.keys()), key_path_dict
        return key_path_dict

    def get_resolution_correction_info_path(self, res_corr_setup_name: str) -> PathType:
        return self.res_corr_info_input_dict[res_corr_setup_name]

    def apply_resolution_correction_for_key(
        self,
        key_path_tuple: Tuple[str, PathType],
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        input_key, input_path = key_path_tuple
        output_key, temp_output_path, output_path = self.get_output_key_temp_path_path_triplet(
            input_key=input_key, input_path=input_path
        )

        if dry_run:
            return output_key, output_path

        assert os.path.isfile(input_path), key_path_tuple
        sample = pd.read_hdf(input_path)
        assert isinstance(sample, pd.DataFrame)

        for setup_name, setup_definition in ResolutionCorrectionSetupManager.setup_definitions.items():
            assert setup_name in self.res_corr_info_input_dict.keys(), (setup_name, self.res_corr_info_input_dict.keys())
            original_col_name = setup_definition.primary_hist_var.df_label  # type: str
            new_col_name = MissingMassResolutionCorrector.get_corrected_col_name(correction_setup_name=setup_name)

            check_orig_name = MissingMassResolutionCorrector.get_original_col_name(correction_setup_name=setup_name)
            assert original_col_name == check_orig_name, (original_col_name, check_orig_name)
            assert original_col_name in sample.columns, original_col_name
            assert new_col_name not in sample.columns, new_col_name

            res_corrector = MissingMassResolutionCorrector(
                resolution_correction_setup_name=setup_name,
                resolution_info_path=self.get_resolution_correction_info_path(res_corr_setup_name=setup_name),
            )
            res_corrector.add_required_columns(sample=sample)

            if len(np.unique(sample.index)) == 0:
                MissingMassResolutionCorrector.add_corrected_column_to_empty_data_frame(
                    sample=sample, correction_setup_name=setup_name
                )
            elif input_key.startswith("Data"):
                MissingMassResolutionCorrector.add_corrected_column_for_data(
                    sample=sample, correction_setup_name=setup_name
                )
            else:
                res_corrector.apply_resolution_correction(sample=sample, correct_overflows=True)

        sample.to_hdf(temp_output_path, key=output_key)
        os.rename(temp_output_path, output_path)
        return output_key, output_path

    def apply_resolution_correction(self, dry_run: bool = False) -> Dict[str, PathType]:
        with multiprocessing.Pool(processes=10) as pool:
            output_tuples = pool.map(
                partial(ResolutionCorrectionTask.apply_resolution_correction_for_key, self, dry_run=dry_run),
                self.get_input_files().items(),
            )

        return {output_key: output_path for output_key, output_path in output_tuples}

    def get_output_key_temp_path_path_triplet(
        self,
        input_key: str,
        input_path: PathType,
    ) -> Tuple[str, PathType, PathType]:
        dir_path, file_name = os.path.split(input_path)
        output_name_1, output_name_2 = file_name.split(f"_{WeightManager.finalized_weights_appendix}_", 1)
        output_name = f"{output_name_1}_{WeightManager.finalized_weights_appendix}_{self.appendix}_{output_name_2}"
        output_path = os.path.join(dir_path, output_name)
        temp_output_path = os.path.join(dir_path, f"temp_{output_name}")
        output_key_1, output_key_2 = input_key.split(f"_{WeightManager.finalized_weights_appendix}_", 1)
        output_key = f"{output_key_1}_{WeightManager.finalized_weights_appendix}_{self.appendix}_{output_key_2}"

        return output_key, temp_output_path, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.apply_resolution_correction(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.apply_resolution_correction(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Dict[str, utilities.TaskType]:
        return {
            self.input_data_paths_key: self.clone(WeightFinalizationTask),
            self.input_res_corr_infos_key: ResolutionCorrectionManagerTask(
                working_dir=self.working_dir,
                selection_name=SelectionSetups.resolution_correction_mapping[self.selection_name],
                cs_tag=self.cs_tag,
                config_combinations=self.config_combinations,
            ),
        }


@inherits(ResolutionEvaluationTask)
class ResolutionCorrectionValidationTask(utilities.BaseOfflineTask):
    scope_setting = luigi.Parameter(positional=False)  # type: str
    normed_to_data = luigi.BoolParameter(positional=False)  # type: bool
    number_of_validation_bins = luigi.IntParameter(positional=False)  # type: int

    var_scope_option_name = "scope_setting"  # type: str
    n_bins_option_name = "number_of_validation_bins"  # type: str
    normed_option_name = "normed_to_data"  # type: str
    normed_options = (True, False)  # type: Tuple[bool, bool]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert self.number_of_validation_bins > 0, self.number_of_validation_bins

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )

        self.mc_sample_keys = rdstar_samples.main_sample_keys  # type: List[str]
        self.data_sample_key = "DataY4S"  # type: str

        self.__scope_tuple = None  # type: Optional[Tuple[float, float]]

        self._res_corr_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=self.resolution_correction_setup_name
        )  # type: ResolutionCorrectionSetup

        self.before_col = MissingMassResolutionCorrector.get_original_col_name(
            correction_setup_name=self.resolution_correction_setup_name
        )  # type: str
        self.after_col = MissingMassResolutionCorrector.get_corrected_col_name(
            correction_setup_name=self.resolution_correction_setup_name
        )  # type: str

        self.weight_col_name = WeightManager.final_weight_with_additions_col  # type: str

        self.sideband_key = "q2_sideband"  # type: str

        self.key_prefix = (
            f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}_"
            f"{self.selection_name}_with_CS_{self.cs_tag}_"
        )  # type: str

        self.figure_size = (8.0, 4.5)  # type: Tuple[float, float]

        sub_dir_name1 = f"validation_{self.scope_tuple_str}_in_{self.number_of_validation_bins}_bins"  # type: str
        sub_dir_name2 = "normed_to_data" if self.normed_to_data else "unscaled"  # type: str

        self.base_path = os.path.join(
            self.working_dir,
            ResolutionEvaluationTask.base_dir_name,
            self.main_sel_tag,
            self.resolution_correction_setup_name,
            sub_dir_name1,
        )  # type: PathType

        self.comparison_plots_path = os.path.join(self.base_path, sub_dir_name2)  # type: PathType
        self.composition_plots_path = os.path.join(self.base_path, "bin_composition")  # type: PathType
        self.migration_plots_path = os.path.join(self.base_path, "bin_migration")  # type: PathType

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}  # type: IDType

    @property
    def is_tight_scope_setting(self) -> bool:
        return self.scope_setting.startswith("tight")

    @property
    def scope_tuple(self) -> Tuple[float, float]:
        if self.__scope_tuple is not None:
            return self.__scope_tuple

        special_option_cases = ResolutionCorrectionSetupManager.get_special_options_for_setup(
            setup_name=self.resolution_correction_setup_name
        )  # type: Dict[Tuple[str, int], Tuple[float, float]]

        if (self.scope_setting, self.number_of_validation_bins) in special_option_cases.keys():
            self.__scope_tuple = special_option_cases[(self.scope_setting, self.number_of_validation_bins)]
        else:
            scope_options = ResolutionCorrectionSetupManager.get_scope_options_for_setup(
                setup_name=self.resolution_correction_setup_name
            )  # type: Dict[str, Tuple[float, float]]
            self.__scope_tuple = scope_options[self.scope_setting]
        return self.__scope_tuple

    @property
    def scope_tuple_str(self) -> str:
        scope_str = f"from_{self.scope_tuple[0]}_to_{self.scope_tuple[1]}"  # type: str
        scope_str = scope_str.replace(".", "p")
        scope_str = scope_str.replace("-", "minus")
        return scope_str

    def get_hist_variables(self) -> List[Tuple[HistVariable, bool]]:
        return [
            (
                HistVariable(
                    df_label=var,
                    n_bins=self.number_of_validation_bins,
                    scope=self.scope_tuple,
                    var_name=self._res_corr_setup.primary_hist_var.variable_name,
                    unit=self._res_corr_setup.primary_hist_var.unit,
                ),
                var == self.after_col,
            )
            for var in [self.before_col, self.after_col]
        ]

    def get_sideband_infos(self) -> List[Tuple[SidebandInfo, bool]]:
        title_end = f"Correction via {self._res_corr_setup.latex_title_appendix}"
        q2_sideband_title = rf"{LatexObsNames.fitting_q2} Sideband"
        q2_sideband_text = q2_sideband_title + "\n" + LatexObsNames.fitting_q2[:-1] + r" < 4.0" + LatexUnits.gev_sq[1:]
        sideband_infos = [
            (
                SidebandInfo(
                    key="q2_sideband",
                    hist_var=var,
                    title=f"{'After' if is_corrected else 'Before'} {var.variable_name} {title_end}",
                    sb_text=q2_sideband_text,
                    legend_loc=(1, 1),
                    legend_cols=(4, 1),
                    y_scale=(1.2, 1.6),
                    text_h_pos=(0.74, 0.56 if not self.is_tight_scope_setting else 0.02),
                    text_v_pos=(0.81, 0.98),
                    systematics=None,
                ),
                is_corrected,
            )
            for var, is_corrected in self.get_hist_variables()
        ]  # type: List[Tuple[SidebandInfo, bool]]

        sideband_infos_with_full_sys = [
            (
                reset_systematics_in_sideband_info(
                    sideband_info=sbi, new_systematics=full_systematics_cols, sys_name_tag=full_systematics_name
                ),
                is_corrected,
            )
            for sbi, is_corrected in sideband_infos
        ]  # type: List[Tuple[SidebandInfo, bool]]

        return sideband_infos + sideband_infos_with_full_sys

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"
        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (k == f"{mc_key}_{full_key}" and os.path.splitext(p)[1] == ".h5")
        ]  # type: List[PathType]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (k == f"{self.data_sample_key}_{full_key}" and os.path.splitext(p)[1] == ".h5")
        ]  # type: List[PathType]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        return mc_sample, data_sample

    def do_resolution_correction_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:
        if dry_run:
            mc_sample, data_sample = pd.DataFrame(), pd.DataFrame()
        else:
            mc_sample, data_sample = self.get_mc_and_data_sample()
            WeightManager.add_additional_weights(df=mc_sample)

        output_dict = {}  # type: Dict[str, PathType]

        for sideband_info, is_corrected in self.get_sideband_infos():
            for style in sideband_info.styles:
                new_output = plot_data_mc_comparison(
                    mc_df=mc_sample,
                    data_df=data_sample,
                    weight_col=self.weight_col_name,
                    sideband_info=sideband_info,
                    dec_modes=None,
                    dec_titles=get_short_dec_titles(dec_title_dict=rdstar_dec_titles),
                    export_path=self.comparison_plots_path,
                    style=style,
                    ratio_type="vs_uncert",
                    gof_check_method="toys",
                    fig_size=self.figure_size,
                    normalize_mc_to_data=self.normed_to_data,
                    adaptive_binning=True,
                    dry_run=dry_run,
                )
                assert new_output is not None
                assert all(k not in output_dict for k in new_output.keys())
                assert all(v not in output_dict.values() for v in new_output.values())
                output_dict.update(new_output)

        if self.normed_to_data:
            for main_hist_var, is_corrected in self.get_hist_variables():
                new_output_entries = plot_bin_composition(
                    mc_df=mc_sample,
                    main_hist_var=main_hist_var,
                    secondary_hist_var=self._res_corr_setup.secondary_hist_var,
                    secondary_binning=self._res_corr_setup.secondary_var_binning,
                    secondary_hist_var_label_mapping=self._res_corr_setup.tick_label_mapping,
                    weight_col=self.weight_col_name,
                    plot_shape=True,
                    dec_modes=rdstar_dec_modes,
                    dec_titles=rdstar_dec_titles,
                    file_name_suffix="corrected" if is_corrected else "uncorrected",
                    export_path=os.path.join(self.composition_plots_path, "corrected" if is_corrected else "uncorrected"),
                    dry_run=dry_run,
                )
                assert all(k not in output_dict for k in new_output_entries.keys())
                assert all(v not in output_dict.values() for v in new_output_entries.values())
                output_dict.update(new_output_entries)

            before_and_after_hist_vars = tuple(
                [v for v, _ in self.get_hist_variables()]
            )  # type: Tuple[HistVariable, ...]
            new_output_entries = plot_bin_migration(
                mc_df=mc_sample,
                from_hist_var=before_and_after_hist_vars[0],
                to_hist_var=before_and_after_hist_vars[1],
                from_to_label_appendix=("", "(Corrected)"),
                weight_col=self.weight_col_name,
                normalize_to_origin=True,
                dec_modes=rdstar_dec_modes,
                dec_titles=rdstar_dec_titles,
                file_name_suffix=None,
                show_color_bar=False,
                fig_size=(12.0, 10.0),
                export_path=self.migration_plots_path,
                dry_run=dry_run,
            )
            assert all(k not in output_dict for k in new_output_entries.keys())
            assert all(v not in output_dict.values() for v in new_output_entries.values())
            output_dict.update(new_output_entries)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_resolution_correction_evaluation(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_resolution_correction_evaluation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )

    @staticmethod
    def get_parameter_combinations(res_corr_setup_name: str) -> Tuple[Dict[str, Any], ...]:
        n_bin_options = ResolutionCorrectionSetupManager.get_n_bin_options_for_setup(setup_name=res_corr_setup_name)
        scope_options = ResolutionCorrectionSetupManager.get_scope_options_for_setup(setup_name=res_corr_setup_name)

        options = {
            ResolutionCorrectionValidationTask.n_bins_option_name: n_bin_options,
            ResolutionCorrectionValidationTask.var_scope_option_name: tuple(scope_options.keys()),
            ResolutionCorrectionValidationTask.normed_option_name: ResolutionCorrectionValidationTask.normed_options,
        }  # type: Dict[str, Tuple[Union[str, int, bool], ...]]

        value_combinations = list(itertools.product(*options.values()))  # type: List[Tuple[Any, ...]]

        assert all(isinstance(vc, tuple) for vc in value_combinations)
        assert all(len(vc) == 3 for vc in value_combinations)
        assert all(isinstance(vc[0], int) for vc in value_combinations)
        assert all(isinstance(vc[1], str) for vc in value_combinations)
        assert all(isinstance(vc[2], bool) for vc in value_combinations)

        return tuple(
            [
                dict(zip(options.keys(), vc))
                for vc in value_combinations
                if abs(scope_options[vc[1]][1] - scope_options[vc[1]][0]) / (1.0 * vc[0]) >= 0.01
            ]
        )


@inherits(ResolutionCorrectionManagerTask)
class ResolutionCorrectionValidationManagerTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for res_corr_setup_name in ResolutionCorrectionSetupManager.get_setup_definition_names():
            parameter_combinations = ResolutionCorrectionValidationTask.get_parameter_combinations(
                res_corr_setup_name=res_corr_setup_name
            )  # type: Tuple[Dict[str, Any], ...]
            for parameter_combination in parameter_combinations:
                yield self.clone(
                    ResolutionCorrectionValidationTask,
                    resolution_correction_setup_name=res_corr_setup_name,
                    **parameter_combination,
                )
