"""
Tasks performing the Sideband Validation

Felix Metzner 2019
"""

import os
import copy
import logging
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from collections import abc, defaultdict

from typing import Optional, List, Tuple, Dict, Generator

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.systematics.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.validation.sideband_validation import (
    plot_data_mc_comparison,
    sideband_info_collection,
    SidebandInfo,
    create_sideband_infos_with_systematics,
)

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.tasks.resolution_correction_task import (
    ResolutionCorrectionTask,
    ResolutionCorrectionManagerTask,
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "SidebandEvaluationWrapperTask",
    "SidebandEvaluationTask",
]


@inherits(ResolutionCorrectionManagerTask)
class SidebandEvaluationWrapperTask(luigi.WrapperTask):
    systematics_info_names = ("no_sys", full_systematics_name)  # type: Tuple[str, ...]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    @staticmethod
    def get_sorted_sideband_info_dict() -> Dict[str, List[SidebandInfo]]:
        sideband_infos = copy.copy(sideband_info_collection)  # type: List[SidebandInfo]
        sorted_sideband_info_dict = defaultdict(list)
        for sideband_info in sideband_infos:
            sorted_sideband_info_dict[sideband_info.key].append(sideband_info)

        return sorted_sideband_info_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for side_band_key in self.get_sorted_sideband_info_dict().keys():
            for systematics_info_name in self.systematics_info_names:
                yield self.clone(
                    SidebandEvaluationTask,
                    side_band_key=side_band_key,
                    systematics_info_name=systematics_info_name,
                )


@inherits(SidebandEvaluationWrapperTask)
class SidebandEvaluationTask(utilities.BaseOfflineTask):
    side_band_key = luigi.Parameter(positional=False)  # type: str
    systematics_info_name = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "SidebandValidation"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.selection_name,
        )  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.mc_sample_keys = rdstar_samples.main_sample_keys
        self.data_sample_key = "DataY4S"

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"

        self.side_band_infos = self.get_side_band_info_for_systematics()  # type: List[SidebandInfo]

        self.input_appendix = f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"
        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def get_side_band_info_for_systematics(self) -> List[SidebandInfo]:
        side_band_info_dict = SidebandEvaluationWrapperTask.get_sorted_sideband_info_dict()
        assert (
            self.systematics_info_name in SidebandEvaluationWrapperTask.systematics_info_names
        ), self.systematics_info_name

        side_band_info_list = side_band_info_dict[self.side_band_key]  # type: List[SidebandInfo]
        if self.systematics_info_name == "no_sys":
            return side_band_info_list
        elif self.systematics_info_name == full_systematics_name:
            side_band_infos_with_full_sys = create_sideband_infos_with_systematics(
                sideband_info_list=side_band_info_list,
                systematics_collection=full_systematics_cols,
                sys_name_tag=full_systematics_name,
            )  # type: List[SidebandInfo]
            return side_band_infos_with_full_sys
        else:
            raise ValueError(
                f"systematics_info_name '{self.systematics_info_name}' "
                f"is not known to SidebandEvaluationTask! "
                f"Available options are: {SidebandEvaluationWrapperTask.systematics_info_names}."
            )

    def get_mc_and_data_file_paths(self) -> Optional[Tuple[List[PathType], PathType]]:
        full_key = f"{self.key_prefix}{self.side_band_key}"

        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{mc_key}_{self.input_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]  # type: List[PathType]

        if len(mc_paths) == 0:
            return None

        assert len(mc_paths) == len(self.mc_sample_keys), (
            len(mc_paths),
            mc_paths,
            self.mc_sample_keys,
            full_key,
            "\n" + "\n\t- ".join([f"{k}: {v}" for k, v in self.input_dict.items()]) + "\n",
            "\nEntries matching conditions:\n"
            + "\n\t- ".join(
                [
                    f"{k}: {v}"
                    for k, v in self.input_dict.items()
                    if k.endswith(full_key) and os.path.splitext(v)[1] == ".h5"
                ]
            )
            + "\n",
        )

        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{self.data_sample_key}_{self.input_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        return mc_paths, data_path

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        mc_and_data_paths = self.get_mc_and_data_file_paths()
        assert mc_and_data_paths is not None
        mc_paths, data_path = mc_and_data_paths  # type: List[PathType], PathType

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame)

        return mc_sample, data_sample

    @property
    def sideband_samples_are_available(self) -> bool:
        paths = self.get_mc_and_data_file_paths()
        return False if paths is None else True

    def produce_sideband_plot(
        self,
        mc_df: Optional[pd.DataFrame],
        data_df: Optional[pd.DataFrame],
        input_tuple: Tuple[SidebandInfo, str],
        dry_run: bool,
    ) -> Dict[str, PathType]:
        sideband_info, style = input_tuple
        assert sideband_info.key == self.side_band_key, (sideband_info.key, self.side_band_key)

        output_path = os.path.join(self.base_path, sideband_info.hist_var.df_label, self.systematics_info_name)

        if not dry_run:
            assert isinstance(mc_df, pd.DataFrame), type(mc_df)
            assert isinstance(data_df, pd.DataFrame), type(data_df)
            os.makedirs(name=output_path, exist_ok=True)

        plot_output_dict = plot_data_mc_comparison(
            mc_df=mc_df,
            data_df=data_df,
            dry_run=dry_run,
            weight_col=WeightManager.final_weight_with_additions_col,
            sideband_info=sideband_info,
            style=style,
            normalize_mc_to_data=True,
            ratio_type="vs_uncert",
            gof_check_method="toys",
            adaptive_binning=True,
            export_path=output_path,
        )
        assert plot_output_dict is not None

        return plot_output_dict

    def do_sideband_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        if not self.sideband_samples_are_available:
            return output_dict

        mc_df, data_df = (None, None)
        if not dry_run:
            mc_df, data_df = self.get_mc_and_data_sample()
            WeightManager.add_additional_weights(df=mc_df)

        sb_plot_infos = [(sb_info, style) for sb_info in self.side_band_infos for style in sb_info.styles]
        assert all(sb_info.key == self.side_band_key for sb_info, _ in sb_plot_infos), (
            self.side_band_key,
            [sbi.key for sbi, _ in sb_plot_infos],
        )

        for sb_plot_info_tuple in sb_plot_infos:
            eval_output = self.produce_sideband_plot(
                mc_df=mc_df,
                data_df=data_df,
                input_tuple=sb_plot_info_tuple,
                dry_run=dry_run,
            )

            assert all(k not in output_dict.keys() for k in eval_output.keys()), [
                k for k in eval_output.keys() if k in output_dict.keys()
            ]
            assert all(v not in output_dict.values() for v in eval_output.values()), [
                v for v in eval_output.values() if v in output_dict.values()
            ]
            output_dict.update(eval_output)

        assert len(output_dict) % len(sb_plot_infos) == 0, (len(output_dict), len(sb_plot_infos))
        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_sideband_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_sideband_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def complete(self):
        # Catching special case of a SidebandEvaluationTask which cannot produce output.
        if isinstance(self.output(), dict) and len(self.output()) == 0:
            logging.warning(
                f"SidebandEvaluationTask cannot produce output for "
                f"selection name {self.selection_name} and sideband key {self.side_band_key}!"
            )
            return True
        return super().complete()

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )
