"""
Weight Evaluation and Finalization Tasks

Felix Metzner 2019
"""

import os
import b2luigi as luigi

from luigi.util import inherits
from collections import Iterable

from typing import Tuple, List, Dict, Generator

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.combiner_tasks import SampleCombinerTask
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.weight_management.weight_analysis import WeightInfoGatherer
from rdstar.offline_analysis.weight_management.weight_finalization import WeightFinalizingManager, WeightFinalizer
from rdstar.offline_analysis.weight_management.weight_evaluation import (
    FFWeightEvaluator,
    PIDWeightEvaluator,
    PIDWeightEvaluationPreparer,
)


__all__ = [
    "WeightEvaluationTask",
    "WeightFinalizationTask",
    "WeightInfoGathererTask",
    "WeightValidationWrapperTask",
]


class WeightValidationWrapperTask(luigi.WrapperTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType
    selection_name = luigi.Parameter(positional=False)  # type: str
    cs_tag = luigi.Parameter(positional=False)  # type: str
    config_combinations = luigi.ListParameter(positional=False, hashed=True)

    selection_options = [True, False]  # type: List[bool]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.weight_keys = WeightManager.weight_info_keys  # type: List[str]

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for weight_key in self.weight_keys:
            for with_selection in self.selection_options:
                if weight_key == WeightManager.pid_weight_info_key:
                    yield self.clone(PIDWeightEvaluationManagerTask, with_selection=with_selection)
                else:
                    yield self.clone(WeightEvaluationTask, weight_key=weight_key, with_selection=with_selection)


@inherits(WeightValidationWrapperTask)
class WeightInfoGathererTask(utilities.BaseOfflineTask):
    weight_key = luigi.Parameter(positional=False)  # type: str

    base_dir_name = WeightManager.weight_analysis_base_dir_name  # type: str

    @property
    def resources(self) -> Dict[str, int]:
        if self.weight_key == WeightManager.pid_weight_info_key:
            return {utilities.offline_memory_resource_name: 9}
        else:
            return {utilities.offline_memory_resource_name: 3}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.weight_key, str), self.weight_key
        assert isinstance(self.selection_name, str), self.selection_name
        assert isinstance(self.cs_tag, str), self.cs_tag

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.main_sel_tag)  # type: PathType

        self.input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

        self.weight_info_gatherer = WeightInfoGatherer(
            weight_key=self.weight_key,
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
            input_dictionary=self.input_dict,
            output_path=self.base_path,
        )

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def gather_weight_infos(
        self,
        dry_run: bool = False,
    ) -> Generator[Tuple[str, PathType], None, None]:
        for output_tuple in self.weight_info_gatherer.gather_weight_info(dry_run=dry_run):
            yield output_tuple

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        list(self.gather_weight_infos())

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.gather_weight_infos(dry_run=True):
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, Iterable)
        for combo_name, _, relevant_configs in self.config_combinations:
            yield SampleCombinerTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                config_params=relevant_configs,
            )


@inherits(WeightInfoGathererTask)
class WeightEvaluationTask(utilities.BaseOfflineTask):
    with_selection = luigi.BoolParameter(positional=False)  # type: bool

    base_dir_name = WeightManager.weight_analysis_base_dir_name  # type: str
    plots_dir_name = "plots"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.with_selection, bool), type(self.with_selection)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        with_sel_tag = WeightManager.with_selection_suffix if self.with_selection else ""  # type: str
        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.main_sel_tag,
            self.plots_dir_name,
            f"{self.weight_key}{with_sel_tag}",
        )  # type: PathType

        input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]
        df_key = WeightManager.get_weight_output_key(
            weight_key=self.weight_key,
            with_selection=self.with_selection,
        )  # type: str
        df_path = input_dict[df_key]  # type: PathType

        self.weight_info_evaluator = FFWeightEvaluator(
            weight_key=self.weight_key,
            input_df_path=df_path,
            output_dir_path=self.base_path,
            with_preselection=self.with_selection,
        )

    def produce_weight_info_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        return self.weight_info_evaluator.produce_weight_plots(dry_run=dry_run)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.produce_weight_info_plots()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.produce_weight_info_plots(dry_run=True)  # type: Dict[str, PathType]
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(WeightInfoGathererTask)


@inherits(WeightValidationWrapperTask)
class PIDWeightEvaluationManagerTask(luigi.WrapperTask):
    with_selection = luigi.BoolParameter(positional=False)  # type: bool

    base_dir_name = WeightManager.weight_analysis_base_dir_name  # type: str

    @property
    def resources(self) -> Dict[str, int]:
        if self.with_selection:
            return {utilities.offline_memory_resource_name: 4}
        else:
            return {utilities.offline_memory_resource_name: 8}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.main_sel_tag)  # type: PathType

        prerequisites_input_dict = {
            k: v.path for d in self.prerequisites_input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]
        key = PIDWeightEvaluationPreparer.get_weight_info_container_output_key(
            with_selection=self.with_selection
        )  # type: str
        assert key in prerequisites_input_dict.keys(), (key, prerequisites_input_dict.keys())
        self.pid_weight_info_container_path = prerequisites_input_dict[key]  # type: PathType

    def get_pid_weight_info_keys(self) -> Tuple[str, ...]:
        if not os.path.isfile(self.pid_weight_info_container_path):
            return tuple([])
        pid_weight_infos = PIDWeightEvaluationPreparer.load_weight_info_container(
            file_path=self.pid_weight_info_container_path,
        )
        assert len(pid_weight_infos.keys()) > 0, (len(pid_weight_infos.keys()), pid_weight_infos.keys())
        return tuple(pid_weight_infos.keys())

    def get_log_file_dir(self) -> PathType:
        return self.base_path

    def get_task_file_dir(self) -> PathType:
        return self.base_path

    def prerequisites_input(self) -> List[Dict[str, luigi.LocalTarget]]:
        return luigi.task.getpaths(list(self.prerequisites()))

    def prerequisites(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(PIDWeightEvaluationInfoSplitterTask)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for prerequisite in self.prerequisites():
            yield prerequisite

        for pid_weight_info_key in self.get_pid_weight_info_keys():
            assert isinstance(pid_weight_info_key, str), (type(pid_weight_info_key).__name__, pid_weight_info_key)
            yield self.clone(PIDWeightEvaluationTask, pid_weight_key=pid_weight_info_key)


@inherits(PIDWeightEvaluationManagerTask)
class PIDWeightEvaluationInfoSplitterTask(utilities.BaseOfflineTask):

    base_dir_name = WeightManager.weight_analysis_base_dir_name  # type: str
    pid_weight_info_container_dir_name = "split_pid_weight_infos"  # type: str

    @property
    def resources(self) -> Dict[str, int]:
        if self.with_selection:
            return {utilities.offline_memory_resource_name: 2}
        else:
            return {utilities.offline_memory_resource_name: 4}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        with_sel_tag = WeightManager.with_selection_suffix if self.with_selection else ""  # type: str
        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.main_sel_tag,
            f"{self.pid_weight_info_container_dir_name}{with_sel_tag}",
        )  # type: PathType

        input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

        self.pid_weight_info_preparer = PIDWeightEvaluationPreparer(
            input_dict=input_dict,
            output_dir_path=self.base_path,
            with_selection=self.with_selection,
        )

    def split_pid_weight_infos(self) -> None:
        return self.pid_weight_info_preparer.prepare_pid_infos()

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.split_pid_weight_infos()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_key, target_path = PIDWeightEvaluationPreparer.get_weight_info_container_output_key_path_pair(
            output_dir_path=self.base_path,
            with_selection=self.with_selection,
        )
        return {output_key: luigi.LocalTarget(target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(WeightInfoGathererTask, weight_key=WeightManager.pid_weight_info_key)


@inherits(PIDWeightEvaluationManagerTask)
class PIDWeightEvaluationTask(utilities.BaseOfflineTask):
    pid_weight_key = luigi.Parameter(positional=False)  # type: str

    base_dir_name = WeightManager.weight_analysis_base_dir_name  # type: str
    plots_dir_name = f"{WeightManager.pid_weight_info_key}_{WeightEvaluationTask.plots_dir_name}"  # type: str

    @property
    def resources(self) -> Dict[str, int]:
        return {utilities.offline_memory_resource_name: 2}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self.with_selection, bool), type(self.with_selection)
        assert isinstance(self.pid_weight_key, str), (type(self.pid_weight_key).__name__, self.pid_weight_key)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        with_sel_tag = WeightManager.with_selection_suffix if self.with_selection else ""  # type: str
        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.main_sel_tag,
            self.plots_dir_name,
            f"{self.pid_weight_key}{with_sel_tag}",
        )  # type: PathType

        input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        key = PIDWeightEvaluationPreparer.get_weight_info_container_output_key(with_selection=self.with_selection)
        assert key in input_dict.keys(), (key, input_dict.keys())
        info_container_path = input_dict[key]  # type: PathType

        pid_weight_infos = PIDWeightEvaluationPreparer.load_weight_info_container(file_path=info_container_path)
        pid_weight_info = pid_weight_infos[self.pid_weight_key]

        self.weight_info_evaluator = PIDWeightEvaluator(
            weight_key=WeightManager.pid_weight_info_key,
            input_df_path=pid_weight_info.hdf5_path,
            output_dir_path=self.base_path,
            with_preselection=self.with_selection,
            pid_weight_key=self.pid_weight_key,
            pid_weight_info=pid_weight_info,
            pid_data_scaling=None,
        )

    def produce_weight_info_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        return self.weight_info_evaluator.produce_weight_plots(dry_run=dry_run)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.produce_weight_info_plots()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.produce_weight_info_plots(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(PIDWeightEvaluationInfoSplitterTask)


@inherits(SampleCombinerTask)
class WeightFinalizationTask(utilities.BaseOfflineTask):
    cs_tag = luigi.Parameter(positional=False)  # type: str
    config_combinations = luigi.ListParameter(positional=False, hashed=True)

    sample_combiner_key = "sample_combiner"  # type: str
    weight_analysis_key = "weight_analysis"  # type: str
    weight_finalizer_key = "weight_finalizer"  # type: str

    in_keys = [sample_combiner_key, weight_analysis_key]  # type: List[str]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        prerequisites_input_dict = {
            k: v.path
            for input_dict in self.prerequisites_input()
            if isinstance(input_dict, dict)
            for key, d in input_dict.items()
            if isinstance(d, dict) and key in self.in_keys
            for k, v in d.items()
        }  # type: Dict[str, PathType]

        self.finalization_manager = WeightFinalizingManager(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
            input_dict=prerequisites_input_dict,
        )

        self._pid_dict_path = str(self.finalization_manager.get_flag_info_dict_path())  # type: str

    def gather_finalized_samples(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]
        for input_key, input_path in self.finalization_manager.get_input_key_path_pairs():
            output_key = WeightFinalizer.get_output_key(input_key=input_key)
            output_path = WeightFinalizer.get_output_path(input_path=input_path)
            assert output_key not in output_dict.keys(), output_key
            assert output_path not in output_dict.values(), output_path

            if not dry_run:
                input_path = WeightFinalizer.get_output_path(input_path=input_path, intermediate=True)
                os.symlink(src=input_path, dst=output_path)

            output_dict.update({output_key: output_path})
        return output_dict

    def run(self) -> None:
        self.gather_finalized_samples()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {k: luigi.LocalTarget(p) for k, p in self.gather_finalized_samples(dry_run=True).items()}

    def prerequisites_input(self) -> List[Dict[str, luigi.LocalTarget]]:
        return luigi.task.getpaths(list(self.prerequisites()))

    def prerequisites(self) -> Generator[Dict[str, utilities.TaskType], None, None]:
        yield {
            self.sample_combiner_key: self.clone(
                SampleCombinerTask,
                combo_name=self.combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                config_params=self.config_params,
            )
        }

        yield {
            self.weight_analysis_key: self.clone(
                WeightInfoGathererTask,
                weight_key=WeightManager.pid_weight_info_key,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                config_combinations=self.config_combinations,
            )
        }

    def requires(self) -> Generator[Dict[str, utilities.TaskType], None, None]:
        for prerequisite in self.prerequisites():
            yield prerequisite

        for input_key, input_path in self.finalization_manager.get_input_key_path_pairs():
            yield {
                self.weight_finalizer_key: self.clone(
                    WeightFinalizationApplicationTask,
                    file_input_key=input_key,
                    file_input_path=input_path,
                    pid_dict_path=self._pid_dict_path,
                )
            }


@inherits(WeightFinalizationTask)
class WeightFinalizationApplicationTask(utilities.BaseOfflineTask):
    file_input_key = luigi.Parameter(positional=False)  # type: str
    file_input_path = luigi.Parameter(positional=False)  # type: str
    pid_dict_path = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

        self.weight_finalizer = WeightFinalizer(
            input_key=self.file_input_key,
            input_path=self.file_input_path,
            pid_dict_path=self.pid_dict_path,
        )

    def apply_weight_finalization(self, dry_run: bool = False) -> Tuple[str, PathType]:
        return self.weight_finalizer.apply_weight_finalization(combo_name=self.combo_name, dry_run=dry_run)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.apply_weight_finalization()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_key, target_path = self.apply_weight_finalization(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(
            SampleCombinerTask,
            combo_name=self.combo_name,
            working_dir=self.working_dir,
            selection_name=self.selection_name,
            config_params=self.config_params,
        )

        yield self.clone(
            WeightInfoGathererTask,
            weight_key=WeightManager.pid_weight_info_key,
            working_dir=self.working_dir,
            selection_name=self.selection_name,
            config_combinations=self.config_combinations,
        )
