"""
Analysis Flow Validation -- Confusion Matrices

Felix Metzner 2019
"""

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from typing import NamedTuple, Union, Optional, Callable, Tuple, List, Dict, Any

from templatefitter.plotter.plot_utilities import FigureType, AxesType

from rdstar.offline_analysis import mc_ids
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.rdstar_plots import decay_mode_id_dict, u4s_decay_mode_id_dict, rdstar_color_map


__all__ = [
    "ConfusionMatrixInfo",
    "confusion_matrix_info_collection",
    "create_confusion_matrix_plot",
    "get_confusion_matrix",
]


def gen_id_formatter(label: Union[str, int, float]) -> str:
    return mc_ids.gen_id_to_label[int(label)]


def per_gen_id_formatter(label: Union[str, int, float]) -> str:
    assert isinstance(label, str), (label, type(label).__name__)
    return mc_ids.gen_id_to_label[int(label.rsplit("_", 1)[1])]


def sig_id_formatter(label: Union[str, int, float]) -> str:
    return mc_ids.sig_id_to_label[int(label)]


def per_sig_id_formatter(label: Union[str, int, float]) -> str:
    assert isinstance(label, str), (label, type(label).__name__)
    return mc_ids.sig_id_to_label[int(label.rsplit("_", 1)[1])]


def decay_mode_id_formatter(label: Union[str, int, float]) -> str:
    return decay_mode_id_dict()[int(label)]


def u4s_mode_id_formatter(label: Union[str, int, float]) -> str:
    return u4s_decay_mode_id_dict()[int(label)]


class ConfusionMatrixInfo(NamedTuple):
    name: str
    base_info: str
    norm_info: str
    norm_y: bool = True
    x_label: Optional[str] = None
    y_label: Optional[str] = None
    title: Optional[str] = None
    x_format_func: Optional[Callable] = None
    y_format_func: Optional[Callable] = None
    transpose: bool = False
    remove_keys: Optional[List[str]] = None


default_remove_keys = [
    *rdstar_samples.alternative_gap_samples,
    *rdstar_samples.efficiency_samples,
    *rdstar_samples.mva_samples,
    *rdstar_samples.off_resonance_mc_samples,
    *rdstar_samples.unused_samples,
]  # type: List[str]


confusion_matrix_info_collection = [
    # Gen ID vs Sig ID
    ConfusionMatrixInfo(
        name="gen_id_vs_sig_id",
        base_info="gen_id_per_sig_id_",
        norm_info="gen_id_counts",
        norm_y=True,
        x_label="Signal ID",
        y_label="Generator ID",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=gen_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_sig_id",
        base_info="u_gen_id_per_sig_id_",
        norm_info="gen_id_counts_unique",
        norm_y=True,
        x_label="Signal ID",
        y_label="Generator ID",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=gen_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    # Data Type vs Sig ID
    ConfusionMatrixInfo(
        name="data_type_all_vs_sig_id",
        base_info="data_type_per_sig_id_",
        norm_info="data_type_counts",
        norm_y=True,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_all_vs_sig_id",
        base_info="u_data_type_per_sig_id_",
        norm_info="data_type_counts_unique",
        norm_y=True,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="data_type_vs_sig_id",
        base_info="data_type_per_sig_id_",
        norm_info="data_type_counts",
        norm_y=True,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_vs_sig_id",
        base_info="u_data_type_per_sig_id_",
        norm_info="data_type_counts_unique",
        norm_y=True,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="data_type_all_vs_sig_id_n_all",
        base_info="data_type_per_sig_id_",
        norm_info="n_candidates",
        norm_y=True,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_all_vs_sig_id_n_all",
        base_info="u_data_type_per_sig_id_",
        norm_info="n_events",
        norm_y=False,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="data_type_vs_sig_id_n_all",
        base_info="data_type_per_sig_id_",
        norm_info="n_candidates",
        norm_y=True,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_vs_sig_id_n_all",
        base_info="u_data_type_per_sig_id_",
        norm_info="n_events",
        norm_y=False,
        x_label="Signal ID",
        y_label="Data Type",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    # Gen ID vs Data Type
    ConfusionMatrixInfo(
        name="gen_id_vs_data_type_all",
        base_info="gen_id_per_data_type",
        norm_info="gen_id_counts",
        norm_y=True,
        x_label="Generator ID",
        y_label="Data Type",
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_data_type_all",
        base_info="gen_id_per_data_type_unique",
        norm_info="gen_id_counts_unique",
        norm_y=True,
        x_label="Generator ID",
        y_label="Data Type",
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="gen_id_vs_data_type",
        base_info="gen_id_per_data_type",
        norm_info="gen_id_counts",
        norm_y=True,
        x_label="Generator ID",
        y_label="Data Type",
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_data_type",
        base_info="gen_id_per_data_type_unique",
        norm_info="gen_id_counts_unique",
        norm_y=True,
        x_label="Generator ID",
        y_label="Data Type",
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=default_remove_keys,
    ),
    # Decay Mode ID vs Sig ID
    ConfusionMatrixInfo(
        name="decay_mode_id_vs_sig_id",
        base_info="decay_id_vs_sig_id_",
        norm_info="events_per_decay_mode_id",
        norm_y=True,
        x_label="Signal ID",
        y_label="Decay Mode ID",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_decay_mode_id_vs_sig_id",
        base_info="u_decay_id_vs_sig_id_",
        norm_info="unique_events_per_decay_mode_id",
        norm_y=True,
        x_label="Signal ID",
        y_label="Decay Mode ID",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="decay_mode_id_vs_sig_id_n",
        base_info="decay_id_vs_sig_id_",
        norm_info="sig_id_counts",
        norm_y=False,
        x_label="Signal ID",
        y_label="Decay Mode ID",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_decay_mode_id_vs_sig_id_n",
        base_info="u_decay_id_vs_sig_id_",
        norm_info="sig_id_counts_unique",
        norm_y=False,
        x_label="Signal ID",
        y_label="Decay Mode ID",
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    # Sig ID vs Decay Mode ID
    ConfusionMatrixInfo(
        name="sig_id_vs_decay_mode_id",
        base_info="decay_id_vs_sig_id_",
        norm_info="sig_id_counts",
        norm_y=False,
        x_label="Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_decay_mode_id",
        base_info="u_decay_id_vs_sig_id_",
        norm_info="sig_id_counts_unique",
        norm_y=False,
        x_label="Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="sig_id_vs_decay_mode_id_n",
        base_info="decay_id_vs_sig_id_",
        norm_info="events_per_decay_mode_id",
        norm_y=True,
        x_label="Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_decay_mode_id_n",
        base_info="u_decay_id_vs_sig_id_",
        norm_info="unique_events_per_decay_mode_id",
        norm_y=True,
        x_label="Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Gen ID vs Decay Mode ID
    ConfusionMatrixInfo(
        name="gen_id_vs_decay_mode_id",
        base_info="decay_id_vs_gen_id_",
        norm_info="gen_id_counts",
        norm_y=False,
        x_label="Decay Mode ID",
        y_label="Generator ID",
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_gen_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_decay_mode_id",
        base_info="u_decay_id_vs_gen_id_",
        norm_info="gen_id_counts_unique",
        norm_y=False,
        x_label="Decay Mode ID",
        y_label="Generator ID",
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_gen_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Sig ID vs Tag Decay Mode ID
    ConfusionMatrixInfo(
        name="sig_id_vs_tag_mode_id",
        base_info="tag_id_vs_sig_id_",
        norm_info="sig_id_counts",
        norm_y=False,
        x_label="Tag Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=None,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_tag_mode_id",
        base_info="u_tag_id_vs_sig_id_",
        norm_info="sig_id_counts_unique",
        norm_y=False,
        x_label="Tag Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=None,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Sig ID vs U4S Decay Mode ID
    ConfusionMatrixInfo(
        name="sig_id_vs_u4s_mode_id",
        base_info="u4s_id_vs_sig_id_",
        norm_info="sig_id_counts",
        norm_y=False,
        x_label=r"$\Upsilon$(4S) Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=u4s_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_u4s_mode_id",
        base_info="u_u4s_id_vs_sig_id_",
        norm_info="sig_id_counts_unique",
        norm_y=False,
        x_label=r"$\Upsilon$(4S) Decay Mode ID",
        y_label="Signal ID",
        title=None,
        x_format_func=u4s_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Tag Decay Mode ID
    # U4S Decay Mode ID
]  # type: List[ConfusionMatrixInfo]


def create_confusion_matrix_plot(
    confusion_matrix: pd.DataFrame,
    confusion_matrix_info: ConfusionMatrixInfo,
) -> Tuple[FigureType, AxesType]:
    if confusion_matrix_info.transpose:
        confusion_matrix = confusion_matrix.transpose()

    fig_size = (10.0, 4.5)  # type: Tuple[float, float]
    cb_fraction = 0.15
    if len(confusion_matrix.columns) > 15 or len(confusion_matrix.index) > 10:
        new_x = fig_size[0]
        new_y = fig_size[1]
        if len(confusion_matrix.columns) > 15:
            new_x = (1.0 + (len(confusion_matrix.columns) - 15) * 1.0 / 15) * fig_size[0]
        if len(confusion_matrix.index) > 10:
            new_y = (1.0 + (len(confusion_matrix.index) - 10) * 1.0 / 10) * fig_size[1]
            cb_fraction *= new_x / new_y
        fig_size = (new_x, new_y)
    fig = plt.figure(num=None, figsize=fig_size, dpi=100, facecolor="w", edgecolor="k")

    ax = plt.gca()

    # Alternative to seaborn heatmap:
    # Maybe switch from seaborn heatmap to matplotlib implementation, e.g. from here:
    # https://matplotlib.org/3.1.1/gallery/images_contours_and_fields/image_annotated_heatmap.html
    ax = sns.heatmap(
        data=confusion_matrix,
        cmap=rdstar_color_map,
        mask=confusion_matrix == 0.0,
        annot=True,
        ax=ax,
        xticklabels=[confusion_matrix_info.x_format_func(t) for t in confusion_matrix.columns]
        if confusion_matrix_info.x_format_func is not None
        else "auto",
        yticklabels=[confusion_matrix_info.y_format_func(t) for t in list(confusion_matrix.index)]
        if confusion_matrix_info.y_format_func is not None
        else "auto",
        cbar_kws={"fraction": cb_fraction},
    )

    ax.set_yticklabels(ax.get_yticklabels(), rotation=0, va="center", rotation_mode="anchor")
    ax.set_xticklabels(ax.get_xticklabels(), rotation=-35, ha="left", rotation_mode="anchor")

    if confusion_matrix_info.title is not None:
        ax.set_title(label=confusion_matrix_info.title, loc="center")
    if confusion_matrix_info.x_label is not None:
        ax.set_xlabel(confusion_matrix_info.x_label)
    if confusion_matrix_info.y_label is not None:
        ax.set_ylabel(confusion_matrix_info.y_label)

    ax.tick_params(axis="both", which="major", bottom=True, top=False, left=True, right=False, direction="out")
    ax.tick_params(axis="both", which="minor", bottom=False, top=False, left=False, right=False)

    return fig, ax


def get_confusion_matrix(
    sub_info_dict: Dict[str, Any],
    confusion_matrix_info: ConfusionMatrixInfo,
) -> pd.DataFrame:
    assert isinstance(sub_info_dict, dict)

    norm_key_identifier = confusion_matrix_info.norm_info
    key_identifier = confusion_matrix_info.base_info
    assert norm_key_identifier in sub_info_dict, (norm_key_identifier, sub_info_dict.keys())

    if not key_identifier.endswith("_"):
        assert norm_key_identifier != key_identifier
        assert isinstance(sub_info_dict[key_identifier], dict)
        assert all(isinstance(v, dict) for v in sub_info_dict[key_identifier].values())
        assert all(isinstance(vv, int) for v in sub_info_dict[key_identifier].values() for vv in v.values())
        return get_conf_matrix_from_one_dict(
            sub_info_dict=sub_info_dict,
            norm_key_identifier=norm_key_identifier,
            key=key_identifier,
            norm_over_y=confusion_matrix_info.norm_y,
            remove_keys=confusion_matrix_info.remove_keys,
        )
    else:
        keys = [k for k in sub_info_dict.keys() if k.startswith(key_identifier)]
        if confusion_matrix_info.remove_keys is not None:
            assert isinstance(confusion_matrix_info.remove_keys, list)
            keys = [k for k in keys if k not in confusion_matrix_info.remove_keys]
        assert len(keys) > 0 and norm_key_identifier not in keys
        assert all(isinstance(sub_info_dict[k], dict) for k in keys)
        assert all(isinstance(v, int) for k in keys for v in sub_info_dict[k].values())
        return get_conf_matrix_from_multiple(
            sub_info_dict=sub_info_dict,
            norm_key_identifier=norm_key_identifier,
            keys=keys,
            norm_over_y=confusion_matrix_info.norm_y,
            remove_keys=confusion_matrix_info.remove_keys,
        )


def get_conf_matrix_from_multiple(
    sub_info_dict: Dict[str, Any],
    norm_key_identifier: str,
    keys: List[str],
    norm_over_y: bool,
    remove_keys: Optional[List[str]],
) -> pd.DataFrame:
    if remove_keys is None:
        remove_keys = []
    sub_keys = list(set([kk for k in keys for kk in sub_info_dict[k].keys() if kk not in remove_keys]))
    original = sub_info_dict[norm_key_identifier]
    if isinstance(original, int):
        result = {
            k: {kk: 1.0 * sub_info_dict[k][kk] / original if kk in sub_info_dict[k].keys() else 0.0 for kk in sub_keys}
            for k in keys
        }
    elif isinstance(original, dict):
        assert all(isinstance(v, int) for v in original.values())
        if norm_over_y:
            assert len(original) >= len(sub_keys), (original, sub_keys)
            assert all(k in original for k in sub_keys)
            assert all(all(sub_info_dict[k][kk] <= original[kk] for kk in sub_info_dict[k]) for k in keys)
            result = {
                k: {
                    kk: 1.0 * sub_info_dict[k][kk] / original[kk] if kk in sub_info_dict[k].keys() else 0.0
                    for kk in sub_keys
                }
                for k in keys
            }
        else:
            r_keys = [k for k in keys if sub_info_dict[k]]
            assert len(original) >= len(r_keys), (original, r_keys)
            assert all(get_norm_key_for_inverted(k) in original for k in r_keys)
            assert all(
                all(sub_info_dict[k][kk] <= original[get_norm_key_for_inverted(k)] for kk in sub_info_dict[k])
                for k in r_keys
            )
            result = {
                k: {
                    kk: 1.0 * sub_info_dict[k][kk] / original[get_norm_key_for_inverted(k)]
                    if kk in sub_info_dict[k].keys()
                    else 0.0
                    for kk in sub_keys
                }
                for k in r_keys
            }
    else:
        raise ValueError("Unexpected type retrieved for normalization.")

    return pd.DataFrame(result)


def get_conf_matrix_from_one_dict(
    sub_info_dict: Dict[str, Any],
    norm_key_identifier: str,
    key: str,
    norm_over_y: bool,
    remove_keys: Optional[List[str]],
) -> pd.DataFrame:
    if remove_keys is None:
        remove_keys = []
    keys = [k for k in sub_info_dict[key].keys() if k not in remove_keys]
    sub_keys = list(set([kk for k in keys for kk in sub_info_dict[key][k].keys() if kk not in remove_keys]))
    original = sub_info_dict[norm_key_identifier]

    if isinstance(original, int):
        result = {
            k: {
                kk: 1.0 * sub_info_dict[key][k][kk] / original if kk in sub_info_dict[key][k].keys() else 0.0
                for kk in sub_keys
            }
            for k in keys
        }
    elif isinstance(original, dict):
        assert all(isinstance(v, int) for v in original.values())
        if norm_over_y:
            assert len(original) >= len(sub_keys), (original, sub_keys)
            assert all(k in original for k in sub_keys)
            assert all(all(sub_info_dict[key][k][kk] <= original[kk] for kk in sub_info_dict[key][k]) for k in keys)
            result = {
                k: {
                    kk: 1.0 * sub_info_dict[key][k][kk] / original[kk] if kk in sub_info_dict[key][k].keys() else 0.0
                    for kk in sub_keys
                }
                for k in keys
            }
        else:
            r_keys = [k for k in keys if sub_info_dict[k]]
            assert len(original) >= len(r_keys), (original, r_keys)
            assert all(k in original for k in r_keys)
            assert all(all(sub_info_dict[key][k][kk] <= original[k] for kk in sub_info_dict[key][k]) for k in r_keys)
            result = {
                k: {
                    kk: 1.0 * sub_info_dict[key][k][kk] / original[k] if kk in sub_info_dict[key][k].keys() else 0.0
                    for kk in sub_keys
                }
                for k in r_keys
            }
    else:
        raise ValueError("Unexpected type retrieved for normalization.")

    return pd.DataFrame(result)


def get_norm_key_for_inverted(key: str) -> int:
    return int(key.rsplit("_", 1)[1])
