"""
FSR Correction Validation

Patrick Ecker 2020
"""

import os
import pandas as pd
import matplotlib.pyplot as plt

from typing import List, Tuple, Union, Optional, NamedTuple, Dict

from templatefitter.plotter.plot_style import KITColors
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot

from rdstar.utilities import PathType
from rdstar.offline_analysis.histogram_variables import missing_mass, pl_star
from rdstar.latex_labels.latex_particle_names import LatexParticleNames

__all__ = [
    "FSRVariable",
    "fsr_missing_mass",
    "fsr_pl_star",
    "FSRCorrectionValidator",
]


class FSRVariable(NamedTuple):
    variable: HistVariable  # the corresponding Histogram Variable
    fsr_column: str  # the column with the fsr correction applied
    no_fsr_column: str  # the column without the fsr correction applied


fsr_missing_mass = FSRVariable(
    variable=missing_mass,
    fsr_column="m2RecoilSignalSide",
    no_fsr_column="m2RecoilSignalSideNoFSRCorrection",
)

fsr_pl_star = FSRVariable(
    variable=pl_star,
    fsr_column="pStarL",
    no_fsr_column="pStarLNoFSR",
)


class FSRCorrectionValidator:
    default_fig_size = (4, 4)  # type: Tuple[Union[int, float], Union[int, float]]
    default_fig_dpi = 200  # type: int

    def __init__(
        self,
        list_of_df_paths: List[PathType],
        base_output_path: PathType,
        fsr_comparison_vars: List[FSRVariable],
        fsr_vars: List[HistVariable],
        efficiency_dict: Dict[str, Dict[str, float]],
        selection_name: str,
    ) -> None:
        self._list_of_df_paths = list_of_df_paths
        self._base_output_path = base_output_path
        self._fsr_comparison_vars = fsr_comparison_vars
        self._fsr_vars = fsr_vars
        self._efficiency_dict = efficiency_dict  # type: Dict[str, Dict[str, float]]
        self._selection_name = selection_name

    @property
    def figure_size(self) -> Tuple[Union[int, float], Union[int, float]]:
        return self.default_fig_size

    @property
    def figure_dpi(self) -> int:
        return self.default_fig_dpi

    def _get_df_mc(self, dry_run: bool = False) -> pd.DataFrame:
        if dry_run:
            return pd.DataFrame()
        else:
            dfs = []  # type: List[pd.DataFrame]
            for path in self._list_of_df_paths:
                df = pd.read_hdf(path_or_buf=path)
                assert isinstance(df, pd.DataFrame)
                dfs.append(df)
            df_mc = pd.concat(dfs, ignore_index=True)
            return df_mc

    @staticmethod
    def get_fsr_corrected_df(df: pd.DataFrame, dry_run: bool = False) -> pd.DataFrame:
        if dry_run:
            return pd.DataFrame()
        else:
            return df.query("wasFSRCorrected == 1")

    def calculate_corrected_ratio(self) -> float:
        total_number_of_electrons = 0.0  # type: float
        fsr_corrected_electrons = 0.0  # type: float
        for _, eff_dict in self._efficiency_dict.items():
            total_number_of_electrons += eff_dict["total_number_selected"]
            fsr_corrected_electrons += eff_dict["fsr_corrected_number_selected"]
        return fsr_corrected_electrons / total_number_of_electrons

    def plot_variable_comparison(
        self,
        df: pd.DataFrame,
        fsr_variable: FSRVariable,
        appendix: Optional[str] = None,
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        appendix = f"_{appendix}" if appendix else ""
        output_key = f"FSRValidation_{self._selection_name}_{fsr_variable.variable.df_label}_comparison{appendix}"
        output_path = os.path.join(self._base_output_path, f"{output_key}.pdf")

        if dry_run:
            return output_key, output_path

        shp = SimpleHistogramPlot(variable=fsr_variable.variable)
        shp.add_component(
            label=r"With FSR Correction",
            data=df[fsr_variable.fsr_column].values,
            hist_type="step",
            color=KITColors.kit_blue,
        )
        shp.add_component(
            label=r"Without FSR Correction",
            data=df[fsr_variable.no_fsr_column].values,
            hist_type="step",
            color=KITColors.kit_red,
        )
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)
        shp.plot_on(ax=ax)

        if self._efficiency_dict:
            ratio = round(self.calculate_corrected_ratio() * 100, 2)
            title = fr"{ratio} \% of all {LatexParticleNames.electron_pm_str} corrected"
            ax.set_title(label=title, loc="right")

        export(fig=fig, filename=output_key, target_dir=self._base_output_path, close_figure=True)

        return output_key, output_path

    def plot_fsr_variable(
        self,
        df: pd.DataFrame,
        fsr_variable: HistVariable,
        appendix: Optional[str] = None,
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        appendix = f"_{appendix}" if appendix else ""
        output_key = f"FSRValidation_{self._selection_name}_{fsr_variable.df_label}{appendix}"
        output_path = os.path.join(self._base_output_path, f"{output_key}.pdf")

        if dry_run:
            return output_key, output_path

        shp = SimpleHistogramPlot(fsr_variable)
        shp.add_component(
            label="With FSR Correction",
            data=df,
            hist_type="stepfilled",
            color=KITColors.kit_blue,
        )

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)
        shp.plot_on(ax=ax)

        if self._efficiency_dict:
            ratio = round(self.calculate_corrected_ratio() * 100, 2)
            title = fr"{ratio} \% of all {LatexParticleNames.electron_pm_str} corrected"
            ax.set_title(label=title, loc="right")

        export(fig=fig, filename=output_key, target_dir=self._base_output_path, close_figure=True)

        return output_key, output_path

    def do_validation(self, dry_run: bool = False) -> List[Tuple[str, PathType]]:
        output_list = []  # type: List[Tuple[str, PathType]]
        df_mc = self._get_df_mc(dry_run=dry_run)

        for fsr_var in self._fsr_comparison_vars:
            output_key, output_path = self.plot_variable_comparison(
                df=df_mc,
                fsr_variable=fsr_var,
                appendix="all",
                dry_run=dry_run,
            )
            output_list.append((output_key, output_path))

            output_key, output_path = self.plot_variable_comparison(
                df=self.get_fsr_corrected_df(df=df_mc, dry_run=dry_run),
                fsr_variable=fsr_var,
                appendix="corrected_only",
                dry_run=dry_run,
            )
            output_list.append((output_key, output_path))

        for fsr_var in self._fsr_vars:
            output_key, output_path = self.plot_fsr_variable(
                df=self.get_fsr_corrected_df(df=df_mc, dry_run=dry_run),
                fsr_variable=fsr_var,
                appendix="corrected_only",
                dry_run=dry_run,
            )
            output_list.append((output_key, output_path))

        return output_list
