"""
Analysis Flow Validation -- Multiplicity Evaluation

Felix Metzner 2019
"""

import copy
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from typing import Any, NamedTuple, Union, Optional, Tuple, List, Dict, Set

from templatefitter.plotter.plot_style import KITColors
from templatefitter.plotter.plot_utilities import color_fader, AxesType, FigureType

from analysistools.plotting.latex_tables import df_to_split_latex_table

from rdstar.utilities import PathType
from rdstar.offline_analysis import mc_ids
from rdstar.offline_analysis.rdstar_plots import decay_mode_id_dict
from rdstar.offline_analysis.validation.analysis_flow_validation import (
    get_cut_flow_table_col_name,
    update_table_fraction_output_dict,
)


__all__ = [
    "MultiplicityInfo",
    "multiplicity_info_collection",
    "create_multiplicity_plot",
    "create_multiplicity_table",
    "create_multiplicity_table_for_data",
]


class MultiplicityInfo(NamedTuple):
    key: str
    column_name_dict: Union[None, Dict] = None
    table_fractions: Optional[List[int]] = None


multiplicity_info_collection = [
    MultiplicityInfo(key="", column_name_dict=None),
    MultiplicityInfo(key="_per_decay_id", column_name_dict=decay_mode_id_dict(), table_fractions=[2, 4]),
    MultiplicityInfo(key="_is_sig_per_decay_id", column_name_dict=decay_mode_id_dict(), table_fractions=[2, 4]),
    MultiplicityInfo(key="_is_sig_strict_per_decay_id", column_name_dict=decay_mode_id_dict(), table_fractions=[2, 4]),
    MultiplicityInfo(key="_is_sig_strict_b_per_decay_id", column_name_dict=decay_mode_id_dict(), table_fractions=[2, 4]),
    MultiplicityInfo(key="_per_data_type", column_name_dict=None, table_fractions=[3, 6]),
    MultiplicityInfo(key="_per_sig_id", column_name_dict=mc_ids.sig_id_to_label, table_fractions=[1, 3]),
]  # type: List[MultiplicityInfo]


def get_multiplicity_data(multiplicity_dict) -> List[str]:
    return [item for sublist in [[i] * int(j) for i, j in multiplicity_dict.items()] for item in sublist]


def create_multiplicity_plot(
    multiplicity_dict: Dict[str, Any],
    cut_list: List[str],
    selection_name: str,
    max_cands: int = 10,
    log_y_scale: bool = False,
) -> Tuple[FigureType, AxesType]:
    overflow_name = "OF"
    emp_dict = multiplicity_dict
    if max_cands <= 20:
        overflow_name = "Overflow"

    reduced_emp_dict = {k: {kk: vv for kk, vv in v.items() if kk <= max_cands + 1} for k, v in emp_dict.items()}
    # Add overflow bin:
    of_bin_emp_dict = {k: sum([vv for kk, vv in v.items() if kk > max_cands + 1]) for k, v in emp_dict.items()}
    for k in reduced_emp_dict.keys():
        if (max_cands + 1) in reduced_emp_dict[k].keys():
            reduced_emp_dict[k][max_cands + 1] += of_bin_emp_dict[k]
        else:
            reduced_emp_dict[k][max_cands + 1] = 0

    width_elements = 1.0 / len(reduced_emp_dict.keys())
    width = width_elements * 0.8
    width_space = width_elements * 0.15

    fig = plt.figure(num=None, figsize=(9, 3.0), dpi=200, facecolor="w", edgecolor="k")
    ax = plt.gca()

    for i, (key, data) in enumerate(reduced_emp_dict.items()):
        label = get_cut_flow_table_col_name(
            index=int(key.split("_", 1)[0]),
            cut_list=cut_list,
            selection_name=selection_name,
        )
        if i < len(reduced_emp_dict) - 1:
            color = color_fader(KITColors.kit_green, KITColors.kit_black, i / (len(reduced_emp_dict) - 1))
        else:
            color = KITColors.kit_red
        x_pos = np.array(list(data.keys())) + i * (width + width_space)
        ax.bar(x_pos, data.values(), width, label=label, color=color)

    ax.set_xlabel("Event Multiplicity")
    ax.set_xlim(0.6, max_cands + 2)
    ax.set_xticks(np.arange(1, max_cands + 2))
    ax.set_xticks([], minor=True)
    x_tick_labels = [x if x <= max_cands else overflow_name for x in list(np.arange(1, max_cands + 2))]
    ax.set_xticklabels(x_tick_labels, ha="left")

    ax.set_ylabel("Number of Events")
    if log_y_scale:
        ax.set_yscale("log")

    ax.legend(loc=1, ncol=2)

    return fig, ax


def create_multiplicity_table(
    info_dict: Dict[str, Any],
    multiplicity_info: MultiplicityInfo,
    selection_name: str,
    selection_tag: str,
    base_path: PathType,
    dry_run: bool = False,
) -> Dict[str, PathType]:
    assert multiplicity_info.key != "", multiplicity_info.key
    assert multiplicity_info.table_fractions is not None

    info_k = f"event_multiplicities{multiplicity_info.key}"
    em_table_functions = dict([("median", np.median), ("max", np.max), ("min", np.min), ("mean", np.mean)])
    cut_list = None  # type: Optional[List[str]]

    output_dict = {}  # type: Dict[str, PathType]
    em_data_dict = {}  # type: Dict[Set[str], Dict[str, List[str]]]
    if not dry_run:
        cut_list = info_dict[list(info_dict.keys())[len(info_dict) - 1]]["cuts"]
        sub_keys = list(set([kk for k in info_dict.keys() for kk in info_dict[k][info_k].keys()]))  # type: List[Set[str]]
        em_data_dict = {
            kk: {
                k: get_multiplicity_data(multiplicity_dict=info_dict[k][info_k][kk])
                if kk in info_dict[k][info_k].keys()
                else []
                for k in info_dict.keys()
            }
            for kk in sub_keys
        }

    for func_name, func in em_table_functions.items():
        name_prefix = f"{selection_tag}_emt_{info_k}_{func_name}"
        update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=base_path,
            table_fractions=multiplicity_info.table_fractions,
        )

        if dry_run:
            continue

        em_dict = {
            d_id: {k: func(em_data[k]) if em_data[k] else np.nan for k in info_dict.keys()}
            for d_id, em_data in em_data_dict.items()
        }
        em_df = pd.DataFrame.from_dict(em_dict, orient="index")
        em_df.fillna("---", inplace=True)

        assert cut_list is not None
        cut_names = {
            c: get_cut_flow_table_col_name(
                index=int(c.split("_", 1)[0]),
                cut_list=cut_list,
                selection_name=selection_name,
            )
            for c in em_df.columns
        }
        em_df.rename(columns=cut_names, inplace=True)
        em_df = em_df.T
        if multiplicity_info.column_name_dict is None:
            em_df.rename(columns={c: rf"{i}_{c}" for i, c in enumerate(list(em_df.columns))}, inplace=True)
        else:
            col_dict = multiplicity_info.column_name_dict
            assert all(c in col_dict.keys() for c in em_df.columns), (em_df.columns, col_dict.keys())
            em_df.rename(columns={k: rf"{k}_{v}" for k, v in col_dict.items()}, inplace=True)

        df_to_split_latex_table(
            df=em_df,
            table_fractions=multiplicity_info.table_fractions,
            name_prefix=name_prefix,
            output_lookup_dict=output_dict,
            placeholder_element="---",
            remove_column_name_index=True,
            allow_smaller_sub_dicts=True,
            create_pdf=True,
            float_format=lambda s: f"{s:0.2f}" if isinstance(s, float) else "! --- !",
        )

    return output_dict


def create_multiplicity_table_for_data(
    info_dict: Dict[str, Any],
    selection_name: str,
    selection_tag: str,
    base_path: PathType,
    dry_run: bool = False,
) -> Dict[str, PathType]:
    table_fractions = [2, 4]  # type: List[int]
    all_key = -1  # type: int
    all_info_k = "event_multiplicities"  # type: str
    per_decay_id_info_k = "event_multiplicities_per_decay_id"  # type: str

    em_table_functions = dict([("median", np.median), ("max", np.max), ("min", np.min), ("mean", np.mean)])
    cut_list = None  # type: Optional[List[str]]

    output_dict = {}  # type: Dict[str, PathType]
    em_data_dict = {}  # type: Dict[Union[Set[str], int], Dict[str, List[str]]]
    if not dry_run:
        cut_list = info_dict[list(info_dict.keys())[len(info_dict) - 1]]["cuts"]
        sub_keys = list(
            set([kk for k in info_dict.keys() for kk in info_dict[k][per_decay_id_info_k].keys()])
        )  # type: List[Set[str]]
        em_data_dict = {
            kk: {
                k: get_multiplicity_data(multiplicity_dict=info_dict[k][per_decay_id_info_k][kk])
                if kk in info_dict[k][per_decay_id_info_k].keys()
                else []
                for k in info_dict.keys()
            }
            for kk in sub_keys
        }
        assert all_key not in em_data_dict.keys()
        em_data_dict.update(
            {all_key: {k: get_multiplicity_data(multiplicity_dict=info_dict[k][all_info_k]) for k in info_dict.keys()}}
        )

    for func_name, func in em_table_functions.items():
        name_prefix = f"{selection_tag}_emt_data_{func_name}"
        update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=base_path,
            table_fractions=table_fractions,
        )

        if dry_run:
            continue

        em_dict = {
            d_id: {k: func(em_data[k]) if em_data[k] else np.nan for k in info_dict.keys()}
            for d_id, em_data in em_data_dict.items()
        }

        df = pd.DataFrame.from_dict(em_dict, orient="index")
        df.fillna("---", inplace=True)

        assert cut_list is not None
        cut_names = {
            c: get_cut_flow_table_col_name(
                index=int(c.split("_", 1)[0]),
                cut_list=cut_list,
                selection_name=selection_name,
            )
            for c in df.columns
        }
        df.rename(columns=cut_names, inplace=True)
        df = df.T

        column_mapping = copy.copy(decay_mode_id_dict())
        column_mapping.update({all_key: "All"})
        assert all(c in column_mapping for c in df.columns), (df.columns, column_mapping.keys())
        df.rename(columns={k: rf"{k}_{v}" for k, v in column_mapping.items()}, inplace=True)

        df_to_split_latex_table(
            df=df,
            table_fractions=table_fractions,
            name_prefix=name_prefix,
            output_lookup_dict=output_dict,
            placeholder_element="---",
            remove_column_name_index=True,
            allow_smaller_sub_dicts=True,
            create_pdf=True,
            float_format=lambda s: f"{s:0.2f}" if isinstance(s, float) else "! --- !",
        )

    return output_dict
