"""
Additional plotting tools for validation of observables.

Felix Metzner 2021
"""

import os
import pandas as pd

from matplotlib import pyplot as plt
from typing import Optional, Union, List, Tuple, Dict

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_utilities import export, FigureType, AxesType
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params
from templatefitter.plotter.validation_plots import BinCompositionPlot, BinMigrationPlot

from rdstar.utilities import PathType
from rdstar.offline_analysis.rdstar_plots import (
    rdstar_dec_modes,
    rdstar_dec_titles,
    DecayModeDictType,
    DecayTitleDictType,
    rdstar_color_map,
)


__all__ = [
    "plot_bin_migration",
    "plot_bin_composition",
]


def plot_bin_migration(
    mc_df: pd.DataFrame,
    from_hist_var: HistVariable,
    to_hist_var: HistVariable,
    from_to_label_appendix: Optional[Tuple[Optional[str], Optional[str]]] = None,
    weight_col: Optional[str] = None,
    normalize_to_origin: bool = True,
    dec_modes: Optional[DecayModeDictType] = None,
    dec_titles: Optional[DecayTitleDictType] = None,
    file_name_suffix: Optional[str] = None,
    show_color_bar: bool = False,
    fig_size: Tuple[float, float] = (12.0, 10.0),
    export_path: Optional[Union[str, PathType]] = None,
    dry_run: bool = False,
) -> Dict[str, PathType]:

    file_name_end = "" if file_name_suffix is None else f"_{file_name_suffix}"  # type: str
    var_info_for_file = f"{from_hist_var.df_label}_to_{to_hist_var.df_label}"  # type: str
    file_name_base = f"bin_migration_plot_{var_info_for_file}{file_name_end}_"  # type: str

    _dec_modes = rdstar_dec_modes if dec_modes is None else dec_modes  # type: DecayModeDictType
    _dec_titles = rdstar_dec_titles if dec_titles is None else dec_titles  # type: DecayTitleDictType

    output_dict = {}  # type: Dict[str, PathType]

    all_mode_ids = [d_id for d_ids in _dec_modes.values() for d_id in d_ids]  # type: List[int]
    mode_infos = [(k, v) for k, v in _dec_modes.items()]  # type: List[Tuple[str, List[int]]]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        file_name = f"{file_name_base}{mode_name}"  # type: str

        if export_path is not None:
            output_path = os.path.join(export_path, f"{file_name}.pdf")  # type: PathType
            assert file_name not in output_dict.keys(), file_name
            assert output_path not in output_dict.values(), (file_name, output_path)
            output_dict.update({file_name: output_path})

        if dry_run:
            continue

        migration_plot = BinMigrationPlot(
            from_variable=from_hist_var,
            to_variable=to_hist_var,
            weight_column=weight_col,
            color_map=rdstar_color_map,
            from_to_label_appendix=from_to_label_appendix,
            max_number_of_ticks=32,
            max_bins_for_labels=30,
        )

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=fig_size)  # type: FigureType, AxesType

        migration_plot.plot_on(
            df=mc_df.loc[mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)],
            ax=ax,
            normalize_to_origin=normalize_to_origin,
            show_color_bar=show_color_bar,
        )

        ax.set_title(_dec_titles[mode_name], loc="right")

        if export_path is not None:
            export(fig=fig, filename=file_name, target_dir=export_path, close_figure=True)

    return output_dict


def plot_bin_composition(
    mc_df: pd.DataFrame,
    main_hist_var: HistVariable,
    secondary_hist_var: HistVariable,
    secondary_binning: Binning,
    secondary_hist_var_label_mapping: Optional[Dict[int, str]] = None,
    weight_col: Optional[str] = None,
    plot_shape: bool = False,
    dec_modes: Optional[DecayModeDictType] = None,
    dec_titles: Optional[DecayTitleDictType] = None,
    file_name_suffix: Optional[str] = None,
    draw_legend: bool = True,
    legend_columns: int = 1,
    fig_size: Tuple[float, float] = (8.0, 5.0),
    export_path: Optional[Union[str, PathType]] = None,
    dry_run: bool = False,
) -> Dict[str, PathType]:

    _dec_modes = rdstar_dec_modes if dec_modes is None else dec_modes  # type: DecayModeDictType
    _dec_titles = rdstar_dec_titles if dec_titles is None else dec_titles  # type: DecayTitleDictType

    file_name_end = "" if file_name_suffix is None else f"_{file_name_suffix}"  # type: str
    var_info_for_file = f"{main_hist_var.df_label}_for_{secondary_hist_var.df_label}"  # type: str
    file_name_base = f"bin_composition_plot_{var_info_for_file}{file_name_end}_"  # type: str

    output_dict = {}  # type: Dict[str, PathType]

    all_mode_ids = [d_id for d_ids in _dec_modes.values() for d_id in d_ids]  # type: List[int]
    mode_infos = [(k, v) for k, v in _dec_modes.items()]  # type: List[Tuple[str, List[int]]]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        file_name = f"{file_name_base}{mode_name}"  # type: str

        if export_path is not None:
            output_path = os.path.join(export_path, f"{file_name}.pdf")  # type: PathType
            assert file_name not in output_dict.keys(), file_name
            assert output_path not in output_dict.values(), (file_name, output_path)
            output_dict.update({file_name: output_path})

        if dry_run:
            continue

        set_matplotlibrc_params()

        bc_plot = BinCompositionPlot(
            variable=main_hist_var,
            secondary_variable=secondary_hist_var,
            secondary_variable_binning=secondary_binning,
            weight_column=weight_col,
            secondary_tick_label_mapping=secondary_hist_var_label_mapping,
        )

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=fig_size)  # type: FigureType, AxesType

        bc_plot.plot_on(
            df=mc_df.loc[mc_df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids)],
            ax=ax,
            draw_legend=draw_legend,
            legend_columns=legend_columns,
        )

        if plot_shape:
            shape_plot = SimpleHistogramPlot(variable=main_hist_var)
            shape_plot.add_component(
                label="Shape",
                data=mc_df,
                weights=weight_col,
                hist_type="step",
                color=KITColors.light_grey,
            )

            ax2 = ax.twinx()  # type: AxesType

            shape_plot.plot_on(
                ax=ax2,
                draw_legend=False,
                normed=True,
            )
            ax2.set_ylabel(ylabel="")
            ax2.set_yticks(ticks=[])
            ax2.set_yticklabels(labels=[])

        ax.set_title(_dec_titles[mode_name], loc="right")

        if export_path is not None:
            export(fig=fig, filename=file_name, target_dir=export_path, close_figure=True)

    return output_dict
