"""
Manager class providing all the necessary information about the
privately added weights used by the analysis

Felix Metzner 2020
"""

import os
import logging
import numpy as np
import pandas as pd

from collections.abc import Set as SetBaseClass
from typing import Optional, Tuple, List, Dict, NamedTuple, Iterator, Collection

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.mc_reweighting.BFReweighting.bf_reweighting import BFReweighter
from rdstar.mc_reweighting.BFReweighting.d_bf_reweighting import DBFReweighter
from rdstar.offline_analysis.corrections.k_short_corrections import KShortEffCorrector
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection
from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection  # , InclusiveTagCorrection
from rdstar.offline_analysis.rdstar_sample_weights import scale_weight_column_name, off_to_on_res_scale_weight_column_name

from rdstar.offline_analysis.tasks import utilities

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "WeightTypeInfo",
    "WeightTypeCollection",
    "WeightManager",
]


_combined_ff_weight_col_name = "__weight_ff_nom__"  # type: str


class WeightTypeInfo(NamedTuple):
    name: str
    column_name: str
    latex_str: str
    is_additional: bool
    is_for_efficiency: bool
    validate: bool


class _WeightTypeCollectionClass(SetBaseClass):
    def __init__(self, input_iterable: Collection) -> None:
        assert all(isinstance(wti, WeightTypeInfo) for wti in input_iterable)

        names = [w.name for w in input_iterable]  # type: List[str]
        assert len(names) == len(set(names)), (len(names), len(set(names)))
        cols = [w.column_name for w in input_iterable]  # type: List[str]
        assert len(cols) == len(set(cols)), (len(cols), len(set(cols)))

        self.__elements = tuple(input_iterable)  # type: Tuple[WeightTypeInfo, ...]

        assert set(self.additional_weight_columns) | set(self.combined_weight_columns) == set(cols)

    def __iter__(self) -> Iterator[WeightTypeInfo]:
        return iter(self.__elements)

    def __contains__(self, value: object) -> bool:
        if not isinstance(value, str):
            raise TypeError(f"Expected argument of type str but got {type(value).__name__}")
        return value in [w.name for w in self.__elements]

    def __len__(self) -> int:
        return len(self.__elements)

    def __getitem__(self, name: str) -> WeightTypeInfo:
        if name not in self:
            raise KeyError(f"Weight type {name} is not available!")
        return [w for w in self.__elements if w.name == name][0]

    @property
    def combined_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if not w.is_additional])

    @property
    def additional_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if w.is_additional])

    @property
    def efficiency_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if w.is_for_efficiency])

    @property
    def weights_for_validation(self) -> Tuple[WeightTypeInfo, ...]:
        return tuple([w for w in self.__elements if w.validate])


WeightTypeCollection = _WeightTypeCollectionClass(
    [
        WeightTypeInfo(
            name="ff_weight",
            column_name=_combined_ff_weight_col_name,
            latex_str="Form Factor Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="pid_weight",
            column_name=ParticleIDCorrection.get_final_pid_weight_col_name(),
            latex_str="Particle ID Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="slow_pi_weight",
            column_name=TrackingEffCorrector.slow_pi_corr_col_name,
            latex_str="Slow Pion Correction Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="k_short_weight",
            column_name=KShortEffCorrector.k_short_corr_col_name,
            latex_str=LatexParticleNames.k_short_str + r" Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="scale_weight",
            column_name=scale_weight_column_name,
            latex_str="MC Scaling Weights",
            is_additional=False,
            is_for_efficiency=False,
            validate=False,
        ),
        WeightTypeInfo(
            name="b_bf_weight",
            column_name=BFReweighter.weight_total_col_name,
            latex_str=LatexParticleNames.b_str + " Meson BF Weight",
            is_additional=True,
            is_for_efficiency=False,
            validate=True,
        ),
        WeightTypeInfo(
            name="d_bf_weight",
            column_name=DBFReweighter.d_bf_weight_col,
            latex_str=LatexParticleNames.d_str + " Meson BF Weight",
            is_additional=True,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="tag_weight",
            column_name=TagCorrection.tagging_correction_name,
            latex_str="Tagging Calibration Weights",
            is_additional=True,
            is_for_efficiency=False,
            validate=True,
        ),
        # WeightTypeInfo(  # TODO: use instead of simple tag_weight
        #     name="tag_inclusive_weight",
        #     column_name=InclusiveTagCorrection.tagging_correction_name,
        #     latex_str="Tagging Calibration Weights",
        #     is_additional=True,
        #     is_for_efficiency=False,
        #     validate=True,
        # ),
        WeightTypeInfo(
            name="off_res_weight",
            column_name=off_to_on_res_scale_weight_column_name,
            latex_str="Off-Resonance Scaling Weights",
            is_additional=True,  # TODO: Consider not as additional but as combined_weight_column or completely separate?
            is_for_efficiency=False,
            validate=False,
        ),
    ]
)  # type: _WeightTypeCollectionClass


class WeightManager:
    weight_analysis_base_dir_name = "Weight_Analysis"  # type: str

    finalized_weights_appendix = "withFinalWeights"  # type: str

    final_weight_col = "__weight_full__"  # type: str
    final_weight_with_additions_col = "__weight_final_wAdditions_corr__"  # type: str

    combined_ff_weight_col_name = _combined_ff_weight_col_name  # type: str

    weight_collection = WeightTypeCollection  # type: _WeightTypeCollectionClass

    combined_weight_column_names = weight_collection.combined_weight_columns  # type: Tuple[str, ...]
    additional_weight_column_names = weight_collection.additional_weight_columns  # type: Tuple[str, ...]
    efficiency_weight_column_names = weight_collection.efficiency_weight_columns  # type: Tuple[str, ...]

    pid_weight_info_key = "pid_weight_info"  # type: str
    ff_d_weight_info_key = "ff_d_weight_info"  # type: str
    ff_ds_weight_info_key = "ff_ds_weight_info"  # type: str
    ff_dss_weight_info_key = "ff_dss_weight_info"  # type: str

    weight_info_keys = [
        pid_weight_info_key,
        ff_d_weight_info_key,
        ff_ds_weight_info_key,
        ff_dss_weight_info_key,
    ]  # type: List[str]

    pid_dict_output_key = "pid_hadron_for_flag_dict"  # type: str
    combined_ff_weights_prefix = "combined"  # type: str

    evt_id_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()  # type: str
    c_id_col = "unique_candidate_id"  # type: str
    id_col = "combined_candidate_id"  # type: str

    with_selection_suffix = "_with_selection"  # type: str

    @staticmethod
    def add_additional_weights(df: pd.DataFrame) -> None:
        assert WeightManager.final_weight_col in df.columns
        assert WeightManager.final_weight_with_additions_col not in df.columns
        assert all(c in df.columns for c in WeightManager.additional_weight_column_names), [
            c for c in WeightManager.additional_weight_column_names if c not in df.columns
        ]

        col_list = list(WeightManager.additional_weight_column_names)
        col_list.append(WeightManager.final_weight_col)
        df.loc[:, WeightManager.final_weight_with_additions_col] = df[col_list].product(axis=1)

    @staticmethod
    def add_unique_id(
        df: pd.DataFrame,
        check_if_unique: bool = False,
        mask: Optional[pd.Series] = None,
    ) -> np.ndarray:
        id_values = utilities.add_unique_candidate_id(
            df=df,
            id_column_name=WeightManager.id_col,
            event_id_name=WeightManager.evt_id_col,
            candidate_id_name=WeightManager.c_id_col,
            check_if_unique=check_if_unique,
            mask=mask,
        )
        return id_values

    @staticmethod
    def get_weight_output_key(weight_key: str, with_selection: bool) -> str:
        assert weight_key in WeightManager.weight_info_keys, (weight_key, WeightManager.weight_info_keys)
        selection_suffix = WeightManager.with_selection_suffix if with_selection else ""  # type: str
        if weight_key.startswith("ff"):
            return f"{WeightManager.combined_ff_weights_prefix}_{weight_key}{selection_suffix}"
        else:
            return f"{WeightManager.pid_dict_output_key}{selection_suffix}"

    @staticmethod
    def get_flag_info_dict_path_from_task_input_dict(
        input_dict: Dict[str, PathType],
        selection_name: str,
    ) -> PathType:
        return input_dict[WeightManager.get_pid_flag_dict_output_key(selection_name=selection_name)]

    @staticmethod
    def get_pid_weight_info_dict_file_path(
        dir_path: PathType,
        tmp: bool = False,
    ) -> PathType:
        if tmp:
            return os.path.join(dir_path, f"temp_{WeightManager.pid_dict_output_key}.json")
        else:
            return os.path.join(dir_path, f"{WeightManager.pid_dict_output_key}.json")

    @staticmethod
    def get_pid_flag_dict_output_key(selection_name: str) -> str:
        return f"{WeightManager.pid_dict_output_key}_{selection_name}"

    @staticmethod
    def get_weight_info_data_frame_path(
        dir_path: PathType,
        weight_key: str,
        tmp: bool = False,
        with_selection: bool = False,
    ) -> PathType:
        file_name = WeightManager.get_weight_output_key(weight_key=weight_key, with_selection=with_selection)
        if tmp:
            return os.path.join(dir_path, f"temp_{file_name}.h5")
        else:
            return os.path.join(dir_path, f"{file_name}.h5")
