"""
This file contains all the decay chains relevant to the analysis.

It is important to note, that the final Y(4S) particle is always
reconstructed in such a way, that the first daughter (index 0) is
the tag B meson, and the second (index 1) is the signal B meson.

Felix Metzner 2018
"""

from typing import Union, Optional, Dict, List, Callable, Any

from analysistools.processing import decay_channels_utils

from rdstar.online_analysis import roe_masks
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.latex_labels.latex_observable_names import LatexObsNames


VertexingSettingType = Dict[str, Union[str, Dict[str, Union[int, bool, List[int]]]]]
ROESettingType = Dict[str, Union[str, Optional[List[Callable]]]]
ExtraSettingsType = Dict[str, Union[Any, str, VertexingSettingType, ROESettingType]]


######################
# Upsilon(4S) Decays #
######################
U4S_decay_channels = {
    "Upsilon(4S):charged": "Upsilon(4S) -> B+_tag B-_sig",
    "Upsilon(4S):mixed": "Upsilon(4S) -> B0_tag B0_sig",
    "Upsilon(4S):charged_wt": "Upsilon(4S) -> B0_tag B-_sig",
    "Upsilon(4S):charged_wt2": "Upsilon(4S) -> B0_tag B+_sig",
    "Upsilon(4S):mixed_wt": "Upsilon(4S) -> B+_tag B0_sig",
    "Upsilon(4S):mixed_wt2": "Upsilon(4S) -> B-_tag B0_sig",
}  # type: Dict[str, str]

U4S_decay_label_to_decay_id = {
    "Upsilon(4S):charged": 1001,
    "Upsilon(4S):mixed": 1000,
    "Upsilon(4S):charged_wt": 1011,
    "Upsilon(4S):charged_wt2": 1021,
    "Upsilon(4S):mixed_wt": 1010,
    "Upsilon(4S):mixed_wt2": 1020,
}  # type: Dict[str, int]

U4S_decay_label_to_chain = {
    "Upsilon(4S):charged": "Upsilon(4S):charged{suf} -> B+:hadronic_tag B-:sig{suf}",
    "Upsilon(4S):mixed": "Upsilon(4S):mixed{suf} -> anti-B0:hadronic_tag B0:sig{suf}",
    "Upsilon(4S):charged_wt": "Upsilon(4S):charged_wt{suf} -> B0:hadronic_tag B-:sig{suf}",
    "Upsilon(4S):charged_wt2": "Upsilon(4S):charged_wt2{suf} -> B0:hadronic_tag B+:sig{suf}",
    "Upsilon(4S):mixed_wt": "Upsilon(4S):mixed_wt{suf} -> B+:hadronic_tag B0:sig{suf}",
    "Upsilon(4S):mixed_wt2": "Upsilon(4S):mixed_wt2{suf} -> B-:hadronic_tag B0:sig{suf}",
    # What about B0 + B0, which could happen through mixing.
}  # type: Dict[str, str]

U4S_cut_infos = [
    CutInfo(
        cut_str="7.0 < M < 13.0",
        cut_label=LatexObsNames.y4s_mass[:-1] + r"\;\in\;(7.0, 13.0)\," + LatexUnits.gev[1:],
        cut_name="mass_cut",
        cut_var="M",
        cut_var_latex_label=LatexObsNames.y4s_mass,
        cut_type=("<", "<"),
        cut_limits=(7.0, 13.0),
        cut_online=True,
        latex_unit=LatexUnits.gev,
    ),
    CutInfo(
        cut_str="nROE_Tracks(CleanROE) < 1",
        cut_label=LatexObsNames.roe_n_tracks[:-1] + r" == 0$",
        cut_name="n_roe_tracks_cut",
        cut_var="nROE_Tracks(CleanROE)",
        cut_var_latex_label=LatexObsNames.roe_n_tracks,
        cut_type="<",
        cut_limits=1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]  # type: List[CutInfo]
U4S_cuts = [cut_info.cut_str for cut_info in U4S_cut_infos]  # type: List[str]

U4S_decay_label_to_cut_list = {
    "Upsilon(4S):charged": U4S_cuts + [],
    "Upsilon(4S):mixed": U4S_cuts + [],
    "Upsilon(4S):charged_wt": U4S_cuts + [],
    "Upsilon(4S):charged_wt2": U4S_cuts + [],
    "Upsilon(4S):mixed_wt": U4S_cuts + [],
    "Upsilon(4S):mixed_wt2": U4S_cuts + [],
}  # type: Dict[str, List[str]]

vertex_fit_setting_for_U4S = {
    "function": "vertexTree",
    "kwargs": {
        "conf_level": -1,
        "updateAllDaughters": False,
        "ipConstraint": True,
        "massConstraint": [411, 421, 413, 423],
    },
}  # type: VertexingSettingType

roe_setting_for_U4S = {
    "mask_list": roe_masks.roe_mask_list,
    "mask_name": None,
}  # type: ROESettingType

extra_settings_for_U4S = {
    "roe": roe_setting_for_U4S,
    # "vertex_fit": vertex_fit_setting_for_U4S,
}  # type: ExtraSettingsType

U4S_decay_label_to_extra = {
    "Upsilon(4S):charged": extra_settings_for_U4S,
    "Upsilon(4S):mixed": extra_settings_for_U4S,
    "Upsilon(4S):charged_wt": extra_settings_for_U4S,
    "Upsilon(4S):charged_wt2": extra_settings_for_U4S,
    "Upsilon(4S):mixed_wt": extra_settings_for_U4S,
    "Upsilon(4S):mixed_wt2": extra_settings_for_U4S,
}  # type: Dict[str, ExtraSettingsType]

####################
# Tag-Side B Meson #
####################
# Particle lists are produced by the FEI.
# The information defined here is mainly used to define the cuts for B_tag candidates.
BTag_decay_channels = {
    "B+:hadronic_tag": "via FEI",
    "B0:hadronic_tag": "via FEI",
}  # type: Dict[str, str]

BTag_decay_label_to_decay_id = {
    "B+:hadronic_tag": "FEI+",
    "B0:hadronic_tag": "FEI0",
}  # type: Dict[str, str]

BTag_decay_label_to_chain = {
    "B+:hadronic_tag": "B+:generic",
    "B0:hadronic_tag": "B0:generic",
}  # type: Dict[str, str]

BTag_cut_infos = [
    CutInfo(
        cut_str="Mbc > 5.22",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" > 5.22\," + LatexUnits.gev[1:],
        cut_name="mbc_cut",
        cut_var="Mbc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type=">",
        cut_limits=5.22,
        cut_online=True,
        latex_unit=LatexUnits.gev,
    ),
    CutInfo(
        cut_str="-0.15 < deltaE < 0.1",
        cut_label=LatexObsNames.b_tag_delta_e[:-1] + r"\;\in\;(-0.15, 0.1)\," + LatexUnits.gev[1:],
        cut_name="deltaE_cut",
        cut_var="deltaE",
        cut_var_latex_label=LatexObsNames.b_tag_delta_e,
        cut_type=("<", "<"),
        cut_limits=(-0.15, 0.1),
        cut_online=True,
        latex_unit=LatexUnits.gev,
    ),
]  # type: List[CutInfo]
BTag_cuts = [cut_info.cut_str for cut_info in BTag_cut_infos]  # type: List[str]

BTag_decay_label_to_cut_list = {
    "B+:hadronic_tag": BTag_cuts + [],
    "B0:hadronic_tag": BTag_cuts + [],
}  # type: Dict[str, List[str]]

roe_setting_for_BTag = {"mask_list": roe_masks.roe_mask_list, "mask_name": "CleanROE"}  # type: ROESettingType
cs_setting_for_BTag = {"roe_mask": "CleanROE"}  # type: Dict[str, str]

extra_settings_for_BTag = {"roe": roe_setting_for_BTag, "cs": cs_setting_for_BTag}  # type: ExtraSettingsType

BTag_decay_label_to_extra = {
    "B+:hadronic_tag": extra_settings_for_BTag,
    "B0:hadronic_tag": extra_settings_for_BTag,
}  # type: Dict[str, ExtraSettingsType]

############
# B Decays #
############
B_decay_channels = {
    "B0:D_e": "B0 -> D- e+",
    "B0:D_mu": "B0 -> D- mu+",
    "B-:D0_e": "B- -> D0 e-",
    "B-:D0_mu": "B- -> D0 mu-",
    "B0:D*_e": "B0 -> D*- e+",
    "B0:D*_mu": "B0 -> D*- mu+",
    "B-:D*0_e": "B- -> D*0 e-",
    "B-:D*0_mu": "B- -> D*0 mu-",
}  # type: Dict[str, str]

B_decay_label_to_decay_id = {
    "B0:D_e": 11,
    "B0:D_mu": 12,
    "B-:D0_e": 13,
    "B-:D0_mu": 14,
    "B0:D*_e": 15,
    "B0:D*_mu": 16,
    "B-:D*0_e": 17,
    "B-:D*0_mu": 18,
}  # type: Dict[str, int]

B_decay_label_to_chain = {
    "B0:D_e": "B0:D_e{suf} -> D-:all{suf} e+:fsp{suf} {neutrino}",
    "B0:D_mu": "B0:D_mu{suf} -> D-:all{suf} mu+:fsp{suf} {neutrino}",
    "B-:D0_e": "B-:D0_e{suf} -> D0:all{suf} e-:fsp{suf} {neutrino}",
    "B-:D0_mu": "B-:D0_mu{suf} -> D0:all{suf} mu-:fsp{suf} {neutrino}",
    "B0:D*_e": "B0:D*_e{suf} -> D*-:all{suf} e+:fsp{suf} {neutrino}",
    "B0:D*_mu": "B0:D*_mu{suf} -> D*-:all{suf} mu+:fsp{suf} {neutrino}",
    "B-:D*0_e": "B-:D*0_e{suf} -> D*0:all{suf} e-:fsp{suf} {neutrino}",
    "B-:D*0_mu": "B-:D*0_mu{suf} -> D*0:all{suf} mu-:fsp{suf} {neutrino}",
}  # type: Dict[str, str]

sig_b_mass_cut = CutInfo(
    cut_str="1.0 < M < 6.",
    cut_label=LatexObsNames.b_sig_mass[:-1] + r"\;\in\;(1.0, 6.0)\," + LatexUnits.gev[1:],
    cut_name="sig_b_mass_cut",
    cut_var="M",
    cut_var_latex_label=LatexObsNames.b_sig_mass,
    cut_type=("<", "<"),
    cut_limits=(1.0, 6.0),
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
sig_b_r2_cut = CutInfo(
    cut_str="R2 < 0.6",
    cut_label=LatexObsNames.cs_r2[:-1] + r" < 0.6$",
    cut_name="sig_b_r2_cut",
    cut_var="R2",
    cut_var_latex_label=LatexObsNames.cs_r2,
    cut_type="<",
    cut_limits=0.6,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)

B_cut_infos = [sig_b_mass_cut, sig_b_r2_cut]  # type: List[CutInfo]
B_cuts = [cut_info.cut_str for cut_info in B_cut_infos]  # type: List[str]

sig_b_electron_id_cut = CutInfo(
    cut_str="daughter(1, eIDBelle) > 0.6",
    cut_label=LatexObsNames.pid_electron[:-1] + r" > 0.6$",
    cut_name="sig_b_e_id_cut",
    cut_var="daughter(1, eIDBelle)",
    cut_var_latex_label=LatexObsNames.pid_electron,
    cut_type=">",
    cut_limits=0.6,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)
sig_b_electron_p_cut = CutInfo(
    cut_str="daughter(1, p) > 0.3",
    cut_label=LatexObsNames.lepton_p[:-1] + r" > 300\," + LatexUnits.mev[1:],
    cut_name="sig_b_e_p_cut",
    cut_var="daughter(1, p)",
    cut_var_latex_label=LatexObsNames.lepton_p,
    cut_type=">",
    cut_limits=0.3,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
sig_b_muon_id_cut = CutInfo(
    cut_str="daughter(1, muIDBelle) > 0.9",
    cut_label=LatexObsNames.pid_muon[:-1] + r" > 0.9$",
    cut_name="sig_b_mu_id_cut",
    cut_var="daughter(1, muIDBelle)",
    cut_var_latex_label=LatexObsNames.pid_muon,
    cut_type=">",
    cut_limits=0.9,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)
sig_b_muon_quality_cut = CutInfo(
    cut_str="daughter(1, muIDBelleQuality) == 1",
    cut_label=LatexObsNames.pid_muon_quality[:-1] + r" == 1",
    cut_name="sig_b_muon_belle_quality_cut",
    cut_var="daughter(1, muIDBelleQuality)",
    cut_var_latex_label=LatexObsNames.pid_muon_quality,
    cut_type="==",
    cut_limits=1,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)
sig_b_muon_p_cut = CutInfo(
    cut_str="daughter(1, p) > 0.6",
    cut_label=LatexObsNames.momentum_p[:1] + r" > 600\," + LatexUnits.gev[1:],
    cut_name="sig_b_mu_p_cut",
    cut_var="daughter(1, p)",
    cut_var_latex_label=LatexObsNames.momentum_p,
    cut_type=">",
    cut_limits=0.6,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)

sig_b_electron_cut_infos = [sig_b_electron_id_cut, sig_b_electron_p_cut]  # type: List[CutInfo]
sig_b_electron_cuts = [cut_info.cut_str for cut_info in sig_b_electron_cut_infos]  # type: List[str]
sig_b_muon_cut_infos = [sig_b_muon_id_cut, sig_b_muon_quality_cut, sig_b_muon_p_cut]  # type: List[CutInfo]
sig_b_muon_cuts = [cut_info.cut_str for cut_info in sig_b_muon_cut_infos]  # type: List[str]

B_decay_label_to_cut_list = {
    "B0:D_e": B_cuts + sig_b_electron_cuts + [],
    "B0:D_mu": B_cuts + sig_b_muon_cuts + [],
    "B-:D0_e": B_cuts + sig_b_electron_cuts + [],
    "B-:D0_mu": B_cuts + sig_b_muon_cuts + [],
    "B0:D*_e": B_cuts + sig_b_electron_cuts + [],
    "B0:D*_mu": B_cuts + sig_b_muon_cuts + [],
    "B-:D*0_e": B_cuts + sig_b_electron_cuts + [],
    "B-:D*0_mu": B_cuts + sig_b_muon_cuts + [],
}  # type: Dict[str, List[str]]

vertex_fit_setting_for_B = {
    "function": "vertexTree",
    "kwargs": {
        "conf_level": -1,
        "updateAllDaughters": False,
        "ipConstraint": False,
        "massConstraint": [411, 421, 413, 423],
    },
}  # type: VertexingSettingType
roe_setting_for_B = {"mask_list": roe_masks.roe_mask_list, "mask_name": "CleanROE"}  # type: ROESettingType
cs_setting_for_B = {"roe_mask": "CleanROE"}  # type: Dict[str, str]

extra_settings_for_B = {
    "roe": roe_setting_for_B,
    "cs": cs_setting_for_B,
    "vertex_fit": vertex_fit_setting_for_B,
}  # type: ExtraSettingsType

B_decay_label_to_extra = {
    "B0:D_e": extra_settings_for_B,
    "B0:D_mu": extra_settings_for_B,
    "B-:D0_e": extra_settings_for_B,
    "B-:D0_mu": extra_settings_for_B,
    "B0:D*_e": extra_settings_for_B,
    "B0:D*_mu": extra_settings_for_B,
    "B-:D*0_e": extra_settings_for_B,
    "B-:D*0_mu": extra_settings_for_B,
}  # type: Dict[str, ExtraSettingsType]

#####################
# B Decays with Tau #
#####################
# For when B Decays to taus are taken from MC information.
Btau_decay_channels = {
    "B0:D_tau": "B0 -> D- tau+",
    "B-:D0_tau": "B- -> D0 tau-",
    "B0:D*_tau": "B0 -> D*- tau+",
    "B-:D*0_tau": "B- -> D*0 tau-",
}  # type: Dict[str, str]

Btau_decay_label_to_decay_id = {
    "B0:D_tau": 1,
    "B-:D0_tau": 2,
    "B0:D*_tau": 3,
    "B-:D*0_tau": 4,
}  # type: Dict[str, int]

Btau_decay_label_to_chain = {
    "B0:D_tau": "B0:D_tau{suf} -> D-:all{suf} tau+:all{suf} nu_tau:mc{suf}",
    "B-:D0_tau": "B-:D0_tau{suf} -> D0:all{suf} tau-:all{suf} anti-nu_tau:mc{suf}",
    "B0:D*_tau": "B0:D*_tau{suf} -> D*-:all{suf} tau+:all{suf} nu_tau:mc{suf}",
    "B-:D*0_tau": "B-:D*0_tau{suf} -> D*0:all{suf} tau-:all{suf} anti-nu_tau:mc{suf}",
}  # type: Dict[str, str]

Btau_cuts = []  # type: List[str]

Btau_decay_label_to_cut_list = {
    "B0:D_tau": Btau_cuts + [],
    "B-:D0_tau": Btau_cuts + [],
    "B0:D*_tau": Btau_cuts + [],
    "B-:D*0_tau": Btau_cuts + [],
}  # type: Dict[str, List[str]]

roe_setting_for_Btau = {"mask_list": roe_masks.roe_mask_list, "mask_name": "CleanROE"}  # type: ROESettingType
cs_setting_for_Btau = {"roe_mask": "CleanROE"}  # type: Dict[str, str]

extra_settings_for_Btau = {"roe": roe_setting_for_Btau, "cs": cs_setting_for_Btau}  # type: ExtraSettingsType

Btau_decay_label_to_extra = {
    "B0:D_tau": extra_settings_for_Btau,
    "B-:D0_tau": extra_settings_for_Btau,
    "B0:D*_tau": extra_settings_for_Btau,
    "B-:D*0_tau": extra_settings_for_Btau,
}  # type: Dict[str, ExtraSettingsType]

#########################
# B Decays to  D(*) pi0 #
#########################

B_Dlpi0_decay_channels = {
    "B0:D_e_pi0": "B0 -> D- e+ pi0",
    "B0:D_mu_pi0": "B0 -> D- mu+ pi0",
    "B-:D0_e_pi0": "B- -> D0 e- pi0",
    "B-:D0_mu_pi0": "B- -> D0 mu- pi0",
    "B0:D*_e_pi0": "B0 -> D*- e+ pi0",
    "B0:D*_mu_pi0": "B0 -> D*- mu+ pi0",
    "B-:D*0_e_pi0": "B- -> D*0 e- pi0",
    "B-:D*0_mu_pi0": "B- -> D*0 mu- pi0",
}  # type: Dict[str, str]

B_Dlpi0_decay_label_to_decay_id = {
    "B0:D_e_pi0": 61,
    "B0:D_mu_pi0": 62,
    "B-:D0_e_pi0": 63,
    "B-:D0_mu_pi0": 64,
    "B0:D*_e_pi0": 65,
    "B0:D*_mu_pi0": 66,
    "B-:D*0_e_pi0": 67,
    "B-:D*0_mu_pi0": 68,
}  # type: Dict[str, int]

B_Dlpi0_decay_label_to_chain = {
    "B0:D_e_pi0": "B0:D_e_pi0{suf} -> D-:all{suf} e+:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B0:D_mu_pi0": "B0:D_mu_pi0{suf} -> D-:all{suf} mu+:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B-:D0_e_pi0": "B-:D0_e_pi0{suf} -> D0:all{suf} e-:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B-:D0_mu_pi0": "B-:D0_mu_pi0{suf} -> D0:all{suf} mu-:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B0:D*_e_pi0": "B0:D*_e_pi0{suf} -> D*-:all{suf} e+:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B0:D*_mu_pi0": "B0:D*_mu_pi0{suf} -> D*-:all{suf} mu+:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B-:D*0_e_pi0": "B-:D*0_e_pi0{suf} -> D*0:all{suf} e-:fsp{suf} pi0:fsp{suf} {neutrino}",
    "B-:D*0_mu_pi0": "B-:D*0_mu_pi0{suf} -> D*0:all{suf} mu-:fsp{suf} pi0:fsp{suf} {neutrino}",
}  # type: Dict[str, str]

B_Dlpi0_pi0_p_cut = CutInfo(
    cut_str="useCMSFrame(daughter(2, p)) > 0.2",
    cut_label=LatexObsNames.d_st_st_meson_pi0_cms_p[:-1] + r" > 200\," + LatexUnits.mev[1:],
    cut_name="pi0_from_dstst_p_cut",
    cut_var="useCMSFrame(daughter(2, p))",
    cut_var_latex_label=LatexObsNames.d_st_st_meson_pi0_cms_p,
    cut_type=">",
    cut_limits=0.2,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)

B_Dlpi0_pi0_inv_m_cut = CutInfo(
    cut_str="0.12 < daughter(2, InvM) < 0.15",
    # cut_str="0.11 < daughter(2, InvM) < 0.16",
    cut_label=LatexObsNames.d_st_st_meson_pi0_m_inv[:-1] + r"\;\in\;(120, 150)\," + LatexUnits.mev[1:],
    cut_name="pi0_from_dstst_inv_m_cut",
    cut_var="daughter(2, InvM)",
    cut_var_latex_label=LatexObsNames.d_st_st_meson_pi0_m_inv,
    cut_type=("<", "<"),
    cut_limits=(0.12, 0.15),
    cut_online=True,
    latex_unit=LatexUnits.gev,
)

B_Dlpi0_cut_infos = [
    sig_b_mass_cut,
    sig_b_r2_cut,
    # B_Dlpi0_pi0_inv_m_cut,
    B_Dlpi0_pi0_p_cut,
]  # type: List[CutInfo]
B_Dlpi0_cuts = [cut_info.cut_str for cut_info in B_Dlpi0_cut_infos]  # type: List[str]

Dlpi0_electron_cut_infos = [sig_b_electron_id_cut, sig_b_electron_p_cut]  # type: List[CutInfo]
Dlpi0_electron_cuts = [cut_info.cut_str for cut_info in Dlpi0_electron_cut_infos]  # type: List[str]
Dlpi0_muon_cut_infos = [sig_b_muon_id_cut, sig_b_muon_quality_cut, sig_b_muon_p_cut]  # type: List[CutInfo]
Dlpi0_muon_cuts = [cut_info.cut_str for cut_info in Dlpi0_muon_cut_infos]  # type: List[str]

B_Dlpi0_decay_label_to_cut_list = {
    "B0:D_e_pi0": B_Dlpi0_cuts + Dlpi0_electron_cuts + [],
    "B0:D_mu_pi0": B_Dlpi0_cuts + Dlpi0_muon_cuts + [],
    "B-:D0_e_pi0": B_Dlpi0_cuts + Dlpi0_electron_cuts + [],
    "B-:D0_mu_pi0": B_Dlpi0_cuts + Dlpi0_muon_cuts + [],
    "B0:D*_e_pi0": B_Dlpi0_cuts + Dlpi0_electron_cuts + [],
    "B0:D*_mu_pi0": B_Dlpi0_cuts + Dlpi0_muon_cuts + [],
    "B-:D*0_e_pi0": B_Dlpi0_cuts + Dlpi0_electron_cuts + [],
    "B-:D*0_mu_pi0": B_Dlpi0_cuts + Dlpi0_muon_cuts + [],
}  # type: Dict[str, List[str]]

roe_setting_for_B_Dlpi0 = {"mask_list": roe_masks.roe_mask_list, "mask_name": "CleanROE"}  # type: ROESettingType
cs_setting_for_B_Dlpi0 = {"roe_mask": "CleanROE"}  # type: Dict[str, str]
vertex_fit_setting_for_B_Dlpi0 = vertex_fit_setting_for_B  # type: VertexingSettingType

extra_settings_for_B_Dlpi0 = {
    "roe": roe_setting_for_B_Dlpi0,
    "cs": cs_setting_for_B_Dlpi0,
    "vertex_fit": vertex_fit_setting_for_B_Dlpi0,
}  # type: ExtraSettingsType

B_Dlpi0_decay_label_to_extra = {
    "B0:D_e_pi0": extra_settings_for_B_Dlpi0,
    "B0:D_mu_pi0": extra_settings_for_B_Dlpi0,
    "B-:D0_e_pi0": extra_settings_for_B_Dlpi0,
    "B-:D0_mu_pi0": extra_settings_for_B_Dlpi0,
    "B0:D*_e_pi0": extra_settings_for_B_Dlpi0,
    "B0:D*_mu_pi0": extra_settings_for_B_Dlpi0,
    "B-:D*0_e_pi0": extra_settings_for_B_Dlpi0,
    "B-:D*0_mu_pi0": extra_settings_for_B_Dlpi0,
}  # type: Dict[str, ExtraSettingsType]

#########################
# B Decays to  D(*) pi- #
#########################

B_Dlpim_decay_channels = {
    "B0:D_e_pi-": "B0 -> anti-D0 e+ pi-",
    "B0:D_mu_pi-": "B0 -> anti-D0 mu+ pi-",
    "B-:D+_e_pi-": "B- -> D+ e- pi-",
    "B-:D+_mu_pi-": "B- -> D+ mu- pi-",
    "B0:D*_e_pi-": "B0 -> anti-D*0 e+ pi-",
    "B0:D*_mu_pi-": "B0 -> anti-D*0 mu+ pi-",
    "B-:D*+_e_pi-": "B- -> D*+ e- pi-",
    "B-:D*+_mu_pi-": "B- -> D*+ mu- pi-",
}  # type: Dict[str, str]

B_Dlpim_decay_label_to_decay_id = {
    "B0:D_e_pi-": 71,
    "B0:D_mu_pi-": 72,
    "B-:D+_e_pi-": 73,
    "B-:D+_mu_pi-": 74,
    "B0:D*_e_pi-": 75,
    "B0:D*_mu_pi-": 76,
    "B-:D*+_e_pi-": 77,
    "B-:D*+_mu_pi-": 78,
}  # type: Dict[str, int]

B_Dlpim_decay_label_to_chain = {
    "B0:D_e_pi-": "B0:D_e_pi-{suf} -> anti-D0:all{suf} e+:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B0:D_mu_pi-": "B0:D_mu_pi-{suf} -> anti-D0:all{suf} mu+:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B-:D+_e_pi-": "B-:D+_e_pi-{suf} -> D+:all{suf} e-:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B-:D+_mu_pi-": "B-:D+_mu_pi-{suf} -> D+:all{suf} mu-:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B0:D*_e_pi-": "B0:D*_e_pi-{suf} -> anti-D*0:all{suf} e+:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B0:D*_mu_pi-": "B0:D*_mu_pi-{suf} -> anti-D*0:all{suf} mu+:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B-:D*+_e_pi-": "B-:D*+_e_pi-{suf} -> D*+:all{suf} e-:fsp{suf} pi-:fsp{suf} {neutrino}",
    "B-:D*+_mu_pi-": "B-:D*+_mu_pi-{suf} -> D*+:all{suf} mu-:fsp{suf} pi-:fsp{suf} {neutrino}",
}  # type: Dict[str, str]

B_Dlpim_cut_infos = [sig_b_mass_cut, sig_b_r2_cut]  # type: List[CutInfo]
B_Dlpim_cuts = [cut_info.cut_str for cut_info in B_Dlpim_cut_infos]  # type: List[str]

Dlpim_electron_cut_infos = [sig_b_electron_id_cut, sig_b_electron_p_cut]  # type: List[CutInfo]
Dlpim_electron_cuts = [cut_info.cut_str for cut_info in Dlpim_electron_cut_infos]  # type: List[str]
Dlpim_muon_cut_infos = [sig_b_muon_id_cut, sig_b_muon_quality_cut, sig_b_muon_p_cut]  # type: List[CutInfo]
Dlpim_muon_cuts = [cut_info.cut_str for cut_info in Dlpim_muon_cut_infos]  # type: List[str]

B_Dlpim_decay_label_to_cut_list = {
    "B0:D_e_pi-": B_Dlpim_cuts + Dlpim_electron_cuts + [],
    "B0:D_mu_pi-": B_Dlpim_cuts + Dlpim_muon_cuts + [],
    "B-:D+_e_pi-": B_Dlpim_cuts + Dlpim_electron_cuts + [],
    "B-:D+_mu_pi-": B_Dlpim_cuts + Dlpim_muon_cuts + [],
    "B0:D*_e_pi-": B_Dlpim_cuts + Dlpim_electron_cuts + [],
    "B0:D*_mu_pi-": B_Dlpim_cuts + Dlpim_muon_cuts + [],
    "B-:D*+_e_pi-": B_Dlpim_cuts + Dlpim_electron_cuts + [],
    "B-:D*+_mu_pi-": B_Dlpim_cuts + Dlpim_muon_cuts + [],
}  # type: Dict[str, List[str]]

roe_setting_for_B_Dlpim = {"mask_list": roe_masks.roe_mask_list, "mask_name": "CleanROE"}  # type: ROESettingType
cs_setting_for_B_Dlpim = {"roe_mask": "CleanROE"}  # type: Dict[str, str]
vertex_fit_setting_for_B_Dlpim = vertex_fit_setting_for_B  # type: VertexingSettingType

extra_settings_for_B_Dlpim = {
    "roe": roe_setting_for_B_Dlpim,
    "cs": cs_setting_for_B_Dlpim,
    "vertex_fit": vertex_fit_setting_for_B_Dlpim,
}  # type: ExtraSettingsType

B_Dlpim_decay_label_to_extra = {
    "B0:D_e_pi-": extra_settings_for_B_Dlpim,
    "B0:D_mu_pi-": extra_settings_for_B_Dlpim,
    "B-:D+_e_pi-": extra_settings_for_B_Dlpim,
    "B-:D+_mu_pi-": extra_settings_for_B_Dlpim,
    "B0:D*_e_pi-": extra_settings_for_B_Dlpim,
    "B0:D*_mu_pi-": extra_settings_for_B_Dlpim,
    "B-:D*+_e_pi-": extra_settings_for_B_Dlpim,
    "B-:D*+_mu_pi-": extra_settings_for_B_Dlpim,
}  # type: Dict[str, ExtraSettingsType]

#####################
# B Decays with D** #
#####################
# For when B Decays to D** are taken from MC information.

B_D2star_decay_channels = {
    "B0:D1_e": "B0 -> D_1- e+",
    "B0:D1_mu": "B0 -> D_1- mu+",
    "B0:D0*_e": "B0 -> D_0*- e+",
    "B0:D0*_mu": "B0 -> D_0*- mu+",
    "B0:D'1_e": "B0 -> D'_1- e+",
    "B0:D'1_mu": "B0 -> D'_1- mu+",
    "B0:D2*_e": "B0 -> D_2*- e+",
    "B0:D2*_mu": "B0 -> D_2*- mu+",
    "B-:D10_e": "B- -> D_10 e-",
    "B-:D10_mu": "B- -> D_10 mu-",
    "B-:D0*0_e": "B- -> D_0*0 e-",
    "B-:D0*0_mu": "B- -> D_0*0 mu-",
    "B-:D'10_e": "B- -> D'_10 e-",
    "B-:D'10_mu": "B- -> D'_10 mu-",
    "B-:D2*0_e": "B- -> D_2*0 e-",
    "B-:D2*0_mu": "B- -> D_2*0 mu-",
}  # type: Dict[str, str]

B_D2star_decay_label_to_decay_id = {
    "B0:D1_e": 151,
    "B0:D1_mu": 152,
    "B0:D0*_e": 153,
    "B0:D0*_mu": 154,
    "B0:D'1_e": 155,
    "B0:D'1_mu": 156,
    "B0:D2*_e": 157,
    "B0:D2*_mu": 158,
    "B-:D10_e": 161,
    "B-:D10_mu": 162,
    "B-:D0*0_e": 163,
    "B-:D0*0_mu": 164,
    "B-:D'10_e": 165,
    "B-:D'10_mu": 166,
    "B-:D2*0_e": 167,
    "B-:D2*0_mu": 168,
}  # type: Dict[str, int]

B_D2star_decay_label_to_chain = {
    "B0:D1_e": "B0:D1_e{suf} -> D_1-:all{suf} e+:fsp{suf} nu_e:mc{suf}",
    "B0:D1_mu": "B0:D1_mu{suf} -> D_1-:all{suf} mu+:fsp{suf} nu_mu:mc{suf}",
    "B0:D0*_e": "B0:D0*_e{suf} -> D_0*-:all{suf} e+:fsp{suf} nu_e:mc{suf}",
    "B0:D0*_mu": "B0:D0*_mu{suf} -> D_0*-:all{suf} mu+:fsp{suf} nu_mu:mc{suf}",
    "B0:D'1_e": "B0:D'1_e{suf} -> D'_1-:all{suf} e+:fsp{suf} nu_e:mc{suf}",
    "B0:D'1_mu": "B0:D'1_mu{suf} -> D'_1-:all{suf} mu+:fsp{suf} nu_mu:mc{suf}",
    "B0:D2*_e": "B0:D2*_e{suf} -> D_2*-:all{suf} e+:fsp{suf} nu_e:mc{suf}",
    "B0:D2*_mu": "B0:D2*_mu{suf} -> D_2*-:all{suf} mu+:fsp{suf} nu_mu:mc{suf}",
    "B-:D10_e": "B-:D10_e{suf} -> D_10:all{suf} e-:fsp{suf} anti-nu_e:mc{suf}",
    "B-:D10_mu": "B-:D10_mu{suf} -> D_10:all{suf} mu-:fsp{suf} anti-nu_mu:mc{suf}",
    "B-:D0*0_e": "B-:D0*0_e{suf} -> D_0*0:all{suf} e-:fsp{suf} anti-nu_e:mc{suf}",
    "B-:D0*0_mu": "B-:D0*0_mu{suf} -> D_0*0:all{suf} mu-:fsp{suf} anti-nu_mu:mc{suf}",
    "B-:D'10_e": "B-:D'10_e{suf} -> D'_10:all{suf} e-:fsp{suf} anti-nu_e:mc{suf}",
    "B-:D'10_mu": "B-:D'10_mu{suf} -> D'_10:all{suf} mu-:fsp{suf} anti-nu_mu:mc{suf}",
    "B-:D2*0_e": "B-:D2*0_e{suf} -> D_2*0:all{suf} e-:fsp{suf} anti-nu_e:mc{suf}",
    "B-:D2*0_mu": "B-:D2*0_mu{suf} -> D_2*0:all{suf} mu-:fsp{suf} anti-nu_mu:mc{suf}",
}  # type: Dict[str, str]

B_D2star_cuts = []  # type: List[str]

B_D2star_decay_label_to_cut_list = {
    "B0:D1_e": B_D2star_cuts,
    "B0:D1_mu": B_D2star_cuts,
    "B0:D0*_e": B_D2star_cuts,
    "B0:D0*_mu": B_D2star_cuts,
    "B0:D'1_e": B_D2star_cuts,
    "B0:D'1_mu": B_D2star_cuts,
    "B0:D2*_e": B_D2star_cuts,
    "B0:D2*_mu": B_D2star_cuts,
    "B-:D10_e": B_D2star_cuts,
    "B-:D10_mu": B_D2star_cuts,
    "B-:D0*0_e": B_D2star_cuts,
    "B-:D0*0_mu": B_D2star_cuts,
    "B-:D'10_e": B_D2star_cuts,
    "B-:D'10_mu": B_D2star_cuts,
    "B-:D2*0_e": B_D2star_cuts,
    "B-:D2*0_mu": B_D2star_cuts,
}  # type: Dict[str, List[str]]

roe_setting_for_B_D2star = {"mask_list": roe_masks.roe_mask_list, "mask_name": "CleanROE"}  # type: ROESettingType
cs_setting_for_B_D2star = {"roe_mask": "CleanROE"}  # type: Dict[str, str]

extra_settings_for_B_D2star = {
    "roe": roe_setting_for_B_D2star,
    "cs": cs_setting_for_B_D2star,
}  # type: ExtraSettingsType

B_D2star_decay_label_to_extra = {
    "B0:D1_e": extra_settings_for_B_D2star,
    "B0:D1_mu": extra_settings_for_B_D2star,
    "B0:D0*_e": extra_settings_for_B_D2star,
    "B0:D0*_mu": extra_settings_for_B_D2star,
    "B0:D'1_e": extra_settings_for_B_D2star,
    "B0:D'1_mu": extra_settings_for_B_D2star,
    "B0:D2*_e": extra_settings_for_B_D2star,
    "B0:D2*_mu": extra_settings_for_B_D2star,
    "B-:D10_e": extra_settings_for_B_D2star,
    "B-:D10_mu": extra_settings_for_B_D2star,
    "B-:D0*0_e": extra_settings_for_B_D2star,
    "B-:D0*0_mu": extra_settings_for_B_D2star,
    "B-:D'10_e": extra_settings_for_B_D2star,
    "B-:D'10_mu": extra_settings_for_B_D2star,
    "B-:D2*0_e": extra_settings_for_B_D2star,
    "B-:D2*0_mu": extra_settings_for_B_D2star,
}  # type: Dict[str, ExtraSettingsType]

##############
# D** Decays #
##############

D2star_decay_channels = {
    "D_0*+:Dpi0": "D_0*+ -> D+ pi0",
    "D_0*0:D0pi0": "D_0*0 -> D0 pi0",
    "D_1+:D*pi0": "D_1+ -> D*+ pi0",
    "D_10:D*0pi0": "D_10 -> D*0 pi0",
    "D'_1+:D*pi0": "D'_1+ -> D*+ pi0",
    "D'_10:D*0pi0": "D'_10 -> D*0 pi0",
    "D_2*+:D*pi0": "D_2*+ -> D*+ pi0",
    "D_2*+:Dpi0": "D_2*+ -> D+ pi0",
    "D_2*0:D*0pi0": "D_2*0 -> D*0 pi0",
    "D_2*0:D0pi0": "D_2*0 -> D0 pi0",
}  # type: Dict[str, str]

D2star_decay_label_to_decay_id = {
    "D_0*+:Dpi0": 50,
    "D_0*0:D0pi0": 51,
    "D_1+:D*pi0": 52,
    "D_10:D*0pi0": 53,
    "D'_1+:D*pi0": 54,
    "D'_10:D*0pi0": 55,
    "D_2*+:D*pi0": 56,
    "D_2*+:Dpi0": 57,
    "D_2*0:D*0pi0": 58,
    "D_2*0:D0pi0": 59,
}  # type: Dict[str, int]

D2star_decay_label_to_chain = {
    "D_0*+:Dpi0": "D_0*+:Dpi0{suf} -> D+:all{suf} pi0:fsp{suf}",
    "D_0*0:D0pi0": "D_0*0:D0pi0{suf} -> D0:all{suf} pi0:fsp{suf}",
    "D_1+:D*pi0": "D_1+:D*pi0{suf} -> D*+:all{suf} pi0:fsp{suf}",
    "D_10:D*0pi0": "D_10:D*0pi0{suf} -> D*0:all{suf} pi0:fsp{suf}",
    "D'_1+:D*pi0": "D'_1+:D*pi0{suf} -> D*+:all{suf} pi0:fsp{suf}",
    "D'_10:D*0pi0": "D'_10:D*0pi0{suf} -> D*0:all{suf} pi0:fsp{suf}",
    "D_2*+:D*pi0": "D_2*+:D*pi0{suf} -> D*+:all{suf} pi0:fsp{suf}",
    "D_2*+:Dpi0": "D_2*+:Dpi0{suf} -> D+:all{suf} pi0:fsp{suf}",
    "D_2*0:D*0pi0": "D_2*0:D*0pi0{suf} -> D*0:all{suf} pi0:fsp{suf}",
    "D_2*0:D0pi0": "D_2*0:D0pi0{suf} -> D0:all{suf} pi0:fsp{suf}",
}  # type: Dict[str, str]

D2star_cuts = []  # type: List[str]

D2star_decay_label_to_cut_list = {
    "D_0*+:Dpi0": D2star_cuts,
    "D_0*0:D0pi0": D2star_cuts,
    "D_1+:D*pi0": D2star_cuts,
    "D_10:D*0pi0": D2star_cuts,
    "D'_1+:D*pi0": D2star_cuts,
    "D'_10:D*0pi0": D2star_cuts,
    "D_2*+:D*pi0": D2star_cuts,
    "D_2*+:Dpi0": D2star_cuts,
    "D_2*0:D*0pi0": D2star_cuts,
    "D_2*0:D0pi0": D2star_cuts,
}  # type: Dict[str, List[str]]

#############
# D* Decays #
#############
Dstar_decay_channels = {
    "D*0:D0pi0": "D*0 -> D0 pi0",
    "D*0:D0gamma": "D*0 -> D0 gamma",
    "D*+:D0pi+": "D*+ -> D0 pi+",
    "D*+:D+pi0": "D*+ -> D+ pi0",
}  # type: Dict[str, str]

Dstar_decay_label_to_decay_id = {
    "D*0:D0pi0": 21,
    "D*0:D0gamma": 22,
    "D*+:D0pi+": 23,
    "D*+:D+pi0": 24,
}  # type: Dict[str, int]

Dstar_decay_label_to_chain = {
    "D*0:D0pi0": "D*0:D0pi0{suf} -> D0:all{suf} pi0:fsp{suf}",
    "D*0:D0gamma": "D*0:D0gamma{suf} -> D0:all{suf} gamma:fsp{suf}",
    "D*+:D0pi+": "D*+:D0pi+{suf} -> D0:all{suf} pi+:fsp{suf}",
    "D*+:D+pi0": "D*+:D+pi0{suf} -> D+:all{suf} pi0:fsp{suf}",
}  # type: Dict[str, str]

D_Dstar_p_cut = CutInfo(
    cut_str="useCMSFrame(p) < 3.0",
    cut_label=LatexObsNames.momentum_p_cms[:-1] + r" < 3.0\," + LatexUnits.gev[1:],
    cut_name="Dstar_p_cut",
    cut_var="useCMSFrame(p)",
    cut_var_latex_label=LatexObsNames.momentum_p_cms,
    cut_type="<",
    cut_limits=3.0,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
D_Dstar_mass_diff_155_cut = CutInfo(
    cut_str="daughterMotherDiffOf(0, M) < 0.155",
    cut_label=LatexObsNames.d_d_st_delta_m[:-1] + r" < 155\," + LatexUnits.mev[1:],
    cut_name="delta_M_cut",
    cut_var="daughterMotherDiffOf(0, M)",
    cut_var_latex_label=LatexObsNames.d_d_st_delta_m,
    cut_type="<",
    cut_limits=0.155,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
D_Dstar_mass_diff_130_155_cut = CutInfo(
    cut_str="0.13 < daughterMotherDiffOf(0, M) < 0.155",
    cut_label=LatexObsNames.d_d_st_delta_m[:-1] + r"\;\in\;(130, 155)\," + LatexUnits.mev[1:],
    cut_name="delta_M_cut",
    cut_var="daughterMotherDiffOf(0, M)",
    cut_var_latex_label=LatexObsNames.d_d_st_delta_m,
    cut_type=("<", "<"),
    cut_limits=(0.13, 0.155),
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
D_Dstar_mass_diff_160_cut = CutInfo(
    cut_str="daughterMotherDiffOf(0, M) < 0.16",
    cut_label=LatexObsNames.d_d_st_delta_m[:-1] + r" < 160\," + LatexUnits.mev[1:],
    cut_name="delta_M_cut",
    cut_var="daughterMotherDiffOf(0, M)",
    cut_var_latex_label=LatexObsNames.d_d_st_delta_m,
    cut_type="<",
    cut_limits=0.16,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
D_daughter_pi_p_cut = CutInfo(
    cut_str="useCMSFrame(daughter(1, p)) < 0.4",
    cut_label=LatexObsNames.d_st_meson_pi_cms_p[:-1] + r" < 400\," + LatexUnits.mev[1:],
    cut_name="pi_p_cut",
    cut_var="useCMSFrame(daughter(1, p))",
    cut_var_latex_label=LatexObsNames.d_st_meson_pi_cms_p,
    cut_type="<",
    cut_limits=0.4,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)

Dstar_cut_infos = [D_Dstar_p_cut]  # type: List[CutInfo]
Dstar_cuts = [cut_info.cut_str for cut_info in Dstar_cut_infos]  # type: List[str]

Dstar_decay_label_to_cut_info_list = {
    "D*0:D0pi0": Dstar_cut_infos + [D_Dstar_mass_diff_160_cut],
    "D*0:D0gamma": Dstar_cut_infos + [D_Dstar_mass_diff_130_155_cut],
    "D*+:D0pi+": Dstar_cut_infos + [D_Dstar_mass_diff_155_cut, D_daughter_pi_p_cut],
    "D*+:D+pi0": Dstar_cut_infos + [D_Dstar_mass_diff_155_cut],
}  # type: Dict[str, List[CutInfo]]
Dstar_decay_label_to_cut_list = {}  # type: Dict[str, List[str]]
for channel, cut_infos in Dstar_decay_label_to_cut_info_list.items():
    Dstar_decay_label_to_cut_list[channel] = [cut_info.cut_str for cut_info in cut_infos]

############
# D Decays #
############

# "D0 -> K_S0 K_S0"  1.8 E-4
# "D0 -> pi+ pi-"  1.42 E-3
# "D0 -> pi+ pi- pi0"  1.47 %
# "D0 -> pi0 pi0"  8.26 E-4
# "D0 -> K_S0 pi0 pi0"  9.1 E-3

# "D+ -> K- pi+ pi+ pi+ pi-"  4.1 E-4
# "D+ -> pi+ pi0"  1.24 E-3
# "D+ -> pi+ pi+ pi-"  3.29 E-3
# "D+ -> pi+ pi0 pi0"  4.7 E-3

D_decay_channels = {
    "D+:Kpipi": "D+ -> K- pi+ pi+",  # 9.46 %
    "D+:Kpipipi": "D+ -> K- pi+ pi+ pi0",  # 6.14 %
    "D+:Kpipipipi": "D+ -> K- pi+ pi+ pi+ pi-",  # 5.8 E-3
    "D+:KSpi": "D+ -> K_S0 pi+",  # 1.53 %
    "D+:KSpipi": "D+ -> K_S0 pi+ pi0",  # 7.24 %
    "D+:KSpipipi": "D+ -> K_S0 pi+ pi+ pi-",  # 3.05 %
    "D+:KSK": "D+ -> K_S0 K+",  # 2.95 E-3
    "D+:KKpi": "D+ -> K+ K- pi+",  # 9.96 E-3
    "D0:Kpi": "D0 -> K- pi+",  # 3.93 %
    "D0:Kpipi": "D0 -> K- pi+ pi0",  # 14.3 %
    "D0:Kpipipi": "D0 -> K- pi+ pi+ pi-",  # 8.07 %
    "D0:Kpipipipi": "D0 -> K- pi+ pi+ pi- pi0",  # 4.2 %
    "D0:KSpi": "D0 -> K_S0 pi0",  # 1.20 %
    "D0:KSpipi": "D0 -> K_S0 pi+ pi-",  # 2.85 %
    "D0:KSpipipi": "D0 -> K_S0 pi+ pi- pi0",  # 5.2 %
    "D0:KK": "D0 -> K- K+",  # 4.01 E-3
}  # type: Dict[str, str]

D_decay_label_to_decay_id = {
    "D+:Kpipi": 31,
    "D+:Kpipipi": 32,
    "D+:Kpipipipi": 33,
    "D+:KSpi": 34,
    "D+:KSpipi": 35,
    "D+:KSpipipi": 36,
    "D+:KSK": 37,
    "D+:KKpi": 38,
    "D0:Kpi": 41,
    "D0:Kpipi": 42,
    "D0:Kpipipi": 43,
    "D0:Kpipipipi": 44,
    "D0:KSpi": 45,
    "D0:KSpipi": 46,
    "D0:KSpipipi": 47,
    "D0:KK": 48,
}  # type: Dict[str, int]

D_decay_label_to_chain = {
    "D+:Kpipi": "D+:Kpipi{suf} -> K-:fsp{suf} pi+:fsp{suf} pi+:fsp{suf}",
    "D+:Kpipipi": "D+:Kpipipi{suf} -> K-:fsp{suf} pi+:fsp{suf} pi+:fsp{suf} pi0:fsp{suf}",
    "D+:Kpipipipi": "D+:Kpipipipi{suf} -> K-:fsp{suf} pi+:fsp{suf} pi+:fsp{suf} pi+:fsp{suf} pi-:fsp{suf}",
    "D+:KSpi": "D+:KSpi{suf} -> K_S0:fsp{suf} pi+:fsp{suf}",
    "D+:KSpipi": "D+:KSpipi{suf} -> K_S0:fsp{suf} pi+:fsp{suf} pi0:fsp{suf}",
    "D+:KSpipipi": "D+:KSpipipi{suf} -> K_S0:fsp{suf} pi+:fsp{suf} pi+:fsp{suf} pi-:fsp{suf}",
    "D+:KSK": "D+:KSK{suf} -> K_S0:fsp{suf} K+:fsp{suf}",
    "D+:KKpi": "D+:KKpi{suf} -> K+:fsp{suf} K-:fsp{suf} pi+:fsp{suf}",
    "D0:Kpi": "D0:Kpi{suf} -> K-:fsp{suf} pi+:fsp{suf}",
    "D0:Kpipi": "D0:Kpipi{suf} -> K-:fsp{suf} pi+:fsp{suf} pi0:fsp{suf}",
    "D0:Kpipipi": "D0:Kpipipi{suf} -> K-:fsp{suf} pi+:fsp{suf} pi+:fsp{suf} pi-:fsp{suf}",
    "D0:Kpipipipi": "D0:Kpipipipi{suf} -> K-:fsp{suf} pi+:fsp{suf} pi+:fsp{suf} pi-:fsp{suf} pi0:fsp{suf}",
    "D0:KSpi": "D0:KSpi{suf} -> K_S0:fsp{suf} pi0:fsp{suf}",
    "D0:KSpipi": "D0:KSpipi{suf} -> K_S0:fsp{suf} pi+:fsp{suf} pi-:fsp{suf}",
    "D0:KSpipipi": "D0:KSpipipi{suf} -> K_S0:fsp{suf} pi+:fsp{suf} pi-:fsp{suf} pi0:fsp{suf}",
    "D0:KK": "D0:KK{suf} -> K-:fsp{suf} K+:fsp{suf}",
}  # type: Dict[str, str]

D_cut_infos = [D_Dstar_p_cut]  # type: List[CutInfo]
D_cuts = [cut_info.cut_str for cut_info in D_cut_infos]  # type: List[str]

d_sig_Mbf_cut_info = CutInfo(
    cut_str="abs(SigMBF) <= 4.",
    cut_label=LatexObsNames.mass_abs_sig_mbf[:-1] + r" < 4.0$",
    cut_name="sig_Mbf_cut",
    cut_var="abs(SigMBF)",
    cut_var_latex_label=LatexObsNames.mass_abs_sig_mbf,
    cut_type="<=",
    cut_limits=4.0,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)
d_dM_04_cut_info = CutInfo(
    cut_str="abs(dM) < 0.04",
    cut_label=LatexObsNames.mass_abs_delta_m[:-1] + r" < 40\," + LatexUnits.mev[1:],
    cut_name="delta_M_cut",
    cut_var="abs(dM)",
    cut_var_latex_label=LatexObsNames.mass_abs_delta_m,
    cut_type="<",
    cut_limits=0.04,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
d_dM_05_cut_info = CutInfo(
    cut_str="abs(dM) < 0.05",
    cut_label=LatexObsNames.mass_abs_delta_m[:-1] + r" < 50\," + LatexUnits.mev[1:],
    cut_name="delta_M_cut",
    cut_var="abs(dM)",
    cut_var_latex_label=LatexObsNames.mass_abs_delta_m,
    cut_type="<",
    cut_limits=0.05,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
d_1st_daughter_p_cut_info = CutInfo(
    cut_str="useCMSFrame(daughter(1, p)) > 0.2",
    cut_label=LatexObsNames.d_st_meson_pi_1_cms_p[:-1] + r" > 200\," + LatexUnits.mev[1:],
    cut_name="pion_p_cut",
    cut_var="useCMSFrame(daughter(1, p))",
    cut_var_latex_label=LatexObsNames.d_st_meson_pi_1_cms_p,
    cut_type=">",
    cut_limits=0.2,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
d_2nd_daughter_p_cut_info = CutInfo(
    cut_str="useCMSFrame(daughter(2, p)) > 0.2",
    cut_label=LatexObsNames.d_st_meson_pi_2_cms_p[:-1] + r" > 200\," + LatexUnits.mev[1:],
    cut_name="pion_p_cut",
    cut_var="useCMSFrame(daughter(2, p))",
    cut_var_latex_label=LatexObsNames.d_st_meson_pi_2_cms_p,
    cut_type=">",
    cut_limits=0.2,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
d_3rd_daughter_p_cut_info = CutInfo(
    cut_str="useCMSFrame(daughter(3, p)) > 0.2",
    cut_label=LatexObsNames.d_st_meson_pi_3_cms_p[:-1] + r" > 200\," + LatexUnits.mev[1:],
    cut_name="pion_p_cut",
    cut_var="useCMSFrame(daughter(3, p))",
    cut_var_latex_label=LatexObsNames.d_st_meson_pi_3_cms_p,
    cut_type=">",
    cut_limits=0.2,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)

D_decay_label_to_cut_info_list = {
    "D+:Kpipi": D_cut_infos + [d_sig_Mbf_cut_info],
    "D+:Kpipipi": D_cut_infos + [d_dM_04_cut_info, d_3rd_daughter_p_cut_info],
    "D+:Kpipipipi": D_cut_infos + [d_sig_Mbf_cut_info],
    "D+:KSpi": D_cut_infos + [d_dM_04_cut_info],
    "D+:KSpipi": D_cut_infos + [d_dM_04_cut_info, d_2nd_daughter_p_cut_info],
    "D+:KSpipipi": D_cut_infos + [d_dM_04_cut_info],
    "D+:KSK": D_cut_infos + [d_dM_04_cut_info],
    "D+:KKpi": D_cut_infos + [d_dM_04_cut_info],
    "D0:Kpi": D_cut_infos + [d_sig_Mbf_cut_info],
    "D0:Kpipi": D_cut_infos + [d_dM_05_cut_info, d_2nd_daughter_p_cut_info],
    "D0:Kpipipi": D_cut_infos + [d_sig_Mbf_cut_info],
    "D0:Kpipipipi": D_cut_infos + [d_dM_04_cut_info],
    "D0:KSpi": D_cut_infos + [d_dM_05_cut_info, d_1st_daughter_p_cut_info],
    "D0:KSpipi": D_cut_infos + [d_dM_04_cut_info],
    "D0:KSpipipi": D_cut_infos + [d_dM_04_cut_info, d_3rd_daughter_p_cut_info],
    "D0:KK": D_cut_infos + [d_sig_Mbf_cut_info],
}  # type: Dict[str, List[CutInfo]]
D_decay_label_to_cut_list = {}  # type: Dict[str, List[str]]
for channel, cut_infos in D_decay_label_to_cut_info_list.items():
    D_decay_label_to_cut_list[channel] = [cut_info.cut_str for cut_info in cut_infos]


##############
# Tau Decays #
##############
tau_decay_channels = {
    "tau-:e": "tau- -> e- anti-nu_e nu_tau",
    "tau-:mu": "tau- -> mu- anti-nu_mu nu_tau",
}  # type: Dict[str, str]

tau_decay_label_to_decay_id = {
    "tau-:e": 2021,
    "tau-:mu": 2022,
}  # type: Dict[str, int]

tau_decay_label_to_chain = {
    "tau-:e": "tau-:e{suf} -> e-:fsp{suf} anti-nu_e:mc{suf} nu_tau:mc{suf}",
    "tau-:mu": "tau-:mu{suf} -> mu-:fsp{suf} anti-nu_mu:mc{suf} nu_tau:mc{suf}",
}  # type: Dict[str, str]

tau_decay_label_to_cut_list = {
    "tau-:e": [],
    "tau-:mu": [],
}  # type: Dict[str, List[str]]

###############
# pi0 Decays #
###############
pi0_decay_channels = {
    "pi0:gammagamma": "pi0 -> gamma gamma",
}  # type: Dict[str, str]

pi0_decay_label_to_decay_id = {
    "pi0:gammagamma": 3011,
}  # type: Dict[str, int]

pi0_decay_label_to_chain = {
    "pi0:gammagamma": "pi0:gammagamma{suf} -> gamma:fsp{suf} gamma:fsp{suf}",
}  # type: Dict[str, str]

pi0_reco_dm_cut = CutInfo(
    cut_str="abs(dM) < 0.03",
    cut_label=LatexObsNames.pi0_abs_delta_m[:-1] + r" < 30\," + LatexUnits.mev[1:],
    cut_name="pi0_reco_dm_cut",
    cut_var="abs(dM)",
    cut_var_latex_label=LatexObsNames.pi0_abs_delta_m,
    cut_type="<",
    cut_limits=0.03,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
pi0_reco_sig_mbf_cut = CutInfo(
    cut_str="abs(SigMBF) < 3.0",
    cut_label=LatexObsNames.pi0_abs_sig_mbf[:-1] + r" < 3.0$",
    cut_name="pi0_reco_sig_mbf_cut",
    cut_var="abs(SigMBF)",
    cut_var_latex_label=LatexObsNames.pi0_abs_sig_mbf,
    cut_type="<",
    cut_limits=3.0,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)
pi0_reco_energy_cut = CutInfo(
    cut_str="E > 0.1",
    cut_label=LatexObsNames.pi0_e[:-1] + r" > 100\," + LatexUnits.mev[1:],
    cut_name="pi0_reco_energy_cut",
    cut_var="E",
    cut_var_latex_label=LatexObsNames.pi0_e,
    cut_type=">",
    cut_limits=0.1,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)


pi0_reco_cut_infos = [pi0_reco_dm_cut, pi0_reco_sig_mbf_cut, pi0_reco_energy_cut]  # type: List[CutInfo]
pi0_reco_cuts = [cut_info.cut_str for cut_info in pi0_reco_cut_infos]  # type: List[str]
pi0_decay_label_to_cut_list = {
    "pi0:gammagamma": pi0_reco_cuts,
}  # type: Dict[str, List[str]]

###############
# K_S0 Decays #
###############
KS0_decay_channels = {
    "K_S0:pipi": "K_S0 -> pi+ pi-",
}  # type: Dict[str, str]

KS0_decay_label_to_decay_id = {
    "K_S0:pipi": 2011,
}  # type: Dict[str, int]

KS0_decay_label_to_chain = {
    "K_S0:pipi": "K_S0:pipi{suf} -> pi+:fsp{suf} pi-:fsp{suf}",
}  # type: Dict[str, str]

KS0_reco_dm_cut = CutInfo(
    cut_str="abs(dM) < 0.1",
    cut_label=LatexObsNames.k_short_abs_delta_m[:-1] + r" < 100\," + LatexUnits.mev[1:],
    cut_name="KS0_reco_dm_cut",
    cut_var="abs(dM)",
    cut_var_latex_label=LatexObsNames.k_short_abs_delta_m,
    cut_type="<",
    cut_limits=0.1,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
KS0_reco_cut_infos = [KS0_reco_dm_cut]  # type: List[CutInfo]
KS0_reco_cuts = [cut_info.cut_str for cut_info in KS0_reco_cut_infos]  # type: List[str]
KS0_decay_label_to_cut_list = {
    "K_S0:pipi": KS0_reco_cuts,
}  # type: Dict[str, List[str]]

############################
# Merging the dictionaries #
############################

# Decay Channels that can only be reconstructed from generator level information
generator_level_only_channels = [
    Btau_decay_channels,
    tau_decay_channels,
    B_D2star_decay_channels,
    D2star_decay_channels,
]  # type: List[Dict[str, str]]

# Decay Channels that contain additional neutrinos and therefore require isSignalAcceptMissingNeutrino
channels_with_neutrinos = [
    B_decay_channels,
    B_Dlpi0_decay_channels,
    B_Dlpim_decay_channels,
    BTag_decay_channels,
    U4S_decay_channels,
]  # type: List[Dict[str, str]]

decay_channel_dict = decay_channels_utils.prepare_decay_channel_dict(
    local_vars=locals(),
    channels_with_neutrinos=channels_with_neutrinos,
    gen_level_only_channels=generator_level_only_channels,
)

if __name__ == "__main__":
    decay_channels_utils.print_decay_channels_table(decay_channel_dict)
