"""
Steering file for testing the n-tuple file generation for the B -> D(*) tau nu analysis.

Contains the reconstruction analog to reconstruction.py but produces additional output.

Felix Metzner 2018
"""

import basf2

from typing import List

from modularAnalysis import setAnalysisConfigParams, statistics, massVertexKFit, fillParticleListsFromMC

from analysistools.processing import job_interface
from analysistools.processing import path_utilities

from rdstar.online_analysis import b2bii_utils
from rdstar.online_analysis.decay_utilities import Reconstructor
from rdstar.online_analysis import observables, decay_utilities, decay_channels


cmd_args = job_interface.get_commandline_args()  # type: job_interface.BatchJobCMDLineArgs

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config=cmd_args.config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()  # type: basf2.Path

b2bii_utils.b2bii_read_and_convert(
    path=main_path,
    input_files=cmd_args.input_files,
    input_sequences=cmd_args.input_sequences,
)

setAnalysisConfigParams(
    configParametersAndValues={"mcMatchingVersion": "Belle"},
    path=main_path,
)

main_path.add_module(module="CurlerCloneMarker")
decay_utilities.fsp.fsp_collection.get_final_state_particles(
    path=main_path,
    use_mc_info=cmd_args.config["use_mc_info"],
)

reconstructor = Reconstructor(
    path=main_path,
    config=cmd_args.config,
    decay_channels_module=decay_channels,
)

#######
# pi0 #
#######

piNeutralVars = [
    "PDG",
    "matchedMC(PDG)",
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcMother(mcMother(PDG)))",
    "mcMother(mcMother(mcMother(mcMother(PDG))))",
    "mcMother(mcDaughter(0,PDG))",
    "isSignal",
    "M",
    "dM",
    "InvM",
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "daughterInvariantMass(0, 1)",
    "chiProb",
    "abs(SigMBF)",
    "daughterAngle(0,1)",
    "useRestFrame(daughterAngle(0,1))",
    "daughterAngleInBetween(0,1)",
    "useRestFrame(daughterAngleInBetween(0,1))",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    "daughter(0, goodBelleGamma)",
    "daughter(0, clusterE9E21)",
    "daughter(0, clusterE1E9)",
    "daughter(0, clusterE9E25)",
    "daughter(0, clusterNHits)",
    "daughter(0, clusterHighestE)",
    "daughter(0, E)",
    "daughter(1, goodBelleGamma)",
    "daughter(1, clusterE9E25)",
    "daughter(1, clusterNHits)",
    "daughter(1, clusterHighestE)",
    "daughter(1, E)",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "atcPIDBelle(2,3)",
    "extraInfo(ksnbStandard)",
    "extraInfo(goodKs)",
    "dr",
    "dz",
    "daughterDiffOf(0, 1, E)",
    "daughterSumOf(E)",
    "formula(daughterDiffOf(0, 1, E) / daughterSumOf(E))",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["piNeutral_from_list_ntuple"],
    particleList="pi0:fsp",
    variables=piNeutralVars,
)

reconstructor.reconstruct_pi0()

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["piNeutral_reco_ntuple"],
    particleList="pi0:reco",
    variables=piNeutralVars,
)

massVertexKFit(
    list_name="pi0:reco",
    conf_level=-2.0,
    decay_string="",
    path=main_path,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["piNeutral_reco_vertex_ntuple"],
    particleList="pi0:reco",
    variables=piNeutralVars,
)

#######
# pi+ #
#######

piChargedVars = [
    "PDG",
    "matchedMC(PDG)",
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcMother(mcMother(PDG)))",
    "mcMother(mcMother(mcMother(mcMother(PDG))))",
    "mcMother(mcDaughter(0,PDG))",
    "isSignal",
    "M",
    "dM",
    "InvM",
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "atcPIDBelle(2,3)",
    "extraInfo(ksnbStandard)",
    "extraInfo(goodKs)",
    "dr",
    "dz",
    "trackCurlerCloneIndicator",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["piCharged_ntuple"],
    particleList="pi+:fsp",
    variables=piChargedVars,
)

############
# e and mu #
############

leptonVars = [
    "PDG",
    "matchedMC(PDG)",
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcDaughter(0,PDG))",
    "mcDaughter(0,PDG)",
    "isSignal",
    "M",
    "dM",
    "InvM",
    "daughterInvariantMass(0, 1)",
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "atcPIDBelle(2,3)",
    "extraInfo(ksnbStandard)",
    "extraInfo(goodKs)",
    "dr",
    "dz",
    "pValue",
    "nCDCHits",
    "nSVDHits",
    "daughter(0, pValue)",
    "daughter(0, nCDCHits)",
    "daughter(0, nSVDHits)",
    "daughter(0, dr)",
    "daughter(0, dz)",
    "matchedMC(E)",
    "matchedMC(p)",
    "matchedMC(px)",
    "matchedMC(py)",
    "matchedMC(pz)",
    "trackCurlerCloneIndicator",
    "extraInfo(fsrCorrectedE)",
    "extraInfo(fsrCorrectedPx)",
    "extraInfo(fsrCorrectedPy)",
    "extraInfo(fsrCorrectedPz)",
    "extraInfo(fsrCorrectedP)",
    "extraInfo(fsrGammaE)",
    "extraInfo(fsrGammaPx)",
    "extraInfo(fsrGammaPy)",
    "extraInfo(fsrGammaPz)",
    "extraInfo(fsrGammaP)",
    "extraInfo(fsrCorrected)",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["e_ntuple"],
    particleList="e-:fsp",
    variables=leptonVars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["mu_ntuple"],
    particleList="mu-:fsp",
    variables=leptonVars,
)

########
# K_S0 #
########

KS0Vars = [
    "PDG",
    "matchedMC(PDG)",
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcMother(mcMother(PDG)))",
    "mcMother(mcMother(mcMother(mcMother(PDG))))",
    "mcMother(mcDaughter(0,PDG))",
    "isSignal",
    "M",
    "dM",
    "InvM",
    "abs(SigMBF)",
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "atcPIDBelle(2,3)",
    "extraInfo(ksnbStandard)",
    "extraInfo(goodKs)",
    "dr",
    "dz",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["K_S0_ntuple"],
    particleList="K_S0:fsp",
    variables=KS0Vars,
)

######
# K+ #
######

KVars = [
    "PDG",
    "matchedMC(PDG)",
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcMother(mcMother(PDG)))",
    "mcMother(mcMother(mcMother(mcMother(PDG))))",
    "mcMother(mcDaughter(0,PDG))",
    "isSignal",
    "M",
    "dM",
    "InvM",
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "atcPIDBelle(2,3)",
    "extraInfo(ksnbStandard)",
    "extraInfo(goodKs)",
    "dr",
    "dz",
    "trackCurlerCloneIndicator",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["K_ntuple"],
    particleList="K+:fsp",
    variables=KVars,
)

#########
# Gamma #
#########

gammaVars = [
    "PDG",
    "matchedMC(PDG)",
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcMother(mcMother(PDG)))",
    "mcMother(mcMother(mcMother(mcMother(PDG))))",
    "mcMother(mcDaughter(0,PDG))",
    "isSignal",
    "goodBelleGamma",
    "clusterE9E25",
    "clusterNHits",
    "clusterHighestE",
    "M",
    "dM",
    "InvM",
    "abs(SigMBF)",
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "atcPIDBelle(2,3)",
    "extraInfo(ksnbStandard)",
    "extraInfo(goodKs)",
    "dr",
    "dz",
    "matchedMC(E)",
    "matchedMC(px)",
    "matchedMC(py)",
    "matchedMC(pz)",
    "matchedMC(p)",
    # "extraInfo(fsrCorrected)",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["gamma_ntuple"],
    particleList="gamma:fsp",
    variables=gammaVars,
)

if cmd_args.config["use_mc_info"] is True:
    b_direct_gen_coll = observables.get_direct_gen_level_b_collection(cmd_args.config)

    fillParticleListsFromMC([("B-:mc", "")], addDaughters=True, path=main_path)
    fillParticleListsFromMC([("B0:mc", "")], addDaughters=True, path=main_path)
    main_path.add_module(
        module="VariablesToNtuple",
        fileName=cmd_args.output_dict["B_minus_generator_direct_info_ntuple"],
        particleList="B-:mc",
        variables=b_direct_gen_coll.as_list(),
    )
    main_path.add_module(
        module="VariablesToNtuple",
        fileName=cmd_args.output_dict["B_zero_generator_direct_info_ntuple"],
        particleList="B0:mc",
        variables=b_direct_gen_coll.as_list(),
    )

basf2.print_path(path=main_path)
basf2.process(path=main_path)
print(statistics)
