import os
import copy
import uproot
import logging
import numpy as np
import pandas as pd
import b2luigi as luigi

from typing import Union, Optional, Tuple, List, Dict, Generator, Any

from analysistools.dataframe_tools.uproot_compatibility import uproottree2df
from analysistools.batch_tools.file_handler import Belle_url_mc_types, load_json_config

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.rdstar_samples import private_data_types
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.studies.low_level_cut_study.utilities.task_utilities import get_config_base_path
from rdstar.studies.low_level_cut_study.utilities.cut_study_plotter import LowLevelCutPlotter
from rdstar.studies.low_level_cut_study.utilities.data_mc_comparison_plotter import DataMCComparisonPlotter


logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
mpl_logger = logging.getLogger("matplotlib")
mpl_logger.setLevel(logging.WARNING)


__all__ = [
    "MergeNTuplesTask",
    "LowLevelCutsEvaluationTask",
    "LowLevelCutStudyMaster",
]


OfflineConfigDictType = Dict[str, Union[str, PathType, List[str], Dict[str, Union[str, PathType]]]]
BatchConfigDictType = Dict[str, Union[str, int, bool, PathType, List[str], Dict[str, Dict[str, Union[float, int, None]]]]]


class MergeNTuplesTask(utilities.BaseOfflineTask):
    config_path = luigi.Parameter(positional=False, hashed=True)  # type: PathType
    sample_type = luigi.Parameter(positional=False)  # type: str
    n_tuple_name = luigi.Parameter(positional=False)  # type: str
    n_tuple_samples = luigi.ListParameter(positional=False, hashed=True)  # type: List[str]

    base_dir_name = "CombinedNTuples"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.config_path)  # type: OfflineConfigDictType
        _working_dir = offline_config["working_dir"]  # type: PathType
        assert isinstance(_working_dir, str), (type(_working_dir).__name__, _working_dir)
        self.working_dir = _working_dir  # type: PathType

        self.input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

    @staticmethod
    def get_n_tuple_output_key(n_tuple_name: str) -> str:
        return f"{n_tuple_name}_total"

    def get_sample_input_key(self, sample_name: str) -> str:
        stream_tag = ""  # type: str
        if LowLevelCutStudyMaster.get_sample_type(sample_name=sample_name) == "Generic":
            stream_number = LowLevelCutStudyMaster.get_generic_sample_stream(offline_config_path=self.config_path)
            stream_tag = f"_{stream_number}"

        return f"{sample_name}{stream_tag}_{self.n_tuple_name}"

    def run_merging(self, dry_run: bool = False) -> Tuple[str, PathType]:
        n_tuple_dir = os.path.join(self.working_dir, self.base_dir_name, self.sample_type)  # type: PathType
        output_key = self.get_n_tuple_output_key(n_tuple_name=self.n_tuple_name)  # type: str
        output_path = os.path.join(n_tuple_dir, f"{output_key}.h5")  # type: PathType
        tmp_output_path = os.path.join(n_tuple_dir, f"temp_{output_key}.h5")  # type: PathType

        combos = [self.get_sample_input_key(sample_name=sample) for sample in self.n_tuple_samples]  # type: List[str]
        if not dry_run:
            if os.path.exists(path=output_path):
                raise RuntimeError(f"The output file\n\t{output_path}\nexists already")

            dfs = []  # type: List[pd.DataFrame]
            number_of_events = 0  # type: int
            for combo in combos:
                assert os.path.isfile(path=self.input_dict[combo]), self.input_dict[combo]
                assert os.path.splitext(p=self.input_dict[combo])[1] == ".root", os.path.splitext(
                    p=self.input_dict[combo]
                )
                data_file = uproot.open(self.input_dict[combo])
                tree = data_file[data_file.keys()[0]]
                df = uproottree2df(tree)
                df = df.loc[:, ~df.columns.duplicated()]
                number_of_events += len(np.unique(df.index))
                dfs.append(df)

            if len(dfs) > 1:
                total_n_tuple = pd.concat(dfs, ignore_index=True, sort=False)  # type: pd.DataFrame
                total_n_tuple.reset_index(drop=True, inplace=True)
                assert len(total_n_tuple.index) == len(np.unique(total_n_tuple.index)), (
                    len(total_n_tuple.index),
                    len(np.unique(total_n_tuple.index)),
                )
            elif len(dfs) == 1:
                total_n_tuple = dfs[0]
            else:
                raise RuntimeError(f"No DataFrames to be merged for n-tuple {self.n_tuple_name}...")

            assert len(total_n_tuple.index) == number_of_events, (len(total_n_tuple.index), number_of_events)

            total_n_tuple.to_hdf(path_or_buf=tmp_output_path, key="low_level")
            os.rename(src=tmp_output_path, dst=output_path)

        return output_key, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_merging()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_key, target_path = self.run_merging(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield DataCopyTask(offline_config_path=self.config_path)


class LowLevelCutsEvaluationTask(utilities.BaseOfflineTask):
    config_path = luigi.Parameter(positional=False)  # type: PathType
    n_tuple = luigi.Parameter(positional=False)  # type: str
    apply_pre_cuts = luigi.Parameter(positional=False)  # type: str
    variable_set = luigi.Parameter(positional=False)  # type: str
    remove_outliers = luigi.BoolParameter(positional=False)  # type: bool

    base_dir_name = "CutEvaluation"  # type: str
    relevant_sample_type = "Signal"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.config_path)  # type: OfflineConfigDictType
        _working_dir = offline_config["working_dir"]  # type: PathType
        assert isinstance(_working_dir, str), (type(_working_dir).__name__, _working_dir)
        self.working_dir = _working_dir  # type: PathType

        self.cut_evaluation_dir = os.path.join(self.working_dir, self.base_dir_name)  # type: PathType

        self.input_dict = {
            k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()
        }  # type: Dict[str, PathType]

    def run_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dir = os.path.join(self.cut_evaluation_dir, self.n_tuple)  # type: PathType

        plotter = LowLevelCutPlotter(
            n_tuple=self.n_tuple,
            target_path=output_dir,
            apply_pre_cuts=self.apply_pre_cuts,
            variable_set=self.variable_set,
            remove_outliers=self.remove_outliers,
        )
        output = plotter.generate_output_dict()  # type: Dict[str, PathType]

        if not dry_run:
            df = pd.read_hdf(self.input_dict[MergeNTuplesTask.get_n_tuple_output_key(n_tuple_name=self.n_tuple)])
            assert isinstance(df, pd.DataFrame)
            plotter.add_df(df=df)
            plot_dict = plotter.create_plots()
            assert plot_dict == output

        return output

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.run_evaluation(dry_run=True).items()
        }

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        sample_mapping = LowLevelCutStudyMaster.get_n_tuple_to_sample_name_mapping(offline_config_path=self.config_path)
        n_tuple_samples = sample_mapping[self.n_tuple][self.relevant_sample_type]  # type: List[str]
        yield MergeNTuplesTask(
            config_path=self.config_path,
            n_tuple_name=self.n_tuple,
            sample_type=self.relevant_sample_type,
            n_tuple_samples=n_tuple_samples,
        )


@luigi.inherits(LowLevelCutsEvaluationTask)
class LowLevelCutsDataMCComparisonTask(utilities.BaseOfflineTask):
    base_dir_name = "DataMCComparison"  # type: str

    data_sample_type_name = "DataY4S"  # type: str
    mc_sample_type_name = "Generic"  # type: str
    relevant_sample_types = [data_sample_type_name, mc_sample_type_name]  # type: List[str]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.config_path)  # type: OfflineConfigDictType
        _working_dir = offline_config["working_dir"]  # type: PathType
        assert isinstance(_working_dir, str), (type(_working_dir).__name__, _working_dir)
        self.working_dir = _working_dir  # type: PathType

        self.cut_evaluation_dir = os.path.join(self.working_dir, self.base_dir_name)  # type: PathType

        self.input_dict = {
            k: {kk: vv.path for kk, vv in vd.items()} for d in self.input() if isinstance(d, dict) for k, vd in d.items()
        }  # type: Dict[str, Dict[str, PathType]]

    def run_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dir = os.path.join(self.cut_evaluation_dir, self.n_tuple)  # type: PathType

        plotter = DataMCComparisonPlotter(
            n_tuple=self.n_tuple,
            target_path=output_dir,
            apply_pre_cuts=self.apply_pre_cuts,
            variable_set=self.variable_set,
            remove_outliers=self.remove_outliers,
        )
        output = plotter.generate_output_dict()  # type: Dict[str, PathType]

        if not dry_run:
            n_tuple_name = MergeNTuplesTask.get_n_tuple_output_key(n_tuple_name=self.n_tuple)
            data_df = pd.read_hdf(self.input_dict[self.data_sample_type_name][n_tuple_name])
            assert isinstance(data_df, pd.DataFrame)
            mc_df = pd.read_hdf(self.input_dict[self.mc_sample_type_name][n_tuple_name])
            assert isinstance(mc_df, pd.DataFrame)

            plotter.add_df(df=mc_df)
            plotter.add_data_df(data_df=data_df)
            plot_dict = plotter.create_plots()
            assert plot_dict == output

        return output

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.run_evaluation(dry_run=True).items()
        }

    def requires(self) -> Generator[Dict[str, utilities.TaskType], None, None]:
        sample_mapping = LowLevelCutStudyMaster.get_n_tuple_to_sample_name_mapping(offline_config_path=self.config_path)

        for sample_type in self.relevant_sample_types:
            yield {
                sample_type: MergeNTuplesTask(
                    config_path=self.config_path,
                    sample_type=sample_type,
                    n_tuple_name=self.n_tuple,
                    n_tuple_samples=sample_mapping[self.n_tuple][sample_type],
                )
            }


class LowLevelCutStudyMaster(luigi.WrapperTask):
    config_path = luigi.Parameter(positional=False)  # type: PathType

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.config_path)  # type: OfflineConfigDictType
        _working_dir = offline_config["working_dir"]  # type: PathType
        assert isinstance(_working_dir, str), (type(_working_dir).__name__, _working_dir)
        self.working_dir = _working_dir  # type: PathType

    @staticmethod
    def get_batch_config_dicts(offline_config_path: PathType) -> List[BatchConfigDictType]:
        offline_config = utilities.load_yaml_config(yaml_config_path=offline_config_path)  # type: OfflineConfigDictType
        _batch_configs = offline_config["batch_configs"]
        assert isinstance(_batch_configs, list), type(_batch_configs).__name__
        batch_configs = [p for p in _batch_configs if isinstance(p, str)]  # type: List[PathType]

        batch_config_paths = [
            os.path.join(get_config_base_path(), config) for config in batch_configs
        ]  # type: List[PathType]
        batch_config_dicts = [load_json_config(c) for c in batch_config_paths]  # type: List[BatchConfigDictType]
        return batch_config_dicts

    @staticmethod
    def get_n_tuple_to_sample_name_mapping(offline_config_path: PathType) -> Dict[str, Dict[str, List[str]]]:
        batch_config_dicts = LowLevelCutStudyMaster.get_batch_config_dicts(
            offline_config_path=offline_config_path
        )  # type: List[BatchConfigDictType]

        combinations = {}  # type: Dict[str, Dict[str, List[str]]]
        for batch_config_dict in batch_config_dicts:
            _data_dict = batch_config_dict["data"]
            assert isinstance(_data_dict, dict), type(_data_dict).__name__
            data_dict = _data_dict  # type: Dict[str, Any]
            _output_list = batch_config_dict["output"]
            assert isinstance(_output_list, list), type(_output_list).__name__

            samples = tuple(data_dict.keys())  # type: Tuple[str, ...]
            n_tuples = _output_list  # type: List[str]
            for n_tuple in n_tuples:
                if n_tuple in combinations.keys():
                    sample_mapping = LowLevelCutStudyMaster.get_type_to_samples_map(samples=samples)
                    new_sample_mapping = copy.copy(combinations[n_tuple])
                    for sample_type, new_samples in sample_mapping.items():
                        if sample_type in new_sample_mapping:
                            existing_samples = copy.copy(new_sample_mapping[sample_type])
                            new_sample_mapping[sample_type] = list(set(existing_samples + new_samples))
                        else:
                            new_sample_mapping[sample_type] = new_samples

                    assert all(k in new_sample_mapping for k in combinations[n_tuple].keys())
                    assert all(all(s in new_sample_mapping[k] for s in v) for k, v in combinations[n_tuple].items())
                    combinations[n_tuple].update(new_sample_mapping)
                else:
                    combinations.update({n_tuple: LowLevelCutStudyMaster.get_type_to_samples_map(samples=samples)})

        return combinations

    @staticmethod
    def get_sample_type(sample_name: str) -> str:
        if sample_name == "DataY4S":
            return sample_name
        elif sample_name in Belle_url_mc_types.keys():
            return "Generic"
        elif sample_name in private_data_types:
            return "Signal"
        else:
            raise RuntimeError(f"Encountered unknown sample name {sample_name}")

    @staticmethod
    def get_generic_sample_stream(offline_config_path: PathType) -> int:
        batch_config_dicts = LowLevelCutStudyMaster.get_batch_config_dicts(
            offline_config_path=offline_config_path
        )  # type: List[BatchConfigDictType]

        __streams = [bcd["data"][s]["streams"] for bcd in batch_config_dicts for s in bcd["data"]]  # type: ignore
        _streams = [s for s in __streams if isinstance(s, list) or s is None]  # type: List[Optional[List[int]]]

        assert all(isinstance(s, list) or s is None for s in _streams), [type(s) for s in _streams]
        assert all(len(s) == 1 for s in _streams if s is not None), [len(s) for s in _streams if s is not None]
        streams = list(set([s[0] for s in _streams if s is not None]))
        assert len(streams) == 1, streams
        stream = streams[0]
        assert isinstance(stream, int), (stream, type(stream))
        return stream

    @staticmethod
    def get_type_to_samples_map(samples: Tuple[str, ...]) -> Dict[str, List[str]]:
        sample_types = list(set([LowLevelCutStudyMaster.get_sample_type(sample_name=s) for s in samples]))
        return {
            ts: [s for s in samples if ts == LowLevelCutStudyMaster.get_sample_type(sample_name=s)] for ts in sample_types
        }

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for n_tuple in self.get_n_tuple_to_sample_name_mapping(offline_config_path=self.config_path).keys():
            for option_kwargs in LowLevelCutPlotter.get_settings_combinations():
                yield LowLevelCutsEvaluationTask(
                    config_path=self.config_path,
                    n_tuple=n_tuple,
                    **option_kwargs,
                )
                yield LowLevelCutsDataMCComparisonTask(
                    config_path=self.config_path,
                    n_tuple=n_tuple,
                    **option_kwargs,
                )


if __name__ == "__main__":
    config_base_path = get_config_base_path()  # type: PathType
    offline_config_file_name = "low_level_cut_study_conf.yaml"  # type: str

    my_offline_config_path = os.path.join(config_base_path, offline_config_file_name)  # type: PathType
    assert os.path.isfile(my_offline_config_path), my_offline_config_path

    num_workers = 10  # type: int
    luigi.process(
        LowLevelCutStudyMaster(config_path=my_offline_config_path),
        workers=num_workers,
    )
