"""
Steering file for the n-tuple production for the final state particles low level cut study.

Patrick Ecker 2020
"""

import basf2

from modularAnalysis import setAnalysisConfigParams, statistics

from analysistools.processing import job_interface
from analysistools.processing import path_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import b2bii_utils

from rdstar.studies.low_level_cut_study.utilities.final_state_particles_cut_study import fsp_collection
from rdstar.studies.low_level_cut_study.reconstruction.low_level_study_observables import LowLevelStudyObservables


cmd_args = job_interface.get_commandline_args()  # type: job_interface.BatchJobCMDLineArgs

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config=cmd_args.config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()  # type: basf2.Path

b2bii_utils.b2bii_read_and_convert(
    path=main_path,
    input_files=cmd_args.input_files,
    input_sequences=cmd_args.input_sequences,
)

setAnalysisConfigParams(configParametersAndValues={"mcMatchingVersion": "Belle"}, path=main_path)

skim_cuts_utils.apply_sampling_skim(
    path=main_path,
    sampling_rate=cmd_args.sampling_rate,
)

main_path.add_module(module="CurlerCloneMarker")

observables = LowLevelStudyObservables(use_mc_info=cmd_args.config["use_mc_info"])
fsp_collection.get_final_state_particles(path=main_path, use_mc_info=cmd_args.config["use_mc_info"])

#######
# pi0 #
#######

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["piNeutral_from_list_ntuple"],
    particleList="pi0:fsp",
    variables=observables.get_neutral_pion_observables(),
)

#######
# pi+ #
#######

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["piCharged_ntuple"],
    particleList="pi+:fsp",
    variables=observables.get_charged_pion_observables(),
)

############
# e and mu #
############

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["e_ntuple"],
    particleList="e-:fsp",
    variables=observables.get_lepton_observables(include_fsr_vars=True),
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["mu_ntuple"],
    particleList="mu-:fsp",
    variables=observables.get_lepton_observables(include_fsr_vars=False),
)

########
# K_S0 #
########

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["K_S0_ntuple"],
    particleList="K_S0:fsp",
    variables=observables.get_kaon_short_observables(),
)

######
# K+ #
######

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["K_ntuple"],
    particleList="K+:fsp",
    variables=observables.get_charged_kaon_observables(),
)

#########
# Gamma #
#########

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["gamma_ntuple"],
    particleList="gamma:fsp",
    variables=observables.get_gamma_observables(),
)

basf2.print_path(path=main_path)
basf2.process(path=main_path)
print(statistics)
