"""
Provides histogram variables for basic observables of low level cut studies.

Felix Metzner 2021
"""

import copy
import pandas as pd

from abc import ABC
from typing import Sequence, Tuple, List

from templatefitter.plotter.histogram_variable import HistVariable

from analysistools.processing.observables_utils import makeROOTCompatible

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames


__all__ = [
    "BasicObservablesCollectionBase",
    "GeneralObservablesCollection",
    "MomentumObservablesCollection",
    "EnergyObservablesCollection",
    "MassObservablesCollection",
    "ImpactParameterObservablesCollection",
    "TrackObservablesCollection",
    "ClusterObservablesCollection",
    "PIDObservablesCollection",
    "FSRCorrectionObservablesCollection",
]


class BasicObservablesCollectionBase(ABC):
    collection_name = ""  # type: str

    default_binning = 50  # type: int

    _histogram_variables = ()  # type: Tuple[HistVariable, ...]

    @classmethod
    def get_histogram_variables(cls, root_compatible: bool) -> Tuple[HistVariable, ...]:
        assert len(cls._histogram_variables) > 0, [v.df_label for v in cls._histogram_variables]
        assert len(set([v.df_label for v in cls._histogram_variables])) == len(cls._histogram_variables), (
            len(set([v.df_label for v in cls._histogram_variables])),
            len(cls._histogram_variables),
            [v.df_label for v in cls._histogram_variables],
        )

        if root_compatible:
            return tuple(
                [
                    HistVariable(
                        df_label=makeROOTCompatible([hv.df_label])[0],
                        n_bins=hv.n_bins,
                        scope=hv.scope,
                        var_name=hv.variable_name,
                        unit=hv.unit,
                        use_log_scale=hv.use_log_scale,
                    )
                    for hv in cls._histogram_variables
                ]
            )

        return copy.copy(cls._histogram_variables)

    @staticmethod
    def get_histogram_variable(
        base_hist_var: HistVariable,
        dfs: Sequence[pd.DataFrame],
        round_scope_precision: int = 0,
    ) -> HistVariable:
        return HistVariable.get_scoped_histogram_variable(
            base_hist_var=base_hist_var,
            dfs=dfs,
            round_scope_precision=round_scope_precision,
        )

    @classmethod
    def get_hist_vars(cls) -> List[HistVariable]:
        return list(copy.copy(cls._histogram_variables))


class GeneralObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "general_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    general_pdg_code_hist_var = HistVariable(
        df_label="PDG",
        n_bins=default_binning,
        var_name=LatexObsNames.reco_pdg_code,
        unit=LatexUnits.no_unit,
    )

    general_decay_mode_id_hist_var = HistVariable(
        df_label="extraInfo(decayModeID)",
        n_bins=default_binning,
        var_name=LatexObsNames.reco_decay_mode_id,
        unit=LatexUnits.no_unit,
    )

    _histogram_variables = (
        general_pdg_code_hist_var,
        general_decay_mode_id_hist_var,
    )  # type: Tuple[HistVariable, ...]


class MomentumObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "momentum_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    momentum_p_hist_var = HistVariable(
        df_label="p",
        n_bins=default_binning,
        var_name=LatexObsNames.momentum_p,
        unit=LatexUnits.gev,
    )

    momentum_pt_hist_var = HistVariable(
        df_label="pt",
        n_bins=default_binning,
        var_name=LatexObsNames.momentum_pt,
        unit=LatexUnits.gev,
    )

    momentum_pz_hist_var = HistVariable(
        df_label="pz",
        n_bins=default_binning,
        var_name=LatexObsNames.momentum_pz,
        unit=LatexUnits.gev,
    )

    momentum_theta_hist_var = HistVariable(
        df_label="theta",
        n_bins=default_binning,
        var_name=LatexObsNames.momentum_theta,
        unit=LatexUnits.rad,
    )

    momentum_cos_theta_hist_var = HistVariable(
        df_label="cosTheta",
        n_bins=default_binning,
        var_name=LatexObsNames.momentum_cos_theta,
        unit=LatexUnits.no_unit,
    )

    momentum_phi_hist_var = HistVariable(
        df_label="phi",
        n_bins=default_binning,
        var_name=LatexObsNames.momentum_phi,
        unit=LatexUnits.rad,
    )

    _histogram_variables = (
        momentum_p_hist_var,
        momentum_pt_hist_var,
        momentum_pz_hist_var,
        momentum_theta_hist_var,
        momentum_cos_theta_hist_var,
        momentum_phi_hist_var,
    )  # type: Tuple[HistVariable, ...]


class EnergyObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "energy_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    energy_e_hist_var = HistVariable(
        df_label="E",
        n_bins=default_binning,
        var_name=LatexObsNames.energy_e,
        unit=LatexUnits.gev,
    )

    energy_q_hist_var = HistVariable(
        df_label="Q",
        n_bins=default_binning,
        var_name=LatexObsNames.energy_q,
        unit=LatexUnits.gev,
    )

    energy_delta_q_hist_var = HistVariable(
        df_label="dQ",
        n_bins=default_binning,
        var_name=LatexObsNames.energy_delta_q,
        unit=LatexUnits.gev,
    )

    _histogram_variables = (
        energy_e_hist_var,
        energy_q_hist_var,
        energy_delta_q_hist_var,
    )  # type: Tuple[HistVariable, ...]


class MassObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "mass_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    mass_m_hist_var = HistVariable(
        df_label="M",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_m,
        unit=LatexUnits.gev,
    )

    mass_inv_m_hist_var = HistVariable(
        df_label="InvM",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_inv_m,
        unit=LatexUnits.gev,
    )

    mass_delta_m_hist_var = HistVariable(
        df_label="dM",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_delta_m,
        unit=LatexUnits.gev,
    )

    mass_abs_delta_m_hist_var = HistVariable(
        df_label="abs(dM)",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_abs_delta_m,
        unit=LatexUnits.gev,
    )

    mass_sig_mbf_hist_var = HistVariable(
        df_label="SigMBF",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_sig_mbf,
        unit=LatexUnits.no_unit,
    )

    mass_abs_sig_mbf_hist_var = HistVariable(
        df_label="abs(SigMBF)",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_abs_sig_mbf,
        unit=LatexUnits.no_unit,
    )

    _histogram_variables = (
        mass_m_hist_var,
        mass_inv_m_hist_var,
        mass_delta_m_hist_var,
        mass_abs_delta_m_hist_var,
        mass_sig_mbf_hist_var,
        mass_abs_sig_mbf_hist_var,
    )  # type: Tuple[HistVariable, ...]


class ImpactParameterObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "impact_parameter_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    impact_delta_r_hist_var = HistVariable(
        df_label="dr",
        n_bins=default_binning,
        var_name=LatexObsNames.impact_delta_r,
        unit=LatexUnits.cm,
    )

    impact_delta_z_hist_var = HistVariable(
        df_label="dz",
        n_bins=default_binning,
        var_name=LatexObsNames.impact_delta_z,
        unit=LatexUnits.cm,
    )

    impact_abs_delta_z_hist_var = HistVariable(
        df_label="abs(dz)",
        n_bins=default_binning,
        var_name=LatexObsNames.impact_abs_delta_z,
        unit=LatexUnits.cm,
    )

    _histogram_variables = (
        impact_delta_r_hist_var,
        impact_delta_z_hist_var,
        impact_abs_delta_z_hist_var,
    )  # type: Tuple[HistVariable, ...]


class TrackObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "track_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    track_p_value_hist_var = HistVariable(
        df_label="pValue",
        n_bins=default_binning,
        var_name=LatexObsNames.tracks_p_value,
        unit=LatexUnits.no_unit,
    )

    track_n_cdc_hits_hist_var = HistVariable(
        df_label="nCDCHits",
        n_bins=default_binning,
        var_name=LatexObsNames.tracks_n_cdc_hits,
        unit=LatexUnits.no_unit,
    )

    track_n_svd_hits_hist_var = HistVariable(
        df_label="nSVDHits",
        n_bins=default_binning,
        var_name=LatexObsNames.tracks_n_svd_hits,
        unit=LatexUnits.no_unit,
    )

    track_curler_clone_indicator_hist_var = HistVariable(
        df_label="trackCurlerCloneIndicator",
        n_bins=default_binning,
        var_name=LatexObsNames.tracks_curler_clone_indicator,
        unit=LatexUnits.no_unit,
    )

    _histogram_variables = (
        track_p_value_hist_var,
        track_n_cdc_hits_hist_var,
        track_n_svd_hits_hist_var,
        track_curler_clone_indicator_hist_var,
    )  # type: Tuple[HistVariable, ...]


class ClusterObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "cluster_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    cluster_good_belle_gamma_hist_var = HistVariable(
        df_label="goodBelleGamma",
        n_bins=default_binning,
        var_name=LatexObsNames.cluster_good_belle_gamma,
        unit=LatexUnits.no_unit,
    )

    cluster_e1_over_e9_hist_var = HistVariable(
        df_label="clusterE1E9",
        n_bins=default_binning,
        var_name=LatexObsNames.cluster_e1_over_e9,
        unit=LatexUnits.no_unit,
    )

    cluster_e9_over_e21_hist_var = HistVariable(
        df_label="clusterE9E21",
        n_bins=default_binning,
        var_name=LatexObsNames.cluster_e9_over_e21,
        unit=LatexUnits.no_unit,
    )

    cluster_e9_over_e25_hist_var = HistVariable(
        df_label="clusterE9E25",
        n_bins=default_binning,
        var_name=LatexObsNames.cluster_e9_over_e25,
        unit=LatexUnits.no_unit,
    )

    cluster_n_hits_hist_var = HistVariable(
        df_label="clusterNHits",
        n_bins=default_binning,
        var_name=LatexObsNames.cluster_n_hits,
        unit=LatexUnits.no_unit,
    )

    cluster_highest_e_hist_var = HistVariable(
        df_label="clusterHighestE",
        n_bins=default_binning,
        var_name=LatexObsNames.cluster_highest_e,
        unit=LatexUnits.gev,
    )

    _histogram_variables = (
        cluster_good_belle_gamma_hist_var,
        cluster_e1_over_e9_hist_var,
        cluster_e9_over_e21_hist_var,
        cluster_e9_over_e25_hist_var,
        cluster_n_hits_hist_var,
        cluster_highest_e_hist_var,
    )  # type: Tuple[HistVariable, ...]


class PIDObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "particle_id_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    pid_electron_hist_var = HistVariable(
        df_label="eIDBelle",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_electron,
        unit=LatexUnits.no_unit,
    )

    pid_muon_hist_var = HistVariable(
        df_label="muIDBelle",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_muon,
        unit=LatexUnits.no_unit,
    )

    pid_muon_quality_hist_var = HistVariable(
        df_label="muIDBelleQuality",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_muon_quality,
        unit=LatexUnits.no_unit,
    )

    pid_atc23_hist_var = HistVariable(
        df_label="atcPIDBelle(2,3)",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_atc23,
        unit=LatexUnits.no_unit,
    )

    pid_atc32_hist_var = HistVariable(
        df_label="atcPIDBelle(3,2)",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_atc32,
        unit=LatexUnits.no_unit,
    )

    pid_ksnb_standard_hist_var = HistVariable(
        df_label="extraInfo(ksnbStandard)",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_ksnb_standard,
        unit=LatexUnits.no_unit,
    )

    pid_good_ks_hist_var = HistVariable(
        df_label="extraInfo(goodKs)",
        n_bins=default_binning,
        var_name=LatexObsNames.pid_good_ks,
        unit=LatexUnits.no_unit,
    )

    pid_good_belle_gamma_hist_var = ClusterObservablesCollection.cluster_good_belle_gamma_hist_var

    _histogram_variables = (
        pid_electron_hist_var,
        pid_muon_hist_var,
        pid_muon_quality_hist_var,
        pid_atc23_hist_var,
        pid_atc32_hist_var,
        pid_ksnb_standard_hist_var,
        pid_good_ks_hist_var,
        pid_good_belle_gamma_hist_var,
    )  # type: Tuple[HistVariable, ...]


class FSRCorrectionObservablesCollection(BasicObservablesCollectionBase):
    collection_name = "fsr_correction_variables"  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    fsr_daughters_inv_m_hist_var = HistVariable(
        df_label="daughterInvariantMass(0, 1)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_corrected_e_inv_m,
        unit=LatexUnits.gev,
    )

    fsr_corrected_e_hist_var = HistVariable(
        df_label="extraInfo(fsrCorrectedE)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_energy,
        unit=LatexUnits.gev,
    )

    fsr_corrected_p_hist_var = HistVariable(
        df_label="extraInfo(fsrCorrectedP)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_p,
        unit=LatexUnits.gev,
    )

    fsr_corrected_px_hist_var = HistVariable(
        df_label="extraInfo(fsrCorrectedPx)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_px,
        unit=LatexUnits.gev,
    )

    fsr_corrected_py_hist_var = HistVariable(
        df_label="extraInfo(fsrCorrectedPy)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_py,
        unit=LatexUnits.gev,
    )

    fsr_corrected_pz_hist_var = HistVariable(
        df_label="extraInfo(fsrCorrectedPz)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_pz,
        unit=LatexUnits.gev,
    )

    fsr_gamma_e_hist_var = HistVariable(
        df_label="extraInfo(fsrGammaE)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_gamma_energy,
        unit=LatexUnits.gev,
    )

    fsr_gamma_p_hist_var = HistVariable(
        df_label="extraInfo(fsrGammaP)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_gamma_p,
        unit=LatexUnits.gev,
    )

    fsr_gamma_px_hist_var = HistVariable(
        df_label="extraInfo(fsrGammaPx)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_gamma_px,
        unit=LatexUnits.gev,
    )

    fsr_gamma_py_hist_var = HistVariable(
        df_label="extraInfo(fsrGammaPy)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_gamma_py,
        unit=LatexUnits.gev,
    )

    fsr_gamma_pz_hist_var = HistVariable(
        df_label="extraInfo(fsrGammaPz)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_gamma_pz,
        unit=LatexUnits.gev,
    )

    fsr_electron_has_correction_hist_var = HistVariable(
        df_label="extraInfo(hasFSRCorrection)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_e_with_fsr,
        unit=LatexUnits.no_unit,
    )

    fsr_gamma_has_correction_hist_var = HistVariable(
        df_label="extraInfo(hasFSRCorrection)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_photon_used_for_fsr,
        unit=LatexUnits.no_unit,
    )

    fsr_corrected_electron_hist_var = HistVariable(
        df_label="extraInfo(fsrCorrected)",
        n_bins=default_binning,
        var_name=LatexObsNames.fsr_corrected_e,
        unit=LatexUnits.no_unit,
    )

    _histogram_variables = (
        fsr_daughters_inv_m_hist_var,
        fsr_corrected_e_hist_var,
        fsr_corrected_p_hist_var,
        fsr_corrected_px_hist_var,
        fsr_corrected_py_hist_var,
        fsr_corrected_pz_hist_var,
        fsr_gamma_e_hist_var,
        fsr_gamma_p_hist_var,
        fsr_gamma_px_hist_var,
        fsr_gamma_py_hist_var,
        fsr_gamma_pz_hist_var,
        fsr_electron_has_correction_hist_var,
        fsr_gamma_has_correction_hist_var,
        fsr_corrected_electron_hist_var,
    )  # type: Tuple[HistVariable, ...]
