"""
Plotter class for low level data-MC comparison.

Felix Metzner 2021
"""


import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, Optional, Tuple, List, Dict, Any

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import DataMCHistogramPlot
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors

from rdstar.utilities import PathType
from rdstar.cut_information_objects.cut_info import CutInfo

from rdstar.studies.low_level_cut_study.utilities.low_level_plotter_base import LowLevelPlotterBaseClass
from rdstar.studies.low_level_cut_study.utilities.special_observables_handler import SpecialVariableCollections


__all__ = [
    "DataMCComparisonPlotter",
]


class DataMCComparisonPlotter(LowLevelPlotterBaseClass):
    plot_name_suffix = "mc_data_comparison"  # type: str

    default_number_of_bins = 50  # type: int

    cut_line_color = KITColors.kit_red  # type: str
    mc_color = KITColors.kit_cyan  # type: str

    def __init__(
        self,
        n_tuple: str,
        target_path: PathType,
        apply_pre_cuts: str,
        variable_set: str,
        remove_outliers: bool,
    ) -> None:
        super().__init__(
            n_tuple=n_tuple,
            target_path=target_path,
            apply_pre_cuts=apply_pre_cuts,
            variable_set=variable_set,
            remove_outliers=remove_outliers,
        )

        self.data_df = None  # type: Optional[pd.DataFrame]

    def add_data_df(self, data_df: pd.DataFrame) -> None:
        self.data_df = self.prepare_df(df=data_df)
        self.check_columns(df=self.data_df)

    def get_output_key_path_pair(
        self,
        tuple_key: str,
        cut_info: CutInfo,
        add_chi2_test_info: bool = True,
    ) -> Tuple[str, PathType]:
        chi2_info_tag = "with_chi2_info" if add_chi2_test_info else "without_chi2_info"  # type: str
        escaped_key = self.escape_decay_string_in_tuple_key(tuple_key=tuple_key)  # type: str
        output_key = f"{escaped_key}_{cut_info.cut_name}_{chi2_info_tag}_{self.plot_name_suffix}"  # type: str
        path = os.path.join(self.target_path, chi2_info_tag, f"{output_key}.pdf")  # type: PathType
        return output_key, path

    def generate_output_dict(self) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]
        relevant_tuple_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]
        for tuple_key in relevant_tuple_keys:

            if self._variable_set_name == self.cut_variable_set_option:
                cut_infos = self.cut_dicts[tuple_key]["cuts"]
            elif self._variable_set_name == self.validation_variable_set_option:
                hv_collection = SpecialVariableCollections.get_variable_collection_for_n_tuple(n_tuple_name=tuple_key)
                hist_vars = hv_collection.get_set_of_observables(root_compatible=True, use_mc_info=False)
                cut_infos = [CutInfo.create_fake_cut_info_from_hist_variable(hist_var=hv) for hv in hist_vars]
            else:
                raise RuntimeError(f"Invalid option {self.variable_set_option_name} = {self._variable_set_name}!")

            for cut_info in cut_infos:
                assert isinstance(cut_info, CutInfo), type(cut_info).__name__

                self.update_output_dict(
                    output_dict=output,
                    new_output_dict=dict(
                        [
                            self.get_output_key_path_pair(
                                tuple_key=tuple_key, cut_info=cut_info, add_chi2_test_info=True
                            ),
                            self.get_output_key_path_pair(
                                tuple_key=tuple_key, cut_info=cut_info, add_chi2_test_info=False
                            ),
                        ]
                    ),
                )

        return output

    def add_to_output_dict(
        self,
        output_dict: Dict[str, PathType],
        tuple_key: str,
        cut_info: CutInfo,
        add_chi2_info: bool,
    ) -> Tuple[str, PathType]:
        o_key, o_path = self.get_output_key_path_pair(
            tuple_key=tuple_key,
            cut_info=cut_info,
            add_chi2_test_info=add_chi2_info,
        )
        target_dir, filename_w_ext = os.path.split(o_path)
        filename_wo_ext = os.path.splitext(filename_w_ext)[0]

        self.update_output_dict(output_dict=output_dict, new_output_dict={o_key: o_path})

        return filename_wo_ext, target_dir

    def create_plots(self) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]
        relevant_tuple_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]

        for tuple_key in relevant_tuple_keys:
            channel = self.cut_dicts[tuple_key]["channel"]
            assert isinstance(channel, str), type(channel).__name__

            if self._variable_set_name == self.cut_variable_set_option:
                new_output = self._create_cut_variable_plots(channel=channel, tuple_key=tuple_key)
                self.update_output_dict(output_dict=output, new_output_dict=new_output)
            elif self._variable_set_name == self.validation_variable_set_option:
                new_output = self._create_validation_variable_plots(channel=channel, tuple_key=tuple_key)
                self.update_output_dict(output_dict=output, new_output_dict=new_output)
            else:
                raise RuntimeError(f"Invalid option {self.variable_set_option_name} = {self._variable_set_name}!")

        return output

    def _create_cut_variable_plots(
        self,
        channel: str,
        tuple_key: str,
    ) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]
        cut_infos = self.cut_dicts[tuple_key]["cuts"]

        for cut_info in cut_infos:
            assert isinstance(cut_info, CutInfo), type(cut_info).__name__
            unscoped_hist_var = HistVariable(
                df_label=cut_info.cut_var_root,
                n_bins=self.default_number_of_bins,
                scope=None,
                var_name=cut_info.cut_var_latex_label,
                unit=cut_info.latex_unit,
            )

            new_output = self._create_plots(
                channel=channel,
                tuple_key=tuple_key,
                cut_info=cut_info,
                unscoped_hist_var=unscoped_hist_var,
            )
            self.update_output_dict(output_dict=output, new_output_dict=new_output)

        return output

    def _create_validation_variable_plots(
        self,
        channel: str,
        tuple_key: str,
    ) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]

        hist_var_collection = SpecialVariableCollections.get_variable_collection_for_n_tuple(n_tuple_name=tuple_key)

        for hist_var in hist_var_collection.get_set_of_observables(root_compatible=True, use_mc_info=False):
            cut_info = CutInfo.create_fake_cut_info_from_hist_variable(
                hist_var=hist_var,
                invert_root_compatible_df_label=True,
            )
            new_output = self._create_plots(
                channel=channel,
                tuple_key=tuple_key,
                cut_info=cut_info,
                unscoped_hist_var=hist_var,
            )
            self.update_output_dict(output_dict=output, new_output_dict=new_output)

        return output

    def _create_plots(
        self,
        channel: str,
        tuple_key: str,
        cut_info: CutInfo,
        unscoped_hist_var: HistVariable,
    ) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]

        _cut_infos = self.cut_dicts[tuple_key]["cuts"]
        assert isinstance(_cut_infos, list), type(_cut_infos).__name__
        cut_infos = _cut_infos
        ch_cut = f"{self.decay_mode_id_col_name} == {self.channel_name_to_id_mapping[channel]}" if channel else None

        set_matplotlibrc_params()

        filename_wo_ext, target_dir = self.add_to_output_dict(
            output_dict=output,
            tuple_key=tuple_key,
            cut_info=cut_info,
            add_chi2_info=False,
        )
        filename_chi2_wo_ext, target_dir_chi2 = self.add_to_output_dict(
            output_dict=output,
            tuple_key=tuple_key,
            cut_info=cut_info,
            add_chi2_info=True,
        )

        df_data = self.apply_cuts(df=self.data_df, pivot=cut_info, cut_infos=cut_infos, channel_cut_string=ch_cut)
        df_mc = self.apply_cuts(df=self.df, pivot=cut_info, cut_infos=cut_infos, channel_cut_string=ch_cut)
        self.remove_outliers_from_dfs(pivot=cut_info, dfs=(df_data, df_mc))

        if len(np.unique(df_mc.index)) + len(np.unique(df_data.index)) == 0:
            fig, ax = plt.subplots(figsize=self.default_figure_size)
            ax.set_title(label=self.get_plot_title(channel=channel), loc="right")
            ax.set_title(label=cut_info.cut_label, loc="left")
            ax.text(x=0.5, y=0.5, s="No Data", fontsize=16)
            export(fig=fig, filename=filename_wo_ext, target_dir=target_dir, close_figure=True)
            export(fig=fig, filename=filename_chi2_wo_ext, target_dir=target_dir_chi2, close_figure=True)
            return output

        this_hist_var = HistVariable.get_scoped_histogram_variable(
            base_hist_var=unscoped_hist_var,
            dfs=[df_mc, df_data],
            round_scope_precision=0,
        )

        fig, ax = DataMCHistogramPlot.create_hist_ratio_figure(fig_size=self.default_figure_size)

        dmc_hp = DataMCHistogramPlot(variable=this_hist_var)

        dmc_hp.add_data_component(label="Data", data=df_data)

        dmc_hp.add_mc_component(
            label="MC",
            data=df_mc,
            weights=None,
            systematics=None,
            color=self.mc_color,
        )

        chi2, dof, p_val, toy_output = dmc_hp.plot_on(
            ax1=ax[0],
            ax2=ax[1],
            ratio_type="vs_uncert",
            normalize_to_data=True,
            include_sys=False,
            gof_check_method="pearson",
            y_scale=1.2,
            plot_outlier_indicators=False,
        )

        self.plot_cut_lines(ax=ax[0], cut_info=cut_info)

        ax[0].set_title(label=self.get_plot_title(channel=channel), loc="right")
        use_title = self.get_cut_string_title(cut_info=cut_info)
        if ax[0].get_ylim()[1] > 0.85e4:
            padding = " " * 12
            use_title = padding + cut_info.cut_label
        ax[0].set_title(label=use_title, loc="left")

        export(fig=fig, filename=filename_wo_ext, target_dir=target_dir, close_figure=False)

        if chi2 is not None:
            info_text = "\n" + r"$p$-Value = " + f"{p_val:.2f}"  # type: str
            if chi2 >= 1000:
                info_text += "\n" + r"$\chi^2 / n_\mathrm{dof}$ = " + f"{chi2:.0f}/{int(dof)} = {chi2 / dof:.2f}"
            else:
                info_text += "\n" + r"$\chi^2 / n_\mathrm{dof}$ = " + f"{chi2:.1f}/{int(dof)} = {chi2 / dof:.2f}"

            fig.canvas.draw()
            x_ticks = ax[1].get_xticklabels()  # type: Union[Any, List[plt.text.Text]]
            if x_ticks[0].get_text():
                if len(x_ticks[0].get_text().split("{")[1].split("}")[0]) > 4:
                    new_x_ticks = x_ticks[1::2]
                    if all(tick.get_text().split("{")[1].split("}")[0][-1] == "0" for tick in new_x_ticks):
                        for tick in x_ticks:
                            original_text = tick.get_text()
                            original_value = original_text.split("{")[1].split("}")[0]
                            tick.set_text(original_text.replace(original_value, original_value[:-1]))
                    ax[1].set_xticklabels(x_ticks)
                    plt.setp(x_ticks[::2], visible=False)
                    plt.setp(new_x_ticks, visible=True)

            legend_pos = ax[0].get_legend().get_window_extent()
            legend_left_lower_edge_pos_in_ax_coords = ax[0].transAxes.inverted().transform(legend_pos.min)
            ax[0].text(
                x=legend_left_lower_edge_pos_in_ax_coords[0],
                y=legend_left_lower_edge_pos_in_ax_coords[1],
                s=info_text,
                transform=ax[0].transAxes,
                va="top",
                ha="left",
                linespacing=1.5,
            )

        export(fig=fig, filename=filename_chi2_wo_ext, target_dir=target_dir_chi2, close_figure=True)
        return output
