"""
Base class for the plotters of the low level studies.

Patrick Ecker 2020
"""


import os
import copy
import itertools
import numpy as np
import pandas as pd

from scipy import stats
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Union, Optional, Tuple, List, Dict, Set, Any

from templatefitter.plotter.plot_style import KITColors
from templatefitter.plotter.plot_utilities import AxesType

from analysistools.processing.observables_utils import makeROOTCompatible

from rdstar.utilities import PathType
from rdstar.cut_information_objects.cut_info import CutInfo, CutType
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.offline_analysis.rdstar_plots import d_meson_decay_mode_id_to_latex_label_dict, d_meson_decay_name_to_id_dict

from rdstar.studies.low_level_cut_study.utilities.special_observables_handler import (
    SpecialVariableCollections,
    PhotonObservables,
)
from rdstar.studies.low_level_cut_study.utilities.low_level_cut_defintions import (
    combined_channel_dict,
    fsp_cut_dict,
    D_cut_dict,
    combined_decay_mode_id_dict,
)


__all__ = [
    "LowLevelPlotterBaseClass",
]


class LowLevelPlotterBaseClass(ABC):
    is_signal_col_name = "isSignal"  # type: str
    decay_mode_id_col_name = "extraInfo__bodecayModeID__bc"  # type: str

    cut_line_color = KITColors.kit_red  # type: str
    cut_line_width = 2.0  # type: float
    arrow_length = 3.0  # type: float

    default_figure_size = (6.0, 4.5)  # type: Tuple[float, float]

    cut_option_name = "apply_pre_cuts"  # type: str
    apply_no_cuts_option = "no_cuts"  # type: str
    apply_other_cuts_option = "other_cuts"  # type: str
    apply_all_cuts_option = "all_cuts"  # type: str
    apply_cuts_options = {
        apply_no_cuts_option: 0,
        apply_other_cuts_option: 1,
        apply_all_cuts_option: 2,
    }  # type: Dict[str, int]
    cuts_sub_dir_names = ("no_other_cuts", "other_cuts_applied", "all_cuts_applied")  # type: Tuple[str, ...]
    cuts_title_suffix = (r"no other Cuts", r"other cuts applied", r"all cuts applied")  # type: Tuple[str, ...]

    variable_set_option_name = "variable_set"  # type: str
    cut_variable_set_option = "cut_variables"  # type: str
    validation_variable_set_option = "validation_variables"  # type: str
    variable_set_options = {
        cut_variable_set_option: True,
        validation_variable_set_option: False,
    }  # type: Dict[str, bool]
    special_cut_options = (
        apply_no_cuts_option,
        apply_other_cuts_option,
    )  # type: Tuple[str, ...]

    remove_outlier_option_name = "remove_outliers"  # type: str
    remove_outliers_options = (True, False)  # type: Tuple[bool, bool]
    remove_outliers_sub_dir_name = "outliers_removed"  # type: str
    outlier_sigma = 10  # type: float

    particle_mapping = {
        "dzero_ntuple": LatexParticleNames.d_zero_str,
        "dcharged_ntuple": LatexParticleNames.d_minus_str,
        "dstzero_ntuple": LatexParticleNames.d_st_zero_str,
        "dstcharged_ntuple": LatexParticleNames.d_st_minus_str,
        "piNeutral_from_list_ntuple": LatexParticleNames.pi_zero_str,
        "piCharged_ntuple": LatexParticleNames.pi_plus_str,
        "e_ntuple": LatexParticleNames.electron_minus_str,
        "mu_ntuple": LatexParticleNames.muon_minus_str,
        "K_S0_ntuple": LatexParticleNames.k_short_str,
        "K_ntuple": LatexParticleNames.k_plus_str,
        "gamma_ntuple": LatexParticleNames.gamma_str,
    }  # type: Dict[str, str]

    _escape_decay_str_mapping = OrderedDict(
        [
            ("*", "St"),
            ("-", "Minus"),
            ("+", "Plus"),
            (":", "_to_"),
            ("0", "zero"),
        ]
    )  # type: Dict[str, str]

    nan_check_variable_exceptions = (
        "SigMBF",
        "massDifferenceSignificance",
        "cosTheta",
        "theta",
        "phi",
    )  # type: Tuple[str, ...]

    def __init__(
        self,
        n_tuple: str,
        target_path: PathType,
        apply_pre_cuts: str,
        variable_set: str,
        remove_outliers: bool,
    ) -> None:
        if apply_pre_cuts not in self.apply_cuts_options.keys():
            raise ValueError(
                f"Parameter apply_pre_cuts must be one of the strings "
                f"{list(self.apply_cuts_options.keys())}, but {apply_pre_cuts} was provided!"
            )

        if variable_set not in self.variable_set_options.keys():
            raise ValueError(
                f"Parameter variable_set must be one of the strings "
                f"{list(self.variable_set_options.keys())}, but {variable_set} was provided!"
            )

        if apply_pre_cuts in self.special_cut_options and not self.variable_set_options[variable_set]:
            raise ValueError(
                f"The settings combination\n\tapply_pre_cuts = {apply_pre_cuts}\n\t"
                f"variable_set = {variable_set}\nis not allowed!"
            )

        self._pre_cuts = apply_pre_cuts  # type: str
        self._variable_set_name = variable_set  # type: str
        self._remove_outliers = remove_outliers  # type: bool

        self.n_tuple = n_tuple  # type: str
        self._base_target_path = target_path  # type: PathType

        self.cut_dicts = copy.copy(fsp_cut_dict)  # type: Dict[str, Dict[str, Union[str, List[CutInfo]]]]
        self.cut_dicts.update(copy.copy(D_cut_dict))

        self.channel_name_to_id_mapping = copy.copy(combined_decay_mode_id_dict)  # type: Dict[str, int]

        self.__nan_check_failing_vars = set()  # type: Set[str]

        self.df = None  # type: Optional[pd.DataFrame]

    def add_df(self, df: pd.DataFrame) -> None:
        self.df = self.prepare_df(df=df)
        self.check_columns(df=self.df)

    @property
    def target_path(self) -> PathType:
        var_set_dir_name = self._variable_set_name  # type: str
        cuts_sub_dir = self.cuts_sub_dir_names[self.apply_cuts_options[self._pre_cuts]]  # type: str
        if self._remove_outliers:
            cuts_sub_dir += f"_and_{self.remove_outliers_sub_dir_name}"
        return os.path.join(self._base_target_path, var_set_dir_name, cuts_sub_dir)

    @staticmethod
    def prepare_df(df: pd.DataFrame) -> pd.DataFrame:
        base_vars = ["dz"]  # type: List [str]
        for var in base_vars:
            abs_var = f"abs({var})"  # type: str
            root_var = makeROOTCompatible(variables=[var])[0]  # type: str
            root_abs_var = makeROOTCompatible(variables=[abs_var])[0]  # type: str
            if root_var in df.columns and root_abs_var not in df.columns:
                df.loc[:, root_abs_var] = df[root_var].abs()
        return df

    @property
    def particle_label(self) -> str:
        return self.particle_mapping[self.n_tuple]

    def get_plot_title(self, channel: str) -> str:
        if channel:
            if channel in d_meson_decay_name_to_id_dict().keys():
                return d_meson_decay_mode_id_to_latex_label_dict()[d_meson_decay_name_to_id_dict()[channel]]
            return combined_channel_dict[channel]
        else:
            return self.particle_label

    def get_cut_string_title(self, cut_info: CutInfo) -> str:
        return cut_info.cut_label + rf"  ({self.cuts_title_suffix[self.apply_cuts_options[self._pre_cuts]]})"

    def escape_decay_string_in_tuple_key(self, tuple_key: str) -> str:
        assert not any(v in self._escape_decay_str_mapping.keys() for v in self._escape_decay_str_mapping.values()), [
            v for v in self._escape_decay_str_mapping.values() if v in self._escape_decay_str_mapping.keys()
        ]
        new_tuple_key = copy.copy(tuple_key)
        for original_char, escaped_char in self._escape_decay_str_mapping.items():
            new_tuple_key = new_tuple_key.replace(original_char, escaped_char)
        return new_tuple_key

    @abstractmethod
    def generate_output_dict(self) -> Dict[str, PathType]:
        raise NotImplementedError()

    @abstractmethod
    def create_plots(self) -> Dict[str, PathType]:
        raise NotImplementedError()

    def apply_cuts(
        self,
        df: pd.DataFrame,
        pivot: CutInfo,
        cut_infos: List[CutInfo],
        channel_cut_string: Optional[str] = None,
    ) -> pd.DataFrame:
        if df is None or not isinstance(df, pd.DataFrame):
            raise RuntimeError("You need to get a Dataframe first!")
        if pivot.cut_var_root not in df.columns:
            raise ValueError(f"The column {pivot.cut_var_root} is not available in the dataframe!")

        used_cut_strings = []  # type: List[str]

        if self._pre_cuts == self.apply_other_cuts_option:
            used_cut_strings.extend([ci.cut_str_root for ci in cut_infos if ci != pivot])
        elif self._pre_cuts == self.apply_no_cuts_option:
            pass
        elif self._pre_cuts == self.apply_all_cuts_option:
            used_cut_strings.extend([ci.cut_str_root for ci in cut_infos])
        else:
            raise RuntimeError(
                f"Pre cuts option {self._pre_cuts} does not fit any of the known options "
                f"{list(self.apply_cuts_options.keys())}! This should not happen!"
            )

        if channel_cut_string:
            used_cut_strings.append(channel_cut_string)

        cut_query = " and ".join(used_cut_strings)  # type: str
        if cut_query:
            f = copy.copy(df.query(expr=cut_query, inplace=False))
        else:
            f = copy.copy(df)

        self.__check_df(df=f, pivot=pivot)
        f.reset_index(drop=True, inplace=True)

        return f

    def remove_outliers_from_dfs(
        self,
        pivot: CutInfo,
        dfs: Tuple[pd.DataFrame, ...],
    ) -> None:
        if not self._remove_outliers:
            return

        assert isinstance(dfs, tuple), type(dfs)
        assert all(isinstance(df, pd.DataFrame) for df in dfs), [type(df) for df in dfs]

        is_validation_variable = self._variable_set_name == self.validation_variable_set_option  # type: bool
        is_exceptional_variable = any(v in pivot.cut_var_root for v in self.nan_check_variable_exceptions)  # type: bool
        is_exceptional = is_exceptional_variable or self.n_tuple == PhotonObservables.n_tuple_name()  # type: bool
        is_nan_exception = is_exceptional and is_validation_variable  # type: bool

        max_score = stats.halfnorm.cdf(self.outlier_sigma)  # type: float
        nan_policy = "propagate" if is_nan_exception else "raise"  # type: str

        if len(dfs) == 1:
            drop_conditions = tuple(
                [np.abs(stats.zscore(dfs[0][pivot.cut_var_root], nan_policy=nan_policy)) > max_score]
            )  # type: Tuple[np.ndarray, ...]
        else:
            min_max_values = (+np.inf, -np.inf)  # type: Tuple[float, float]
            for df in dfs:
                keep_cond = np.abs(stats.zscore(df[pivot.cut_var_root], nan_policy=nan_policy)) <= max_score
                current_min_max = (
                    np.min(df.loc[keep_cond, pivot.cut_var_root]),
                    np.max(df.loc[keep_cond, pivot.cut_var_root]),
                )  # type: Tuple[float, float]
                if current_min_max[0] < min_max_values[0] and np.isfinite(current_min_max[0]):
                    min_max_values = (current_min_max[0], min_max_values[1])
                if current_min_max[1] > min_max_values[1] and np.isfinite(current_min_max[1]):
                    min_max_values = (min_max_values[0], current_min_max[1])

            drop_conditions = tuple(
                [
                    ((df[pivot.cut_var_root] < min_max_values[0]) | (df[pivot.cut_var_root] > min_max_values[1])).values
                    for df in dfs
                ]
            )

        assert len(drop_conditions) == len(dfs), (len(drop_conditions), len(dfs))
        assert all(len(f) == len(c) for f, c in zip(dfs, drop_conditions)), [
            (len(f), len(c)) for f, c in zip(dfs, drop_conditions)
        ]

        for df, drop_condition in zip(dfs, drop_conditions):
            df.drop(labels=df[drop_condition].index, axis=0, inplace=True)
            self.__check_df(df=df, pivot=pivot, in_outlier_remover=True)
            df.reset_index(drop=True, inplace=True)

    def __check_df(
        self,
        df: pd.DataFrame,
        pivot: CutInfo,
        in_outlier_remover: bool = False,
    ) -> None:
        is_validation_variable = self._variable_set_name == self.validation_variable_set_option  # type: bool
        is_exceptional_variable = any(v in pivot.cut_var_root for v in self.nan_check_variable_exceptions)  # type: bool
        is_exceptional = is_exceptional_variable or self.n_tuple == PhotonObservables.n_tuple_name()  # type: bool
        is_nan_exception = is_exceptional and is_validation_variable  # type: bool

        if in_outlier_remover:
            columns_with_nans = {c for c in df.columns if df[c].isnull().values.any()}  # type: Set[str]
            if columns_with_nans and columns_with_nans.difference(self.__nan_check_failing_vars):
                self.__nan_check_failing_vars |= columns_with_nans
                print(
                    f"WARNING: Encountered NaNs after outlier removal in the following columns: "
                    f"{list(self.__nan_check_failing_vars)}"
                )

        if is_nan_exception:
            df[pivot.cut_var_root].replace(to_replace=[np.inf, -np.inf], value=np.nan, inplace=True)

        if df[pivot.cut_var_root].isnull().any():
            if is_nan_exception:
                df.dropna(axis=0, subset=[pivot.cut_var_root], inplace=True)
            else:
                raise ValueError(f"NaN values encountered in relevant column {pivot.cut_var_root}!")

    def plot_cut_lines(
        self,
        ax: AxesType,
        cut_info: CutInfo,
    ) -> None:
        if self._pre_cuts == self.apply_all_cuts_option:
            return

        if cut_info.cut_type.cut_direction is None:
            return

        if isinstance(cut_info.cut_limits, tuple):
            cuts = list(cut_info.cut_limits)
            cut_directions = list(cut_info.cut_type.cut_direction)
            assert len(cut_directions) == len(cuts)
        else:
            cuts = [cut_info.cut_limits]
            assert isinstance(cut_info.cut_type.cut_direction, str), cut_info.cut_type.type_str
            cut_directions = [cut_info.cut_type.cut_direction]

        for cut, cut_dir in zip(cuts, cut_directions):
            ax.axvline(cut, color=self.cut_line_color, lw=self.cut_line_width, linestyle="--")
            arrow_y_start = 0.9 * ax.get_ylim()[1]
            x_scale = 0.02 * abs(ax.get_xlim()[1] - ax.get_xlim()[0])

            dir_fac = 1.0 if cut_dir is CutType.greater_than_cut_direction else -1.0  # type: float

            a_params = {
                "color": self.cut_line_color,
                "ec": self.cut_line_color,
                "fc": self.cut_line_color,
                "lw": self.cut_line_width,
                "arrowstyle": "-|>",
            }  # type: Dict[str, Any]

            ax.annotate(
                s="",
                xy=(cut + dir_fac * (1 + self.arrow_length) * x_scale, arrow_y_start),
                xytext=(cut + dir_fac * x_scale, arrow_y_start),
                xycoords="data",
                arrowprops=a_params,
            )

    def check_columns(self, df: pd.DataFrame) -> None:
        relevant_tuple_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]
        missing_vars_dict = {}  # type: Dict[str, List[str]]

        for tuple_key in relevant_tuple_keys:
            if self._variable_set_name == self.cut_variable_set_option:
                __cis = [ci for ci in self.cut_dicts[tuple_key]["cuts"] if isinstance(ci, CutInfo)]  # type: List[CutInfo]
                expected_vars = [ci.cut_var_root for ci in __cis]  # type: List[str]
            elif self._variable_set_name == self.validation_variable_set_option:
                hvc = SpecialVariableCollections.get_variable_collection_for_n_tuple(n_tuple_name=tuple_key)
                expected_vars = [
                    hv.df_label for hv in hvc.get_set_of_observables(root_compatible=True, use_mc_info=False)
                ]
            else:
                raise RuntimeError(f"Invalid option {self.variable_set_option_name} = {self._variable_set_name}!")

            missing_vars = [v for v in expected_vars if v not in df.columns]  # type: List[str]
            if missing_vars:
                missing_vars_dict.update({tuple_key: missing_vars})

        if missing_vars_dict:
            info_str = ""
            for key, value in missing_vars_dict.items():
                info_str += f"\t{key}:\n\t\t-" + "\n\t\t-".join(value)
            raise RuntimeError("The following expected variables are not available in the data frame:\n" + info_str)

    @staticmethod
    def update_output_dict(
        output_dict: Dict[str, PathType],
        new_output_dict: Dict[str, PathType],
    ) -> None:
        for o_key, o_path in new_output_dict.items():
            assert o_key not in output_dict.keys(), o_key
            assert o_path not in output_dict.values(), o_path
            output_dict.update({o_key: o_path})

    @staticmethod
    def get_settings_combinations() -> Tuple[Dict[str, str], ...]:
        options = {
            LowLevelPlotterBaseClass.cut_option_name: tuple(LowLevelPlotterBaseClass.apply_cuts_options.keys()),
            LowLevelPlotterBaseClass.variable_set_option_name: tuple(LowLevelPlotterBaseClass.variable_set_options),
            LowLevelPlotterBaseClass.remove_outlier_option_name: LowLevelPlotterBaseClass.remove_outliers_options,
        }  # type: Dict[str, Tuple[Union[str, bool], ...]]

        value_combinations = list(itertools.product(*options.values()))

        return tuple(
            [
                dict(zip(options.keys(), vc))
                for vc in value_combinations
                if not (
                    vc[0] in LowLevelPlotterBaseClass.special_cut_options
                    and not LowLevelPlotterBaseClass.variable_set_options[vc[1]]
                )
            ]
        )


if __name__ == "__main__":
    for option_combination in LowLevelPlotterBaseClass.get_settings_combinations():
        print(option_combination)
