"""
Covers additional observables of interest, which are not used for cuts,
but can be used to evaluate the quality of the reconstruction.

Felix Metzner 2021
"""

import copy
import pandas as pd

from abc import ABC, abstractmethod
from typing import Sequence, Iterable, List, Tuple, Dict, Type

from templatefitter.plotter.histogram_variable import HistVariable

from analysistools.processing.observables_utils import makeROOTCompatible

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.latex_labels.latex_string_methods import LatexObsNameMethods

from rdstar.studies.low_level_cut_study.utilities.basic_observables_collections import (
    BasicObservablesCollectionBase,
    GeneralObservablesCollection,
    MomentumObservablesCollection,
    EnergyObservablesCollection,
    MassObservablesCollection,
    ImpactParameterObservablesCollection,
    TrackObservablesCollection,
    ClusterObservablesCollection,
    FSRCorrectionObservablesCollection,
)

__all__ = [
    "LowLevelStudySpecialObservablesBase",
    "NeutralPionObservables",
    "ChargedPionObservables",
    "ElectronObservables",
    "MuonObservables",
    "KShortObservables",
    "KaonObservables",
    "PhotonObservables",
    "NeutralDMesonObservables",
    "ChargedDMesonObservables",
    "NeutralDStarMesonObservables",
    "ChargedDStarMesonObservables",
    "SpecialVariableCollections",
]


class LowLevelStudySpecialObservablesBase(ABC):
    particle_name = ""  # type: str
    particle_label = ""  # type: str
    default_binning = BasicObservablesCollectionBase.default_binning  # type: int

    @classmethod
    def particle(cls) -> str:
        return cls.particle_name

    @classmethod
    def n_tuple_name(cls) -> str:
        return f"{cls.particle()}_ntuple"

    @classmethod
    def get_set_of_observables(cls, root_compatible: bool, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        hist_var_set = tuple(set(cls._get_set_of_observables(use_mc_info=use_mc_info)))
        assert len(hist_var_set) > 0, [v.df_label for v in hist_var_set]

        if root_compatible:
            return tuple(
                [
                    HistVariable(
                        df_label=makeROOTCompatible([hv.df_label])[0],
                        n_bins=hv.n_bins,
                        scope=hv.scope,
                        var_name=hv.variable_name,
                        unit=hv.unit,
                        use_log_scale=hv.use_log_scale,
                    )
                    for hv in hist_var_set
                ]
            )

        return hist_var_set

    @staticmethod
    def get_histogram_variable(
        base_hist_var: HistVariable,
        dfs: Sequence[pd.DataFrame],
        round_scope_precision: int = 0,
    ) -> HistVariable:
        return HistVariable.get_scoped_histogram_variable(
            base_hist_var=base_hist_var, dfs=dfs, round_scope_precision=round_scope_precision
        )

    @classmethod
    def get_for_daughter(cls, daughter_index: int, variable_list: Iterable[HistVariable]) -> List[HistVariable]:
        assert 0 <= daughter_index <= 20, daughter_index
        assert cls.particle_label, cls.particle_label

        new_hist_vars = []  # type: List[HistVariable]
        for hist_var in variable_list:
            new_df_label = f"daughter({daughter_index}, {hist_var.df_label})"  # type: str
            new_latex_label = LatexObsNameMethods.create_x_of_nth_daughter_of_y(
                x=hist_var.variable_name, y=cls.particle_label, n=daughter_index
            )  # type: str

            new_hist_vars.append(
                HistVariable(
                    df_label=new_df_label,
                    n_bins=hist_var.n_bins,
                    var_name=new_latex_label,
                    unit=hist_var.unit,
                )
            )

        return new_hist_vars

    @staticmethod
    def get_in_cms_frame(hist_var_list: Sequence[HistVariable]) -> List[HistVariable]:
        return [
            HistVariable(
                df_label=f"useCMSFrame({hist_var.df_label})",
                n_bins=hist_var.n_bins,
                var_name=LatexObsNameMethods.add_super_script(
                    var_str=hist_var.variable_name, script=LatexObsNames.cms_tag
                ),
                unit=hist_var.unit,
            )
            for hist_var in hist_var_list
        ]

    @staticmethod
    def get_in_rest_frame(hist_var_list: Sequence[HistVariable]) -> List[HistVariable]:
        return [
            HistVariable(
                df_label=f"useRestFrame({hist_var.df_label})",
                n_bins=hist_var.n_bins,
                var_name=LatexObsNameMethods.add_super_script(
                    var_str=hist_var.variable_name, script=LatexObsNames.rest_tag
                ),
                unit=hist_var.unit,
            )
            for hist_var in hist_var_list
        ]

    @staticmethod
    def get_mc_matched(hist_var_list: Sequence[HistVariable]) -> List[HistVariable]:
        return [
            HistVariable(
                df_label=f"matchedMC({hist_var.df_label})",
                n_bins=hist_var.n_bins,
                var_name=LatexObsNameMethods.add_super_script(
                    var_str=hist_var.variable_name, script=LatexObsNames.mc_tag
                ),
                unit=hist_var.unit,
            )
            for hist_var in hist_var_list
        ]

    @classmethod
    @abstractmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        raise NotImplementedError()


class NeutralPionObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "piNeutral_from_list"  # type: str

    particle_label = LatexParticleNames.pi_zero_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    pi0_daughters_inv_m = HistVariable(
        df_label="daughterInvariantMass(0, 1)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_inv_m,
        unit=LatexUnits.gev,
    )

    pi0_daughters_angle = HistVariable(
        df_label="daughterAngle(0, 1)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_cos_angle,
        unit=LatexUnits.rad,
    )

    pi0_daughters_angle_rest = HistVariable(
        df_label="useRestFrame(daughterAngle(0, 1))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_rest_cos_angle,
        unit=LatexUnits.rad,
    )

    pi0_daughters_angle_btw = HistVariable(
        df_label="daughterAngleInBetween(0, 1)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_angle,
        unit=LatexUnits.rad,
    )

    pi0_daughters_angle_btw_rest = HistVariable(
        df_label="useRestFrame(daughterAngleInBetween(0, 1))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_rest_cos_angle,
        unit=LatexUnits.rad,
    )

    pi0_daughters_sum_e = HistVariable(
        df_label="daughterSumOf(E)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_sum_e,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_e = HistVariable(
        df_label="daughterDiffOf(0, 1, E)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_diff_e,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_e_normed = HistVariable(
        df_label="formula(daughterDiffOf(0, 1, E) / daughterSumOf(E))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_diff_e_normed,
        unit=LatexUnits.no_unit,
    )

    pi0_daughters_diff_phi = HistVariable(
        df_label="daughterDiffOf(0, 1, phi)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_delta_phi,
        unit=LatexUnits.rad,
    )

    pi0_daughters_diff_phi_cms = HistVariable(
        df_label="useCMSFrame(daughterDiffOf(0, 1, phi))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_cms_delta_phi,
        unit=LatexUnits.rad,
    )

    pi0_daughters_diff_theta = HistVariable(
        df_label="daughterDiffOf(0, 1, theta)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_delta_theta,
        unit=LatexUnits.rad,
    )

    pi0_daughters_diff_theta_cms = HistVariable(
        df_label="useCMSFrame(daughterDiffOf(0, 1, theta))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_cms_delta_theta,
        unit=LatexUnits.rad,
    )

    pi0_daughters_diff_p = HistVariable(
        df_label="daughterDiffOf(0, 1, p)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_diff_p,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_p_cms = HistVariable(
        df_label="useCMSFrame(daughterDiffOf(0, 1, p))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_cms_diff_p,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_pt = HistVariable(
        df_label="daughterDiffOf(0, 1, pt)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_diff_pt,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_pt_cms = HistVariable(
        df_label="useCMSFrame(daughterDiffOf(0, 1, pt))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_cms_diff_pt,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_pz = HistVariable(
        df_label="daughterDiffOf(0, 1, pz)",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_diff_pz,
        unit=LatexUnits.gev,
    )

    pi0_daughters_diff_pz_cms = HistVariable(
        df_label="useCMSFrame(daughterDiffOf(0, 1, pz))",
        n_bins=default_binning,
        var_name=LatexObsNames.pi0_daughters_cms_diff_pz,
        unit=LatexUnits.gev,
    )

    __pi0_hist_vars = [
        pi0_daughters_inv_m,
        pi0_daughters_angle,
        pi0_daughters_angle_rest,
        pi0_daughters_angle_btw,
        pi0_daughters_angle_btw_rest,
        pi0_daughters_sum_e,
        pi0_daughters_diff_e,
        pi0_daughters_diff_e_normed,
        pi0_daughters_diff_phi,
        pi0_daughters_diff_phi_cms,
        pi0_daughters_diff_theta,
        pi0_daughters_diff_theta_cms,
        pi0_daughters_diff_p,
        pi0_daughters_diff_p_cms,
        pi0_daughters_diff_pt,
        pi0_daughters_diff_pt_cms,
        pi0_daughters_diff_pz,
        pi0_daughters_diff_pz_cms,
    ]  # type: List[HistVariable]

    __mom_hist_vars = MomentumObservablesCollection.get_hist_vars()
    __pi0_hist_vars.extend(__mom_hist_vars)
    __pi0_hist_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(hist_var_list=__mom_hist_vars))
    __pi0_hist_vars.extend(MassObservablesCollection.get_hist_vars())
    __pi0_hist_vars.extend(EnergyObservablesCollection.get_hist_vars())

    pi0_hist_vars = tuple(__pi0_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        hist_vars = [hv for hv in cls.pi0_hist_vars]  # type: List[HistVariable]
        for d_index in [0, 1]:
            hist_vars.extend(
                cls.get_for_daughter(daughter_index=d_index, variable_list=ClusterObservablesCollection.get_hist_vars())
            )
        return tuple(hist_vars)


class ChargedPionObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "piCharged"  # type: str

    particle_label = LatexParticleNames.pi_pm_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    __pi_var_list = []  # type: List[HistVariable]

    __mom_hist_vars = MomentumObservablesCollection.get_hist_vars()
    __pi_var_list.extend(__mom_hist_vars)
    __pi_var_list.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(__mom_hist_vars))
    __pi_var_list.extend(EnergyObservablesCollection.get_hist_vars())
    __pi_var_list.extend(TrackObservablesCollection.get_hist_vars())
    __pi_var_list.extend(ImpactParameterObservablesCollection.get_hist_vars())

    pi_hist_vars = tuple(__pi_var_list)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        return cls.pi_hist_vars


class LeptonObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "lepton"  # type: str

    particle_label = LatexParticleNames.lepton_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    __lepton_hist_vars = []  # type: List[HistVariable]

    __mom_hist_vars = MomentumObservablesCollection.get_hist_vars()
    __lepton_hist_vars.extend(__mom_hist_vars)
    __lepton_hist_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(__mom_hist_vars))
    __lepton_hist_vars.extend(TrackObservablesCollection.get_hist_vars())
    __lepton_hist_vars.extend(ImpactParameterObservablesCollection.get_hist_vars())
    __lepton_hist_vars.extend(EnergyObservablesCollection.get_hist_vars())

    lepton_hist_vars = tuple(__lepton_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        return cls.lepton_hist_vars


class ElectronObservables(LeptonObservables):
    particle_name = "e"  # type: str

    particle_label = LatexParticleNames.electron_str  # type: str
    default_binning = LeptonObservables.default_binning  # type: int

    __e_hist_vars = []  # type: List[HistVariable]

    __e_hist_vars.extend(
        [
            hv
            for hv in FSRCorrectionObservablesCollection.get_hist_vars()
            if hv != FSRCorrectionObservablesCollection.fsr_gamma_has_correction_hist_var
        ]
    )

    e_hist_vars = tuple(__e_hist_vars)  # type: Tuple[HistVariable, ...]

    __mc_hist_vars = [EnergyObservablesCollection.energy_e_hist_var]  # type: List[HistVariable]
    __mc_hist_vars.extend(MomentumObservablesCollection.get_hist_vars())
    __mc_hist_vars.extend(ImpactParameterObservablesCollection.get_hist_vars())
    __mc_e_hist_vars = LowLevelStudySpecialObservablesBase.get_mc_matched(
        hist_var_list=__mc_hist_vars
    )  # type: List[HistVariable]
    mc_e_hist_vars = tuple(__mc_e_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        h_vars = list(copy.copy(LeptonObservables._get_set_of_observables(use_mc_info=use_mc_info)))
        h_vars.extend(list(copy.copy(cls.e_hist_vars)))
        h_vars.extend(cls.get_for_daughter(daughter_index=0, variable_list=TrackObservablesCollection.get_hist_vars()))
        h_vars.extend(
            cls.get_for_daughter(
                daughter_index=0,
                variable_list=ImpactParameterObservablesCollection.get_hist_vars(),
            )
        )

        if use_mc_info:
            daughter_0_mc_hist_vars = cls.get_for_daughter(daughter_index=0, variable_list=cls.mc_e_hist_vars)
            return tuple(h_vars + list(cls.mc_e_hist_vars) + daughter_0_mc_hist_vars)

        return tuple(h_vars)


class MuonObservables(LeptonObservables):
    particle_name = "mu"  # type: str

    particle_label = LatexParticleNames.mu_str  # type: str
    default_binning = LeptonObservables.default_binning  # type: int

    __mu_hist_vars = []  # type: List[HistVariable]
    mu_hist_vars = tuple(__mu_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        h_vars = list(copy.copy(LeptonObservables._get_set_of_observables(use_mc_info=use_mc_info)))
        h_vars.extend(list(copy.copy(cls.mu_hist_vars)))
        return tuple(h_vars)


class KShortObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "K_S0"  # type: str

    particle_label = LatexParticleNames.k_short_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    __ks0_hist_vars = []  # type: List[HistVariable]

    __mom_hist_vars = MomentumObservablesCollection.get_hist_vars()
    __ks0_hist_vars.extend(__mom_hist_vars)
    __ks0_hist_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(hist_var_list=__mom_hist_vars))
    __ks0_hist_vars.extend(MassObservablesCollection.get_hist_vars())
    __ks0_hist_vars.extend(EnergyObservablesCollection.get_hist_vars())
    __ks0_hist_vars.extend(ImpactParameterObservablesCollection.get_hist_vars())

    ks0_hist_vars = tuple(__ks0_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        return cls.ks0_hist_vars


class KaonObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "K"  # type: str

    particle_label = LatexParticleNames.k_pm_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    __k_hist_vars = []  # type: List[HistVariable]

    __mom_hist_vars = MomentumObservablesCollection.get_hist_vars()
    __k_hist_vars.extend(__mom_hist_vars)
    __k_hist_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(hist_var_list=__mom_hist_vars))
    __k_hist_vars.extend(EnergyObservablesCollection.get_hist_vars())
    __k_hist_vars.extend(TrackObservablesCollection.get_hist_vars())
    __k_hist_vars.extend(ImpactParameterObservablesCollection.get_hist_vars())

    k_hist_vars = tuple(__k_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        return cls.k_hist_vars


class PhotonObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "gamma"  # type: str

    particle_label = LatexParticleNames.gamma_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    __gamma_hist_vars = [
        FSRCorrectionObservablesCollection.fsr_gamma_has_correction_hist_var,
        MomentumObservablesCollection.momentum_phi_hist_var,
        MomentumObservablesCollection.momentum_theta_hist_var,
        MomentumObservablesCollection.momentum_cos_theta_hist_var,
    ]  # type: List[HistVariable]

    __gamma_hist_vars.extend(EnergyObservablesCollection.get_hist_vars())
    __gamma_hist_vars.extend(ClusterObservablesCollection.get_hist_vars())

    gamma_hist_vars = tuple(__gamma_hist_vars)  # type: Tuple[HistVariable, ...]

    __mc_hist_vars = [
        EnergyObservablesCollection.energy_e_hist_var,
        MomentumObservablesCollection.momentum_phi_hist_var,
        MomentumObservablesCollection.momentum_theta_hist_var,
        MomentumObservablesCollection.momentum_cos_theta_hist_var,
    ]  # type: List[HistVariable]
    __mc_gamma_hist_vars = LowLevelStudySpecialObservablesBase.get_mc_matched(
        hist_var_list=__mc_hist_vars
    )  # type: List[HistVariable]

    mc_gamma_hist_vars = tuple(__mc_gamma_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        if use_mc_info:
            return tuple(list(cls.gamma_hist_vars) + list(cls.mc_gamma_hist_vars))
        return cls.gamma_hist_vars


class DMesonObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "d"  # type: str

    particle_label = LatexParticleNames.d_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    __d_hist_vars = [
        GeneralObservablesCollection.general_pdg_code_hist_var,
        GeneralObservablesCollection.general_decay_mode_id_hist_var,
    ]  # type: List[HistVariable]

    __mom_hist_vars = MomentumObservablesCollection.get_hist_vars()
    __d_hist_vars.extend(__mom_hist_vars)
    __d_hist_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(hist_var_list=__mom_hist_vars))
    __d_hist_vars.extend(MassObservablesCollection.get_hist_vars())
    __d_hist_vars.extend(EnergyObservablesCollection.get_hist_vars())

    d_hist_vars = tuple(__d_hist_vars)  # type: Tuple[HistVariable, ...]

    __d_daughter_hist_vars = [
        # GeneralObservablesCollection.general_pdg_code_hist_var,
        MassObservablesCollection.mass_inv_m_hist_var,
        TrackObservablesCollection.track_curler_clone_indicator_hist_var,
    ]  # type: List[HistVariable]

    d_daughter_hist_vars = tuple(__d_daughter_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def get_set_of_observables_wo_daughter_vars(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        return copy.copy(cls.d_hist_vars)

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        h_vars = list(cls.get_set_of_observables_wo_daughter_vars(use_mc_info=use_mc_info))

        for d_i in range(5):
            h_vars.extend(cls.get_for_daughter(daughter_index=d_i, variable_list=cls.d_daughter_hist_vars))

            mom_hv_l = [MomentumObservablesCollection.momentum_p_hist_var]
            daughter_mom_hv_l = cls.get_for_daughter(daughter_index=d_i, variable_list=mom_hv_l)
            h_vars.extend(LowLevelStudySpecialObservablesBase.get_in_rest_frame(hist_var_list=daughter_mom_hv_l))
            h_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(hist_var_list=daughter_mom_hv_l))

        return tuple(h_vars)


class NeutralDMesonObservables(DMesonObservables):
    particle_name = "dzero"  # type: str

    particle_label = LatexParticleNames.d_zero_str  # type: str
    default_binning = DMesonObservables.default_binning  # type: int


class ChargedDMesonObservables(DMesonObservables):
    particle_name = "dcharged"  # type: str

    particle_label = LatexParticleNames.d_pm_str  # type: str
    default_binning = DMesonObservables.default_binning  # type: int


class DStarMesonObservables(LowLevelStudySpecialObservablesBase):
    particle_name = "dst"  # type: str

    particle_label = LatexParticleNames.d_st_str  # type: str
    default_binning = LowLevelStudySpecialObservablesBase.default_binning  # type: int

    dst_daughters_inv_m = HistVariable(
        df_label="daughterInvariantMass(0, 1)",
        n_bins=default_binning,
        var_name=LatexObsNames.d_st_daughters_inv_m,
        unit=LatexUnits.gev,
    )

    dst_daughter_mother_diff_m = HistVariable(
        df_label="daughterMotherDiffOf(0, M)",
        n_bins=default_binning,
        var_name=LatexObsNames.d_d_st_delta_m,
        unit=LatexUnits.gev,
    )

    dst_daughter_mother_diff_inv_m = HistVariable(
        df_label="daughterMotherDiffOf(0, InvM)",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_delta_m_1st_daughter_mother,
        unit=LatexUnits.gev,
    )

    dst_abs_mass_diff_significance = HistVariable(
        df_label="abs(massDifferenceSignificance(0))",
        n_bins=default_binning,
        var_name=LatexObsNames.mass_sigma_delta_m_1st_daughter_mother,
        unit=LatexUnits.rad,
    )

    dst_daughters_angle_btw_rest = HistVariable(
        df_label="useLabFrame(daughterAngleInBetween(0, 1))",
        n_bins=default_binning,
        var_name=LatexObsNames.d_st_daughters_lab_cos_angle,
        unit=LatexUnits.rad,
    )

    __dst_hist_vars = [
        dst_daughters_inv_m,
        dst_daughter_mother_diff_m,
        dst_daughter_mother_diff_inv_m,
        dst_abs_mass_diff_significance,
        dst_daughters_angle_btw_rest,
    ]  # type: List[HistVariable]

    dst_hist_vars = tuple(__dst_hist_vars)  # type: Tuple[HistVariable, ...]

    __dst_daughter_hist_vars = [
        # GeneralObservablesCollection.general_pdg_code_hist_var,
        EnergyObservablesCollection.energy_e_hist_var,
        MomentumObservablesCollection.momentum_p_hist_var,
    ]  # type: List[HistVariable]
    __dst_daughter_hist_vars.extend(MassObservablesCollection.get_hist_vars())

    dst_daughter_hist_vars = tuple(__dst_daughter_hist_vars)  # type: Tuple[HistVariable, ...]

    __dst_first_daughter_hist_vars = []  # type: List[HistVariable]

    dst_first_daughter_hist_vars = tuple(__dst_first_daughter_hist_vars)  # type: Tuple[HistVariable, ...]

    __dst_second_daughter_hist_vars = [
        NeutralPionObservables.pi0_daughters_inv_m,
        TrackObservablesCollection.track_curler_clone_indicator_hist_var,
    ]  # type: List[HistVariable]

    dst_second_daughter_hist_vars = tuple(__dst_second_daughter_hist_vars)  # type: Tuple[HistVariable, ...]

    @classmethod
    def _get_set_of_observables(cls, use_mc_info: bool = False) -> Tuple[HistVariable, ...]:
        h_vars = list(copy.copy(cls.dst_hist_vars))

        h_vars.extend(DMesonObservables.get_set_of_observables_wo_daughter_vars(use_mc_info=use_mc_info))

        for d_i in range(2):
            h_vars.extend(cls.get_for_daughter(daughter_index=d_i, variable_list=cls.dst_daughter_hist_vars))

            mom_hv_l = [MomentumObservablesCollection.momentum_p_hist_var]
            daughter_mom_hv_l = cls.get_for_daughter(daughter_index=d_i, variable_list=mom_hv_l)
            h_vars.extend(LowLevelStudySpecialObservablesBase.get_in_rest_frame(hist_var_list=daughter_mom_hv_l))
            h_vars.extend(LowLevelStudySpecialObservablesBase.get_in_cms_frame(hist_var_list=daughter_mom_hv_l))

        h_vars.extend(cls.get_for_daughter(daughter_index=0, variable_list=cls.dst_first_daughter_hist_vars))
        h_vars.extend(cls.get_for_daughter(daughter_index=1, variable_list=cls.dst_second_daughter_hist_vars))

        return tuple(h_vars)


class NeutralDStarMesonObservables(DStarMesonObservables):
    particle_name = "dstzero"  # type: str

    particle_label = LatexParticleNames.d_st_zero_str  # type: str
    default_binning = DStarMesonObservables.default_binning  # type: int


class ChargedDStarMesonObservables(DStarMesonObservables):
    particle_name = "dstcharged"  # type: str

    particle_label = LatexParticleNames.d_st_pm_str  # type: str
    default_binning = DStarMesonObservables.default_binning  # type: int


class SpecialVariableCollections:
    special_var_collections = {
        NeutralPionObservables.n_tuple_name(): NeutralPionObservables,
        ChargedPionObservables.n_tuple_name(): ChargedPionObservables,
        LeptonObservables.n_tuple_name(): LeptonObservables,
        ElectronObservables.n_tuple_name(): ElectronObservables,
        MuonObservables.n_tuple_name(): MuonObservables,
        KShortObservables.n_tuple_name(): KShortObservables,
        KaonObservables.n_tuple_name(): KaonObservables,
        PhotonObservables.n_tuple_name(): PhotonObservables,
        NeutralDMesonObservables.n_tuple_name(): NeutralDMesonObservables,
        ChargedDMesonObservables.n_tuple_name(): ChargedDMesonObservables,
        NeutralDStarMesonObservables.n_tuple_name(): NeutralDStarMesonObservables,
        ChargedDStarMesonObservables.n_tuple_name(): ChargedDStarMesonObservables,
    }  # type: Dict[str, Type[LowLevelStudySpecialObservablesBase]]

    @staticmethod
    def get_variable_collection_for_n_tuple(n_tuple_name: str) -> Type[LowLevelStudySpecialObservablesBase]:
        key = n_tuple_name  # type: str
        if n_tuple_name.startswith("d") and ":" in n_tuple_name:
            key = n_tuple_name.split(sep="_ntuple", maxsplit=1)[0] + "_ntuple"

        assert key in SpecialVariableCollections.special_var_collections.keys(), (key, n_tuple_name)
        return SpecialVariableCollections.special_var_collections[key]
